/***************************************************************************
    file	         : kb_sqlite.cpp
    copyright            : (C) 2004 by Brian Harring
			   (C) 1999,2000,2001,2002,2003 by Mike Richardson
			   (C) 2000,2001,2002,2003 by theKompany.com
			   (C) 2001,2002,2003 by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : ferringb@gentoo.org
***************************************************************************/


#include	"kb_sqlite3.h"
#include	"kb_basequery.h"

#define		__DEBUG_SQLITE3	(0)
#if		__DEBUG_SQLITE3
#include	"kb_dprintf.h"
#endif

#include	<cstdlib>
#include	<unistd.h>

static	QIntDict<SQLite3TypeMap>	dIdentToType	;
static	QDict	<KBSQLite3Type>		typesDict	;


static	SQLite3TypeMap typeMap[] =
{
/*	ident			itype		mtype		flags			*/
{	SQLITE_INTEGER,		KB::ITFixed,	"integer",	0			},
{	SQLITE_FLOAT,		KB::ITFloat,	"float",	FF_LENGTH|FF_PREC	},
{	SQLITE_BLOB,		KB::ITBinary,	"blob",		0			},
{	SQLITE_TEXT,		KB::ITString,	"text",		0			},
}	;


/*  KBSQLite3Type							*/
/*  KBSQLite3Type	: Constructor for Sqlite type object		*/
/*  typeInfo		: SQLite3TypeMap * : Type information		*/
/*  length		: uint		  : Underlying length		*/
/*  prec		: uint		  : Underlying precision	*/
/*  nullOK		: bool		  : True if null is OK		*/
/*  (returns)		: KBSqliteType	  :				*/

KBSQLite3Type::KBSQLite3Type
	(	SQLite3TypeMap	*typeInfo,
		uint		length,
		uint		prec,
		bool		nullOK
	)
	:
	KBType
	(	"SQLite3",
		typeInfo == 0 ? KB::ITString : typeInfo->itype,
		length,
		prec,
		nullOK
	),
	m_typeInfo( typeInfo )
{
}

/*  KBSqlite3Type							*/
/*  isValid	: Test if value is valid for type			*/
/*  value	: const QString & : Value to check			*/
/*  pError	: KBValue &	  : Error return			*/
/*  where	: const QString & : Caller				*/
/*  (returns)	: bool		  : Valid				*/

bool	KBSQLite3Type::isValid
	(	const QString	&value,
		KBError		&pError,
		const QString	&where
	)
{
	//**
	return KBType::isValid (value, pError, where) ;
}

/*  KBSQLite3Type							*/
/*  getQueryText:							*/
/*  value	: KBDataArray  * : Raw text of value to convert		*/
/*  d		: KBShared     * : Decoded representation		*/
/*  buffer	: KBDataBuffer & : Results buffer			*/
/*  (returns)	: void		 :					*/

void	KBSQLite3Type::getQueryText
	(	KBDataArray	*value,
		KBShared	*d,
		KBDataBuffer	&buffer,
		QTextCodec	*codec
	)
{
	XPRINTF
	((	"KBSQLite3Type::getQueryText: v=%p t=%d (binary=%d)\n",
		(void *)value,
		m_iType,
		KB::ITBinary
	))	;

	if ((value != 0) && (m_iType == KB::ITBinary))
	{
		const char *escaped = sqlite3_mprintf("%q", value->m_data) ;
		buffer.append (escaped) ;
		sqlite3_free  ((char *)escaped) ;
		return	;
	}

	KBType::getQueryText (value, d, buffer, codec) ;
}

/*  KBSQLite3Type							*/
/*  escapeText	: Escape text characters				*/
/*  value	: KBDataArray  * : Raw text to escape			*/
/*  buffer	: KBDataBuffer & : Escaped text				*/
/*  (returns)	: void		 :					*/

void	KBSQLite3Type::escapeText
	(	KBDataArray	*value,
		KBDataBuffer	&buffer
	)
{
	const char *escaped = sqlite3_mprintf("%q", value->m_data) ;
	buffer.append (escaped) ;
	sqlite3_free  ((char *)escaped) ;

//	char	   ch	;
//
//	for (uint idx = 0 ; idx < value->m_length ; idx += 1)
//		switch (ch = value->m_data[idx])
//		{
//			case '\'' :
//			case '\\' :
//				buffer.append('\\') ;
//				buffer.append(ch  ) ;
//				break	;
//
//			default	:
//				buffer.append(ch  ) ;
//				break	;
//		}
}

/*  KBSQLite3Type							*/
/*  escapeText	: Escape text characters				*/
/*  value	: QCString     & : Raw text to escape			*/
/*  buffer	: KBDataBuffer & : Escaped text				*/
/*  (returns)	: void		 :					*/

void	KBSQLite3Type::escapeText
	(	QCString	&value,
		KBDataBuffer	&buffer
	)
{
	const char *escaped = sqlite3_mprintf("%q", value.data()) ;
	buffer.append (escaped) ;
	sqlite3_free  ((char *)escaped) ;

//	char	   ch	;
//
//	for (uint idx = 0 ; idx < value.length() ; idx += 1)
//		switch (ch = value.at(idx))
//		{
//			case '\'' :
//			case '\\' :
//				buffer.append('\\') ;
//				buffer.append(ch  ) ;
//				break	;
//
//			default	:
//				buffer.append(ch  ) ;
//				break	;
//		}
}

void	KBSQLite3Type::retype
	(	int		type
	)
{
	switch (type)
	{
		case SQLITE_INTEGER :
			m_iType = KB::ITFixed	;
			break	;
		case SQLITE_FLOAT   :
			m_iType = KB::ITFloat	;
			break	;
		case SQLITE_BLOB    :
			m_iType = KB::ITBinary	;
			break	;
		case SQLITE_TEXT    :
			m_iType = KB::ITString	;
			break	;
		default	:
			break	;
	}
}


/*  ------------------------------------------------------------------  */

/*  KBSQLite															*/
/*  KBSQLite	: Constructor for Sqlite database connection class	*/
/*  (returns)	: KBServer	:											*/

KBSQLite3::KBSQLite3() : KBServer()
{
}

/*  KBSQLite3															*/
/*  ~KBSQLite3	: Destructor for Sqlite database connection class	*/
/*  (returns)	:		:												*/

KBSQLite3::~KBSQLite3 ()
{
}

/*  KBSQLite3								*/
/*  optionFlags	: Get server option flags				*/
/*  (returns)	: uint		: Flags					*/

uint	KBSQLite3::optionFlags ()
{
	return	0	;
}

/*  KBSQLite3								*/
/*  getQueryText: write the SQL expression for value into buffer	*/
/*  value	: KBValue*	: value to be converted			*/
/*  buffer	: KBDataBuffer&	: destination buffer			*/
/*  codec	: QTextCodec*	: non-default text codec		*/

void	KBSQLite3::getQueryText
	(	const KBValue  *value,
		KBDataBuffer   &buffer,
		QTextCodec     *codec
	)
{
	if (value->getType()->getIType() == KB::ITBinary)
	{
		const char *escaped = sqlite3_mprintf("%q", value->dataArray()->m_data) ;
		buffer.append (escaped) ;
		sqlite3_free  ((char *)escaped) ;
		return	;
	}

	value->getQueryText(buffer, codec) ;
}

/*  KBSQLite3								*/
/*  doConnect	: Open connection to database				*/
/*  svInfo	: KBServerInfo *: Server information			*/
/*  (returns)	: bool		: Success				*/

bool	KBSQLite3::doConnect
	(	KBServerInfo	*svinfo
	)
{
	DPRINTF
	((	"KBSQLite3::doConnect: [%s]\n",
		(cchar *)m_database
	))	;

	m_readOnly = svinfo->readOnly () ;

	QString	database = m_database ;

	if (database[0] == '$')
	{
		if (getenv(database.mid(1)) == 0)
		{
			m_lError = KBError
				   (	KBError::Error,
					QString("No '%1' environment parameter").arg(database),
					QString::null,
					__ERRLOCN
				   )	;
			return	false	;
		}

		database = getenv(database.mid(1)) ;
	}

	if (database[0] != '/')
		database = svinfo->getDBPath() + "/" + database ;

	DPRINTF
	((	"KBSQLite3::doConnect: [%s]\n",
		database.latin1()
	))	;

	if (sqlite3_open (database, &m_sqlite3) != SQLITE_OK)
	{
		m_lError = KBError
			   (	KBError::Error,
				TR("Unable to open SQLite database"),
				QString("%1: %2").arg(database).arg(sqlite3_errmsg(m_sqlite3)),
				__ERRLOCN
			   )	;
		sqlite3_close	(m_sqlite3) ;
		m_sqlite3 = 0 ;
		return	false ;
	}

	if (access (database, W_OK) != 0)
		m_readOnly = true ;

	return	true ;
}

/*  KBSQLite3								*/
/*  qrySelect	: Open a select query					*/
/*  data	: bool		  : Query for data			*/
/*  select	: const QString & : Select query			*/
/*  (returns)	: KBSQLSelect *	  : Select query class or NULL on error	*/

KBSQLSelect
	*KBSQLite3::qrySelect
	(	bool		data,
		const QString	&select,
		bool
	)
{
	return	new KBSQLite3QrySelect (this, data, select) ;
}


/*  KBSQLite3								*/
/*  qryUpdate	: Open an update query					*/
/*  data	: bool		  : Query for data			*/
/*  update	: const QString & : Update query			*/
/*  tabName	: const QString & : Table being updated			*/
/*  (returns)	: KBSQLUpdate *	  : Update query class or NULL on error	*/

KBSQLUpdate
	*KBSQLite3::qryUpdate
	(	bool		data,
		const QString	&update,
		const QString	&tabName
	)
{
	if (m_readOnly) 
	{
		m_lError = KBError
			   (	KBError::Error,
				TR("Database is read-only"),
				TR("Attempting update query"),
				__ERRLOCN
			   )	;
		return	0	;
	}

	return	new KBSQLite3QryUpdate (this, data, update, tabName) ;
}

/*  KBSQLite3								*/
/*  qryInsert	: Open an insert query					*/
/*  data	: bool		  : Query for data			*/
/*  insert	: const QString & : Insert query			*/
/*  tabName	: const QString & : Table being updated			*/
/*  (returns)	: KBSQLInsert *	  : Insert query class or NULL on error	*/

KBSQLInsert
	*KBSQLite3::qryInsert
	(	bool		data,
		const QString	&insert,
		const QString	&tabName
	)
{
	if (m_readOnly) 
	{
		m_lError = KBError
			   (	KBError::Error,
				TR("Database is read-only"),
				TR("Attempting insert query"),
				__ERRLOCN
			   )	;
		return	0	;
	}

	return	new KBSQLite3QryInsert (this, data, insert, tabName) ;
}

/*  KBSQLite3								*/
/*  qryDelete	: Open an delete query					*/
/*  data	: bool		  : Query for data			*/
/*  _delete	: const QString & : Delete query			*/
/*  tabName	: const QString & : Table being updated			*/
/*  (returns)	: KBSQLDelete *	  : Delete query class or NULL on error	*/

KBSQLDelete
	*KBSQLite3::qryDelete
	(	bool		data,
		const QString	&deleteStr,
		const QString	&tabName
	)
{
	if (m_readOnly)
	{
		m_lError = KBError
			   (	KBError::Error,
				TR("Database is read-only"),
				TR("Attempting delete query"),
				__ERRLOCN
			   )	;
		return 	0	;
	}

	return	new KBSQLite3QryDelete (this, data, deleteStr, tabName) ;
}

/*  KBSQLite3								*/
/*  command	: Execute arbitrary SQL					*/
/*  data	: bool		  : Querying for data			*/
/*  rawqry	: const QString & : Query text				*/
/*  nvals	: uint		  : Number of substitution values	*/
/*  values	: KBValue *	  : Substitution values			*/
/*  select	: KBSQLSelect **  : Return for result queries		*/
/*  (returns)	: bool		  : Success				*/

bool	KBSQLite3::command
	(	bool		,
		const QString	&,
		uint		,
		KBValue		*,
		KBSQLSelect	**
	)
{
	m_lError = KBError
		   (	KBError::Error,
			TR("Not implemented"),
			TR("Command execution"),
			__ERRLOCN
		   )	;
	return	false	;
}

/*  KBSQLite3								*/
/*  listTypes	: Get list of types with information flags		*/
/*  (returns)	: QString	: List as bar-separated string		*/

QString	KBSQLite3::listTypes()
{
	static QString typeList ;

	if (typeList.isNull())
	{
		typeList = "Primary Key,0|Foreign Key,0" ;

		for (uint idx = 0 ; idx < sizeof(typeMap)/sizeof(SQLite3TypeMap) ; idx += 1)
		{
			SQLite3TypeMap *m = &typeMap[idx] ;
			if ((m->flags & FF_NOCREATE) == 0)
				typeList += QString("|%1,%2").arg(m->mtype).arg(m->flags) ;
		}
	}

	return	typeList ;
}

/*  KBSQLite3								*/
/*  tableExists	: See if named table exists				*/
/*  table	: const QString & : Table name				*/
/*  exists	: bool &	  : True if table exists		*/
/*  (returns)	: bool		  : Success				*/

bool	KBSQLite3::tableExists
	(	const QString	&table,
		bool		&exists
	)
{
	char	**results;
	int	rc	 ;
	int	rows	 ;
	int	col	 ;

	QString	query	= QString("select name from sqlite_master where type = 'table' and name= '%1'")
				.arg(table) ;

	rc	= sqlite3_get_table
		(	m_sqlite3,
			query.latin1(),
			&results,
			&rows,
			&col,
			0
		)	;

	if (rc != SQLITE_OK)
		m_lError = KBError
			   (	KBError::Error,
				TR("Error checking if table exists"),
				sqlite3_errmsg(m_sqlite3),
				__ERRLOCN
			  )	;

	exists	= (rc == SQLITE_OK) && (rows > 0) ;
	sqlite3_free_table (results) ;
	return	rc == SQLITE_OK ;
}

/*  KBSQLite3								*/
/*  doListTables: List tables in database				*/
/*  tabList	: KBTableDetailsList & : Result list			*/
/*  type	: uint		       : Type flags			*/
/*  (returns)	: bool		       : Success			*/

bool	KBSQLite3::doListTables
	(	KBTableDetailsList	&tablist,
		uint			type
	)
{
	int	rc	 ;
	int 	rows	 ;
	int 	cols	 ;
	char	**results;

	QString	query	= QString("select name from sqlite_master where type = '%1'")
				.arg
				(	(type & KB::IsTable   ) != 0 ? "table"	   :
					(type & KB::IsView    ) != 0 ? "view"     :
					(type & KB::IsSequence) != 0 ? "sequence" :
								       "unknown"
				)	;

	rc	= sqlite3_get_table
			(	m_sqlite3,
				query.latin1(),
				&results,
				&rows,
				&cols,
				0
			)	;

	if (rc != SQLITE_OK)
	{
		m_lError = KBError
			   (	KBError::Error,
				TR("Error getting list of tables"),
				sqlite3_errmsg(m_sqlite3),
				__ERRLOCN
			  )	;
		sqlite3_free_table( results );
		return false	;
	}
		
		
	for (int row = 0 ; row < rows ; row += 1)
		tablist.append
		( 	KBTableDetails
			(	QString(results[row + 1]), 
				KB::IsTable,
				QP_SELECT | QP_INSERT | QP_UPDATE | QP_DELETE
		)	)	;
	
	sqlite3_free_table (results) ;
	return	true	;
}

/*  KBSQLite3								*/
/*  doListFields: List fields in table					*/
/*  tabSpec	: KBTableSpec &	: Table specification			*/
/*  (returns)	: bool		: Success				*/

bool	KBSQLite3::doListFields
	(	KBTableSpec	&tabspec
	)
{
	int	rc	  ;
	int	rows	  ;
	int	cols	  ;
	char	**results ;
	bool	anyAI	  = false	;
	
	tabspec.m_keepsCase =  1 ;
	tabspec.m_prefKey   = -1 ;
	
	rc = sqlite3_get_table
			(	m_sqlite3,
				QString("pragma table_info(%1)").arg(tabspec.m_name).latin1(),
				&results,
				&rows,
				&cols,
				0
			)	;
	
	if (rc != SQLITE_OK) 
	{
		m_lError = KBError
			   (	KBError::Error,
				TR("Error getting list of fields"),
				sqlite3_errmsg(m_sqlite3),
				__ERRLOCN
			  )	;
		return	false	;
	}

	
	for (int row = 1 ; row <= rows ; row += 1)
	{
		int		length	= 0	;
		int		prec	= 0	;
		uint		flags	= 0	;
		QCString      	s	= QCString(results[(row * cols) + 2]) ;
		SQLite3TypeMap	*t	= 0	;

		cchar		*name	= results[(row * cols ) + 1] ;
		cchar		*mtype	= results[(row * cols ) + 2] ;

		if (results[(row * cols) + 5][0] != '0') 
			flags |= KBFieldSpec::Primary|KBFieldSpec::Unique ;
		
		if (results[(row * cols) + 3][0] != '0')
			flags |= KBFieldSpec::NotNull ;


		struct	SQLiteType
		{	const char	*name	;
			int		code	;
		}	;
		static	SQLiteType	sqltypes[]	=
		{
			{	"INT",		SQLITE_INTEGER		},
			{	"CHAR",		SQLITE_TEXT		},
			{	"TEXT",		SQLITE_TEXT		},
			{	"VARCHAR",	SQLITE_TEXT		},
			{	"BINARY",	SQLITE_BLOB		},
			{	"BLOB",		SQLITE_BLOB		},
			{	"FLOAT",	SQLITE_FLOAT		},
			{	"NUMERIC",	SQLITE_FLOAT		},
			{	"NUMBER",	SQLITE_FLOAT		},
			{	"DECIMAL",	SQLITE_FLOAT		},
			{	"REAL",		SQLITE_FLOAT		},
			{	"DOUBLE",	SQLITE_FLOAT		},
//			{	"TIMESTAMP",	SQLITE_TIMESTAMP	},
//			{	"TIME",		SQLITE_TIMESTAMP	},
//			{	"DATE",		SQLITE_DATE		},
			{	0,		0			}
		}	;

		for (SQLiteType *st = &sqltypes[0] ; st->name != 0 ; st += 1)
			if (s.find(st->name, 0, false) >= 0)
			{	
				t = dIdentToType.find (st->code) ;
				break	;
			}
		if (t == 0)
		{
			t = dIdentToType.find (SQLITE_TEXT) ;
			length = 0 ;
		}

		if ((flags & KBFieldSpec::Primary) != 0)
			if (t->ident == SQLITE_INTEGER)
			{
				mtype		  = "Primary Key"	;
				flags		 |= KBFieldSpec::Serial ;
				flags		 |= KBFieldSpec::NotNull;
				anyAI		  = true    ;
				tabspec.m_prefKey = row - 1 ;
			}

		KBFieldSpec *fspec = new KBFieldSpec
					 (	row - 1,
						name,
						mtype,
						t->itype,
						flags,
						length,
						prec
					 )	;
		
		fspec->m_dbType = new KBSQLite3Type (t, length, prec, 0) ;
		tabspec.m_fldList.append (fspec) ;	
	}

	sqlite3_free_table (results) ;

	/* If there is any auto-increment column then we can retrieve	*/
	/* a unique key value from any unique column.			*/
	if (anyAI)
		LITER
		(	KBFieldSpec,
			tabspec.m_fldList,
			fSpec,

			if ((fSpec->m_flags & KBFieldSpec::Unique) != 0)
				fSpec->m_flags |= KBFieldSpec::InsAvail ;
		)

	/* If there is no primary key column then the preferred key can	*/
	/* be a unique key column.					*/
	if (tabspec.m_prefKey < 0)
		for (uint idx = 0 ; idx < tabspec.m_fldList.count() ; idx += 1)
		{	KBFieldSpec *fSpec = tabspec.m_fldList.at(idx) ;
			if ((fSpec->m_flags & KBFieldSpec::Unique) != 0)
			{	tabspec.m_prefKey = idx ;
				break	;
			}
		}


	return	true	;
}

bool	KBSQLite3::tblCreateSQL
	(	QList<KBFieldSpec>	&kbf, 
		const QString		&table_name,
		QString			&sql,
		bool			best
	)
{
	QString 	sep		= " " ;

	sql	= QString ("create table '%1' (").arg(table_name) ;

	LITER
	(	KBFieldSpec,
		kbf,
		fspec,

		QString		ftype	= fspec->m_typeName ;
		KB::IType	itype	= fspec->m_typeIntl ;


		if (ftype == "Primary Key")
		{
			sql    += sep + fspec->m_name + " integer primary key " ;
			sep	= ", "	;
			continue	;
		}
		if (ftype == "Foreign Key")
		{
			sql    += sep + fspec->m_name + " integer not null "    ;
			sep	= ", "	;
			continue	;
		}

		/* Map the types used when creating the object and	*/
		/* design dictionary tables.				*/
		if	(ftype == "_Text"   ) ftype = "text"	;
		else if (ftype == "_Integer") ftype = "integer" ;
		else if (ftype == "_Binary" ) ftype = "blob"	;

		SQLite3TypeMap *pTypeMap = 0 ;
		for (uint idx1 = 0 ; idx1 < sizeof(typeMap)/sizeof(SQLite3TypeMap) ; idx1 += 1)
			if (typeMap[idx1].mtype == ftype)
			{
				pTypeMap = &typeMap[idx1] ;
				break	 ;
			}

		if ((pTypeMap == 0) && best) 
			for (uint idx2 = 0; idx2 < sizeof(typeMap)/sizeof(SQLite3TypeMap) ; idx2 += 1)
				if (typeMap[idx2].itype == itype)
					if ((typeMap[idx2].flags & FF_NOCREATE) == 0)
					{	
						pTypeMap = &typeMap[idx2] ;
						break	 ;
					}

		if (pTypeMap == 0) 
		{	
			m_lError = KBError
				   (	KBError::Error,
					TR("Error mapping column type"),
					QString(TR("Type %1 for column %2 unknown")).arg(ftype).arg(fspec->m_name),
					__ERRLOCN
				   )	;
			
			return	false	;
		}

		sql    += QString("%1\t%2 %3").arg(sep).arg(fspec->m_name).arg(pTypeMap->mtype) ;
		if ((fspec->m_flags & KBFieldSpec::NotNull) != 0)
			sql	+= " not null" ;

		sep 	= ", " ;
	)

	sql	+= ")"	;

	DPRINTF
	((	"KBSQLite3::tblCreateSQL: [%s]\n",
		sql.latin1()
	))	;

	return	true	;
}
	
	
/*  KBSQLite3								*/
/*  doCreateTable: Create a new table					*/
/*  tabSpec	 : KBTableSpec &	: Table specification		*/
/*  (returns)	 : bool			: Success			*/

bool	KBSQLite3::doCreateTable
	(	KBTableSpec	&kbt,
		bool		,
		bool		best
	)
{
	QString	sql	;
	
	if (!tblCreateSQL (kbt.m_fldList, kbt.m_name, sql, best)) 
		return	false	;
	
	char	*errmsg	;
	int	ret	= sqlite3_exec(m_sqlite3, sql.latin1(), 0, 0, &errmsg) ;

	if (ret != SQLITE_OK) 
	{
		fprintf
		(	stderr,
			"KBSQLite3::doCreateTable: [%s]\n",
			sql.latin1()
		)	;
		m_lError = KBError
			   (	KBError::Error,
				TR("Unexpected error creating table"),
				errmsg,
				__ERRLOCN
			   )	;
		free	(errmsg);
		return	false	;
	}

	return	true	;
}

/*  KBSQLite3								*/
/*  doRenameTable: Rename a table					*/
/*  oldName	 : cchar *	: Current table name			*/
/*  newName	 : cchar *	: New table name			*/
/*  assoc	 : bool		: Rename associated stuff - none here	*/
/*  (returns)	 : bool		: Success				*/

bool	KBSQLite3::doRenameTable
	(	cchar		*oldName,
		cchar		*newName,
		bool
	)
{
	int 	ret		;
	char	*errmsg		;
	QString	sql		;

	sql	= QString("alter table %1 rename to %2").arg(oldName).arg(newName) ;
	ret	= sqlite3_exec
			(	m_sqlite3,
				sql.latin1(),
				0,
				0,
				&errmsg
			)	;
	if (ret != SQLITE_OK) 
	{
		m_lError = KBError
			   (	KBError::Error,
				TR("Rename table failed"),
				errmsg,
				__ERRLOCN
			   )	 ;
		free	(errmsg) ;
		return	false	 ;
	}

	return	true	;
}

/*  KBSQLite3								*/
/*  doDropTable	: Drop a table						*/
/*  table	: cchar *	: Table name				*/
/*  assoc	: bool		: Drop associated stuff - none here	*/
/*  (returns)	: bool		: Success				*/

bool	KBSQLite3::doDropTable
	(	cchar		*table,
		bool
	)
{
	int 	ret		;
	char	*errmsg		;
	QString	sql		;

	sql	= QString("drop table %1").arg(table) ;
	ret	= sqlite3_exec
			(	m_sqlite3,
				sql.latin1(),
				0,
				0,
				&errmsg
			)	;
	if (ret != SQLITE_OK) 
	{
		m_lError = KBError
			   (	KBError::Error,
				TR("Drop table failed"),
				errmsg,
				__ERRLOCN
			   )	 ;
		free	(errmsg) ;
		return	false	 ;
	}

	return	true	;
}

bool	KBSQLite3::bindParameters
	(	sqlite3_stmt	*stmt,
		uint		nvals,
		const KBValue	*values,
		KBError		&pError
	)
{
	for (uint idx = 0 ; idx < nvals ; idx += 1)
	{
		const KBValue &value = values[idx]     ;
		KB::IType     itype  = value.getType()->getIType() ;
		const char    *dptr  = value.dataPtr() ;

		if (value.isNull())
		{
			sqlite3_bind_null (stmt, idx + 1) ;
			continue ;
		}

		switch (itype)
		{
			case KB::ITFixed   :
				sqlite3_bind_int    (stmt, idx + 1, atoi(dptr)) ;
				break	;

			case KB::ITFloat   :
				sqlite3_bind_double (stmt, idx + 1, atof(dptr)) ;
				break	;

			case KB::ITDate    :
			case KB::ITTime    :
			case KB::ITDateTime:
			case KB::ITString  :
				sqlite3_bind_text   (stmt, idx + 1, dptr, value.dataLength(), SQLITE_STATIC) ;
				break	;

			case KB::ITBinary  :
				sqlite3_bind_blob   (stmt, idx + 1, dptr, value.dataLength(), SQLITE_STATIC) ;
				break	;

			default :
				pError = KBError
					 (	KBError::Error,
						QString(TR("Unrecognised type while binding paramaters: %1")).arg(itype),
						QString(TR("Internal type %1")).arg((int)itype),
						__ERRLOCN
					 )	;
				return	false	;
		}
	}

	return	true	;
}

bool	KBSQLite3::execSQL
	(	const QString	&rawSql,
		const QString	&tag,
		QString		&subSql,
		uint		nvals,
		const KBValue	*values,
		QTextCodec	*codec,
		cchar		*emsg,
		KBError		&pError
	)
{
	sqlite3_stmt	*stmt	;
	const char	*zSql	= rawSql.latin1() ;
	const char	*zTail	;
	if (sqlite3_prepare (m_sqlite3, zSql, qstrlen(zSql), &stmt, &zTail) != SQLITE_OK)
	{
		pError = KBError
			 (	KBError::Error,
				TR("Error executing SQL"),
				sqlite3_errmsg(m_sqlite3),
				__ERRLOCN
			 )	;
		return	false	;
	}

	if (!bindParameters (stmt, nvals, values, pError))
	{
		sqlite3_finalize(stmt) ;
		return	false	;
	}

	switch (sqlite3_step (stmt))
	{
		case SQLITE_DONE   :
			printQuery	 (rawSql, tag, nvals, values, true ) ;
			sqlite3_finalize (stmt) ;
			return	true	;

		case SQLITE_ERROR  :
		case SQLITE_MISUSE :
			pError = KBError
				 (	KBError::Error,
					TR("Error executing SQL"),
					sqlite3_errmsg(m_sqlite3),
					__ERRLOCN
				 )	;
			break	;

		case SQLITE_BUSY :
			pError = KBError
				 (	KBError::Error,
					TR("Error executing SQL"),
					TR("Database is busy"),
					__ERRLOCN
				 )	;
			break	;

		case SQLITE_ROW	 :
			pError = KBError
				 (	KBError::Error,
					TR("Error executing SQL"),
					TR("Unexpected data"),
					__ERRLOCN
				 )	;
			break	;

		default	:
			pError = KBError
				 (	KBError::Error,
					TR("Error executing SQL"),
					TR("Unknown SQLite3 return code"),
					__ERRLOCN
				 )	;
			break	;
	}

	printQuery	 (rawSql, tag, nvals, values, false) ;
	sqlite3_finalize (stmt) ;
	return	false	;
}


/*  ------------------------------------------------------------------- */

/*  KBSQLite3QrySelect							*/
/*  KBSQLite3QrySelect							*/
/*		: Constructor for select query object			*/
/*  server	: KBSQLite3 *		: Server connection object	*/
/*  data	: bool			: Query for data		*/
/*  query	: const QString &	: Select query			*/
/*  (returns)	: KBSQLite3QrySelect	:				*/

KBSQLite3QrySelect::KBSQLite3QrySelect
	(	KBSQLite3		*server,
		bool			data,
		const QString		&query
	)	
	:
	KBSQLSelect	(server, data, query),
	m_server	(server)
{
	m_nRows 	= 0	;
	m_nFields	= 0	;
}

/*  KBSQLite3QrySelect							*/
/*  ~KBSQLite3QrySelect							*/
/*		: Destructor for select query object			*/
/*  (returns)	:		:					*/

KBSQLite3QrySelect::~KBSQLite3QrySelect ()
{
}

/*  KBSQLite3QrySelect							*/
/*  execute	: Execute query						*/
/*  nvals	: uint		: Number of substitution values		*/
/*  values	: KBValue *	: Substitution values			*/
/*  (returns)	 bool		: Success				*/

bool	KBSQLite3QrySelect::execute
	(	uint		nvals,
		const KBValue	*values
	)
{
	sqlite3_stmt	*stmt	;
	const char	*zSql	= m_rawQuery.latin1() ;
	const char	*zTail	;
	bool		rv	= false	;

	DPRINTF
	((	"KBSQLite3QrySelect::execute: [%s]\n",
		zSql
	))	;
//	for (uint idx = 0 ; idx < nvals ; idx += 1)
//		XPRINTF
//		((	"KBSQLite3QrySelect::execute: [%s]\n",
//			values[idx].getRawText().latin1()
//		))	;

	if (sqlite3_prepare (m_server->handle(), zSql, qstrlen(zSql), &stmt, &zTail) != SQLITE_OK)
	{
		m_lError = KBError
			   (	KBError::Error,
				TR("Error executing SQL"),
				sqlite3_errmsg(m_server->handle()),
				__ERRLOCN
			   )	;
		return	false	;
	}

	if (!m_server->bindParameters (stmt, nvals, values, m_lError))
	{
		sqlite3_finalize(stmt) ;
		return	false	;
	}

	
	m_nRows		= 0	;
//	m_nFields	= 0	;

	int	rc3	= sqlite3_step (stmt) ;

	/* Nasty part 1. If the select returns no records then SQLite3	*/
	/* returns no column information. In this case we fall back on	*/
	/* parsing the query ourselves and using that to provide the	*/
	/* count and the column names; assume all columns are text.	*/
	if (rc3 == SQLITE_DONE)
	{
		DPRINTF
		((	"KBSQLite3QrySelect::execute: no records returned\n"
		))	;

		KBBaseSelect	select	;
		if (!select.parse (m_rawQuery, 0))
		{
			m_lError = select.lastError() ;
			return	 false	;
		}

		m_nFields = select.fetchList().count() ;
		m_types	  = new KBType *[m_nFields] ;

		for (uint col = 0 ; col < m_nFields ; col += 1)
		{
			m_colnames.append  (select.fetchList()[col].expr()) ;
                        m_types[col] = new KBSQLite3Type (0, 0, 0, true) ;
		}

		sqlite3_finalize(stmt) ;
		return	true	;
	}

	/* Nasty part 2: SQLite3 does not actually return column type	*/
	/* information, only column-row types, and these may come back	*/
	/* as null. So initially assume all are strings, and track any	*/
	/* not null types. Our column types are updated later.		*/
	if ((rc3 == SQLITE_ROW) && (m_types == 0))
	{
		m_nFields = sqlite3_column_count(stmt) ;
		m_types	  = new KBType *[m_nFields] ;
		memset (m_types, 0, sizeof(KBType *) * m_nFields) ;

		for (uint col = 0 ; col < m_nFields ; col += 1)
		{
			m_colnames.append (sqlite3_column_name(stmt, col)) ;
			m_coltypes.append (sqlite3_column_type(stmt, col)) ;
			m_types[col] = new KBSQLite3Type(0, 0, 0, true) ;
		}

		DPRINTF
		((	"KBSQLite3QrySelect::execute: setup types m_nFields=%d\n",
			m_nFields
		))	;
	}

	while (rc3 == SQLITE_ROW)
	{
		for (uint col = 0 ; col < m_nFields ; col += 1)
		{
			int	stype	= sqlite3_column_type(stmt, col) ;
			KBValue	value	;

			switch (stype)
			{
				case SQLITE_INTEGER :
					value	= KBValue
						  (	sqlite3_column_int   (stmt, col),
						  	m_types[col]
						  )	;
					break	;
				case SQLITE_FLOAT   :
					value	= KBValue
						  (	sqlite3_column_double(stmt, col),
						  	m_types[col]
						  )	;
					break	;
				case SQLITE_BLOB    :
					value	= KBValue
						  (	(cchar *)
							sqlite3_column_blob  (stmt, col),
							sqlite3_column_bytes (stmt, col),
							m_types[col]
						  )	;
					break	;
				case SQLITE_TEXT    :
					value	= KBValue
						  (	(cchar *)
							sqlite3_column_text  (stmt, col),
						  	m_types[col]
						  )	;
					break	;
				default	:
					break	;
			}
			putInCache (m_nRows, col, value) ;

			if (stype != SQLITE_NULL) m_coltypes[col] = stype ;
		}

		m_nRows += 1 ;
		rc3 = sqlite3_step (stmt) ;
	}

	/* If any row data arrived the retype our types with whatever	*/
	/* non-null types we spotted.					*/
	if (m_nRows > 0)
		for (uint col = 0 ; col < m_nFields ; col += 1)
			((KBSQLite3Type *)m_types[col])->retype (m_coltypes[col]) ;

	switch (rc3)
	{
		case SQLITE_DONE   :
			rv	= true	;
			break	;
//			m_server->printQuery (m_rawQuery, m_tag, nvals, values, true ) ;
//			sqlite3_finalize     (stmt) ;
//			return	true	;

		case SQLITE_ERROR  :
		case SQLITE_MISUSE :
			m_lError = KBError
				   (	KBError::Error,
					TR("Error executing SQL"),
					sqlite3_errmsg(m_server->handle()),
					__ERRLOCN
				   )	;
			break	;

		case SQLITE_BUSY :
			m_lError = KBError
				   (	KBError::Error,
					TR("Error executing SQL"),
					TR("Database is busy"),
					__ERRLOCN
				   )	;
			break	;

		default	:
			m_lError = KBError
				   (	KBError::Error,
					TR("Error executing SQL"),
					TR("Unknown SQLite3 return code"),
					__ERRLOCN
				   )	;
			break	;
	}

	DPRINTF
	((	"KBSQLite3QrySelect::execute: done: rv=%d m_nRows=%d m_nFields=%d\n",
		rv,
		m_nRows,
		m_nFields
	))	;

	m_server->printQuery (m_rawQuery, m_tag, nvals, values, rv) ;
	sqlite3_finalize     (stmt) ;
	return	rv ;
}

/*  KBSQLite3QrySelect							*/
/*  getField	: Get a specified field from the query results		*/
/*  qrow	: uint		  : Row number				*/
/*  qcol	: uint		  : Column number			*/
/*  vtrans	: KBValue::VTrans : Translation mode			*/
/*  (returns)	: KBValue	  : Value				*/

KBValue	KBSQLite3QrySelect::getField
	(	uint		qrow,
		uint		qcol,
		KBValue::VTrans
	)
{
	KBValue	value	;
	getFromCache	(qrow, qcol, value) ;
	return	value	;
}

/*  KBSQLite3QrySelect							*/
/*  getFieldname: Get a specified field name from the query results	*/
/*  qcol	: uint		: Column number				*/
/*  (returns)	: QString	: Name					*/

QString	KBSQLite3QrySelect::getFieldName
	(	uint		qcol
	)
{
	/* First check that the request is within the range of rows	*/
	/* and fields returned by the query.				*/
	if (qcol >= m_nFields) 
		return	QString::null	;

	return	m_colnames[qcol] ;
}


/*  KBSQLite3QryUpdate							*/
/*  KBSQLite3QryUpdate							*/
/*		: Constructor for update query object			*/
/*  server	: KBSQLite3 *		: Server connection object	*/
/*  data	: bool			: Query for data		*/
/*  query	: const QString &	: Update query			*/
/*  tabName	: const QString &	: Table being updated		*/
/*  (returns)	: KBSQLite3QryUpdate	:				*/

KBSQLite3QryUpdate::KBSQLite3QryUpdate
	(	KBSQLite3	*server,
		bool		data, 
		const QString	&query,
		const QString	&tabName
	)
	:
	KBSQLUpdate	(server, data, query, tabName),
	m_server	(server)
{
	m_nRows	= 0	;
}

/*  KBSQLite3QryUpdate							*/
/*  execute	: Execute query						*/
/*  nvals	: uint		: Number of substitution values		*/
/*  values	: KBValue *	: Substitution values			*/
/*  (returns)	: bool		: Success				*/

bool	KBSQLite3QryUpdate::execute
	(	uint		nvals,
		const KBValue	*values
	)
{
	DPRINTF
	((	"KBSQLite3QryUpdate::execute: %s\n",
		m_rawQuery.latin1()
	))	;

	if (!m_server->execSQL
		(	m_rawQuery,
			m_tag,
			m_subQuery,
			nvals,
			values,
			m_codec,
			"Update query failed",
			m_lError
		)) 
	{
		return	false	;
	}
	
	m_nRows	= sqlite3_changes (m_server->handle()) ;
	return	true	;
}

/*  KBSQLite3QryUpdate							*/
/*  ~KBSQLite3QryUpdate							*/
/*		: Destructor for update query object			*/
/*  (returns)	:		:					*/

KBSQLite3QryUpdate::~KBSQLite3QryUpdate ()
{
}


/*  KBSQLite3QryInsert							*/
/*  KBSQLite3QryInsert							*/
/*		: Constructor for insert query object			*/
/*  server	: KBSQLite3 *		: Server connection object	*/
/*  data	: bool			: Query for data		*/
/*  tabName	: const QString &	: Table being updated		*/
/*  query	: const QString &	: Insert query			*/
/*  (returns)	: KBSQLite3QryInsert	:				*/

KBSQLite3QryInsert::KBSQLite3QryInsert
	(	KBSQLite3	*server,
		bool		data,
		const QString	&query,
		const QString	&tabName
	)
	:
	KBSQLInsert	(server, data, query, tabName),
	m_server	(server)
{
	m_nRows	= 0	;
	m_rowid	= 0	;
}

/*  KBSQLite3QryInsert							*/
/*  ~KBSQLite3QryInsert							*/
/*		: Destructor for insert query object			*/
/*  (returns)	:		:					*/

KBSQLite3QryInsert::~KBSQLite3QryInsert ()
{
}

/*  KBSQLite3QryInsert							*/
/*  execute	: Execute query						*/
/*  nvals	: uint		: Number of substitution values		*/
/*  values	: KBValue *	: Substitution values			*/
/*  (returns)	: bool		: Success				*/

bool	KBSQLite3QryInsert::execute
	(	uint		nvals,
		const KBValue	*values
	)
{
	DPRINTF
	((	"KBSQLite3QryInsert::execute: %s\n",
		m_rawQuery.latin1()
	))	;

	if (!m_server->execSQL
		(	m_rawQuery,
			m_tag,
			m_subQuery,
			nvals,
			values,
			m_codec,
			"Insert query failed",
			m_lError
		)) 
	{
		return	false	;
	}
	
	m_nRows = sqlite3_changes	   (m_server->handle()) ;
	m_rowid = sqlite3_last_insert_rowid(m_server->handle()) ;
	return	true	;
}

/*  KBSQLite3QryInsert							*/
/*  getNewKey	: Get new insert key					*/
/*  primary	: const QString &	: Key column name		*/
/*  _newKey	: KBValue &		: New key			*/
/*  prior	: bool			: Pre-insert call		*/
/*  (returns)	: bool			: Success			*/

bool	KBSQLite3QryInsert::getNewKey
	(	const QString	&keycol, 
		KBValue		&newKey,
		bool		prior
	)
{
	/* The first time we execute this insert we need to get the	*/
	/* auto-increment column name, if any. We do this here rather	*/
	/* than on creation since it is more difficult to report errors	*/
	/* at query creation time.					*/
	if (m_autocol.isNull())
	{
		KBTableSpec tabSpec (m_tabName) ;
		if (!m_server->listFields (tabSpec))
		{	m_lError = m_server->lastError() ;
			return	 false ;
		}

		/* Set the column name to empty. This allows us to tell	*/
		/* later whether there really is no auto-increment	*/
		/* column, or whether there was an error above.		*/
		m_autocol = "" ;

		LITER
		(	KBFieldSpec,
			tabSpec.m_fldList,
			fSpec,

			if ((fSpec->m_flags & KBFieldSpec::Serial) != 0)
			{	m_autocol = fSpec->m_name ;
				break	;
			}
		)
	}

	if (prior) 
	{
		newKey	= KBValue() ;
		return	true	    ;
	}
 
	if ((keycol == m_autocol) || (keycol == "__autocol__"))
	{
		newKey	= m_rowid   ;
		return	true	    ;
	}

	m_lError = KBError
		   (	KBError::Error,
			"Asking for insert key",
			QString ("%1, %2:%3").arg(m_tabName).arg(keycol).arg(m_autocol),
			__ERRLOCN
		   )	;
	return	false	;
}

/*  KBSQLite3QryDelete							*/
/*  KBSQLite3QryDelete							*/
/*		: Constructor for delete query object			*/
/*  server	: KBSQLite3 *		: Server connection object	*/
/*  data	: bool			: Query for data		*/
/*  query	: const QString &	: Delete query			*/
/*  tabName	: const QString &	: Table being updated		*/
/*  (returns)	: KBSQLite3QryDelete	:				*/

KBSQLite3QryDelete::KBSQLite3QryDelete
	(	KBSQLite3	*server,
		bool		data,
		const QString	&query,
		const QString	&tabName
	)
	:
	KBSQLDelete	(server, data, query, tabName),
	m_server	(server)
{
	m_nRows	= 0;
	m_whereHackNeeded = false;
	
	if ( ! query.contains( "where", false ) )
		m_whereHackNeeded = true;
}

/*  KBSQLite3QryDelete							*/
/*  ~KBSQLite3QryDelete							*/
/*		: Destructor for delete query object			*/
/*  (returns)	:		:					*/

KBSQLite3QryDelete::~KBSQLite3QryDelete()
{
}

/*  KBSQLite3QryDelete							*/
/*  execute	: Execute query						*/
/*  nvals	: uint		: Number of substitution values		*/
/*  values	: KBValue *	: Substitution values			*/
/*  (returns)	: bool		: Success				*/

bool	KBSQLite3QryDelete::execute
	(	uint		nvals,
		const KBValue	*values
	)
{
	DPRINTF
	((	"KBSQLite3QryDelete::execute: %s\n",
		m_rawQuery.latin1()
	))	;
	
	if (!m_server->execSQL
		(	m_rawQuery,
			m_tag,
			m_subQuery,
			nvals,
			values,
			m_codec,
			"Delete Query Failed",
			m_lError
		)) 
	{
		return	false	;
	}
	
	m_nRows	= sqlite3_changes (m_server->handle()) ;
	return	true ;
}


KBFACTORY
(	KBSQLite3Factory,
	"driver_sqlite3"
)

KBFACTORYIMPL
(	KBSQLite3Factory,
	driver_sqlite3
)

QObject	*KBSQLite3Factory::create
	(	QObject			*parent,
		cchar			*object,
		cchar			*,
		const	QStringList	&
	)
{
	if (dIdentToType.count() == 0)
		for (uint idx = 0 ; idx < sizeof(typeMap)/sizeof(SQLite3TypeMap) ; idx += 1)
			if (typeMap[idx].ident != FIELD_TYPE_UNSET)
			{
				SQLite3TypeMap *m = &typeMap[idx] ;
				dIdentToType.insert (m->ident, m) ;
			}

	if ((parent != 0) && !parent->inherits("QWidget"))
	{
		fprintf
		(	stderr,
			"KBSQLite3Factory: parent does not inherit QWidget\n"
		)	;
		return	0 ;
	}

	if (qstrcmp (object, "driver") == 0)
		return	new KBSQLite3() ;

	return	0 ;
}

QString	KBSQLite3Factory::ident ()
{
	return	QString("SQLite3: %1%2 %3")
			.arg(__KB_BUILD_RELEASE)
			.arg(__KB_BUILD_BUILD  )
			.arg(__KB_BUILD_TIME   )
			;
}
