/***************************************************************************
    file	         : kb_viewer.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/

#include	<stdio.h>
#include	<stdarg.h>
#include	<errno.h>
#include	<time.h>

#include	<qfile.h>
#include	<qtextstream.h>

#include	<qptrlist.h>
#include	<qdict.h>
#include	<qevent.h>
#include	<qwidget.h>
#include	<qwidgetlist.h>


#include	"kb_classes.h"
#include	"kb_type.h"
#include	"kb_value.h"
#include	"kb_database.h"
#include	"kb_location.h"
#include	"kb_attr.h"
#include	"kb_dblink.h"
#include	"kb_dbinfo.h"
#include	"kb_node.h"
#include	"kb_display.h"
#include	"kb_prompt.h"
#include	"kb_gui.h"
#include	"kb_notifier.h"
#include	"kb_formcopier.h"

#include	"kb_partwidget.h"
#include	"kb_part.h"

#include	"kb_pluginaction.h"
#include	"kb_libloader.h"
#include	"kb_desktop.h"
#include	"kb_locator.h"

#if		__KB_AUTOMOC
#include	"kb_viewer.moc"
#else
#include 	"kb_viewer.h"
#endif

/*  KBViewer								*/
/*  KBViewer	: Constructor for viewer base class			*/
/*  parent	: KBObjBase *	: Parent object				*/
/*  embed	: QWidget *	: Embedding window			*/
/*  flags	: int		: Initial widget flags			*/
/*  modal	: bool		: Runs as modal				*/
/*  (returns)	: KBViewer	:					*/

KBViewer::KBViewer
	(	KBObjBase	*parent,
		QWidget		*embed,
		int		flags,
		bool		modal
	)
	:
	KBasePart (parent, embed, flags, modal),
	m_objBase (parent)
{
}

/*  KBViewer								*/
/*  ~KBViewer	: Destructor for viewer base class			*/
/*  (returns)	:		:					*/

KBViewer::~KBViewer ()
{
}

/*  KBViewer								*/
/*  showAs	: Show in specified mode				*/
/*  _showAs	: int		: Mode					*/
/*  (returns)	: void		:					*/

void	KBViewer::showAs
	(	int	_showAs
	)
{
	showAs	((KB::ShowAs)_showAs) ;
}

/*  KBViewer								*/
/*  setCaption	: Set caption						*/
/*  caption	: const QString & : Caption				*/
/*  (returns)	: void		  :					*/

void	KBViewer::setCaption
	(	const QString	&caption
	)
{
	if (m_partWidget)
	{
		if (caption.isEmpty())
			m_partWidget->setCaption (getLocation().title()) ;
		else	m_partWidget->setCaption (caption) ;
	}
}


/*  ------------------------------------------------------------------	*/

void	KBViewer::getPluginActionInfo
	(	KBPluginActionInfo	&info
	)
{
	info.m_location = getLocation () ;
}

/*  KBViewer								*/
/*  getPluginActionFactory						*/
/*		: On-demand plugin loader				*/
/*  name	: const QString & : Action plugin name			*/
/*  (returns)	: KBFactory *	  : Action plugin factory		*/

KBFactory
	*KBViewer::getPluginActionFactory
	(	const QString	&name
	)
{
	KBFactory	*factory ;

	/* Action plugins factories are stored in a dictionary as they	*/
	/* are loaded so create the dictionary first time and then see	*/
	/* if the plugin has already been loaded.			*/
	static	QDict<KBFactory> *actionPlugins ;

	if (actionPlugins == 0)
		actionPlugins	= new QDict<KBFactory> ;

	if ((factory = actionPlugins->find (name)) != 0)
		return	factory	;


	/* Not loaded so locate and load it. Start by accessing the	*/
	/* corresponding desktop file and make a sanity check that this	*/
	/* does actually describe a plugin.				*/
	QString	  file = locateFile ("appdata", "services/rekall_plugin_" + name + ".desktop") ;
	if (file.isEmpty())
	{
		fprintf
		(	stderr,
			"KBViewer::getPluginActionFactory: %s: no desktop file\n",
			name.latin1()
		)	;
		return	0	;
	}

	KBDesktop desktop (file) ;
	if (desktop.property ("ServiceTypes") != "Rekall/PluginAction")
	{
		fprintf
		(	stderr,
			"KBViewer::getPluginActionFactory: %s: not Rekall/PluginAction\n",
			name.latin1()
		)	;
		return	0	;
	}


	/* Desktop file OK to load the specified library and then get	*/
	/* the library factory. Once we have the factory we can create	*/
	/* the plugin object itself.					*/
	KBLibLoader *loader = KBLibLoader::self () ;
	QString	    libname = desktop.property("X-KDE-Library") ;
	KBLibrary   *lib    = loader->getLibrary(libname) ;
	if (lib == 0)
	{
		fprintf
		(	stderr,
			"KBViewer::getPluginActionFactory: %s: no library: %s\n",
			name   .latin1(),
			libname.latin1()
		)	;
		return	0	;
	}

	if ((factory = (KBFactory *)lib->factory()) == 0)
	{	
		fprintf
		(	stderr,
			"KBViewer::getPluginActionFactory: %s: no factory\n",
			name.latin1()
		)	;
		return	0	;
	}

	/* Success, plugin loaded. Store in the action plugin factory	*/
	/* in the dictionary so we do not need to load it when it is	*/
	/* subsequently accessed.					*/
	fprintf
	(	stderr,
		"KBViewer::getPluginActionFactory: done\n"
	)	;

	actionPlugins->insert (name, factory) ;
	return	factory	;
}

/*  KBViewer								*/
/*  slotPluginAction							*/
/*		: Invode plugin action					*/
/*  name	: const QString &	: Plugin name			*/
/*  found	: bool &		: Returns whether plugin found	*/
/*  (returns)	: void			:				*/

void	KBViewer::slotPluginAction
	(	const QString	&name,
		bool		&found
	)
{
	KBFactory *factory = getPluginActionFactory (name) ;
	if (factory == 0)
	{
		found	= false	;
		return	;
	}

	KBPluginActionBase *plugin = (KBPluginActionBase *)factory->create (this, 0, name.latin1()) ;
	if (plugin == 0)
	{
		fprintf
		(	stderr,
			"KBPluginAction::slotActivated: %s: create failed\n",
			name.latin1()
		)	;
		found	= false	;
		return	;
	}

	KBPluginActionInfo   info	;
	getPluginActionInfo (info)	;
	plugin->execute	    (info)	;
	delete	plugin	;

	found	= true	;
}
