/***************************************************************************
    file	         : kb_basequery.h
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/



#ifndef	__KB_BASEQUERY_H
#define	__KB_BASEQUERY_H

#if		defined(_WIN32) && !defined(__MINGW32__)
#pragma		warning(push)
#pragma 	warning(disable : 4251)
#endif

#include	<qstring.h>
#include	<qstringlist.h>
#include	<qvaluelist.h>

#include	"kb_classes.h"
#include 	"libcommon_exports.h"


class	KBDBLink	;
class	QDomElement	;
class	KBServer	;

/*  KBBaseQueryValue							*/
/*  ----------------							*/
/*  Class used to store value for a field to be inserted or updated.	*/
/*  The "m_type" member is the type of the value, as below:		*/
/*		X 	: Used only in copy constructor			*/
/*		V	: Value at execution, uses placeholder		*/
/*		S	: Text string					*/
/*		D	: Integer					*/
/*		F	: Float/double					*/
/*		P	: Value is explicit null			*/
/*  These are based on the "printf" format characters.			*/

class	LIBCOMMON_API	KBBaseQueryValue
{
	QString		m_field		;
	int		m_type		;
	QString		m_text		;
	int		m_fixed		;
	double		m_float		;

public	:

	KBBaseQueryValue	()	;
	KBBaseQueryValue	(const QString  &)	;
	KBBaseQueryValue	(const QString  &, const QString &)	;
	KBBaseQueryValue	(const QString  &, int   )		;
	KBBaseQueryValue	(const QString  &, double)		;
	KBBaseQueryValue	(const QString  &, void *)		;
	KBBaseQueryValue	(const QDomElement &)	;

	uint	addToUpdate	(KBServer *, uint, QStringList &) 		 const ;
	uint	addToInsert	(KBServer *, uint, QStringList &, QStringList &) const ;
}	;


/*  KBBaseQueryExpr							*/
/*  ---------------							*/
/*  Class used to store query expressions for select, update and delete	*/
/*  queries (ie., for where, group by, etc). The "m_type" is the type	*/
/*  as above, with the addition of 'A' for "asis", in which case the	*/
/*  expression is self-contained.					*/

class	LIBCOMMON_API	KBBaseQueryExpr
{
	QString		m_field		;
	int		m_type		;
	QString		m_text		;
	int		m_fixed		;
	double		m_float		;
	QString		m_oper		;

public	:

	KBBaseQueryExpr		()	;
	KBBaseQueryExpr		(const QString  &, 		    const char * = 0) ;
	KBBaseQueryExpr		(const QString  &, const QString &, const char * = 0) ;
	KBBaseQueryExpr		(const QString  &, int,		    const char * = 0) ;
	KBBaseQueryExpr		(const QString  &, double,	    const char * = 0) ;
	KBBaseQueryExpr		(const QString  &, void *,	    const char * = 0) ;
	KBBaseQueryExpr		(const QDomElement &)	;

	uint	addToQuery	(KBServer *, uint, QStringList &) const ;

	QString	expr  		(KBServer *, uint &) const ;

	static	const char	*m_asis	 ;	
}	;


/*  KBBaseQueryFetch							*/
/*  ----------------							*/
/*  Contains an expression to be retrieved, comprising the expression	*/
/*  itself plus a possible column alias.				*/

class	LIBCOMMON_API	KBBaseQueryFetch
{
	QString		m_expr		;
	QString		m_alias		;

public	:

	KBBaseQueryFetch()	;

	KBBaseQueryFetch
	(	const QString	&,
		const QString	& = QString::null
	)	;
	KBBaseQueryFetch	(const QDomElement &)	;

	QString		exprText	(KBServer *)	const ;
	const QString	&expr  		()		const ;
	const QString	&alias 		()		const ;
}	;


/*  KBBaseQueryTable							*/
/*  -------------							*/
/*  Information about a table; its name, a possible alias, and a	*/
/*  possible join condition to another table in the query.		*/
 
class	LIBCOMMON_API	KBBaseQueryTable
{
public	:

	enum	JoinType
	{
		None,
		Inner,
		LeftOuter,
		RightOuter
	}	;

private	:

	QString		m_tabName	;
	QString		m_alias		;
	JoinType	m_joinType	;
	QString		m_joinExpr	;
	QString		m_primary	;

public	:

	KBBaseQueryTable	()	;

	KBBaseQueryTable
	(	const QString	&,
		const QString	&,
		const QString	& = QString::null
	)	;

	KBBaseQueryTable
	(	const QString	&,
		const QString	&,
		JoinType,
		const QString	&,
		const QString	& = QString::null
	)	;

	KBBaseQueryTable
	(	const QString	&,
		const QString	&,
		const QString	&,
		const QString	&,
		const QString	& = QString::null
	)	;

	KBBaseQueryTable	(const QDomElement &)	;

	bool	hasOuterJoin 	()		const	;
	bool	hasInnerJoin 	()		const	;
	bool	hasAnyJoin 	()		const	;

	QString	joinType	()		const	;
	QString	joinExpr	(KBServer *) 	const	;
	QString	tableText	(KBServer *) 	const	;

	const QString		&tableName() 	const	;
}	;


/*  KBBaseQuery								*/
/*  -----------								*/
/*  Base class for building insert/update queries. Contains the table	*/
/*  name and the update/insert values.					*/

class	LIBCOMMON_API	KBBaseQuery
{
protected	:

	/* Store values and where terms in the base class. Not ideal	*/
	/* but the combination of values, where and group by, etc.,	*/
	/* over the four query types makes it tricky to do better.	*/
	/* These handle insert, update and delete; the derived select	*/
	/* class stores group by, etc.					*/
	QValueList<KBBaseQueryTable>	m_tableList	;
	QValueList<KBBaseQueryValue>	m_valueList	;
	QValueList<KBBaseQueryExpr >	m_whereList	;

	uint				m_buffPtr	;
	QString				m_buffer	;
	QString				m_token		;
	QString				m_white		;
	KBError				m_error		;

	void		setParseError	(const QString &) 	;
	bool		isKeyword	()			;
	bool		nextToken	()			;
	QString		parseExpr	(bool, bool) 		;
	void		parseExprList	(QValueList<KBBaseQueryExpr>  &, cchar *, bool) ;
	virtual	void	reset		()			;

	virtual QString	makeQueryText	(KBServer *) const = 0	;

public	:

	KBBaseQuery			(const QString & = QString::null) ;
	KBBaseQuery			(const QDomElement &)	;
	virtual	~KBBaseQuery		()	;

	void		setTable	(const QString &)	;

	void		addValue	(const QString &field)	;
	void		addValue	(const QString &field, const QString &)	;
	void		addValue	(const QString &field, int   )		;
	void		addValue	(const QString &field, double)		;
	void		addValue	(const QString &field, void *)		;
	void		addValue	(const QDomElement &)	;

	void		addWhere	(const QString &field,			const char * = 0) ;
	void		addWhere	(const QString &field, const QString &, const char * = 0) ;
	void		addWhere	(const QString &field, int,		const char * = 0) ;
	void		addWhere	(const QString &field, double,		const char * = 0) ;
	void		addWhere	(const QString &field, void *,		const char * = 0) ;
	void		addWhere	(const QDomElement &)	;

	QString		tableName	() const ;

	QString		getQueryText	(KBDBLink *) const	;
	QString		getQueryText	(KBServer *) const	;

	const QValueList<KBBaseQueryExpr> &whereList () const	;

	const KBError	&lastError	() const ;
}	;


/*  KBBaseUpdate							*/
/*  ------------							*/
/*  Derived class for constructing update queries. Adds support for an	*/
/*  arbitrary number of "where" conditions.				*/

class	LIBCOMMON_API	KBBaseUpdate	: public KBBaseQuery
{
protected :

	virtual	QString	makeQueryText	(KBServer *)	const	;

public	:

	KBBaseUpdate			(const QString & = QString::null)	;
	KBBaseUpdate			(const QDomElement &)			;
}	;

/*  KBBaseInsert							*/
/*  ------------							*/
/*  Derived class for constructing insert queries.			*/

class	LIBCOMMON_API	KBBaseInsert	: public KBBaseQuery
{
protected :

	virtual	QString	makeQueryText	(KBServer *)	const	;

public	:

	KBBaseInsert			(const QString & = QString::null)	;
	KBBaseInsert			(const QDomElement &)			;
}	;

/*  KBBaseSelect							*/
/*  ------------							*/
/*  Derived class for constructing select queries.			*/

class	LIBCOMMON_API	KBBaseSelect	: public KBBaseQuery
{
	QValueList<KBBaseQueryFetch>	m_fetchList	;
	QValueList<KBBaseQueryExpr >	m_groupList	;
	QValueList<KBBaseQueryExpr >	m_havingList	;
	QValueList<KBBaseQueryExpr >	m_orderList	;
	bool				m_distinct	;
	bool				m_forUpdate	;

	int				m_offset	;
	int				m_limit		;

protected :

	void		parseFetchList	(QValueList<KBBaseQueryFetch>  &, cchar *, bool) ;
	bool		parseTableList	(KBDBLink *) 		;
	virtual	void	reset		()			;

	virtual	QString	makeQueryText	(KBServer *)	const	;

public	:

	KBBaseSelect		(const QString & = QString::null)	;
	KBBaseSelect		(const QDomElement &)			;

	void	addTable
	(	const QString		&,
		const QString		&
	)	;

	void	addTable
	(	const QString		&,
		const QString		&,
		KBBaseQueryTable::JoinType,
		const QString		&
	)	;

	void	addTable
	(	const QString		&,
		const QString		&,
		const QString		&,
		const QString		&
	)	;

	void	addTable
	(	const QDomElement	&
	)	;

	void		addFetch		(const QString &, const QString & = QString::null) ;
	void		addGroup		(const QString &) ;
	void		addHaving		(const QString &) ;
	void		addOrder		(const QString &) ;

	void		addFetch		(const QDomElement &) ;
	void		addGroup		(const QDomElement &) ;
	void		addHaving		(const QDomElement &) ;
	void		addOrder		(const QDomElement &) ;

	void		setDistinct		(bool)	;
	void		setForUpdate		(bool)	;
	void		setLimit		(int, int) ;

	bool		forUpdate		() const ;
	bool		distinct		() const ;
	int		limit			() const ;
	int		offset			() const ;

	bool		parse			(const QString &, KBDBLink *) ;

	const QValueList<KBBaseQueryTable> &tableList  () const ;
	const QValueList<KBBaseQueryFetch> &fetchList  () const ;
	const QValueList<KBBaseQueryExpr > &groupList  () const ;
	const QValueList<KBBaseQueryExpr > &havingList () const ;
	const QValueList<KBBaseQueryExpr > &orderList  () const ;
}	;

/*  KBBaseDelete							*/
/*  ------------							*/
/*  Derived class for constructing delete queries.			*/

class	LIBCOMMON_API	KBBaseDelete	: public KBBaseQuery
{
protected :

	virtual	QString	makeQueryText	(KBServer *)	const	;

public	:

	KBBaseDelete			(const QString & = QString::null)	;
	KBBaseDelete			(const QDomElement &)			;
}	;

#if		defined(_WIN32) && !defined(__MINGW32__)
#pragma		warning(pop)
#endif

#endif	// __KB_BASEQUERY_H
