/***************************************************************************
    file	         : kb_error.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/

#include	<stdio.h>
#include	<stdlib.h>
#include	<qglobal.h>

#ifdef		Q_OS_LINUX
#include	<unistd.h>
#endif

#ifdef		Q_OS_MACX
#include	<unistd.h>
#endif

#include	"kb_classes.h"
#include	"kb_callback.h"
#include	"kb_appptr.h"



uint 	LIBCOMMON_API		KBError::errDebug	;
static	KBErrorBlockBase	*currentBlockBase	;


/*  KBErrorInfo								*/
/*  KBErrorInfo	: Constructor for error information class		*/
/*  (returns)	: KBErrorInfo	:					*/

KBErrorInfo::KBErrorInfo ()
	:
	m_etype		(KBError::None),
	m_file		(0),
	m_lno		(0)
{
	m_errno		= -1 ;
}

/*  KBErrorInfo								*/
/*  KBErrorInfo	: Constructor for error information class		*/
/*  etype	: int		  : Error type (really KBError::EType)	*/
/*  message	: const QString & : Error message			*/
/*  details	: const QString & : Additional details, may be empty	*/
/*  file	: cchar *	  : File in which error originated ...	*/
/*  ln		: uint		  : ... and line number			*/
/*  (returns)	: KBErrorInfo	  :					*/

KBErrorInfo::KBErrorInfo
	(	int		etype,
		const QString	&message,
		const QString	&details,
		cchar		*file,
		uint		lno
	)
	:
	m_etype		(etype),
	m_message	(message),
	m_details	(details),
	m_file		(file),
	m_lno		(lno)
{
	m_errno		= -1 ;
}


/*  ------------------------------------------------------------------	*/

/*  KBError								*/
/*  KBError	: Constructor for error object				*/
/*  (returns)	: KBError	:					*/

KBError::KBError ()
{
}

/*  KBError								*/
/*  KBError	: Constructor for error object				*/
/*  etype	: EType		  : Error type				*/
/*  message	: const QString & : Main error message			*/
/*  details	: const QString & : Extra error details			*/
/*  file	: cchar *	  : Originating file			*/
/*  lno		: uint		  : Oringinationg line number		*/
/*  (returns)	: KBError	  :					*/

KBError::KBError
	(	EType		etype,
		const QString	&message,
		const QString	&details,
		cchar		*file,
		uint		lno
	)
{
	m_errors.append	(KBErrorInfo (etype, message, details, file, lno)) ;
}

/*  KBError								*/
/*  KBError	: Constructor for error object				*/
/*  e		: const KBError & : Extant error			*/
/*  (returns)	: KBError	  :					*/

KBError::KBError
	(	const KBError	&e
	)
{
	m_errors = e.m_errors ;
}

/*  KBError								*/
/*  operator =	: Error assignment					*/
/*  e		: const KBError & : Error being assigned		*/
/*  (returns)	: void		:					*/

void	KBError::operator =
	(	const KBError	&e
	)
{
	m_errors = e.m_errors	;

	/* Display the errors to stderr if error tracing it on. Only	*/
	/* bother with the first (if any) since assignment with		*/
	/* multiple errors is unlikely.					*/
	if ((errDebug >= 2) && (m_errors.count() > 0))
	{
		KBErrorInfo ei = m_errors[0] ;

		fprintf
		(	kbDPrintfGetStream(),
			"KBError::operator=: %d [%s] [%s] %s:%d\n",
			ei.m_etype,
			ei.m_message.latin1(),
			ei.m_details.latin1(),
			ei.m_file,
			ei.m_lno
		)	;
	}
}

/*  KBError								*/
/*  append	: Append errors						*/
/*  e		: const KBError & : Extant errors			*/
/*  (returns)	: void		  :					*/

void	KBError::append
	(	const KBError	&e
	)
{
	/* Called when accruing errors in the error blocker, append to	*/
	/* errors we have accred so far.				*/
	for (uint idx = 0 ; idx < e.m_errors.count() ; idx += 1)
		m_errors.append(e.m_errors[idx]) ;
}

/*  KBError								*/
/*  display	: Display error						*/
/*  caption	: const QString & : Optional caption			*/
/*  dfile	: cchar *	  : Display source file			*/
/*  dlno	: uint		  : Display line number			*/
/*  mode	: uint		  : Display mode			*/
/*  (returns)	: bool		  : Error displayed			*/

bool	KBError::display
	(	const QString	&caption,
		cchar		*dfile,
		uint		dlno
	)	const
{
	KBCallback *cb	= KBAppPtr::getCallback() ;

	if (cb != 0)
	{
		/* If there is a GUI callback then check for error	*/
		/* blocking. The blocker may accrue or supress the	*/
		/* error; it returns true if the error should be shown	*/
		/* now.							*/
		if (currentBlockBase != 0)
			if (!currentBlockBase->showError (this))
				return	false	;

		if (cb->displayError (*this, caption, dfile, dlno))
			return	true	;
	}

	/* No callback so errors go to stderr. This will be used when	*/
	/* running tests without a real GUI.				*/
	display	(stderr) ;
	return	false	 ;
}

/*  KBError								*/
/*  setErrno	: Set system error code					*/
/*  _errno	: int		: Error code				*/
/*  (returns)	: void		:					*/

void	KBError::setErrno
	(	int		_errno
	)
{
	if (m_errors.count() > 0)
		m_errors[0].m_errno = _errno ;
}

/*  KBError								*/
/*  display	: Display error						*/
/*  fd		: FILE *	: Error output stream			*/
/*  dfile	: cchar *	: Display source file			*/
/*  dlno	: uint		: Display line number			*/
/*  (returns)	: void		:					*/

void	KBError::display
	(	FILE		*fd
	)	const
{
	for (uint idx = 0 ; idx < m_errors.count() ; idx += 1)
	{
		const KBErrorInfo &ei = m_errors[idx] ;

		cchar	*type	;

		switch (ei.m_etype)
		{
			case None    : type = "No error"	; break ;
			case Info    : type = "Information"	; break ;
			case Warning : type = "Warning"		; break ;
			case Error   : type = "Error"		; break ;
			case Fault   : type = "Fault"		; break ;
			default	     : type = "Unknown"		; break ;
		}

		fprintf	(kbDPrintfGetStream(), "KBError: %s\n", type) ;
		fprintf	(kbDPrintfGetStream(), "       : %s\n", ei.m_message.latin1()) ;
		fprintf	(kbDPrintfGetStream(), "       : %s\n", ei.m_details.latin1()) ;
	}
}

void	KBError::setDetails
	(	const QString	&details
	)
{
	m_errors[0].m_details = details ;
}

KBError::EType
	KBError::getEType   	() const
{
	return	(KBError::EType)m_errors[0].m_etype   ;
}

const QString
	&KBError::getMessage	() const
{
	return	m_errors[0].m_message ;
}

const QString
	&KBError::getDetails	() const
{
	return	m_errors[0].m_details ;
}

cchar	*KBError::getFile   	() const
{
	return	m_errors[0].m_file    ;
}

uint	KBError::getLineno  	() const
{
	return	m_errors[0].m_lno     ;
}

int	KBError::getErrno   	() const
{
	return	m_errors[0].m_errno   ;
}


/*  ------------------------------------------------------------------	*/

int	KBError::EInfo
	(	const QString	&text,
		const QString	&details,
		cchar		*file,
		uint		lno
	)
{
	return	KBError	(KBError::Info, text, details, file, lno).DISPLAY() ;
}

int	KBError::EWarning
	(	const QString	&text,
		const QString	&details,
		cchar		*file,
		uint		lno
	)
{
	return	KBError	(KBError::Warning, text, details, file, lno).DISPLAY() ;
}

int	KBError::EError
	(	const QString	&text,
		const QString	&details,
		cchar		*file,
		uint		lno
	)
{
	return	KBError	(KBError::Error, text, details, file, lno).DISPLAY() ;
}

int	KBError::EFault
	(	const QString	&text,
		const QString	&details,
		cchar		*file,
		uint		lno
	)
{
	return	KBError	(KBError::Fault, text, details, file, lno).DISPLAY() ;
}

void	KBError::EFatal
	(	const QString	&text,
		const QString	&details,
		cchar		*file,
		uint		lno
	)
{
	KBError	(KBError::Fault, text, details, file, lno).DISPLAY() ;
	exit	(1) ;
}

/*  ------------------------------------------------------------------	*/


/*  KBErrorBlockBase							*/
/*  KBErrorBlockBase							*/
/*		: Constructor for error blocker				*/
/*  block	: EBlock	: Blocking mode				*/
/*  (returns)	: KBErrorBlock	:					*/

KBErrorBlockBase::KBErrorBlockBase
	(	EBlock		block
	)
{
	/* This instance is chained onto the front of the error block	*/
	/* queue and becomes current. We expect that instance of	*/
	/* the class will be on the stack, so the destructor is called	*/
	/* when the instance goes out of scope.				*/
	m_prior		= currentBlockBase	;
	m_accrue	= 0			;
	m_block		= block			;
	m_errorCount	= 0			;
	currentBlockBase= this			;
}

/*  KBErrorBlockBase							*/
/*  ~KBErrorBlockBase							*/
/*		: Destructor for error blocker				*/
/*  (returns)	:		:					*/

KBErrorBlockBase::~KBErrorBlockBase ()
{
	display	()	;
	currentBlockBase = m_prior	;
}

void	KBErrorBlockBase::display ()
{
	/* If error accrual was selected and any error occurred then	*/
	/* "m_accrue" points at the errors. We switch blocking mode to	*/
	/* "All" else we end up looping.				*/
	if (m_accrue != 0)
	{	m_block	 = All		;
		m_accrue->DISPLAY()	;
		m_accrue = 0		;
	}
}

/*  KBErrorBlockBase							*/
/*  showError	: See whether error should be displayed			*/
/*  error	: const KBError & : Error in question			*/
/*  (returns)	: bool		  : True to show			*/

bool	KBErrorBlockBase::showError
	(	const KBError	*error
	)
{
	/* Always display queryies!					*/
	if (error->m_errors[0].m_etype == KBError::Query)
		return	true	;

	/* If showing (or accrueing) distinct errors the see if this	*/
	/* error has been shown already. This is based on the mesage	*/
	/* and the details.						*/
	bool	prior	= false	;

	if ((m_block == Distinct) || (m_block == AccrueDistinct))
	{	
		QString	ident	= QString("%1::%2")
					.arg(error->m_errors[0].m_message)
					.arg(error->m_errors[0].m_details)
						;
		if (m_occured.find(ident) == 0)
		{
			static	int	one	= 1	;
			m_occured.insert(ident, &one)	;
		}
		else	prior	= true	;
	}
	
	switch (m_block)
	{
		case All	:
			/* Trivial. Barely worthy of a comment.		*/
			m_errorCount += 1 ;
			return	true	;

		case None	:
			/* Trivial. Ditto. Visual balance counts though	*/
			m_errorCount += 1 ;
			return	false	;

		case Single	:
			/* Show a single (ie., the first) error only.	*/
			if (m_errorCount == 0)
			{	m_errorCount += 1 ;
				return	true	;
			}
			return	false	;

		case Distinct	:
			/* Show distinct errors only. The "prior" flag	*/
			/* will have been set above.			*/
			if (!prior)	
			{	m_errorCount += 1 ;
				return	true	;
			}

			return	false	;

		case AccrueDistinct	:
			/* Accrue distinct errors for later display.	*/
			/* Check for prior display, if not then drop	*/
			/* into the straight accrual case.		*/
			if (prior)
				return	false	;

		case Accrue	:
			/* First time for accrual, create a new error	*/
			/* object based on the error passed in, after	*/
			/* that the errors are appended.		*/
			if (m_accrue == 0)
			{
				m_accrue = new KBError (*error)	;
				m_errorCount += 1 ;
				return	 false	;
			}

			m_accrue->append (*error) ;
			return	false	;

		default	:
			break	;
	}

	/* Should not get here so treat it as "All" and display the	*/
	/* error.							*/
	return	true	;
}

uint	KBErrorBlockBase::errorCount ()
{
	return	m_errorCount	;
}
