/***************************************************************************
    file	         : kb_location.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/




#include	<errno.h>

#include	<qstring.h>
#include	<qfile.h>
#include	<qfileinfo.h>
#include	<qtextstream.h>
#include	<qtextcodec.h>
#include	<qcache.h>
#include	<qptrdict.h>

#if		__KB_KDE
#include	<kglobal.h>
#include	<kcharsets.h>
#endif

#include	"kb_classes.h"
#include	"kb_type.h"
#include	"kb_value.h"
#include	"kb_database.h"
#include	"kb_dbinfo.h"
#include	"kb_dblink.h"
#include	"kb_basequery.h"
#include	"kb_notifier.h"
#include	"kb_file.h"
#include	"kb_locator.h"

#include	"kb_location.h"
#include	"kb_location_p.h"


#define		__DEBUG_LOCATION	(0)
#if		__DEBUG_LOCATION
#include	"kb_dprintf.h"
#endif



class	LIBCOMMON_API	KBLocnTypeMap
{
public	:

	QString		type	;
	QString		extn	;
	QString		legend	;
	KBPartFactory	*factory;

	inline	KBLocnTypeMap
		(	cchar		*type,
			cchar		*extn,
			const QString	&legend,
			KBPartFactory	*factory
		)
		:
		type	(type),
		extn	(extn),
		legend	(legend),
		factory	(factory)
	{
	}
}	;


/*  ------------------------------------------------------------------	*/

/*  Column names and "usual" types. Must be orderer to match the RKT__*	*/
/*  define's in "kb_location_p.h". See the ::columnTypes method for	*/
/*  additional comments.						*/

static	struct
	{	const char	*m_colName	;
		KBType		*m_colType	;
	}
	rkColumnInfo[] =
{
	{	"Id",			&_kbFixed	},
	{	"Description",		&_kbBinary	},
	{	"Definition",		&_kbBinary	},
	{	"SaveDate",		&_kbString	},
	{	"Type",			&_kbString	},
	{	"Name",			&_kbString	},
	{	"Extension",		&_kbString	}
}	;

/*  ------------------------------------------------------------------	*/

cchar	*KBLocation::m_pFile	= "!Files"		;
cchar	*KBLocation::m_pInline	= "!Inline"		;
cchar	*KBLocation::m_pStock	= "!Stock"		;
cchar	*KBLocation::m_pLocal	= "!Local"		;

static 	QPtrList<KBLocnTypeMap>	locnTypeMap		;
static	QCache<QByteArray>	*contentsCache		;
static	KBLocation::CacheOption	cacheOption		= KBLocation::CacheNone ;
static	int			cacheSize		;
static	bool			cacheAll		;

/*  ------------------------------------------------------------------	*/

/*  findInCache	: Find object in cache					*/
/*  graphics	: bool		  : Object is a graphic			*/
/*  key		: const QString & : Entry key				*/
/*  data	: QByteArray &	  : Result array			*/
/*  (returns)	: bool		  : Entry found				*/

static	bool	findInCache
	(	bool		,
		const QString	&key,
		QByteArray	&data
	)
{
	/* Ignore the "graphics" flag: the object will either be in the	*/
	/* cache or it won't, and the key includes the object type.	*/
	if (contentsCache != 0)
	{
		QByteArray *cptr = contentsCache->find (key) ;
		if (cptr != 0)
		{
			DPRINTF
			((	"findInCache: %s: found in cache\n",
				key.latin1()
			))	;
			data	= *cptr	;
			return	true	;
		}
	}

	return	false	;
}

/*  saveToCache	: Save object to cache					*/
/*  graphics	: bool		  : Object is a graphic			*/
/*  key		: const QString & : Entry key				*/
/*  data	: QByteArray &	  : Data array				*/
/*  (returns)	: void		  : 					*/

static	void	saveToCache
	(	bool		graphic,
		const QString	&key,
		QByteArray	&data
	)
{
	/* Check cache options; the import here is to allow caching of	*/
	/* graphics (which probably don't change and are less critical)	*/
	/* but not to cache anything else (which another user may be	*/
	/* working on).							*/
	switch (cacheOption)
	{
		case KBLocation::CacheNone	:
			if (!cacheAll) return	;
			break	;

		case KBLocation::CacheAll	:
			break	;

		case KBLocation::CacheGraphics	:
			if (!graphic && !cacheAll) return ;
			break	;

		default	:
			break	;
	}
	
	if (contentsCache != 0)
	{
		QByteArray *cptr = new QByteArray() ;
		cptr->duplicate (data) ;
		contentsCache->insert  (key, cptr, data.size()) ;

		DPRINTF
		((	"saveToCache: %s: saved to cache: %d: cost now: %d\n",
			key.latin1(),
			data .size(),
			contentsCache->totalCost()
		))	;
	}
}

/*  removeFromCache							*/
/*		: Remove object from cache				*/
/*  key		: const QString & : Entry key				*/
/*  (returns)	: void		  : 					*/

static	void	removeFromCache
	(	const QString	&key
	)
{
	if (contentsCache != 0)
		if (contentsCache->remove (key))
		{
			DPRINTF
			((	"removeFromCache: %s: removed\n",
				key.latin1()
			))	;
		}
}

/*  KBLocation								*/
/*  setCacheSize: Set cache size					*/
/*  size	: uint		: Cache size in MB			*/
/*  option	: CacheOption	: Caching option			*/
/*  (returns)	: void		: 					*/

void	KBLocation::setCacheSize 
	(	uint			size,
		KBLocation::CacheOption	option
	)
{
	DPRINTF
	((	"KBLocation::setCacheSize: set to %dMB: opt=%d\n",
		size,
		option
	))	;

	DELOBJ	(contentsCache) ;
	cacheSize	= size	;
	cacheOption	= option;
	cacheAll	= false	;

	if (cacheSize > 0)
	{
		contentsCache = new QCache<QByteArray> (cacheSize * 1024 * 1024, 1009) ;
		contentsCache->setAutoDelete (true) ;
	}
}

void	KBLocation::setCacheAll
	(	bool		all
	)
{
	cacheAll = all ;

	if (cacheAll)
	{
		DELOBJ	(contentsCache) ;
		contentsCache = new QCache<QByteArray> (1024 * 1024 * 1024, 1009) ;
		contentsCache->setAutoDelete (true) ;
		return	;
	}

	setCacheSize (cacheSize, cacheOption) ;
}

/*  KBLocation								*/
/*  getCacheUsed: Get current cache usage rounded up to MB units	*/
/*  (returns)	:							*/

uint	KBLocation::getCacheUsed ()
{
	return	contentsCache == 0 ? 0 : (contentsCache->totalCost() + 1024 * 1024 - 1) / (1024 * 1024) ;
}


/*  ------------------------------------------------------------------	*/


/*  KBLocation								*/
/*  KBLocation	: Constructor for location object			*/
/*  dbInfo	: KBDBInfo *	  : Database information		*/
/*  type	: cchar *	  : Object type				*/
/*  server	: const QString & : Document location			*/
/*  name	: const QString & : Document name			*/
/*  extension	: const QString & : Document file extension		*/
/*  (returns)	: KBLocation	  :					*/

KBLocation::KBLocation
	(	KBDBInfo	*dbInfo,
		cchar		*type,
		const QString	&server,
		const QString	&name,
		const QString	&extension
	)
	:
	m_dbInfo	(dbInfo	  ),
	m_type		(type	  ),
	m_server	(server	  ),
	m_name		(name	  ),
	m_extension	(extension)
{
	if (name.left(2) == "//")
	{
		m_name	 = name.mid(2)	;
		m_server = m_pFile	;
		m_dbInfo = 0		;
	}

	DPRINTF
	((	"KBLocation::KBLocation: set  [%p] [%s][%s][%s][%s]\n",
		(void  *)this,
		(cchar *)m_type,
		(cchar *)m_server,
		(cchar *)m_name,
		(cchar *)m_extension
	))	;
}

/*  KBLocation								*/
/*  KBLocation	: Constructor for location object			*/
/*  other	: KBLocation &	: Extant location object		*/
/*  (returns)	: KBLocation	:					*/

KBLocation::KBLocation
	(	const KBLocation	&other
	)
	:
	m_dbInfo	(other.m_dbInfo	   ),
	m_type		(other.m_type	   ),
	m_server	(other.m_server	   ),
	m_name		(other.m_name	   ),
	m_extension	(other.m_extension ),
	m_dataServer	(other.m_dataServer)
{
	DPRINTF
	((	"KBLocation::KBLocation: copy [%p] [%s][%s][%s][%s]\n",
		(void  *)this,
		(cchar *)m_type,
		(cchar *)m_server,
		(cchar *)m_name,
		(cchar *)m_extension
	))	;
}

/*  KBLocation								*/
/*  KBLocation	: Constructor for location object			*/
/*  (returns)	: KBLocation	  :					*/

KBLocation::KBLocation ()
{
	m_dbInfo = 0 		;
	m_type	 = "unknown"	;
}

/*  KBLocation								*/
/*  findByType	: Find type map entry by type				*/
/*  type	: const QString & : Type				*/
/*  (returns)	: KBLocnTypeMap * : Entry or null if not found		*/

KBLocnTypeMap
	*KBLocation::findByType
	(	const QString	&type
	)
{
	LITER
	(	KBLocnTypeMap,
		locnTypeMap,
		tmp,

		if (tmp->type == type) return tmp ;
	)

	return	0 ;
}

/*  KBLocation								*/
/*  KBLocation								*/
/*  findByExtn	: Find type map entry by extension			*/
/*  extn	: const QString & : Extension				*/
/*  (returns)	: KBLocnTypeMap * : Entry or null if not found		*/

KBLocnTypeMap
	*KBLocation::findByExtn
	(	const QString	&extn
	)
{
	LITER
	(	KBLocnTypeMap,
		locnTypeMap,
		tmp,

		if (tmp->extn == extn) return tmp ;
	)

	return	0 ;
}

/*  KBLocation								*/
/*  operator ==	: Test equality						*/
/*  other	: KBLocation &	: Other location			*/
/*  (returns)	: bool		: Equality				*/

bool	KBLocation::operator ==
	(	const KBLocation	&other
	)
{
	return	(m_dbInfo    == other.m_dbInfo   ) &&
		(m_type      == other.m_type     ) &&
		(m_server    == other.m_server   ) &&
		(m_name	     == other.m_name	 ) &&
		(m_extension == other.m_extension) ;
}

/*  KBLocation								*/
/*  path	: Get path for file objects				*/
/*  altName	: const QString & : Alternate name			*/
/*  (returns)	: Qstring 	  : File path				*/

QString	KBLocation::path
	(	const QString	&altName
	)
	const
{
	/* Note that we return a path unless the location is inline.	*/
	/* If the location is actually in a database server, we still	*/
	/* return a notional path so that "saveDocumentAs" works as	*/
	/* expected.							*/
	if (isInline())
		return	QString::null ;

	if (isLocal ())
	{
		QString	res = altName.isEmpty() ? m_name : altName ;
		if (!m_extension.isEmpty()) res = res + "." + m_extension ;
		return	res ;
	}

	if (isStock ())
	{
		QString	where	=
				QString	("stock/%1/%2.%3")
					.arg(m_type)
					.arg(altName.isEmpty() ? m_name : altName)
					.arg(m_extension) ;
		DPRINTF
		((	"KBLocation::path: stock [%s]\n",
			where.latin1()
		))	;
		return	locateFile ("appdata", where) ;
	}

	QString	name = altName.isNull() ? m_name : altName ;
	name	+= "." + extnForType (m_dbInfo, m_type, m_extension) ;

	/* If we have a database then the path is resolved relative to	*/
	/* the database directory ....					*/
	if (m_dbInfo != 0)
	{
		DPRINTF
		((	"KBLocation::path: [%s]\n",
			(m_dbInfo->getDBPath() + "/"  + name ).latin1()
		))	;
		return	m_dbInfo->getDBPath() + "/"  + name  ;
	}

	/* ... otherwise look in the application global directory.	*/
	return	locateFile ("appdata", name) ;
}

/*  KBLocation								*/
/*  filename	: Get file name for object				*/
/*  altName	: const QString & : Alternate name			*/
/*  (returns)	: Qstring 	  : File path				*/

QString	KBLocation::filename
	(	const QString	&altName
	)
	const
{
	QString	name = altName.isNull() ? m_name : altName ;
	name	+= "." + extnForType (m_dbInfo, m_type, m_extension) ;
	return	name ;
}

/*  KBLocation								*/
/*  ident	: Get location identifier				*/
/*  (returns)	: QString	: Location identifier			*/

QString	KBLocation::ident () const
{
	QString	dbname	= m_dbInfo == 0 ? QString("") : m_dbInfo->getDBName() ;
	return	QString("%1:%2:%3").arg(dbname  )
				   .arg(m_server)
				   .arg(m_name  ) ;
}

/*  KBLocation								*/
/*  title	: Get title for window headers, etc			*/
/*  (returns)	: QString	: Title					*/

QString	KBLocation::title () const
{
	KBLocnTypeMap	*ltm	= findByType (m_type) ;
	QString		legend	;

	if ((ltm != 0) && !ltm->legend.isEmpty())
		legend	= QString("%1: ").arg(ltm->legend) ;

	if (m_server == m_pInline)
		return	QString("%1Inline:%3").arg(legend).arg(m_name) ;
	if (m_server == m_pFile  )
		return	QString("%1!Files:%3").arg(legend).arg(m_name) ;

	return	QString	("%1%2:%3").arg(legend).arg(m_server).arg(m_name) ;
}

/*  KBLocation								*/
/*  columnTypes	: Load column type information				*/
/*  dbLink	: KBDBLink &	  : Database link			*/
/*  pError	: KBError &	  : Error return			*/
/*  (returns)	: RKColumnTypes * : Type information or null on error	*/

const RKColumnTypes
		*KBLocation::columnTypes
		(	KBDBLink	&dbLink,
			KBError		&pError
		)
{
	/* NOTE:							*/
	/* We need to know column types when inserting or updating the	*/
	/* objects table. Previously this was done by a select, but it	*/
	/* feels better to use get a field (column) list. The results	*/
	/* are kept in a per-database dictionary so we only do it once.	*/

	static	QDict<RKColumnTypes>  *colTypes = 0 ;
	if (colTypes == 0)
		colTypes = new QDict<RKColumnTypes> ;


	/* Get the database tag from the database link; we assume that	*/
	/* this is already connected. This tag should be unique over	*/
	/* all databases that we can connect to.			*/
	QString	      dbTag	= dbLink.databaseTag () ;
	RKColumnTypes *rkt	= colTypes->find(dbTag) ;

	DPRINTF
	((	"KBLocation::columnTypes: looking up [%s]->[%p]\n",
		dbTag.latin1(),
		(void *)rkt
	))	;

	if (rkt != 0)
		return rkt ;


	/* Not known so get the table information for the objects table	*/
	/* Then extract the column types into a newly allocated type	*/
	/* object and save it in the dictionary.			*/
	KBTableSpec tabSpec (dbLink.rekallPrefix (OBJECTTABNAME)) ;
	if (!dbLink.listFields (tabSpec))
	{
		pError	= dbLink.lastError() ;
		return	0	;
	}

	rkt	= new RKColumnTypes ;
	for (uint idx = 0 ; idx < RKT_COUNT ; idx += 1)
	{
		KBFieldSpec *fSpec = tabSpec.findField(rkColumnInfo[idx].m_colName) ;
		if ((fSpec != 0) && (fSpec->m_dbType != 0))
		{
			fSpec->m_dbType->ref() ;
			rkt->m_types[idx] = fSpec->m_dbType ;
			continue ;
		}

		/* No field or no field type. Fall back on the "usual"	*/
		/* type.						*/
		rkt->m_types[idx] = rkColumnInfo[idx].m_colType ;

		fprintf
		(	kbDPrintfGetStream(),
			"KBLocation::columnTypes: %s: %s: NO TYPE: %p/%p\n",
			dbTag.latin1(),
			rkColumnInfo[idx].m_colName,
			(void *)fSpec,
			fSpec == 0 ? (void *)0 : (void *)fSpec->m_dbType
		)	;
	}

	/* Also find out whether the "Id" column is read-only. Needed	*/
	/* when building insert queries.				*/
	KBFieldSpec *fSpec = tabSpec.findField("Id") ;
	rkt->m_idRO 	   = fSpec != 0 ?
				(fSpec->m_flags & KBFieldSpec::ReadOnly) != 0 :
				false	;

	DPRINTF
	((	"KBLocation::columnTypes: got [%s] -> (%s/%s/%s/%s/%s/%s/%s)\n",
		dbTag.latin1(),
		rkt->m_types[RKT_ID	    ]->getDescrip().latin1(),
		rkt->m_types[RKT_DESCRIPTION]->getDescrip().latin1(),
		rkt->m_types[RKT_DEFINITION ]->getDescrip().latin1(),
		rkt->m_types[RKT_SAVEDATE   ]->getDescrip().latin1(),
		rkt->m_types[RKT_TYPE	    ]->getDescrip().latin1(),
		rkt->m_types[RKT_NAME	    ]->getDescrip().latin1(),
		rkt->m_types[RKT_EXTENSION  ]->getDescrip().latin1()
	))	;

	colTypes->insert (dbTag, rkt) ;
	return	rkt ;
}

/*  KBLocation								*/
/*  getData	: Construct and execute select for specified field	*/
/*  field	: cchar *	: Required field			*/
/*  pError	: KBError &	: Error return				*/
/*  data	: QByteArray &	: Result buffer				*/
/*  (returns)	: bool		: Success				*/

bool	KBLocation::getData
	(	cchar		*field,
		KBError		&pError,
		QByteArray	&data
	)
	const
{
	/* Check in case we have been called without a database		*/
	/* information structure; script errors seem to be a major	*/
	/* source of this bug :-)					*/
	if (m_dbInfo == 0)
	{
		pError	= KBError
			  (	KBError::Fault,
				"KBLocation::getData called without database information",
				QString("%1.%2").arg(m_name).arg(m_server),
				__ERRLOCN
			  )	;
		return	false	;
	}

	KBSQLSelect	*select	;
	KBDBLink	dbLink 	;
	bool		exists	;

	/* Open a link to the database and very that the objects table	*/
	/* exists. If it does then open a select query to retrieve the	*/
	/* specified field for this location.				*/
	if (!dbLink.connect (*this, m_server, true))
	{	pError	= dbLink.lastError() ;
		return	false	;
	}

	QString	objectTab = dbLink.rekallPrefix (OBJECTTABNAME) ;

	if (!dbLink.tableExists (objectTab, exists))
	{	pError	= dbLink.lastError() ;
		return	false	;
	}

	if (!exists)
	{	pError	= KBError
			  (	KBError::Fault,
				QString(TR("Server %1 does not have a objects table: cannot load")).arg(m_server),
				QString(TR("Type %1, name %2")).arg(m_type).arg(m_name),
				__ERRLOCN
			  )	;
		return	false	;
	}



	/* Execute the select query for the named document of the	*/
	/* appropriate type. Note that we select all columns at the	*/
	/* same time in order to get the column types that the actual	*/
	/* server database is using.					*/
	KBValue		args[3]	;
	int		nargs	= 2 ;
	bool		script	= m_type == "script" ;
	QString		qtext	= buildDataQuery (dbLink, field, script) ;
	KBLocnTypeMap	*tmp	= findByType	 (m_type) ;

	if (tmp == 0)
	{
		pError	= KBError
			  (	KBError::Fault,
			        "KBLocation::getData called with inappropriate type",
				QString("Type code %1").arg(m_type),
				__ERRLOCN
			  )	;
		return	false	;
	}



	if ((select = dbLink.qrySelect (false, qtext)) == 0)
	{
		pError	= dbLink.lastError() ;
		return	false	;
	}

	args[0]	= m_name	;
	args[1]	= m_type	;

	if (script)
	{
		args[2] = m_extension	;
		nargs = 3	;
	}

	if (!select->execute (nargs, args))
	{	pError	= select->lastError() ;
		delete	select	;
		return	false	;
	}

	// fprintf (stderr, "[%s]\n", (cchar *)select->getRawQuery()) ;
	// fprintf (stderr, "[%s]\n", (cchar *)select->getSubQuery()) ;

	/* ... which should return exactly one row. If it does then	*/
	/* we will return the contents of the field as raw text.	*/
	if (!select->rowExists (0))
	{	pError	= KBError
			  (	KBError::Error,
				"Cannot load document",
				QString	("Document %1 (%2) not found").arg(m_name).arg(m_type),
				__ERRLOCN
			  )	;
		delete	select	;
		return	false	;
	}

	KBValue	value	= select->getField(0, 0) ;
	data.duplicate	(value.dataPtr(), value.dataLength()) ;
	delete	select	;
	return	true	;
}

/*  KBLocation								*/
/*  exists	: See if the document exists				*/
/*  (returns)	: QString	: Location contents or null on error	*/

bool	KBLocation::exists() const
{
	DPRINTF
	((	"KBLocation::exists: t=[%s] s=[%s] n=[%s] e=[%s]\n",
		m_type		.latin1(),
		m_server	.latin1(),
		m_name		.latin1(),
		m_extension	.latin1()
	))	;

	/* First case is if document is in the database directory, in	*/
	/* which just check for its existance.				*/
	if (isFile() || isLocal())
		return	QFileInfo(path()).exists () ;

	/* Next case is inline .... easy				*/
	if (isInline())
		return	true	;

	/* Otherwise, we may as well retrieve the contents as this is	*/
	/* going to be almost as quick.					*/
	KBError		error	;
	QByteArray	bytes	;
	return	getData ("Definition", error, bytes) ;
}

/*  KBLocation								*/
/*  contents	: Get the contents of the location			*/
/*  data	: QByteArray &	: Data buffer				*/
/*  pError	: KBError &	: Error return				*/
/*  (returns)	: bool		: Success				*/

bool	KBLocation::contents
	(	QByteArray	&data,
		KBError		&pError
	)
	const
{
	DPRINTF
	((	"KBLocation::contents [%s][%s][%s][%s]\n",
		(cchar *)m_type,
		(cchar *)m_server,
		(cchar *)m_name,
		(cchar *)m_extension
	))	;


	/* First case is if document is in the database directory, is a	*/
	/* stock component, or is in a local file, in which case read	*/
	/* the document from the appropriate file. The ::path() method	*/
	/* returns the correct path in each case.			*/
	if (isFile() || isStock() || isLocal())
	{
		KBFile	file (path()) ;

		DPRINTF
		((	"KBLocation::contents: path=[%s]\n",
			path().latin1()
		))	;

		if (!file.open (IO_ReadOnly))
		{
			pError	= file.lastError() ;
			return	false	;
		}

		data	= file.readAll();
		return	true	;
	}

	/* Next case is inline .... easy. The fudge is that the		*/
	/* definition is stored in the extension!			*/
	if (isInline())
	{
		data	= QCString(m_extension) ;
		return	true	    ;
	}

	/* Otherwise, we need to read the document source from the	*/
	/* objects table for the specified server. The object may be	*/
	/* cached.							*/
	bool	graphic	= m_type == "graphic" ;
	QString	key	= QString("%1//%2//%3//%4").arg(m_server).arg(m_name).arg(m_type).arg(m_extension) ;
	if (findInCache (graphic, key, data))
		return	true	;

	if (getData ("Definition", pError, data))
	{
		saveToCache  (graphic, key, data) ;
		return	true ;
	}

	return	false	;
}

/*  KBLocation								*/
/*  contents	: Get the contents of the location			*/
/*  pError	: KBError &	: Error return				*/
/*  (returns)	: QString	: Location contents or null on error	*/

QString	KBLocation::contents
	(	KBError	&pError
	)
	const
{
	/* Pick the inline case out specially to save conversions ...	*/
	if (isInline()) return m_extension ;

	/* ... otherwise go via the QByteArray method ...		*/
	QByteArray data ;
	if (!contents (data, pError)) return QString::null ;

	return	QString::fromUtf8 (data.data(), data.count()) ;
}

/*  KBLocation								*/
/*  contentsAsDom							*/
/*		: Get the contents of the location as a DOM document	*/
/*  what	: const QString & : Used for error message		*/
/*  pError	: KBError &	  : Error return			*/
/*  (returns)	: QDomDocument	  : Location contents or null on error	*/

QDomDocument
	KBLocation::contentsAsDom
	(	const QString	&what,
		KBError		&pError
	)
	const
{
	QString		doc	= contents (pError) ;
	if (doc.isNull()) return QDomDocument () ;

	QDomDocument	dom	;
	dom.setContent	(doc)	;
	QDomElement	root	= dom.documentElement () ;
	if (root.isNull())
	{
		pError	= KBError
			  (	KBError::Error,
				QString(TR("%1 definition has no root element")).arg(what),
				QString::null,
				__ERRLOCN
			  )	;
		return	QDomDocument() ;
	}

	return	dom	;
}

/*  KBLocation								*/
/*  timestamp	: Get location timestamp				*/
/*  pError	: KBError &	: Error return				*/
/*  (returns)	: QString	: Timestamp or null on error		*/

QString	KBLocation::timestamp
	(	KBError	&pError
	)
	const
{
	/* If the location is a file then grab the modification time of	*/
	/* the file.							*/
	if (isFile() || isLocal())
	{
		QFileInfo info (path()) ;
		if (!info.exists())
		{
			pError	= KBError
				  (	KBError::Error,
				  	"Cannot determine object modification time",
				  	QString	("%1.%2 (%3)")
						.arg(m_name)
						.arg(m_extension)
						.arg(path ()),
				  	__ERRLOCN
				  )	;
			return	QString::null ;
		}

		return	QString("%1").arg(info.lastModified().toTime_t()) ;
	}

	/* The timestamp of an inline location is arbitrary, so return	*/
	/* the current time.						*/
	if (isInline())
		return	QString::number(QDateTime::currentDateTime().toTime_t()) ;

	/* Lastly, for database-stored locations, return the time	*/
	/* stamp from the objects table for this location.		*/
	QByteArray bytes ;
	return	   getData ("SaveDate", pError, bytes) ? bytes : QString::null ;
}

/*  KBLocation								*/
/*  getFactory	: Get part factory for this object type			*/
/*  (returns)	: KBPartFactory * : Factory for the object type		*/

KBPartFactory
	*KBLocation::getFactory () const
{
	KBLocnTypeMap *tmp = findByType (m_type) ;

	DPRINTF
	((	"KBLocation::getFactory: [%s]->[%p]\n",
		(cchar *)m_type,
		(void  *)tmp
	))	;
	return	tmp == 0 ? 0 : tmp->factory ;
}

KBServerInfo
	*KBLocation::getServerInfo () const
{
	return	m_dbInfo == 0 ? 0 : m_dbInfo->findServer (m_server) ;
}

/*  KBLocation								*/
/*  registerType: Register a new type					*/
/*  type	: cchar *	  : Type				*/
/*  extn	: cchar *	  : Extension				*/
/*  legend	: const QString & : Legend for titles			*/
/*  factory	: KBPartFactory * : Factory for the object type		*/
/*  (returns)	: void		  :					*/

void	KBLocation::registerType
	(	cchar		*type,
		cchar		*extn,
		const QString	&legend,
		KBPartFactory	*factory
	)
{
	DPRINTF
	((	"KBLocation::registerType(%s,%s)\n",
		type,
		extn
	))	;
	locnTypeMap.append (new KBLocnTypeMap (type, extn, legend, factory)) ;
}

QString	KBLocation::extnForType
	(	KBDBInfo	*dbInfo,
		const QString 	&type,
		const QString	&extension
	)
{
	DPRINTF
	((	"KBLocation::extnForType(%s,%s)\n",
		(cchar *)type,
		(cchar *)extension
	))	;

	KBLocnTypeMap	*tmp	= findByType (type) ;

	if (tmp != 0)
	{
		if (tmp->extn.isEmpty())
			return	extension  ;
		else	return	dbInfo->getDBExtn() + "." + tmp->extn ;
	}

	return	extension ;
}


/*  KBLocation								*/
/*  saveToFile	: Save document to a file				*/
/*  file	: const QString & : File location			*/
/*  name	: const QString & : Document name			*/
/*  buffer	: cchar *	  : Data buffer				*/
/*  bufflen	: uint		  : Buffer length			*/
/*  pError	: KBError &	  : Error return			*/
/*  (returns)	: bool		  : Success				*/

bool	KBLocation::saveToFile
	(	const QString	&file,
		const QString	&,
		cchar		*buffer,
		uint		bufflen,
		KBError		&pError
	)
{
	KBFile	out	(file) ;
	if (!out.open (IO_WriteOnly))
	{
		pError	= out.lastError() ;
		return	false	;
	}

	if (out.writeBlock (buffer, bufflen) != (int)bufflen)
	{
		pError	= KBError
			  (	KBError::Error,
				QString ("Error writing file \"%1\"").arg(file),
				strerror(errno),
				__ERRLOCN
			  )	;
		return	false	;
	}

	return	true	;
}

/*  KBLocation								*/
/*  buildInfoQuery							*/
/*		: Build select query for object information retrieval	*/
/*  dbLink	: KBDBLink &	: Database link				*/
/*  (returns)	: QString	: Query text				*/

QString	KBLocation::buildInfoQuery
	(	KBDBLink	&dbLink
	)
	const
{
	KBBaseSelect	select	(dbLink.rekallPrefix (OBJECTTABNAME)) ;

	select.addFetch	("Id"		) ;
	select.addFetch	("Description"	) ;
	select.addFetch	("Definition"	) ;
	select.addFetch	("SaveDate"	) ;
	select.addFetch	("Type"		) ;
	select.addFetch	("Name"		) ;
	select.addFetch	("Extension"	) ;
	select.addWhere ("Type"		) ;
	select.addWhere ("Name"		) ;

	return	select.getQueryText(&dbLink) ;
}

/*  KBLocation								*/
/*  buildDataQuery							*/
/*		: Build select query for object data retrieval		*/
/*  dbLink	: KBDBLink &	: Database link				*/
/*  field	: cchar *	: Required data field			*/
/*  addExt	: bool		: Add extension to query filter		*/
/*  (returns)	: QString	: Query text				*/

QString	KBLocation::buildDataQuery
	(	KBDBLink	&dbLink,
		cchar		*field,
		bool		addExt
	)
	const
{
	KBBaseSelect	select	(dbLink.rekallPrefix (OBJECTTABNAME)) ;

	select.addFetch	(field	) ;
	select.addWhere ("Name"	) ;
	select.addWhere ("Type"	) ;
	if (addExt)
		select.addWhere ("Extension") ;

	return	select.getQueryText(&dbLink) ;
}

/*  KBLocation								*/
/*  buildInsertQuery							*/
/*		: Build query for object insert				*/
/*  dbLink	: KBDBLink &	: Database link				*/
/*  idRO	: bool		: ID column is read only		*/
/*  (returns)	: QString	: Query text				*/

QString	KBLocation::buildInsertQuery
	(	KBDBLink	&dbLink,
		bool		idRO
	)
	const
{
	KBBaseInsert	insert	(dbLink.rekallPrefix (OBJECTTABNAME)) ;

	if (!idRO)
		insert.addValue("Id" ) ;

	insert.addValue ("Description", (void *)0) ;
	insert.addValue ("Definition") ;
	insert.addValue ("SaveDate"  ) ;
	insert.addValue ("Type"	  ) ;
	insert.addValue ("Name"	  ) ;

	if (!m_extension.isEmpty())
			insert.addValue("Extension") ;

	return	insert.getQueryText(&dbLink) ;
}

/*  KBLocation								*/
/*  buildUpdateQuery							*/
/*		: Build query for object update				*/
/*  dbLink	: KBDBLink &	: Database link				*/
/*  (returns)	: QString	: Query text				*/

QString	KBLocation::buildUpdateQuery
	(	KBDBLink	&dbLink
	)
	const
{
	KBBaseUpdate	update	(dbLink.rekallPrefix (OBJECTTABNAME)) ;

	update.addValue ("Definition") ; 
	update.addValue ("SaveDate"  ) ; 
	update.addWhere	("Type"	     ) ; 
	update.addWhere	("Name"	     ) ;

	if (!m_extension.isEmpty())
		update.addWhere("Extension") ; 

	return	update.getQueryText(&dbLink) ;
}

/*  KBLocation								*/
/*  buildDeleteQuery							*/
/*		: Build query for object update				*/
/*  dbLink	: KBDBLink &	: Database link				*/
/*  (returns)	: QString	: Query text				*/

QString	KBLocation::buildDeleteQuery
	(	KBDBLink	&dbLink
	)
	const
{
	QString	objectTab = dbLink.rekallPrefix (OBJECTTABNAME) ;
	QString qryString = QString("delete from %1 where %2 = %3 and %4 = %5")
				   .arg(dbLink.mapExpression(objectTab))
				   .arg(dbLink.mapExpression("Name"   ))
				   .arg(dbLink.placeHolder  (0	      ))
				   .arg(dbLink.mapExpression("Type"   ))
				   .arg(dbLink.placeHolder  (1	      ))
				   ;
	KBBaseDelete	delrec	(dbLink.rekallPrefix (OBJECTTABNAME)) ;

	delrec.addWhere ("Name"	) ;
	delrec.addWhere ("Type"	) ;

	return	delrec.getQueryText(&dbLink) ;
}

/*  KBLocation								*/
/*  saveToDB	: Save document to database table			*/
/*  server	: const QString & : Table location			*/
/*  name	: const QString & : Document name			*/
/*  buffer	: cchar *	  : Data buffer				*/
/*  bufflen	: uint		  : Buffer length			*/
/*  pError	: KBError &	  : Error return			*/
/*  (returns)	: bool		  : Success				*/

bool	KBLocation::saveToDB
	(	const QString	&server,
		const QString	&name,
		cchar		*buffer,
		uint		bufflen,
		KBError		&pError
	)
{
	KBDBLink dbLink	;
	bool	 exists	;
	bool	 rv	= false ;

	if (!dbLink.connect (*this, server, true))
	{	pError	= dbLink.lastError() ;
		return	false ;
	}

	QString	    objectTab  = dbLink.rekallPrefix (OBJECTTABNAME) ;
	if (!dbLink.tableExists (objectTab, exists))
	{	pError	= dbLink.lastError() ;
		return	false ;
	}

	if (!exists)
	{
		pError	= KBError
			  (	KBError::Error,
				QString(TR("Server %1 does not have a objects table: cannot save")).arg(m_server),
				QString(TR("Type %1, name %2")).arg(m_type).arg(name),
				__ERRLOCN
			  )	;
		return	false	;
	}

//**	const RKColumnTypes *rkt = columnTypes (dbLink, pError) ;

	/* This is an update-or-insert so we need to find out if there	*/
	/* is already an entry. The select query also serves to get	*/
	/* type information for each of the columns in the objects	*/
	/* table (which will vary from server to server).		*/
	KBSQLSelect *qrySelect = 0 ;
	KBSQLInsert *qryInsert = 0 ;
	KBSQLUpdate *qryUpdate = 0 ;
	KBSQLQuery  *save      ;
	int	    nRows      ;

	uint	    argc   = 0 ;
	KBValue	    args[8]    ;


	QString selString = buildInfoQuery (dbLink) ;
	if ((qrySelect = dbLink.qrySelect(false, selString)) == 0)
	{	pError	= dbLink.lastError() ;
		goto	error ;
	}
	args[0] = QString(m_type) ;
	args[1]	= name	;
	if (!qrySelect->execute (2, args))
	{	pError	= qrySelect->lastError() ;
		goto	error ;
	}

	/* No row exists then insert data, otherwise update an extant	*/
	/* entry.							*/
	if (!qrySelect->rowExists(0))
	{
		KBValue		newKey    ;
		KBTableSpec	tabSpec	  (objectTab) ;
		bool		idRO	  ;

		if (!dbLink.listFields (tabSpec))
		{
			pError	= dbLink.lastError() ;
			return	false	;
		}
		idRO	= (tabSpec.m_fldList.at(0)->m_flags & KBFieldSpec::ReadOnly) != 0 ;

		QString	insString = buildInsertQuery (dbLink, idRO) ;

		/* Construct the insert query and report an error if	*/
		/* this fails. Then if the ID column is not read-only	*/
		/* then see if we get a pre-insert key value.		*/
		if ((qryInsert = dbLink.qryInsert (false, insString, objectTab)) == 0)
		{
			pError	= dbLink.lastError() ;
			goto	error ;
		}

		if (!idRO)
		{
			if (!qryInsert->getNewKey ("Id", newKey, true))
			{
				pError	= dbLink.lastError() ;
				goto	error ;
			}

			args[argc++] = newKey ;
		}

		save	     = qryInsert ;
	}
	else
	{
		QString	updString = buildUpdateQuery (dbLink) ;

		if ((qryUpdate = dbLink.qryUpdate(false, updString, objectTab)) == 0)
		{
			pError	= dbLink.lastError() ;
			goto	error ;
		}

		save	= qryUpdate ;
	}


	args[argc++] = KBValue
		       (	buffer,
				bufflen,
				qrySelect->getFieldType(2)
		       )	;

	{
	QString	dstr = QDateTime::currentDateTime().toString("yyyyMMddhhmmss") ;
	args[argc++] = KBValue (dstr,	qrySelect->getFieldType(3)) ;
	args[argc++] = KBValue (m_type, qrySelect->getFieldType(4)) ;
	args[argc++] = KBValue (name,	qrySelect->getFieldType(5)) ;
	}

	if (!m_extension.isEmpty())
		args[argc++] = KBValue (m_extension, qrySelect->getFieldType(6)) ;


	/* Execute the query. This should succeed with a sing row being	*/
	/* affected in either case.					*/
	if (!save->execute (argc, args))
	{	pError	= save->lastError() ;
		goto	error ;
	}

	nRows	= qryInsert != 0 ? qryInsert->getNumRows() : qryUpdate->getNumRows() ;

	if (nRows != 1)
	{
		pError	= KBError
			  (	KBError::Error,
				QString	("Error saving %1"	       ).arg(m_type),
				QString	("Unexpectedly updated %1 rows").arg(nRows ),
				__ERRLOCN
		)	;
		goto error ;
	}

	rv		 = true	;

	error	:
		DELOBJ	(qrySelect) ;
		DELOBJ	(qryInsert) ;
		DELOBJ	(qryUpdate) ;
		return	rv	 ;
}

/*  KBLocation								*/
/*  save	: Save document						*/
/*  server	: const QString & : Document location			*/
/*  name	: const QString & : Document name			*/
/*  buffer	: cchar *	  : Data buffer				*/
/*  bufflen	: uint		  : Buffer length			*/
/*  pError	: KBError &	  : Error return			*/
/*  (returns)	: bool		  : Success				*/

bool	KBLocation::save
	(	const QString	&__server,
		const QString	&__name,
		cchar		*buffer,
		uint		bufflen,
		KBError		&pError
	)
{
	bool	rv	;
	QString	server	= __server.isNull() ? m_server : __server ;
	QString	name	= __name  .isNull() ? m_name   : __name   ;
	
	DPRINTF
	((	"KBLocation::save       [%d] [%s][%s][%s][%s]\n",
		this,
		(cchar *)m_type,
		(cchar *)m_server,
		(cchar *)m_name,
		(cchar *)m_extension
	))	;

	removeFromCache (QString("%1//%2//%3//%4").arg(server).arg(name).arg(m_type).arg(m_extension)) ;

	if ((server == KBLocation::m_pFile) || (server == KBLocation::m_pLocal))
		rv	 = saveToFile (path(name), name, buffer, bufflen, pError) ;
	else	rv	 = saveToDB   (server, 	   name, buffer, bufflen, pError) ;

	if (rv)
	{	m_server = server ;
		m_name	 = name	  ;
	}

	KBNotifier::self()->nObjectChanged (*this) ;
	return	rv	;
}

/*  KBLocation								*/
/*  save	: Save document						*/
/*  __server	: const QString & : Document location			*/
/*  __name	: const QString & : Document name			*/
/*  text	: const QString & : Document text			*/
/*  pError	: KBError &	  : Error return			*/
/*  (returns)	: bool		  : Success				*/

bool	KBLocation::save
	(	const QString	&__server,
		const QString	&__name,
		const QString	&text,
		KBError		&pError
	)
{
	QCString data = text.utf8() ;
	return	 save (__server, __name, data.data(), data.length(), pError) ;
}

/*  KBLocation								*/
/*  renameFile	: Rename file object					*/
/*  newName	: const QString & : New name				*/
/*  pError	: KBError &	  : Error return			*/
/*  (returns)	: bool		  : Success				*/

bool	KBLocation::renameFile
	(	const QString	&newName,
		KBError		&pError
	)
{
	QString	objPath	= path ()	;
	QString	newPath	= path (newName);

	if (::rename ((cchar *)objPath, (cchar *)newPath) != 0)
	{
		pError	= KBError
			  (	KBError::Error,
				QString	(TR("Failed to rename %1 %2 as %3"))
					.arg(m_type )
					.arg(m_name )
					.arg(newName),
			 	QString	(TR("System error: %1 -> %2: %3"))
					.arg(m_name )
					.arg(newName)
					.arg(strerror(errno)),
				__ERRLOCN
			   )	;
		pError.setErrno (errno) ;

		return	false	;
	}

	return	true	;
}

/*  KBLocation								*/
/*  renameDB	: Rename database object				*/
/*  newName	: const QString & : New name				*/
/*  pError	: KBError &	  : Error return			*/
/*  (returns)	: bool		  : Success				*/

bool	KBLocation::renameDB
	(	const QString	&newName,
		KBError		&pError
	)
{
	KBDBLink dbLink	 ;
	bool	 exists  ;
	KBValue	 args	 [3]  ;

	if (!dbLink.connect (m_dbInfo, m_server, true))
	{
		pError	= dbLink.lastError() ;
		return	false ;
	}

	QString	objectTab = dbLink.rekallPrefix (OBJECTTABNAME) ;

	if (!dbLink.tableExists (objectTab, exists))
	{
		pError	= dbLink.lastError() ;
		return	false ;
	}
	if (!exists) return true ;

	KBSQLUpdate *update ;
	QString qryString = QString("update %1 set %2 = %3 where %4 = %5 and %6 = %7")
				   .arg(dbLink.mapExpression(objectTab))
			  	   .arg(dbLink.mapExpression("Name"   ))
				   .arg(dbLink.placeHolder  (0))
			  	   .arg(dbLink.mapExpression("Name"   ))
				   .arg(dbLink.placeHolder  (1))
			  	   .arg(dbLink.mapExpression("Type"   ))
				   .arg(dbLink.placeHolder  (2))
				   ;

	if ((update = dbLink.qryUpdate(false, qryString, objectTab)) == 0)
	{
		pError	= dbLink.lastError() ;
		return	false ;
	}

	args[0] = QString(newName) ;
	args[1] = QString(m_name ) ;
	args[2] = QString(m_type ) ;

	if (!update->execute (3, args))	
	{
		pError	= update->lastError() ;
		delete	update	;
		return	false	;
	}

	delete	update	;
	return	true	;
}

/*  KBLocation								*/
/*  rename	: Rename object						*/
/*  newName	: const QString & : New name				*/
/*  pError	: KBError &	  : Error return			*/
/*  (returns)	: bool		  : Success				*/

bool	KBLocation::rename
	(	const QString	&newName,
		KBError		&pError
	)
{
	removeFromCache (QString("%1//%2//%3//%4").arg(m_server).arg(m_name ).arg(m_type).arg(m_extension)) ;
	removeFromCache (QString("%1//%2//%3//%4").arg(m_server).arg(newName).arg(m_type).arg(m_extension)) ;

	if (m_server == KBLocation::m_pFile)
		return	renameFile (newName, pError) ;
	else	return	renameDB   (newName, pError) ;
}

/*  KBLocation								*/
/*  removeFile	: Remove file object					*/
/*  pError	: KBError &	  : Error return			*/
/*  (returns)	: bool		  : Success				*/

bool	KBLocation::removeFile
	(	KBError		&pError
	)
{
	QString	objPath	= path () ;

	if (!QFile(objPath).remove())
	{
		pError	= KBError
			  (	KBError::Error,
				QString(TR("Failed to delete %1"))
					.arg(m_name),
		 		QString(TR("System error: %1: %1"   ))
		 			.arg(objPath)
					.arg(strerror(errno)),
				__ERRLOCN
			  )	;
		pError.setErrno	(errno) ;

		return	false	;
	}

	return	true	;
}

/*  KBLocation								*/
/*  removeDB	: Remove database object				*/
/*  pError	: KBError &	  : Error return			*/
/*  (returns)	: bool		  : Success				*/

bool	KBLocation::removeDB
	(	KBError		&pError
	)
{
	KBDBLink dbLink	 ;
	bool	 exists  ;
	KBValue	 args	 [2]  ;

	if (!dbLink.connect (m_dbInfo, m_server, true))
	{
		pError	= dbLink.lastError() ;
		return	false ;
	}

	QString	objectTab = dbLink.rekallPrefix (OBJECTTABNAME) ;

	if (!dbLink.tableExists (objectTab, exists))
	{
		pError	= dbLink.lastError() ;
		return	false ;
	}
	if (!exists) return true ;

	KBSQLDelete *delrec ;

	QString qryString = QString("delete from %1 where %2 = %3 and %4 = %5")
				   .arg(dbLink.mapExpression(objectTab))
				   .arg(dbLink.mapExpression("Name"   ))
				   .arg(dbLink.placeHolder  (0	      ))
				   .arg(dbLink.mapExpression("Type"   ))
				   .arg(dbLink.placeHolder  (1	      ))
				   ;

	if ((delrec = dbLink.qryDelete
			(	false,
				qryString,
				objectTab
			)) == 0)
	{
		pError	= dbLink.lastError() ;
		return	false ;
	}

	args[0] = QString(m_name) ;
	args[1] = QString(m_type) ;

	if (!delrec->execute (2, args))	
	{
		pError	= delrec->lastError() ;
		delete	delrec	;
		return	false	;
	}

	delete	delrec	;
	return	true	;
}

/*  KBLocation								*/
/*  remove	: Remove object						*/
/*  pError	: KBError &	  : Error return			*/
/*  (returns)	: bool		  : Success				*/

bool	KBLocation::remove
	(	KBError		&pError
	)
{
	removeFromCache (QString("%1//%2//%3//%4").arg(m_server).arg(m_name).arg(m_type).arg(m_extension)) ;

	if (m_server == KBLocation::m_pFile)
		return	removeFile (pError) ;
	else	return	removeDB   (pError) ;
}

QString	KBLocation::samePlace
	(	const QString	&name
	)
	const
{
	if (isLocal())
	{
		QStringList bits = QStringList::split('/', m_name) ;
		bits[bits.count() - 1] = name ;
		QString	    res  = bits.join("/") ;
		if (m_name[0] == '/') res = "/" + res ;
		return	    res  ;
	}

	return	name	;
}


KBDBInfo	*KBLocation::dbInfo	() const { return	m_dbInfo	; }
const QString	&KBLocation::type	() const { return	m_type		; }
const QString	&KBLocation::server	() const { return	m_server	; }
const QString	&KBLocation::name	() const { return	m_name		; }
const QString	&KBLocation::extension	() const { return	m_extension	; }
const QString	&KBLocation::dataServer	() const { return	m_dataServer	; }

void	KBLocation::setServer
	(	const QString	&server
	)
{
	m_server	= server ;
}

void	KBLocation::setDataServer
	(	const QString	&server
	)
{
	m_dataServer	= server ;
}

void	KBLocation::setName
	(	const QString	&name
	)
{
	m_name		= name	 ;
}


bool	KBLocation::isFile  () const { return m_server == m_pFile	   ; }
bool	KBLocation::isInline() const { return m_server == m_pInline	   ; }
bool	KBLocation::isStock () const { return m_server == m_pStock	   ; }
bool	KBLocation::isLocal () const { return m_server == m_pLocal	   ; }

bool	KBLocation::isInDB  () const
	{
		return !isFile() && !isInline() && !isStock() ;
	}


/*  ------------------------------------------------------------------  */

/*  kbXMLEncoding: Return XML encoding used				*/
/*  (returns)	 : QString	: Encoding				*/

LIBCOMMON_API QString	kbXMLEncoding ()
{
	return	"UTF-8"	;
}

