/***************************************************************************
    file	         : kb_attrexpr.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/

#include	"kb_classes.h"
#include	"kb_type.h"
#include	"kb_value.h"
#include	"kb_database.h"
#include	"kb_location.h"
#include	"kb_attr.h"
#include	"kb_script.h"
#include	"kb_object.h"
#include	"kb_attrexpr.h"
#include	"kb_callback.h"
#include	"kb_location.h"
#include	"kb_callback.h"
#include	"kb_appptr.h"
#include	"kb_docroot.h"

/*  KBAttrExpr								*/
/*  KBAttrExpr: Constructor for expression class			*/
/*  node	: KBNode *	: Owning node				*/
/*  name	: cchar *	: Attribute name			*/
/*  aList	: const QDict<QString> &				*/
/*				: List of attributes			*/
/*  (returns)	: KBAttrExpr	:					*/

KBAttrExpr::KBAttrExpr
	(	KBNode			*node,
		cchar			*name,
		const QDict<QString>	&aList
	)
	:
	KBAttrStr (node, name, aList)
{
	m_disable	= false	;
	m_code		= 0	;
	m_isexpr	= getValue().at(0) == '=' ;
}

/*  KBAttrExpr								*/
/*  KBAttrExpr: Constructor for expression class			*/
/*  node	: KBNode *	: Owning node				*/
/*  name	: cchar *	: Attribute name			*/
/*  text	: cchar *	: Initial text value			*/
/*  (returns)	: KBAttrExpr	:					*/

KBAttrExpr::KBAttrExpr
	(	KBNode	*node,
		cchar	*name,
		cchar	*text
	)
	:
	KBAttrStr (node, name, text)
{
	m_disable	= false	;
	m_code		= 0	;
	m_isexpr	= getValue().at(0) == '=' ;
}

/*  KBAttrExpr								*/
/*  KBAttrExpr: Constructor for expression class			*/
/*  node	: KBNode *	: Owning node				*/
/*  name	: cchar *	: Attribute name			*/
/*  srce	: KBNode *	: Extant source				*/
/*  (returns)	: KBAttrExpr	:					*/

KBAttrExpr::KBAttrExpr
	(	KBNode	*node,
		cchar	*name,
		KBNode	*srce
	)
	:
	KBAttrStr (node, name, srce)
{
	m_disable	= false	;
	m_code		= 0	;
	m_isexpr	= getValue().at(0) == '=' ;
}

/*  KBAttrExpr								*/
/*  ~KBAttrExpr	: Destructor for expression class			*/
/*  (returns)	:		:					*/

KBAttrExpr::~KBAttrExpr ()
{
	if (m_code != 0) delete m_code ;
}

/*  KBAttrExpr								*/
/*  setValue	: Set event text value					*/
/*  value	: QString &	: Value					*/
/*  (returns)	: void		:					*/

void	KBAttrExpr::setValue
	(	const QString	&value
	)
{
	/* This method is overridden so that we can clear any code	*/
	/* compiled for an earlier value, and clear the disabled flag	*/
	DELOBJ	  (m_code)	;
	m_disable = false	;
	m_isexpr  = value.at(0) == '=' ;

	KBAttrStr::setValue (value)  ;
}

/*  KBAttrExpr								*/
/*  evaluate	: Evaluate value					*/
/*  argval	: const KBValue &  : Argument value			*/
/*  eRC		: KBScriptError *& : Script error or null if OK		*/
/*  fSubs	: bool &	   : Substitution flag			*/
/*  (returns)	: KBValue	   : Resultant value			*/

KBValue	KBAttrExpr::evaluate
	(	const KBValue	&argval,
		KBScriptError	*&eRC,
		bool		&fSubs
	)
{
	eRC	= 0	;
	if (!m_isexpr) return argval ;

	QString		expr	  = getValue () ;
	KBDocRoot	*root	  = m_owner->getDocRoot ()    ;
	KBScriptIF	*scrIface = root->loadScripting (eRC) ;
	KBValue		resval	  ;
	KBScript::ExeRC rc 	  ;

	/* A script interface is needed to be able to execute the	*/
	/* expression code.						*/
	if (scrIface == 0)
		return	resval	;

	/* If the expression is flagged as disabled then there has	*/
	/* been an earlier error ...					*/
	if (m_disable)
	{
		eRC	= new KBScriptError
			  (	KBError
				(	KBError::Error,
					TR("Expression has been disabled due to earlier error"),
					QString::null,
					__ERRLOCN
				),
				this
			  )	;
		return	resval	;
	}


	/* We compile the code and keep a copy. First see if the code	*/
	/* has already been compiled and if not then so so now.		*/
	if (m_code == 0)
	{
		KBError	error	;
		if ((m_code = scrIface->compileExpr
			      (		m_owner,
					expr.mid(1),
					m_owner->isObject()->getPath   (),
					root->getImports(),
					error
			      )	) == 0)
		{
			m_disable = true ;
			eRC	  = new KBScriptError (error, this) ;
			return	  resval ;
		}
	}

	/* Code was or has been compiled successfully, in which case	*/
	/* execute it. Check for an execution time error ....		*/
	rc = m_code->execute (0, 0, resval) ;
	switch (rc)
	{
		case KBScript::ExeAbort :
		case KBScript::ExeTest  :
			eRC	= new KBScriptError () ;
			return	resval	;

		case KBScript::ExeError	:
		case KBScript::ExeFail	:
			{
			QString	   errMsg   ;
			uint	   errLno   ;
			QString	   errText  ;
			KBLocation errLocn  = scrIface->lastError (errMsg, errLno, errText) ;

			m_disable  = true   ;

			if (errLocn.name() != KBLocation::m_pInline)
				eRC   = new KBScriptError
					(	KBError
						(	KBError::Error,
							errMsg,
							QString (TR("%1, line %2\n%3"))
								.arg(errLocn.name())
								.arg(errLno)
							.arg(errText),
							__ERRLOCN
						),
						m_owner->isObject(),
						errLocn,
						errText,
						errLno,
						rc == KBScript::ExeFail
				  	)	;
			else
				eRC   = new KBScriptError
					(	KBError
						(	KBError::Error,
							errMsg,
							QString (TR("%1, line %2\n%3"))
								.arg(errLocn.name())
								.arg(errLno)
							.arg(errText),
							__ERRLOCN
						),
						this
					)	;
			}
			return	resval	;

		default	:
			break	;
	}

	/* Success ....							*/
	eRC	= 0	;
	fSubs	= resval.getRawText().find("%{") >= 0 ;
	return	resval	;
}
