/***************************************************************************
    file	         : kb_attrgeom.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/


#include	<qfont.h>
#include	<qfontmetrics.h>
#include	<qapplication.h>

#include	"kb_attrgeom.h"
#include	"kb_object.h"
#include	"kb_nodemonitor.h"
#include	"kb_display.h"
#include	"kb_options.h"
#include	"kb_ascii.h"

#if		! __KB_RUNTIME
#include	"kb_attrgeomdlg.h"
#endif


#define		__DEBUG_ATTRGEOM	(0)
#if		__DEBUG_ATTRGEOM
#include	"kb_dprintf.h"
#endif

static	QString	modeToText
	(	KBAttrGeom::FloatMode	mode
	)
{
	switch (mode)
	{
		case KBAttrGeom::FMFixed   : return __TR("fixed"  )	;
		case KBAttrGeom::FMFloat   : return __TR("float"  )	;
		case KBAttrGeom::FMStretch : return __TR("stretch")	;
		default			   : break  ;
	}

	return	__TR("unknown") ;
}

static	void	printGridSetup
	(	const QValueList<KBGridSetup>	&setup
	)
{
	for (uint idx = 0 ; idx < setup.count() ; idx += 1)
		DPRINTF
		((	"printGridSetup: %3d: %3d,%3d\n",
			idx,
			setup[idx].m_stretch,
			setup[idx].m_spacing
		))	;
}

static	void	dumpGridSetup
	(	QString				&text,
		int				indent,
		const QValueList<KBGridSetup>	&setup,
		int				limit,
		cchar				*type
	)
{
	for
	(	int index = 0 ;
		(index < (int)setup.count()) && (index < limit) ;
		index += 1
	)
	{
		const KBGridSetup &gs = setup[index] ;
		if ((gs.m_stretch > 0) || (gs.m_spacing > 0))
		text	+= QString("%1<extra name=\"_geometry\" type=\"%2\" index=\"%3\" stretch=\"%4\" spacing=\"%5\"/>\n")
				   .arg("", indent)
				   .arg(type)
				   .arg(index)
				   .arg(gs.m_stretch)
				   .arg(gs.m_spacing)
				;
	}
}

/*  KBAttrGeom								*/
/*  KBAttrGeom	: Constructor for geometry attribute			*/
/*  owner	: KBObject *		: Owning object			*/
/*  item	: KBAttrItem *		: Associated item		*/
/*  attrDict	: QDict<KBAttrItem> &	: Attribute dictionary		*/
/*  (returns)	: KBAttrGeom		:				*/

KBAttrGeom::KBAttrGeom
	(	KBObject		*owner,
		const QDict<QString>	&aDict,
		uint			flags
	)
	:
	KBAttr	(owner, KBAttr::Base, "_geometry", aDict, flags|KAF_GRPFORMAT),
	m_owner	(owner)
{
	m_d.m_x		= getAttrValue	(aDict, "x",	      0) ;
	m_d.m_y		= getAttrValue	(aDict, "y",	      0) ;
	m_d.m_w		= getAttrValue	(aDict, "w",	      0) ;
	m_d.m_h		= getAttrValue	(aDict, "h",	      0) ;

	m_d.m_nRows	= getAttrValue	(aDict, "m_rows",     0) ;
	m_d.m_nCols	= getAttrValue	(aDict, "m_cols",     0) ;
	m_d.m_spacing	= getAttrValue	(aDict, "m_spacing", -1) ;
	m_d.m_margin	= getAttrValue	(aDict, "m_margin",  -1) ;

	m_d.m_minw	= getAttrValue	(aDict, "minw",	      0) ;
	m_d.m_maxw	= getAttrValue	(aDict, "maxw",       0) ;
	m_d.m_minh	= getAttrValue	(aDict, "minh",       0) ;
	m_d.m_maxh	= getAttrValue	(aDict, "maxh",       0) ;

	m_d.m_byChars	= (bool	    )getAttrValue (aDict, "bychars", 0) ;
	m_d.m_align	= (Qt::AlignmentFlags
				    )getAttrValue (aDict, "galign",  0) ;

	m_d.m_xmode	= (FloatMode)getAttrValue (aDict, "xmode",    FMFixed    ) ;
	m_d.m_ymode	= (FloatMode)getAttrValue (aDict, "ymode",    FMFixed    ) ;
	m_d.m_manage	= (MgmtMode )getAttrValue (aDict, "manage",   MgmtStatic ) ;

	m_d.m_overflow	= (OverFlow )getAttrValue (aDict, "overflow", OverFlowOff) ;
	m_d.m_mask	= 0 ;
	m_charWidth	= 0 ;
	m_charHeight	= 0 ;

	DPRINTF
	((	"KBAttrGeom::KBAttrGeom: (%d,%d,%d,%d) (%d,%d)\n",
		m_d.m_x,
		m_d.m_y,
		m_d.m_w,
		m_d.m_h,
		(int)m_d.m_xmode,
		(int)m_d.m_ymode
	))	;

	setupRowColSetup () ;

	m_ps	= QSize(0,0);
}

/*  KBAttrGeom								*/
/*  KBAttrGeom	: Constructor for geometry attribute			*/
/*  owner	: KBObject *		: Owning object			*/
/*  x, ...	: int, ...		: Settings			*/
/*  attrDict	: QDict<KBAttrItem> &	: Attribute dictionary		*/
/*  (returns)	: KBAttrGeom		:				*/

KBAttrGeom::KBAttrGeom
	(	KBObject		*owner,
		int			x,
		int			y,
		int			w,
		int			h,
		FloatMode		xmode,
		FloatMode		ymode,
		uint			flags
	)
	:
	KBAttr	(owner, KBAttr::Base, "_geometry", "", flags|KAF_GRPFORMAT),
	m_owner	(owner)
{
	m_d.m_x		= x	;
	m_d.m_y		= y	;
	m_d.m_w		= w	;
	m_d.m_h		= h	;

	m_d.m_nRows	= 0	;
	m_d.m_nCols	= 0	;
	m_d.m_spacing	= -1	;
	m_d.m_margin	= -1	;

	m_d.m_minw	= 0	;
	m_d.m_maxw	= 0	;
	m_d.m_minh	= 0	;
	m_d.m_maxh	= 0	;

	m_d.m_byChars	= false	;
	m_d.m_align	= Qt::AlignAuto	;
	m_d.m_xmode	= xmode	;
	m_d.m_ymode	= ymode	;
	m_d.m_manage	= MgmtStatic	;

	m_d.m_mask	= 0 	;
	m_d.m_overflow	= OverFlowOff	;

	m_charWidth	= 1 	;
	m_charHeight	= 1	;

	m_ps	= QSize	(0,0)	;
}

/*  KBAttrGeom								*/
/*  KBAttrGeom	: Constructor for geometry attribute			*/
/*  owner	: KBObject *		: Owning object			*/
/*  extant	: KBObject *		: Extant object			*/
/*  (returns)	: KBAttrGeom		:				*/

KBAttrGeom::KBAttrGeom
	(	KBObject		*owner,
		KBObject		*extant
	)
	:
	KBAttr		(owner, "_geometry", extant, extant->attrGeom().getFlags()),
	m_owner		(owner),
	m_rowSetup	(extant->attrGeom().m_rowSetup),
	m_colSetup	(extant->attrGeom().m_colSetup)
{
	m_d	= extant->attrGeom().m_d ;
	setupRowColSetup()	;

	m_ps	= QSize (0,0)	;
}

/*  KBAttrGeom								*/
/*  ~KBAttrGeom	: Destructor for geometry attribute			*/
/*  (returns)	:		:					*/

KBAttrGeom::~KBAttrGeom ()
{
}

/*  KBAttrGeom								*/
/*  hidden	: Get extra hidden flags				*/
/*  (returns)	: int		: Extra flags				*/

uint	KBAttrGeom::hidden () const
{
	uint	hide	= 0 ;

	if (m_owner->parentIsDynamic())
		hide	|= HideX|HideY|HideW|HideH|HideXMode|HideYMode ;
	else	hide	|= HideRange ;

	/* The geometry management options are hidden unless the	*/
	/* object handles this. Should be defined in the object but ...	*/
	if
	(	(m_owner->isContainer () == 0) &&
		(m_owner->isFormBlock () == 0) &&
		(m_owner->isTabberPage() == 0) &&
		(m_owner->isComponent () == 0)
	)
		hide	|= HideMgmt ;

	return	hide	;
}

QRect	KBAttrGeom::mapGeometry
	(	const QRect	&ps,
		const QRect	&r
	)
	const
{
	int	x	= r.left  () ;
	int	y	= r.top   () ;
	int	w	= r.width () ;
	int	h	= r.height() ;

	switch (m_d.m_xmode)
	{
		case FMFloat	: x = ps.width () - x	  ; break ;
		case FMStretch	: w = ps.width () - x - w ; break ;
		default		:			  ; break ;
	}
	switch (m_d.m_ymode)
	{
		case FMFloat	: y = ps.height() - y	  ; break ;
		case FMStretch	: h = ps.height() - y - h ; break ;
		default		:			  ; break ;
	}

	return	QRect (x, y, w, h) ;
}

QRect	KBAttrGeom::mapGeometry
	(	const QRect	&ps
	)
	const
{
	return	mapGeometry (ps, QRect(m_d.m_x, m_d.m_y, m_d.m_w, m_d.m_h)) ;
}

/*  KBObject								*/
/*  convGeometry: Convert geometry to/from absolute			*/
/*  x, ...	: int, ...	: Geometry to covert			*/
/*  hint	: bool		: Return as hint rather than actual	*/
/*  (returns)	: QRect		: Converted geometry			*/

QRect	KBAttrGeom::convGeometry
	(	int		x,
		int		y,
		int		w,
		int		h,
		bool		hint
	)
	const
{
	KBNode	 *pnode	 = m_owner->getParent() ;
	KBObject *parent = pnode == 0 ? 0 : pnode->isObject() ;
	QSize	 ps	 ;

	if (parent != 0)
	{
		if (parent->isDynamic())
			return	QRect (x, y, w, h) ;

		if ((parent->getContainer() == 0) || hint)
			ps	= parent->sizeHint     () ;
		else	ps	= parent->getContainer()->getTopSize() ;
	}
	else	ps	= m_ps	;

	if (ps.width() > 0)
		switch (m_d.m_xmode)
		{
			case FMFloat	: x = ps.width () - x	  ; break ;
			case FMStretch	: w = ps.width () - x - w ; break ;
			default		:			  ; break ;
		}

	if (ps.height() > 0)
		switch (m_d.m_ymode)
		{
			case FMFloat	: y = ps.height() - y	  ; break ;
			case FMStretch	: h = ps.height() - y - h ; break ;
			default		:			  ; break ;
		}

	return	QRect (x, y, w, h) ;
}

/*  KBAttrGeom								*/
/*  geometry	: Get screen geometry					*/
/*  hint	: bool		: Return as hint rather than actual	*/
/*  (returns)	: QRect		: Screen geometry			*/

QRect	KBAttrGeom::geometry
	(	bool		hint
	)
{
	DPRINTF
	((	"KBAttrGeom::geometry: hint=%d: %s\n",
		hint,
		_TEXT(convGeometry(m_d.m_x, m_d.m_y, m_d.m_w, m_d.m_h, hint))
	))	;

	return	convGeometry
		(	m_d.m_x,
			m_d.m_y,
			m_d.m_w,
			m_d.m_h,
			hint
		)	;
}

/*  KBAttrGeom								*/
/*  geometry	: Get screen geometry with offset			*/
/*  offset	: const QPoint & : Offset				*/
/*  hint	: bool		: Return as hint rather than actual	*/
/*  (returns)	: QRect		 : Screen geometry			*/

QRect	KBAttrGeom::geometry
	(	const QPoint	&offset,
		bool		hint
	)
{
	return	convGeometry
		(	m_d.m_x + offset.x(),
			m_d.m_y + offset.y(),
			m_d.m_w,
			m_d.m_h,
			hint
		)	;
}

/*  KBAttrGeom								*/
/*  setMask	: Set visibility mask					*/
/*  mask	: uint		: Mask					*/
/*  (returns)	: void		:					*/

void	KBAttrGeom::setMask
	(	uint		mask
	)
{
	m_d.m_mask	= mask ;
}

/*  KBAttrGeom								*/
/*  mask	: Get visibility mask					*/
/*  (returns)	: uint		: Mask					*/

uint	KBAttrGeom::mask () const
{
	return	m_d.m_mask | hidden () ;
}

KBAttrGeom::
	OverFlow KBAttrGeom::overflow () const
{
	return	m_d.m_overflow	;
}


/*  KBAttrGeom								*/
/*  printAttr	: Print attribute					*/
/*  attrText	: QString &	: Result text for attribute output	*/
/*  nodeText	: QString &	: Result text for node output		*/
/*  indent	: int		: Indent depth for node output		*/
/*  flat	: bool		: Generate flatened version		*/
/*  (returns)	: void		:					*/

void	KBAttrGeom::printAttr
	(	QString		&attrText,
		QString		&nodeText,
		int		indent,
		bool		flat
	)
{
	if ((getFlags() & (KAF_HIDDEN|KAF_USER)) != 0)
		return	;

	if (flat)
	{
		QRect	r = m_owner->realGeometry () ;

		addAttrText (attrText, "x",  r.x     (), true) ;
		addAttrText (attrText, "y",  r.y     (), true) ;
		addAttrText (attrText, "w",  r.width (), true) ;
		addAttrText (attrText, "h",  r.height(), true) ;

		addAttrText (attrText, "ox", m_d.m_x,    true) ;
		addAttrText (attrText, "oy", m_d.m_y,    true) ;
		addAttrText (attrText, "ow", m_d.m_w,    true) ;
		addAttrText (attrText, "oh", m_d.m_h,    true) ;

		if (m_d.m_manage == KBAttrGeom::MgmtDynamic)
		{
			addAttrText (attrText, "xmode",     KBAttrGeom::FMFixed,	true) ;
			addAttrText (attrText, "ymode",     KBAttrGeom::FMFixed,	true) ;
			addAttrText (attrText, "manage",    KBAttrGeom::MgmtStatic,	true) ;
		}
		else
		{
			addAttrText (attrText, "xmode",     m_d.m_xmode,  		true) ;
			addAttrText (attrText, "ymode",     m_d.m_ymode,  		true) ;
			addAttrText (attrText, "manage",    m_d.m_manage, 		true) ;
		}

		addAttrText (attrText, "overflow", m_d.m_overflow, true) ;
		return	;
	}

	/* XYWH and modes are always output, even if zero. Not		*/
	/* significant but makes the XML files a bit easier to read.	*/
	addAttrText (attrText, "x",	    m_d.m_x,     true) ;
	addAttrText (attrText, "y",	    m_d.m_y,     true) ;
	addAttrText (attrText, "w",	    m_d.m_w,     true) ;
	addAttrText (attrText, "h",	    m_d.m_h,     true) ;
	addAttrText (attrText, "xmode",     m_d.m_xmode, true) ;
	addAttrText (attrText, "ymode",     m_d.m_ymode, true) ;

	addAttrText (attrText, "manage",    m_d.m_manage == MgmtDynamic ? 2 : 1) ;
	addAttrText (attrText, "m_rows",    m_d.m_nRows   ) ;
	addAttrText (attrText, "m_cols",    m_d.m_nCols   ) ;
	addAttrText (attrText, "m_margin",  m_d.m_margin, true) ;
	addAttrText (attrText, "m_spacing", m_d.m_spacing,true) ;

	addAttrText (attrText, "galign",    m_d.m_align	  ) ;
	addAttrText (attrText, "minw",	    m_d.m_minw	  ) ;
	addAttrText (attrText, "maxw",	    m_d.m_maxw	  ) ;
	addAttrText (attrText, "minh",	    m_d.m_minh	  ) ;
	addAttrText (attrText, "maxh",	    m_d.m_maxh	  ) ;
	addAttrText (attrText, "bychars",   m_d.m_byChars ) ;

	addAttrText (attrText, "overflow",  m_d.m_overflow) ;

	dumpGridSetup (nodeText, indent, m_rowSetup, m_d.m_nRows, "row") ;
	dumpGridSetup (nodeText, indent, m_colSetup, m_d.m_nCols, "col") ;
}

QString	KBAttrGeom::getValue () const
{
	return	asText (m_d, mask()) ;
}

/*  KBAttrGeom								*/
/*  isHidden	: Check whether attribute should be hidden		*/
/*  (returns)	: bool		: True to hide				*/

bool	KBAttrGeom::isHidden () const
{
	return	mask() == HideAll ;
}

void	KBAttrGeom::setGeometry
	(	const QRect	&rect
	)
{
	QRect	r = convGeometry
		    (	rect.x	    (),
			rect.y	    (),
			rect.width  (),
			rect.height (),
		    	false
		    ) ;

	DPRINTF
	((	"KBAttrGeom::setGeometry: %s -> %s\n",
		_TEXT(rect),
		_TEXT(r)
	))	;


	m_d.m_x	= r.x	  () ;
	m_d.m_y	= r.y	  () ;
	m_d.m_w	= r.width () ;
	m_d.m_h	= r.height() ;
}

#if	! __KB_RUNTIME

KBAttrItem
	*KBAttrGeom::getAttrItem ()
{
	return	new KBAttrGeomItem (this) ;
}

#endif

/*  KBAttrGeom								*/
/*  numRows	: Return number of rows					*/
/*  asSet	: bool		: True to force value as set		*/
/*  (returns)	: uint		: NUmber of rows			*/

uint	KBAttrGeom::numRows
	(	bool		asSet
	)
	const
{
	/* Normally, return zero if we are not using dynamic management	*/
	/* except we can force the value via the argument.		*/
	return	(m_d.m_manage == MgmtDynamic) || asSet ? m_d.m_nRows : 0 ;
}

uint	KBAttrGeom::numCols
	(	bool		asSet
	)
	const
{
	return	(m_d.m_manage == MgmtDynamic) || asSet ? m_d.m_nCols : 0 ;
}

/*  KBAttrGeom								*/
/*  set		: Set geometry						*/
/*  x, ...	: int, ...	: Values				*/
/*  (returns)	: void		:					*/

void	KBAttrGeom::set
	(	int			x,
		int			y,
		int			w,
		int			h
	)
{
	if (x	 != XYNoChange) m_d.m_x	   = x ;
	if (y	 != XYNoChange) m_d.m_y	   = y ;
	if (w	 != XYNoChange) m_d.m_w	   = w ;
	if (h 	 != XYNoChange) m_d.m_h	   = h ;

	if (m_monitor != 0) m_monitor->setText (2, getValue()) ;
}

/*  KBAttrGeom								*/
/*  set		: Set float modes					*/
/*  xmode, ...	: FloatMode, ... : Values				*/
/*  (returns)	: void		 :					*/

void	KBAttrGeom::set
	(	FloatMode		xmode,
		FloatMode		ymode
	)
{
	if (xmode != FMNoChange) m_d.m_xmode = xmode ;
	if (ymode != FMNoChange) m_d.m_ymode = ymode ;

	if (m_monitor != 0) m_monitor->setText (2, getValue()) ;
}

/*  KBAttrGeom								*/
/*  set		: Set management					*/
/*  dynamic	: MgmtMode	: Mode					*/
/*  nRows	: uint		: Number of rows			*/
/*  nCols	: uint		: Number of columns			*/
/*  (returns)	: void		:					*/

void	KBAttrGeom::set
	(	MgmtMode		dynamic,
		uint			nRows,
		uint			nCols
	)
{
	m_d.m_manage	= dynamic	;
	m_d.m_nRows	= nRows		;
	m_d.m_nCols	= nCols		;
	setupRowColSetup() ;

	if (m_monitor != 0) m_monitor->setText (2, getValue()) ;
}


/*  KBAttrGeom								*/
/*  set		: Set management					*/
/*  dynamic	: MgmtMode	: Mode					*/
/*  nRows	: uint		: Number of rows			*/
/*  nCols	: uint		: Number of columns			*/
/*  margin	: int		: Layout margin				*/
/*  spacing	: int		: Layout spacing			*/
/*  (returns)	: void		:					*/

void	KBAttrGeom::set
	(	MgmtMode		dynamic,
		uint			nRows,
		uint			nCols,
		int			margin,
		int			spacing
	)
{
	m_d.m_manage	= dynamic	;
	m_d.m_nRows	= nRows		;
	m_d.m_nCols	= nCols		;
	m_d.m_margin	= margin	;
	m_d.m_spacing	= spacing	;
	setupRowColSetup() ;

	if (m_monitor != 0) m_monitor->setText (2, getValue()) ;
}

/*  KBAttrGeom								*/
/*  setLimits	: Get size limits					*/
/*  minw, ...	: int		: Limits				*/
/*  byChars	: bool		: Limits are in character usits		*/
/*  (returns)	: void		:					*/

void	KBAttrGeom::setLimits
	(	int			minw,
		int			maxw,
		int			minh,
		int			maxh,
		bool			byChars
	)
{
	m_d.m_minw	= minw		;
	m_d.m_maxw	= maxw		;
	m_d.m_minh	= minh		;
	m_d.m_maxh	= maxh		;
	m_d.m_byChars	= byChars	;
}

void	KBAttrGeom::setAlignment
	(	Qt::AlignmentFlags	align
	)
{
	m_d.m_align	= align		;
}


/*  KBAttrGeom								*/
/*  asText	: Get description for use in property dialogs		*/
/*  d		: KBAttrGeomValues & : Geometry information		*/
/*  mask	: uint		     : Hide mask to be used		*/
/*  (returns)	: QString	     : Description			*/

QString	KBAttrGeom::asText
	(	const KBAttrGeomValues	&d,
		uint			mask
	)
{
	QStringList	dim	;
	QStringList	mode	;
	QString		range	;

	if ((mask & KBAttrGeom::HideX)     == 0) dim .append(QString::number(d.m_x)) ;
	if ((mask & KBAttrGeom::HideY)     == 0) dim .append(QString::number(d.m_y)) ;
	if ((mask & KBAttrGeom::HideW)     == 0) dim .append(QString::number(d.m_w)) ;
	if ((mask & KBAttrGeom::HideH)     == 0) dim .append(QString::number(d.m_h)) ;
	if ((mask & KBAttrGeom::HideXMode) == 0) mode.append(TR(modeToText(d.m_xmode))) ;
	if ((mask & KBAttrGeom::HideYMode) == 0) mode.append(TR(modeToText(d.m_ymode))) ;

	if ((mask & KBAttrGeom::HideRange) == 0)
		range	= QString ("(%1:%2,%3:%4)%5")
				.arg(d.m_minw)
				.arg(d.m_maxw)
				.arg(d.m_minh)
				.arg(d.m_maxh)
				.arg(d.m_byChars ? TR(" (chars)") : QString::null)
				;


	QString	res	= (dim .count() > 0 ? QString("(%1)").arg(dim .join(",")) : QString::null) +
			  " "	+
			  range +
			  " "	+
			  (mode.count() > 0 ? QString("(%1)").arg(mode.join(",")) : QString::null) ;

	if ((d.m_mask & KBAttrGeom::HideMgmt) == 0)
		if (d.m_manage == KBAttrGeom::MgmtDynamic)
			res	+= QString(TR(" Dynamic: %1*%2")).arg(d.m_nRows).arg(d.m_nCols) ;

	return	res.simplifyWhiteSpace () ;
}

/*  KBAttrGeom								*/
/*  setFont	: Set font - calculate width and height			*/
/*  font	: const QFont &	: Font					*/
/*  (returns)	: void		:					*/

void	KBAttrGeom::setFont
	(	const QFont	&font
	)
{
	QFontMetrics fm (font) ;

	m_charWidth  = fm.size(0, "IWX").width() / 3 ;
	m_charHeight = fm.lineSpacing() ;

	if (m_charWidth  == 0) m_charWidth = 10 ;
	if (m_charHeight == 0) m_charWidth = 14 ;
}

void	KBAttrGeom::setOverFlow
	(	OverFlow	overflow
	)
{
	m_d.m_overflow = overflow	;
}

/*  KBAttrGeom								*/
/*  fontWidth	: Get font width in pixels				*/
/*  (returns)	: uint		: Width					*/

uint	KBAttrGeom::fontWidth () const
{
	/* Actually, if we are not using character units, return the	*/
	/* height as one, so that multiplying up the height value	*/
	/* leaves it unchanged.						*/
	if (m_d.m_byChars)
	{
		if (m_charWidth == 0)
			((KBAttrGeom *)this)->setFont (QApplication::font()) ;

		return	m_charWidth	;
	}

	return	1 ;
}

/*  KBAttrGeom								*/
/*  fontHeight	: Get font height in pixels				*/
/*  (returns)	: uint		: Height				*/

uint	KBAttrGeom::fontHeight () const
{
	/* Similarly to width above ...					*/
	if (m_d.m_byChars)
	{
		if (m_charHeight == 0)
			((KBAttrGeom *)this)->setFont (QApplication::font()) ;

		return	m_charHeight	;
	}

	return	1 ;
}

/*  KBAttrGeom								*/
/*  minimumWidth : Get minimum width in pixels				*/
/*  (returns)	 : uint		: Maximum height			*/

uint	KBAttrGeom::minimumWidth () const
{
	return	m_d.m_minw * fontWidth () ;
}

/*  KBAttrGeom								*/
/*  maximumWidth : Get maximum width in pixels				*/
/*  (returns)	 : uint		: Maximum height			*/

uint	KBAttrGeom::maximumWidth () const
{
	return	m_d.m_maxw * fontWidth () ;
}

/*  KBAttrGeom								*/
/*  minimumHeight: Get minimum height in pixels				*/
/*  (returns)	 : uint		: Maximum height			*/

uint	KBAttrGeom::minimumHeight () const
{
	return	m_d.m_minh * fontHeight() ;
}

/*  KBAttrGeom								*/
/*  maximumHeight: Get maximum height in pixels				*/
/*  (returns)	 : uint		: Maximum height			*/

uint	KBAttrGeom::maximumHeight () const
{
	return	m_d.m_maxh * fontHeight() ;
}

void	KBAttrGeom::setupRowColSetup ()
{
	while ((int)m_rowSetup.count() < m_d.m_nRows)
		m_rowSetup.append (KBGridSetup(KBOptions::getMinCellHeight(), 0)) ;
	while ((int)m_colSetup.count() < m_d.m_nCols)
		m_colSetup.append (KBGridSetup(KBOptions::getMinCellWidth (), 0)) ;
}

void	KBAttrGeom::setRowSetup
	(	uint		index,
		int		spacing,
		int		stretch
	)
{
	if ((int)index < m_d.m_nRows)
		m_rowSetup[index] = KBGridSetup (spacing, stretch) ;
}

void	KBAttrGeom::setColSetup
	(	uint		index,
		int		spacing,
		int		stretch
	)
{
	if ((int)index < m_d.m_nCols)
		m_colSetup[index] = KBGridSetup (spacing, stretch) ;
}

/*  KBAttrGeom								*/
/*  insertRow	: Inserting a row					*/
/*  index	: uint		: Row index				*/
/*  (returs)	: void		:					*/

void	KBAttrGeom::insertRow
	(	uint		index
	)
{
	DPRINTF
	((	"KBAttrGeom::insertRow: row=%d h=%d\n",
		index,
		KBOptions::getMinCellHeight()
	))	;
	m_d.m_nRows += 1  ;
	m_rowSetup.insert (m_rowSetup.at(index), KBGridSetup(KBOptions::getMinCellHeight(), 0)) ;
	printGridSetup    (m_rowSetup) ;
}

/*  KBAttrGeom								*/
/*  insertCol	: Inserting a column					*/
/*  index	: uint		: Column index				*/
/*  (returs)	: void		:					*/

void	KBAttrGeom::insertCol
	(	uint		index
	)
{
	DPRINTF
	((	"KBAttrGeom::insertCol: col=%d w=%d\n",
		index,
		KBOptions::getMinCellWidth()
	))	;
	m_d.m_nCols += 1  ;
	m_colSetup.insert (m_colSetup.at(index), KBGridSetup(KBOptions::getMinCellWidth (), 0)) ;
	printGridSetup    (m_colSetup) ;
}

/*  KBAttrGeom								*/
/*  removeRow	: Removing a row					*/
/*  index	: uint		: Row index				*/
/*  (returs)	: void		:					*/

void	KBAttrGeom::removeRow
	(	uint		index
	)
{
	m_d.m_nRows -= 1  ;
	m_rowSetup.remove (m_rowSetup.at(index)) ;
	printGridSetup    (m_rowSetup) ;
}

/*  KBAttrGeom								*/
/*  removeCol	: Removing a column					*/
/*  index	: uint		: Column index				*/
/*  (returs)	: void		:					*/

void	KBAttrGeom::removeCol
	(	uint		index
	)
{
	m_d.m_nCols -= 1  ;
	m_colSetup.remove (m_colSetup.at(index)) ;
	printGridSetup    (m_colSetup) ;
}

/*  KBAttrGeom								*/
/*  extendGrid	: Extend grid to specified size				*/
/*  cols	: uint		: Minimum number of columns		*/
/*  rows	: uint		: Minimum number of rows		*/
/*  (returs)	: void		:					*/

void	KBAttrGeom::extendGrid
	(	uint		cols,
		uint		rows
	)
{
	DPRINTF
	((	"KBAttrGeom::extendGrid: (%d,%d) -> (%d,%d)\n",
		m_d.m_nCols,
		m_d.m_nRows,
		cols,
		rows
	))	;

	if ((int)cols > m_d.m_nCols)
	{
		DPRINTF
		((	"KBAttrGeom::extendGrid: EXTEND COLS %d->%d (w=%d\n)\n",
			cols,
			m_d.m_nCols,
			KBOptions::getMinCellWidth()
		))	;

		while ((int)cols > m_d.m_nCols)
		{
			m_colSetup.append (KBGridSetup(KBOptions::getMinCellWidth (), 0)) ;
			m_d.m_nCols += 1 ;
		}

	}

	if ((int)rows > m_d.m_nRows)
	{
		DPRINTF
		((	"KBAttrGeom::extendGrid: EXTEND ROWS %d->%d (h=%d)\n",
			rows,
			m_d.m_nRows,
			KBOptions::getMinCellHeight()
		))	;

		while ((int)rows > m_d.m_nRows)
		{
			m_rowSetup.append (KBGridSetup(KBOptions::getMinCellHeight(), 0)) ;
			m_d.m_nRows += 1 ;
		}
	}
}

/*  KBAttrGeom								*/
/*  setExtraAttribs							*/
/*		: Hook to set extended values from attributes		*/
/*  attribs	: QDict<QString> & : Attributes				*/
/*  (returns)	: bool		   :					*/

bool	KBAttrGeom::setExtraAttribs
	(	const QDict<QString>	&aDict
	)
{
	QString	*type	 = aDict.find   ("type"   ) ;
	int	index	 = getAttrValue (aDict, "index",   -1) ;
	int	spacing	 = getAttrValue (aDict, "spacing",  0) ;
	int	stretch	 = getAttrValue (aDict, "stretch",  0) ;

	DPRINTF
	((	"KBAttrGeom::setExtraAttribs: [%s][%d][%d][%d]\n",
		type->latin1(),
		index,
		spacing,
		stretch
	))	;

	if ((type != 0) && (index >= 0))
	{
		if (*type == "row")
		{
			setRowSetup (index, spacing, stretch) ;
			return true ;
		}
		if (*type == "col")
		{
			setColSetup (index, spacing, stretch) ;
			return true ;
		}
	}

	return	false	;
}

/*  KBAttrGeom								*/
/*  saveParentSize							*/
/* 		: Save parent object size				*/
/*  ps		: const QSize &	: Size					*/
/*  (returns)	: void		:					*/

void	KBAttrGeom::saveParentSize
	(	const QSize	&ps
	)
{
	/* This is called when replicting an object without a parent	*/
	/* for copy and paste: see the KBObject constructors.		*/
	m_ps	= ps	;
}
