/***************************************************************************
    file	         : kb_attrimagedlg.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/


#include	<errno.h>
#include	<qimage.h>
#include	<qgroupbox.h>
#include	<qfileinfo.h>
#include	<qlabel.h>


#include	"rk_hbox.h"
#include	"rk_gridbox.h"
#include	"rk_combobox.h"
#include	"rk_lineedit.h"
#include	"rk_pushbutton.h"

#include	"kb_classes.h"
#include	"kb_attr.h"
#include	"kb_graphic.h"
#include	"kb_docroot.h"
#include	"kb_dbdociter.h"

#if		__KB_AUTOMOC
#include 	"kb_attrimagedlg.moc"
#else
#include 	"kb_attrimagedlg.h"
#endif

#include	"tk_filedialog.h"


#define		__DEBUG_ATTRIMAGEDLG	(0)
#if		__DEBUG_ATTRIMAGEDLG
#include	"kb_dprintf.h"
#endif


static	QString	importImageLoad
	(	const QString		&file,
		KBDBInfo		*dbInfo,
		const QString		&svName,
		KBError			&pError	
	)
{
	/* Load the image file into memory, whence it can be stored as	*/
	/* for any other document.					*/
	QFile  imgFile (file) ;

	if (!imgFile.open (IO_ReadOnly))
	{
		pError	= KBError
			  (	KBError::Error,
				QString  (TR("Cannot open \"%1\"")).arg(file),
				strerror (errno),
				__ERRLOCN
			  )	;
		return	QString::null	;
	}

	QByteArray  data = imgFile.readAll() ;

	/* We will store the image using the final part of the path	*/
	/* with the extension split off ...				*/
	QFileInfo info (file) ;

	KBLocation dest
		   (	dbInfo,
			"graphic",
			svName,
			info.baseName (),
			info.extension(false)
		   )	;

	if (!dest.save
		(	QString::null,
			QString::null,
			data.data(),
			data.size(),
			pError
		))
		return	QString::null	;


	return	info.fileName() ;
}

/*  importImage	: Import an image into the system			*/
/*  dbInfo	: KBDBInfo *	  : Database information object		*/
/*  svName	: const QString & : Server name				*/
/*  name	: QString &	  : Return image name			*/
/*  pError	: KBError &	  : Return error			*/
/*  (returns)	: bool		  : Success				*/

LIBKBASE_API
	bool	importImage
	(	KBDBInfo	*dbInfo,
		const QString	&svName,
		QString		&name,
		KBError		&pError
	)
{
	/* Start with the file open dialog, with a filter set that	*/
	/* covers the extensions handled by Qt (found dynamically, this	*/
	/* adapts to however Qt was built).				*/
	extern	QString	imageFmtList (QStrList) ;
	static	int	lastFilter   = -1	;
	static	QString	lastDir	     ;

	if (lastDir.isEmpty()) lastDir = "."	;

	KBFileDialog fDlg 
		     (	lastDir,
			imageFmtList (QImageIO::inputFormats()),
			qApp->activeWindow(),
			"loadimage",
			true
		     )	;

	fDlg.setMode	  (KBFileDialog::ExistingFile) ;
	fDlg.setCaption   ("Load image ....") ;

	if (lastFilter >= 0)
		fDlg.setCurrentFilter (lastFilter) ;

	/* If the user cancels then that is not an error, but pass the	*/
	/* name back as null to signal such.				*/
	if (!fDlg.exec ())
	{
		name	= QString::null ;
		return true ;
	}


	QString	file	= fDlg.selectedFile () ;

	lastFilter = fDlg.currentFilterIndex () ;
	lastDir	   = QFileInfo(file).dirPath (true) ;

	name	= importImageLoad
		  (	file,
			dbInfo,
			svName,
			pError
		  )	;

	return	!name.isNull()	;
}

/*  importImages: Import images into the system				*/
/*  dbInfo	: KBDBInfo *	  : Database information object		*/
/*  svName	: const QString & : Server name				*/
/*  pError	: KBError &	  : Return error			*/
/*  (returns)	: bool		  : Success				*/

LIBKBASE_API
	bool	importImages
	(	KBDBInfo	*dbInfo,
		const QString	&svName,
		KBError		&pError
	)
{
	/* Start with the file open dialog, with a filter set that	*/
	/* covers the extensions handled by Qt (found dynamically, this	*/
	/* adapts to however Qt was built).				*/
	extern	QString	imageFmtList (QStrList) ;

	KBFileDialog fDlg 
		     (	".",
			imageFmtList (QImageIO::inputFormats()),
			qApp->activeWindow(),
			"loadimage",
			true
		     )	;

	fDlg.setMode	  (KBFileDialog::ExistingFiles) ;
	fDlg.setCaption   ("Load images ....") ;

	/* If the user cancels then that is not an error.		*/
	if (!fDlg.exec ())
	{
		return true ;
	}

	QStringList files = fDlg.selectedFiles() ;
	for (uint idx = 0 ; idx < files.count() ; idx += 1)
	{
		QString	name	= importImageLoad
		  		  (	files[idx],
					dbInfo,
					svName,
					pError
				  )	;
		if (name.isNull()) return false ;
	}

	return	true	;
}

/*  ------------------------------------------------------------------  */

/*  KBAttrImageBaseDlg							*/
/*  KBAttrImageBaseDlg							*/
/*		: Constructor for image selection/load dialog		*/
/*  parent	: QWidget *		: Parent widget			*/
/*  attr	: KBAttr *		: Associated attribute		*/
/*  item	: KBAttrItem *		: Associated item		*/
/*  attrDict	: QDict<KBAttrItem> &	: All attributes		*/
/*  (returns)	: KBImageDlg		:				*/

KBAttrImageBaseDlg::KBAttrImageBaseDlg
	(	QWidget			*parent,
		KBAttr			*attr,
		KBAttrItem		*item,
		QDict<KBAttrItem>	&attrDict
	)
	:
	KBAttrDlg	(parent, attr, item, attrDict)
{
}

void	KBAttrImageBaseDlg::makeSelectors
	(	RKBox		*layout,
		uint		count,
		const char	**legends
	)
{
	RKVBox	  *layMain   = new RKVBox    (layout    ) ; 
	RKGridBox *layGrid   = new RKGridBox (2, layMain) ;

	/* Setup a row for each image required (eq., for pushbuttons	*/
	/* there will be the normal and the rollover images). Each row	*/
	/* gets a groupbox, with a legend if any are avaialble.		*/
	m_count	= count	;
	for (uint idx = 0 ; idx < m_count ; idx += 1)
	{
		QGroupBox    *gb = new QGroupBox
					(	2,
						Qt::Vertical,
						legends == 0 ?
							QString::null :
							TR(legends[idx]),
						layGrid
					)	;

		RKLineEdit   *le = new RKLineEdit  (gb) ;
		RKComboBox   *cb = new RKComboBox  (gb) ;
		RKHBox	     *dh = new RKHBox	   (gb) ;
		dh->addFiller() ;
		RKPushButton *db = new RKPushButton(TR("Preview"), dh) ;

		m_images  .append (le) ;
		m_lists   .append (cb) ;
		m_shows   .append (db) ;

		/* All comboboxes are connected to the same slot. We	*/
		/* will use the "sender()" method to find out which.	*/
		connect	(cb, SIGNAL(activated(int)), SLOT(slotListActive())) ;
		connect (db, SIGNAL(clicked  ()),    SLOT(slotPreview   ())) ;
	}

	if ((m_count % 2) != 0) new QWidget (layGrid) ;

	m_previewGroup	   = new QGroupBox
				 (	1,
					Qt::Vertical,
					QString::null,
					layGrid
				 )	;

	m_previewImage	   = new QLabel (m_previewGroup) ;

	RKHBox	     *hbox = new RKHBox	      (layGrid) ;
	hbox->addFiller () ;
	RKVBox	     *vbox = new RKVBox	      (hbox   ) ;
	vbox->addFiller () ;

	RKPushButton *load = new RKPushButton (TR("Load graphic"), vbox) ;
	connect (load, SIGNAL(clicked()), SLOT(slotClickLoad())) ;

	m_previewGroup->setTitle (" ") ;
}


/*  KBAttrImageBaseDlg							*/
/*  loadImageList: Load list of images in server			*/
/*  (returns)	 : void		:					*/

void	KBAttrImageBaseDlg::loadImageList ()
{
	KBNode		 *node	 = m_item->attr()->getOwner() ;
	const KBLocation &locn	 = node->getDocRoot()->getDocLocation() ;
	KBDBDocIter	 docIter (true) ;
	KBError		 error	 ;

	for (uint idx = 0 ; idx < m_count ; idx += 1)
	{
		m_lists.at(idx)->clear	   ()   ;
		m_lists.at(idx)->insertItem("") ;
	}

	DPRINTF
	((	"KBAttrImageBaseDlg::loadImageList: loading %d lists\n",
		m_count
	))	;

	if (docIter.init
		(	locn.dbInfo(),
			locn.server(),
			"graphic",
			"*",
			error
		))
	{
		QString	name	;
		QString	stamp	;

		while (docIter.getNextDoc (name, stamp))
		{
			DPRINTF
			((	"KBAttrImageBaseDlg::loadImageList: got [%s]\n",
				(cchar *)name
			))	;

			for (uint idx = 0 ; idx < m_count ; idx += 1)
				m_lists.at(idx)->insertItem (name) ;
		}
	}
	else	error.DISPLAY() ;
}

/*  KBAttrImageBaseDlg							*/
/*  slotListActive: User selects image from list			*/
/*  (returns)	  : void	:					*/

void	KBAttrImageBaseDlg::slotListActive ()
{
	for (uint idx = 0 ; idx < m_count ; idx += 1)
		if (m_lists.at(idx) == sender())
		{	m_images.at(idx)->setText (m_lists.at(idx)->currentText()) ;
			break	;
		}
}

void	KBAttrImageBaseDlg::slotPreview ()
{
	for (uint idx = 0 ; idx < m_count ; idx += 1)
		if (m_shows.at(idx) == sender())
		{	previewImage (m_images.at(idx)->text()) ;
			break	;
		}
}

void	KBAttrImageBaseDlg::previewImage
	(	const QString	&image
	)
{
	QStringList	iparts	= QStringList::split ('.', image) ;

	DPRINTF
	((	"KBAttrImageBaseDlg::displayImage: [%s][%s]\n",
		(cchar *)iparts[0],
		(cchar *)iparts[1]
	))	;

	KBError		error	;
	KBNode		*node	= m_item->attr()->getOwner() ;
	QPixmap		pixmap	= KBAttrImage::pixmapFromLocation
				  (	node->getDocRoot(),
					iparts[0],
					iparts[1],
					error
				  )	;


	m_previewGroup->setTitle  (image ) ;
	m_previewImage->setPixmap (pixmap) ;
}

/*  KBAttrImageBaseDlg							*/
/*  slotClickLoad : User requests to load an image			*/
/*  (returns)	  : void	:					*/

void	KBAttrImageBaseDlg::slotClickLoad ()
{
	KBNode		 *node	= m_item->attr()->getOwner() ;
	const KBLocation &locn	= node->getDocRoot()->getDocLocation() ;
	QString		 name	;
	KBError		 error	;

	/* Import the image. Report errors and check for the case where	*/
	/* the user has just cancelled the file dialog.			*/
	if (!importImage
		(	locn.dbInfo(),
			locn.server(),
			name,
			error
		))
	{
		error.DISPLAY() ;
		return		;
	}

	if (!name.isNull()) loadImageList () ;
}

/*  ------------------------------------------------------------------	*/

/*  KBAttrImageDlg							*/
/*  KBAttrImageDlg							*/
/*		: Constructor for image selection/load dialog		*/
/*  parent	: QWidget *		: Parent widget			*/
/*  attr	: KBAttrImage *		: Associated attribute		*/
/*  item	: KBAttrItem *		: Associated item		*/
/*  attrDict	: QDict<KBAttrItem> &	: All attributes		*/
/*  (returns)	: KBImageDlg		:				*/

KBAttrImageDlg::KBAttrImageDlg
	(	QWidget			*parent,
		KBAttrImage		*attr,
		KBAttrItem		*item,
		QDict<KBAttrItem>	&attrDict
	)
	:
	KBAttrImageBaseDlg (parent, attr, item, attrDict)
{
	RKVBox	*layMain   = new RKVBox (parent) ;
	setTopWidget(layMain) ;

	makeSelectors	(layMain, attr->numImages(), attr->legends ()) ;
	layMain->addFiller () ;
}
/*  KBAttrImageDlg							*/
/*  init	: Initialise for editing				*/
/*  value	: const QString & : Value				*/
/*  (returns)	: bool  	  : Always false, no modal dialog	*/

bool	KBAttrImageDlg::init
	(	const QString	&value
	)
{
	QStringList ibits = QStringList::split (';', value) ;
	while (ibits.size() < m_count + 1) ibits.append (QString::null) ;

	loadImageList	   () ;

	for (uint idx1 = 0 ; idx1 < m_count ; idx1 += 1)
	{
		QString	name = ibits[idx1] ;
		m_images.at(idx1)->setText	  (name) ;
		m_lists .at(idx1)->setCurrentText (name) ;
	}

	return	false	;
}

/*  KBAttrImageDlg							*/
/*  value	: Get value						*/
/*  (returns)	: void		:					*/

QString	KBAttrImageDlg::value ()
{
	QStringList list ;

	for (uint idx = 0 ; idx < m_count ; idx += 1)
		list.append (m_images.at(idx)->text()) ;

	return	list.join (";") ;
}

