/***************************************************************************
    file	         : kb_dbspecification.cpp
    copyright            : (C) 1999,2000,2001,2002,2003,2004 by Mike Richardson
			   (C) 2001,2002,2003,2004 by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/


#include	<qobject.h>

#include	"kb_dbspecification.h"
#include	"kb_file.h"


/*  KBDBSpecification							*/
/*  KBDBSpecification							*/
/*		: Constructor for database specification class		*/
/*  (returns)	: KBDBSpecification :					*/

KBDBSpecification::KBDBSpecification ()
{
}

/*  KBDBSpecification							*/
/*  init	: Initialise from loaded DOM document			*/
/*  (returns)	: bool		  : Sucess				*/

bool	KBDBSpecification::init ()
{
	/* Scan the top level of the specification for "database"	*/
	/* elements, and index them by their "name" attribute.		*/
	QDomNodeList nl	 = m_specification.documentElement().elementsByTagName("database") ;
	QStringList  dbl ;

	for (uint idx = 0 ; idx < nl.count() ; idx += 1)
	{
		QDomElement elem = nl.item(idx).toElement() ;
		if (!elem.isNull())
			m_databases.insert (elem.attribute("name"), elem) ;
	}


	return	true	;
}

/*  KBDBSpecification							*/
/*  loadText	: Load specification from text				*/
/*  text	: const QString & : Specification text			*/
/*  (returns)	: bool		  : Sucess				*/

bool	KBDBSpecification::loadText
	(	const QString	&text
	)
{
	m_databases.clear () ;

	if (!m_specification.setContent (text))
	{
		m_error	= KBError
			  (	KBError::Error,
				TR("Cannot parse specification"),
				QString::null,
				__ERRLOCN
			  )	;
		return	false	;
	}

	return	init ()	;
}

/*  KBDBSpecification							*/
/*  loadFile	: Load specification from file				*/
/*  fileName	: const QString & : Specification file name		*/
/*  (returns)	: bool		  : Sucess				*/

bool	KBDBSpecification::loadFile
	(	const QString	&fileName
	)
{
	m_databases.clear () ;

	KBFile	file	(fileName) ;
	if (!file.open (IO_ReadOnly))
	{
		m_error	= file.lastError () ;
		return	false	;
	}

	if (!m_specification.setContent (&file))
	{
		m_error	= KBError
			  (	KBError::Error,
				QString	(TR("Cannot parse \"%1\"")).arg(file.name()),
				QString::null,
				__ERRLOCN
			  )	;
		return	false	;
	}

	return	init ()	;
}

/*  KBDBSpecification							*/
/*  databases	: Get list of databases in specification		*/
/*  (returns)	: QStringList	: Database list				*/

QStringList
	KBDBSpecification::databases ()
{
	QStringList  dbl ;

	for
	(	QMap<QString,QDomElement>::ConstIterator iter = m_databases.begin() ;
		iter != m_databases.end() ;
		++iter
	)
	{
		dbl.append (iter.key()) ;
	}

	return	dbl	;
}

/*  KBDBSpecification							*/
/*  description	: Get description for named database			*/
/*  database	: const QString & : Database name			*/
/*  (returns)	: QString	  : Database description		*/

QString	KBDBSpecification::description
	(	const QString	&database
	)
{
	return	m_databases[database].elementsByTagName("description").item(0).toElement().text() ;
}

/*  KBDBSpecification							*/
/*  url		: Get URL for named database				*/
/*  database	: const QString & : Database name			*/
/*  (returns)	: QString	  : Database URL			*/

QString	KBDBSpecification::url
	(	const QString	&database
	)
{
	return	m_databases[database].elementsByTagName("url").item(0).toElement().text().stripWhiteSpace() ;
}

/*  KBDBSpecification							*/
/*  asfiles	: See if database stores forms, etc., as files		*/
/*  database	: const QString & : Database name			*/
/*  (returns)	: QString	  : Database URL			*/

bool	KBDBSpecification::asfiles
	(	const QString	&database
	)
{
	return	m_databases[database].attribute("asfiles").toInt() ;
}

/*  KBDBSpecification							*/
/*  validForDBs	: Get list of servers for which database is valid	*/
/*  database	: const QString & : Database name			*/
/*  (returns)	: QStringList	  : List of servers			*/

QStringList
	KBDBSpecification::validForDBs
	(	const QString	&database
	)
{
	QDomNodeList nl	 = m_databases[database].elementsByTagName("validdb") ;
	QStringList  vdb ;

	for (uint idx = 0 ; idx < nl.count() ; idx += 1)
		vdb.append (nl.item(idx).toElement().attribute("db")) ;

	return	vdb	;
}
