/***************************************************************************
    file	         : kb_dcopobject.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/


#include	"kb_classes.h"
#include	"kb_type.h"
#include	"kb_value.h"
#include	"kb_block.h"
#include	"kb_display.h"
#include	"kb_docroot.h"
#include	"kb_report.h"
#include	"kb_writer.h"
#include	"kb_dcopobject.h"

#define		__DEBUG_DCOPOBJECT	(0)
#if		__DEBUG_DCOPOBJECT
#include	"kb_dprintf.h"
#endif


/*  KBDCOPObject							*/
/*  KBDCOPObject: Constructor for DCOP wrapper				*/
/*  node	: KBNode *	: Owing node				*/
/*  name	: cchar *	: Object name				*/
/*  (returns)	: KBDCOPObject	:					*/

KBDCOPObject::KBDCOPObject
	(	KBNode		*node,
		cchar		*name
	)
	:
	RKDCOPObject	(name),
	m_node		(node)
{
}

/*  KBDCOPObject							*/
/*  process	: Process a request					*/
/*  func	: const QCString &	: Function requested		*/
/*  data	: const QByteArray &	: Encoded arguments		*/
/*  replyType	: QCString &		: Return reply value type	*/
/*  replayData	: QByteArray &		: Return result(s)		*/
/*  (returns)	: bool			: Success			*/

bool	KBDCOPObject::process
	(	const QCString		&func,
		const QByteArray	&data,
		QCString		&replyType,
		QByteArray		&replyData
	)
{
	DPRINTF
	((	"KBDCOPObject::process: [%s]\n",
		func.data()
	))	;

	/* widgetTree()							*/
	/* ------------							*/
	/* Dump the widget tree. Only works of the node is a block; it	*/
	/* would normally be used at the form/report level.		*/
	if (func == "widgetTree(int)")
	{
		KBBlock	*block	= m_node->isBlock() ;
		if (block == 0) return false ;

		extern	QString	textWidgetTree (QWidget *, uint, int, int = 0) ;

		QDataStream	args	(data,	    IO_ReadOnly ) ;
		QDataStream	reply	(replyData, IO_WriteOnly) ;
		int		flags	;

		if (args.atEnd()) return false;
		args >> flags	;

		replyType	= "QString"	;
		reply	<< textWidgetTree (block->getDisplay()->getTopWidget(), 0, -1, flags) ;
		return	true	;
	}

	/* close(int)							*/
	/* ----------							*/
	/* Applies to forms and reports.				*/
	if (func == "close(int)")
	{
		KBDocRoot *root	= m_node->getDocRoot() ;
		if (root == 0) return false ;

		QDataStream	args	(data,	    IO_ReadOnly ) ;
		QDataStream	reply	(replyData, IO_WriteOnly) ;
		int		rc	;

		if (args.atEnd()) return false;
		args >> rc	;

		replyType	= "void"	;
		reply	<< true	;
		m_node->getDocRoot()->doRequestClose(rc) ;
		return	true	;
	}


	/* executeScript(QString)					*/
	/* ----------------------					*/
	/* Execute an arbitrary script. This is treated as an event	*/
	/* function at the node level.					*/
	if (func == "executeScript(QString)")
	{
		QDataStream	args	(data,	    IO_ReadOnly ) ;
		QDataStream	reply	(replyData, IO_WriteOnly) ;
		QString		code	;

		if (args.atEnd()) return false;
		args >> code	;

		replyType	= "QString"	;

		KBEvent	script	(m_node, "executeScript", code.latin1()) ;
		KBValue	resval	;

		script.execute	(resval, 0, 0) ;
		reply	<< resval.getRawText() ;
		return	true	;
	}

	/* attribute(QString,QString)					*/
	/* --------------------------					*/
	/* Get the value of an attribute on a specified object.		*/
	if (func == "attribute(QString,QString)")
	{
		QDataStream	args	(data,	    IO_ReadOnly ) ;
		QDataStream	reply	(replyData, IO_WriteOnly) ;
		QString		objName	;
		QString		attrName;

		if (args.atEnd()) return false;
		args >> objName	;
		if (args.atEnd()) return false;
		args >> attrName;

		replyType	= "QString"	;

		KBNode	*target	= m_node->getNamedNode (objName, false) ;
		if (target == 0) return false	;
		KBAttr		*attr	= target->getAttr(attrName) ;
		if (attr   == 0) return false	;

		reply	<< attr->getValue ()	;
		return	true	;
	}

	/* widgetTree()							*/
	/* ------------							*/
	/* Dump the KBWriter output for a report.			*/
	if (func == "writerData()")
	{
		KBReport *report = m_node->isReport() ;
		if (report == 0) return false ;

		QDataStream	reply	(replyData, IO_WriteOnly) ;

		replyType	= "QString"	;
		reply	<< report->getWriter()->describe() ;
		return	true	;
	}

	return	RKDCOPObject::process (func, data, replyType, replyData) ;
}

/*  KBDCOPObject							*/
/*  interfaces	: Return interface list					*/
/*  (returns)	: QCStringList	: Interface list			*/

QCStringList
	KBDCOPObject::interfaces()
{
	QCStringList ifaces = RKDCOPObject::interfaces() ;
	ifaces += "kbNode" ;
	return ifaces ;
}

/*  KBDCOPObject							*/
/*  functions	: Return function list					*/
/*  (returns)	: QCStringList	: Function list				*/

QCStringList
	KBDCOPObject::functions ()
{
	QCStringList funcs = RKDCOPObject::functions() ;

	/* Available functions depend on the exact type of node passed	*/
	/* to us (so users should not reply in the interface to 	*/
	/* specify available functions.					*/
	if (m_node->isObject  () != 0)
		funcs += normalizeFunctionSignature ("QString attribute(QString,QString)") ;

	if (m_node->isBlock   () != 0)
		funcs += normalizeFunctionSignature ("QString widgetTree(int)") ;

	if (m_node->isReport   () != 0)
		funcs += normalizeFunctionSignature ("QString writerData()") ;

	if (m_node->getDocRoot() != 0)
		funcs += normalizeFunctionSignature ("bool close(int)") ;

	funcs += normalizeFunctionSignature ("QString executeScript(QString)") ;

	return	funcs	;
}
