/***************************************************************************
    file	         : kb_geometry.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/

#include	<stdio.h>

#include	<qapplication.h>
#include	<qlayout.h>
#include	<qpainter.h>

#include	"kb_object.h"
#include	"kb_display.h"
#include	"kb_docroot.h"
#include	"kb_dialog.h"
#include	"kb_layoutitem.h"
#include	"kb_ascii.h"

#if		__KB_AUTOMOC
#include	"kb_geometry.moc"
#else		
#include	"kb_geometry.h"
#endif


#define		__DEBUG_GEOMETRY	(0)
#if		__DEBUG_GEOMETRY
#include	"kb_dprintf.h"
#endif


/*  KBGridLayout							*/
/*  KBGridLayout: Constructor for grid layout controller		*/
/*  widget	: QWidget *	: Parent widget				*/
/*  geom	: KBAttrGeom &	: Geometry information			*/
/*  displauy	: KBDisplay *	: Associated display object		*/
/*  (returns)	: KBGridLayout	:					*/

KBGridLayout::KBGridLayout
	(	QWidget			*parent,
		const KBAttrGeom	&geom,
		KBDisplay		*display
	)
	:
	QGridLayout	(parent, geom.numRows(false), geom.numCols(false)),
	m_parent	(parent),
	m_display  	(display)
{
	m_reactivate	= false ;

	init		(geom)	;
}

KBGridLayout::~KBGridLayout ()
{
	DPRINTF
	((	"KBGridLayout::~KBGridLayout: called\n"
	))	;

//	QLayoutIterator it = iterator() ;
//	QLayoutItem	*c ;
//	while ((c = it.current()) != 0 )
//	{
//		fprintf	(stderr, "KBGridLayout::~KBGridLayout: i=%p\n", (void *)c) ;
//		++it ;
//	}
}

/*  KBGridLayout							*/
/*  setRowColSetup							*/
/*		: Initialise row and column setup information		*/
/*  rowSetup	: KBGridSetupList &	: Row information		*/
/*  colSetup	: KBGridSetupList &	: Column information		*/
/*  (returns)	: void			:				*/

void	KBGridLayout::setRowColSetup
	(	const KBGridSetupList	&rowSetup,
		const KBGridSetupList	&colSetup
	)
{
	for (int row = 0 ; row < numRows() ; row += 1)
	{
		int	spacing	= 0 ;
		int	stretch	= 0 ;

		if (row < (int)rowSetup.count())
		{	spacing	= rowSetup[row].m_spacing ;
			stretch = rowSetup[row].m_stretch ;
		}

		DPRINTF
		((	"KBGridLayout::setRowColSetup: row %d: sp=%d,st=%d\n",
			row,
			spacing,
			stretch
		))	;
#if	QT_VERSION >= 0x030200
		setRowSpacing (row, spacing) ;
#endif
		setRowStretch (row, stretch) ;
	}

	for (int col = 0 ; col < numCols() ; col += 1)
	{
		int	spacing	= 0 ;
		int	stretch	= 0 ;

		if (col < (int)colSetup.count())
		{	spacing	= colSetup[col].m_spacing ;
			stretch = colSetup[col].m_stretch ;
		}

		DPRINTF
		((	"KBGridLayout::setRowColSetup: col %d: sp=%d,st=%d\n",
			col,
			spacing,
			stretch
		))	;
#if	QT_VERSION >= 0x030200
		setColSpacing (col, spacing) ;
#endif
		setColStretch (col, stretch) ;
	}

	m_reactivate = true   ;
	m_display->getObject()->getDocRoot()->doLayoutChanged() ;
}


/*  KBGridLayout							*/
/*  init	: General initialisation				*/
/*  geom	: KBAttrGeom &	: Geomtry information			*/
/*  (returns)	: void		:					*/

void	KBGridLayout::init
	(	const KBAttrGeom	&geom
	)
{
	int	margin	= geom.margin () ;
	int	spacing	= geom.spacing() ;

	setMargin 	(margin  < 0 ? KBDialog::marginHint () : margin ) ;
	setSpacing	(spacing < 0 ? KBDialog::spacingHint() : spacing) ;
	setRowColSetup	(geom.rowSetup(), geom.colSetup()) ;
}



/*  KBGridLayout							*/
/*  insertWidget: Place a child widget in the layout			*/
/*  item	: KBLayoutItem * : Layout item for widget in question	*/
/*  (returns)	: void		 :					*/

void	KBGridLayout::insertWidget
	(	KBLayoutItem		*item
	)
{
	QRect	rect	= item->designGeometry () ;

	DPRINTF
	((	"KBGridLayout::insertWidget: %p/%p : %s: %08x/%08x\n",
		(void *)item,
		(void *)item->widget(),
		_TEXT(rect),
		item->alignment() & Qt::AlignHorizontal_Mask,
		item->alignment() & Qt::AlignVertical_Mask

	))	;

	/* If the item is not already in place then arrange to hear	*/
	/* when it is destroyed and insert into the map. If it is	*/
	/* present then remove from the layout (prior to reinsertion).	*/
	if (m_gridMap.find (item->widget()) == 0)
	{
		connect
		(	item->widget(),
			SIGNAL	(destroyed	(QObject *)),
			SLOT	(childDestroyed	(QObject *))
		)	;

		m_gridMap.insert (item->widget(), item) ;
	}
	else
		removeItem (item) ;


	/* Insert the widget multi-cell or single as appropriate. Could	*/
	/* do it all as multi-cell (abliet 1x1) but maybe this is more	*/
	/* efficient.							*/
	if ((rect.width() > 1) || (rect.height() > 1))
		addMultiCell
		(	item,
			rect.y(),
			rect.y() + rect.height() - 1,
			rect.x(),
			rect.x() + rect.width () - 1,
			item->alignment()
		)	;
	else	addItem	(item, rect.y(), rect.x()) ;

	m_parent ->update   () ;
	m_display->getObject()->getDocRoot()->doLayoutChanged() ;

	m_reactivate = true    ;
}

/*  KBGridLayout							*/
/*  childDestroyed							*/
/*		: Catch destruction of child widget			*/
/*  object	: QObject *	: Destroyed object, actually a widget	*/
/*  (returns)	: void		:					*/

void	KBGridLayout::childDestroyed
	(	QObject		*object
	)
{
	DPRINTF
	((	"KBGridLayout::slotWidgetDestroyed: %p: %s\n",
		(void *)object,
		object->className()
	))	;
	m_gridMap.remove (object) ;

	QApplication::sendPostedEvents (0, QEvent::ChildRemoved) ;

	m_reactivate = true ;
	m_parent->update () ;
}

/*  KBGridLayout							*/
/*  getCell	: Get cell which contains specfied coordinate		*/
/*  p		: QPoint	: Coordinates				*/
/*  (returns)	: QPoint	: (row, col)				*/

QPoint	KBGridLayout::getCell
	(	QPoint		p
	)
{
	int	cx	= 0	;
	int	cy	= 0	;

	while (cx < numCols() - 1)
	{	if (p.x() < cellGeometry (0, cx + 1).x ()) break ;
		cx += 1	;
	}
	while (cy < numRows() - 1)
	{	if (p.y() < cellGeometry (cy + 1, 0).y()) break ;
		cy += 1	;
	}

	DPRINTF
	((	"KBGridLayout::getCell: (%d,%d) => (%d,%d)\n",
		p.x(), p.y(),
		cx,    cy
	))	;
	/* Note: The result is returned as (col,row) which is		*/
	/* consistent with (x,y) but is the opposite to the		*/
	/* QGridLayout (row,col) interface.				*/
	return	QPoint(cx, cy) ;
}

/*  KBGridLayout							*/
/*  setGeometry	: Set layout geometry					*/
/*  r		: const QRect &	: Geometry				*/
/*  (returns)	: void		:					*/

void	KBGridLayout::setGeometry
	(	const QRect	&r
	)
{
	QSize	sbu	= m_display->showBarUsed() ;
	QGridLayout::setGeometry
	(	QRect
		(	r.x	(),
			r.y	(),
			r.width () - sbu.width (),
			r.height() - sbu.height()
	)	)	;

	m_display->displayResized (r.size()) ;
}

/*  KBGridLayout							*/
/*  outlineCells: Draw cell outlines for design mode			*/
/*  paint	: QPainter &	: Painter				*/
/*  (returns)	: void		:					*/

void	KBGridLayout::outlineCells
	(	QPainter	&paint
	)
{
	/* Nasty hack: After a change is made in design mode which	*/
	/* changes the grid sizes, we need to call the ::activate()	*/
	/* method before outlining the cells, otherwise we still get	*/
	/* the old sizes.						*/
	if (m_reactivate)
	{
		qApp->processEvents()	;
		activate ()		;
		m_reactivate = false	;
	}

	DPRINTF
	((	"KBGridLayout::outlineCells: %d,%d\n",
		numRows(),
		numCols()
	))	;

	paint.setPen (QPen(Qt::DashLine)) ;

	for (int col = 1 ; col < numCols() ; col += 1)
	{
		int	x = cellGeometry(0, col).x() ;
		paint.drawLine (x, 0, x, mainWidget()->height()) ;
	}
	for (int row = 1 ; row < numRows() ; row += 1)
	{
		int	y = cellGeometry(row, 0).y() ;
		paint.drawLine (0, y, mainWidget()->width (), y) ;
	}
}

/*  KBGridLayout							*/
/*  layoutMap	: Get grid widget map					*/
/*  (returns)	: const KBLayoutMap &	: Map				*/

const KBLayoutMap &KBGridLayout::layoutMap ()
{
	/* This is used when recreating a layout for a change in the	*/
	/* number of rows or columns. 					*/
	return	m_gridMap ;
}

/*  ------------------------------------------------------------------	*/

/*  KBStaticLayout							*/
/*  KBStaticLayout							*/
/*		: Constructor for static (stretch/float) layouts	*/
/*  widget	: QWidget *	: Parent widget				*/
/*  geom	: KBAttrGeom &	: Geometry information			*/
/*  displauy	: KBDisplay *	: Associated display object		*/
/*  (returns)	: KBStaticLayout:					*/

KBStaticLayout::KBStaticLayout
	(	QWidget			*parent,
		const KBAttrGeom	&geom,
		KBDisplay		*display
	)
	:
	QLayout		(parent),
	m_parent	(parent),
	m_display	(display)
{
	m_invalid	= true	;
	m_gridItem	= 0	;
}

KBStaticLayout::~KBStaticLayout ()
{
	DPRINTF
	((	"KBStaticLayout::~KBStaticLayout: called\n"
	))	;
}

/*  KBGridLayout							*/
/*  insertWidget: Place a child widget in the layout			*/
/*  item	: KBLayoutItem * : Layout item for widget in question	*/
/*  (returns)	: void		 :					*/

void	KBStaticLayout::insertWidget
	(	KBLayoutItem		*item
	)
{
	DPRINTF
	((	"KBStaticLayout::insertWidget: %d: %p/%p %s [%s]\n",
		item->rtti(),
		(void *)item,
		(void *)item->widget(),
		_TEXT(geometry()),
		item->widget() == 0 ? "/NONE/" : item->widget()->className()
	))	;

	QWidget	*widget	= item->widget() ;

	/* If we don't know about the widget then insert it into the	*/
	/* map, arrange to hear if it is destroyed.			*/
	if (m_staticMap.find (widget) == 0)
	{
		connect
		(	widget,
			SIGNAL	(destroyed	(QObject *)),
			SLOT	(childDestroyed	(QObject *))
		)	;

		m_staticMap.insert (widget, item) ;

		if (item->rtti() == KBLayoutItem::RTTIGrid)
			m_gridItem = item ;
	}

	item->setGeometry (item->mapGeometry (geometry())) ;
	m_display->getObject()->getDocRoot()->doLayoutChanged() ;
}

/*  KBStaticLayout							*/
/*  getCell	: Get cell which contains specfied coordinate		*/
/*  p		: QPoint	: Coordinates				*/
/*  (returns)	: QPoint	: (row, col)				*/

QPoint	KBStaticLayout::getCell
	(	QPoint		p
	)
{
	/* In a static layout the position is returned unchanged. In	*/
	/* effect this method is mapping a physical position to a	*/
	/* logical position.						*/
	return	p	;
}

/*  KBStaticLayout							*/
/*  setGeometry	: Set layout geometry					*/
/*  r		: const QRect &	: Geometry				*/
/*  (returns)	: void		:					*/

void	KBStaticLayout::setGeometry
	(	const QRect	&rect
	)
{
	DPRINTF
	((	"KBStaticLayout::setGeometry: %*s : %s: %s %s->%s: g=%p\n",
		m_display->depth() * 4,
		"",
		m_display->getObject()->className(),
		_TEXT(rect),
		_TEXT(m_size),
		_TEXT(rect.size()),
		m_gridItem
	))	;
	if (rect.size() != m_size)
		QLayout::setGeometry (rect) ;

	m_invalid = false ;

	/* Apply the item-level geometry mapping first (which handles	*/
	/* stretch and float), then display-level mapping (which	*/
	/* handles scroller offsets). The result is applied to the	*/
	/* layout item.							*/
	if (m_gridItem == 0)
		for
		(	QPtrDictIterator<KBLayoutItem> iter (m_staticMap) ;
			iter.current() != 0 ;
			++iter
		)
		{
			KBLayoutItem	*item	= iter.current() ;

			DPRINTF
			((	"KBStaticLayout::setGeometry: %*s : setting: %s\n",
				m_display->depth() * 4,
				"",
				item->widget()->className()
			))	;
			item->setGeometry (item->mapGeometry (rect)) ;
		}
	else
		m_gridItem->setGeometry (m_gridItem->mapGeometry (rect)) ;


	m_size = rect.size() ;
	m_display->displayResized (m_size) ;
}

/*  KBStaticLayout							*/
/*  invalidate	: Force layout recalcuation on next call		*/
/*  (returns)	: void		:					*/

void	KBStaticLayout::invalidate ()
{
	m_invalid = true ;
}


QSize	KBStaticLayout::sizeHint () const
{
	return	m_display->getObject()->sizeHint() ;
}

/*  KBStaticLayout							*/
/*  childDestroyed							*/
/*		: Catch destruction of child widget			*/
/*  object	: QObject *	: Destroyed object, actually a widget	*/
/*  (returns)	: void		:					*/

void	KBStaticLayout::childDestroyed
	(	QObject		*object
	)
{
	DPRINTF
	((	"KBStaticLayout::slotWidgetDestroyed: %p: %s\n",
		(void *)object,
		object->className()
	))	;
	KBLayoutItem *item = m_staticMap.find (object) ;
	if (item != 0)
	{
		m_staticMap.remove (object) ;
		if (item == m_gridItem) m_gridItem = 0 ;
	}
}

QLayoutIterator
	KBStaticLayout::iterator ()
{
	return	QLayoutItem::iterator () ;
}

void	KBStaticLayout::addItem
	(	QLayoutItem		*item
	)
{
}


/*  ------------------------------------------------------------------	*/

/*  KBGeometry								*/
/*  KBGeometry	: Constructor for geometry management class		*/
/*  (returns)	: KBGeometry	:					*/

KBGeometry::KBGeometry	()
{
	m_widget	= 0	;
	m_display	= 0	;
	m_mgmtMode	= KBAttrGeom::MgmtUnknown ;
	m_grid		= 0	;
	m_static	= 0	;
}

KBGeometry::~KBGeometry	()
{
	DPRINTF
	((	"KBGeometry::~KBGeometry: called\n"
	))	;
}

/*  KBGeometry								*/
/*  init	: Initialise geometry management class			*/
/*  widget	: QWidget *	: Widget being managed			*/
/*  display	: KBDisplay *	: Associated display object		*/
/*  (returns)	: void		:					*/

void	KBGeometry::init
	(	QWidget			*widget,
		KBDisplay		*display
	)
{
	m_widget	= widget	;
	m_display	= display	;
	m_grid		= 0		;
	m_static	= 0		;

	const KBAttrGeom &geom = m_display->getObject()->attrGeom() ;

	m_mgmtMode = geom.manage()	;
	if (m_mgmtMode == KBAttrGeom::MgmtDynamic)
		m_grid	 = new KBGridLayout   (m_widget, geom, m_display) ;
	else	m_static = new KBStaticLayout (m_widget, geom, m_display) ;
}

/*  KBGeometry								*/
/*  setRowColSetup							*/
/*		: Initialise row and column setup information		*/
/*  rowSetup	: KBGridSetupList &	: Row information		*/
/*  colSetup	: KBGridSetupList &	: Column information		*/
/*  (returns)	: void			:				*/

void	KBGeometry::setRowColSetup
	(	const KBGridSetupList	&rowSetup,
		const KBGridSetupList	&colSetup
	)
{
	if (m_grid != 0)
		m_grid->setRowColSetup (rowSetup, colSetup) ;
}

/*  KBGeometry								*/
/*  insertWidget: Place a child widget in the layout			*/
/*  item	: KBLayoutItem * : Layout item for widget in question	*/
/*  (returns)	: void		 :					*/

void	KBGeometry::insertWidget
	(	KBLayoutItem		*item
	)
{
	if (m_grid   != 0)
	{
		m_grid  ->insertWidget (item) ;
		return	;
	}
	if (m_static != 0)
	{
		m_static->insertWidget (item) ;
		return	;
	}
}
/*  KBGeometry								*/
/*  resize	: Resize geometry					*/
/*  size	: const QSize &	: New size				*/
/*  (returns)	: void		:					*/

void	KBGeometry::resize
	(	const QSize	&size
	)
{
	DPRINTF
	((	"KBGeometry::resize: [%p] (%d,%d) [%p]\n",
		(void *)this,
		size.width (),
		size.height(),
		m_grid
	))	;

	if (m_grid   != 0)
		m_grid  ->setGeometry (QRect (0, 0, size.width(), size.height())) ;
	if (m_static != 0)
		m_static->setGeometry (QRect (0, 0, size.width(), size.height())) ;
}

/*  KBGeometry								*/
/*  isDynamic	: Check whether layout is dynamic or not		*/
/*  (returns)	: bool		: True if dynamic			*/

bool	KBGeometry::isDynamic ()
{
	return	m_grid	!= 0 ;
}

/*  KBGeometry								*/
/*  outlineCells: Draw cell outlines for design mode			*/
/*  paint	: QPainter &	: Painter				*/
/*  (returns)	: void		:					*/

void	KBGeometry::outlineCells
	(	QPainter	&paint
	)
{
	if (m_grid != 0)
		m_grid->outlineCells (paint) ;
}

/*  KBGeometry								*/
/*  getCell	: Get cell which contains specfied coordinate		*/
/*  p		: QPoint	: Coordinates				*/
/*  (returns)	: QPoint	: (row, col)				*/

QPoint	KBGeometry::getCell
	(	QPoint		p
	)
{
	if (m_grid   != 0)
		return	m_grid  ->getCell (p) ;

	if (m_static != 0)
		return	m_static->getCell (p) ;

	return	p ;
}

/*  KBGeometry								*/
/*  updateDynamic: Update dynamic layout stuff				*/
/*  geom	 : KBAttrGeom &	: Geometry and other information	*/
/*  (returns)	 : void		:					*/

void	KBGeometry::updateDynamic
	(	const KBAttrGeom	&geom
	)
{
	/* If the management mode is known (should always be the case)	*/
	/* and the mode is changing then delete the extant manager and	*/
	/* create the appropriate new one. No need to work about extant	*/
	/* controls since the management mode cannot be changed if	*/
	/* there are any.						*/
	if (m_mgmtMode != KBAttrGeom::MgmtUnknown)
		if (m_mgmtMode != geom.manage())
		{
			DELOBJ	(m_grid  ) ;
			DELOBJ	(m_static) ;

			m_mgmtMode = geom.manage() ;
			if (m_mgmtMode == KBAttrGeom::MgmtDynamic)
				m_grid	 = new KBGridLayout   (m_widget, geom, m_display) ;
			else	m_static = new KBStaticLayout (m_widget, geom, m_display) ;

			return	;
		}


	/* Mode known and unchanging. If it is static then there is	*/
	/* nothing to do.						*/
	if (geom.manage() == KBAttrGeom::MgmtStatic)
		return	;


	/* Must be dynamic; if no grid manager then create one. I think	*/
	/* this case cannot occur but lets be paranoid about it.	*/
	if (m_grid == 0)
		m_grid	 = new KBGridLayout   (m_widget, geom, m_display) ;


	/* Seems to be no way to change the number of rows and		*/
	/* columns, so if these change, delete the current layout and	*/
	/* create a new one, which we repopulate.			*/
	uint	nRows	= geom.numRows(false) ;
	uint	nCols	= geom.numCols(false) ;

	uint	oRows	= m_grid->numRows  () ;
	uint	oCols	= m_grid->numCols  () ;

	if ((nRows != oRows) || (nCols != oCols))
	{
		DPRINTF
		((	"KBGeometry::updateDynamic: grid=%p: (%d,%d) -> (%d,%d)\n",
			(void *)m_grid,
			oRows, oCols,
			nRows, nCols
		))	;

		KBLayoutMap layoutMap = m_grid->layoutMap() ;

		for
		(	QPtrDictIterator<KBLayoutItem> iter (layoutMap) ;
			iter.current() != 0 ;
			++iter
		)
			m_grid->removeItem (iter.current()) ;

		delete	m_grid	;
		m_grid	= new KBGridLayout (m_widget, geom, m_display) ;


		for
		(	QPtrDictIterator<KBLayoutItem> iter (layoutMap) ;
			iter.current() != 0 ;
			++iter
		)
			m_grid->insertWidget (iter.current()) ;

		DPRINTF
		((	"KBGeometry::updateDynamic: grid=%d\n",
			(void *)m_grid
		))	;

	}
	else	/* If there is no change to rows/cols then only need	*/
		/* to ensure margins and spacing and such are correct.	*/
		m_grid->init (geom) ;
}


/*  KBGeometry								*/
/*  sizeHint	: Get size hint for this display			*/
/*  (returns)	: QSize		: Size hint				*/

QSize	KBGeometry::sizeHint ()	const
{
	DPRINTF
	((	"KBGeometry::sizeHint: %s: %p\n",
		(cchar *)m_display->getObject()->getName(),
		(void  *)m_grid
	))	;
	if (m_grid   != 0)
		return	m_grid  ->sizeHint() ;

	if (m_static != 0)
		return	m_static->sizeHint() ;

	return	QSize(0, 0) ;
}

/*  KBGeometry								*/
/*  rodoLayout	: Recalculate layout					*/
/*  invalidate	: bool		: Invalidate first			*/
/*  (returns)	: void		:					*/

void	KBGeometry::redoLayout
	(	bool		invalidate
	)
{
	DPRINTF
	((	"KBGeometry::redoLayout: called: invalidate=%d\n",
		invalidate
	))	;

	if (m_grid   != 0)
	{
		if (invalidate)
			m_grid  ->invalidate () ;
		m_grid  ->activate   () ;
		return	;
	}

	/* For static layouts, simply call "setGeometry" with the	*/
	/* current geometry. This will move any wayward controls into	*/
	/* the right place.						*/
	if (m_static != 0)
	{
		if (invalidate)
			m_static->invalidate () ;
		m_static->setGeometry(m_static->geometry()) ;
		return	;
	}
}
