/***************************************************************************
    file	         : kb_grid.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/

#include	"kb_classes.h"
#include	"kb_type.h"
#include	"kb_value.h"
#include	"kb_location.h"
#include	"kb_rowmark.h"

#include	"kb_sizer.h"
#include	"kb_display.h"
#include	"kb_ctrl.h"

#include	"kb_formblock.h"
#include	"kb_qtgrid.h"
#include	"kb_nodereg.h"
#include	"kb_writer.h"
#include	"kb_layout.h"

#if		__KB_AUTOMOC
#include	"kb_grid.moc"
#else
#include	"kb_grid.h"
#endif

#include	"kb_sorter.h"

/*  KBGrid								*/
/*  KBGrid	: Consructor for grid header node			*/
/*  parent	: KBNode *	: Parent node				*/
/*  aList	: const QDict<QString> &				*/
/*				: List of attributes			*/
/*  ok		: bool *	: Success				*/
/*  (returns)	: KBGrid	:					*/

KBGrid::KBGrid
	(	KBNode			*parent,
		const QDict<QString>	&aList,
		bool			*ok
	)
	:
	KBObject	(parent, "KBGrid", 	aList),
	m_nosort	(this,	 "nosort",	aList)
{
	m_grid		= 0 	;
	m_sortCol	= -1	;
	m_sortAsc	= true	;
	m_rmOffset	= -1	;

#if	! __KB_RUNTIME
	if (ok != 0)
	{
		if (!KBNode::basePropertyDlg ("Grid"))
		{	delete this	;
			*ok	= false ;
			return	    ;
		}

		*ok	= true	;
	}
#endif
}

/*  KBGrid								*/
/*  KBGrid	: Consructor for label node				*/
/*  _parent	: KBNode *	: Parent node				*/
/*  _grid	: KBGrid *	: Extant grid header			*/
/*  (returns)	: KBGrid	:					*/

KBGrid::KBGrid
	(	KBNode		*_parent,
		KBGrid		*_grid
	)
	:
	KBObject	(_parent,		_grid),
	m_nosort	(this,	 "nosort",	_grid)
{
	m_grid		= 0	;
	m_sortCol	= -1	;
	m_sortAsc 	= true	;
	m_rmOffset	= -1	;
}

/*  KBGrid								*/
/*  ~KBGrid	: Desructor for label node				*/
/*  (returns)	:		:					*/

KBGrid::~KBGrid ()
{
}

/*  KBGrid								*/
/*  replicate	: Replicate this grid header				*/
/*  parent	: KBNode *	: Parent of replicant			*/
/*  (returns)	: KBNode *	: New grid header node			*/

KBNode	*KBGrid::replicate
	(	KBNode	*parent
	)
{
	return	replicateBelow (new KBGrid (parent, this)) ;
}

/*  KBGrid								*/
/*  buildDisplay: Build controle for this node				*/
/*  display	: KBDisplay *	: Owning display			*/
/*  (returns)	: void		:					*/

void	KBGrid::buildDisplay
	(	KBDisplay	*display
	)
{
	KBObject::buildDisplay (display) ;

	if (m_grid == 0)
	{
		m_grid	= new KBCtrlGrid (display, this, m_visItems) ;
		setControl (m_grid)  ;
	}

	m_grid->setCtrlGeometry (QRect(m_geom.x(), m_geom.y(), m_geom.w(), m_geom.h()), -1) ;
	m_grid->setVisible	(true) ;
}

/*  KBGrid								*/
/*  clearItems	: Clear all items from grid				*/
/*  all		: bool		: Clear all items list as well		*/
/*  (returns)	: void		:					*/

void	KBGrid::clearItems
	(	bool	all
	)
{
	if (m_grid != 0) m_grid->clearLabels () ;
	if (all	       ) m_allItems.clear    () ;

	m_visItems.clear () ;
}

/*  KBGrid								*/
/*  appendItem	: Append a new item to the grid				*/
/*  item	: KBItem *	: Item to be appended			*/
/*  all		: bool		: Add to all items list as well		*/
/*  (returns)	: void		:					*/

void	KBGrid::appendItem
	(	KBItem		*item,
		bool		all
	)
{
	if (m_grid != 0)
	{
		if (all) m_allItems.append (item) ;

		m_grid->appendLabel
		(	item ->getName  (),
			item ->geometry ().width(),
			!item->isRowMark()
		)	;
		m_visItems.append (item) ;
	}
}

/*  KBGrid								*/
/*  getItems	: Get list of items					*/
/*  ilist	: QPtrList<KBItem> & : List of items			*/
/*  (returns)	: void		  :					*/

void	KBGrid::getItems
	(	QPtrList<KBItem>	&ilist
	)
{
	for (uint idx = 0 ; idx < m_allItems.count() ; idx += 1)
		ilist.append (m_allItems.at(idx)) ;
}

/*  KBGrid								*/
/*  setOrder	: Set columns in specified order			*/
/*  ilist	: QPtrList<KBItem> & : Order list			*/
/*  (returns)	: void		  :					*/

void	KBGrid::setOrder
	(	QPtrList<KBItem>	&ilist
	)
{
	/* This is potentially dangerous, so check the arguments quite	*/
	/* carefully; make sure that exactly the same set of items is	*/
	/* passed as we already have.					*/
	if (ilist.count() != m_visItems.count())
	{
		KBError::EError
		(	TR("Grid ordering error"),
			TR("Incorrect number of grid items passed"),
			__ERRLOCN
		)	;
		return	;
	}

	LITER
	(	KBItem,
		ilist,
		item,

		if (m_visItems.find (item) < 0)
		{
			KBError::EError
			(	TR("Grid ordering error"),
				TR("Unexpected item specified"),
				__ERRLOCN
			)	;
			return	;
		}
	)

	LITER
	(	KBItem,
		m_visItems,
		item,

		if (ilist.find (item) < 0)
		{
			KBError::EError
			(	TR("Grid ordering error"),
				TR("Grid item not specified"),
				__ERRLOCN
			)	;
			return	;
		}
	)

	clearItems () ;

	for (uint idx = 0 ; idx < ilist.count() ; idx += 1)
	{
		KBItem	*item	= ilist.at(idx) ;
		appendItem (item) ;

		if (item->getTabOrder () != 0)
			item->setTabOrder (idx + 1) ;
	}

	m_grid->adjustItems (0) ;

	KBNavigator *navi = getNavigator  () ;
	if (navi != 0) navi->setupTabOrder() ;
}

/*  KBGrid								*/
/*  setColumnWidth: Set width of specified column			*/
/*  item	  : KBItem *	: Item correspoinding to column		*/
/*  width	  : uint	: Required width			*/
/*  (returns)	  : void		:				*/

void	KBGrid::setColumnWidth
	(	KBItem	*item,
		uint	width
	)
{
	if (width != 0)
	{
		int idx	= m_visItems.find (item) ;
		if (idx >= 0) m_grid->setColumnWidth (idx, width) ;
	}
}

/*  KBGrid								*/
/*  columnClicked: Hanle user column clock for sorting			*/
/*  colNo	 : int		: Column number				*/
/*  (returns)	 : void		:					*/

void	KBGrid::columnClicked
	(	int	colNo
	)
{
	if (m_sortCol != colNo)
	{
		m_sortCol = colNo	;
		m_sortAsc = true	;
	}
	else	m_sortAsc = !m_sortAsc	;

	columnSort () ;
}

/*  KBGrid								*/
/*  columnSort	: Sort by column if required				*/
/*  (returns)	: void		:					*/

void	KBGrid::columnSort ()
{
	if ((m_sortCol >= 0) && (m_sortCol < (int)m_visItems.count()))
	{
		getFormBlock()->sortByColumn (m_visItems.at(m_sortCol), m_sortAsc) ;
		m_grid->setSortIndicator (m_sortCol, m_sortAsc) ;
	}
}

/*  KBGrid								*/
/*  resetSort	: Reset ordering					*/
/*  (returns)	: void		:					*/

void	KBGrid::resetSort ()
{
	m_sortCol = -1	 ;
	m_sortAsc = true ;
	m_grid->setSortIndicator (-1, false) ;
}

/*  KBGrid								*/
/*  showAs	: Switch in or out of design mode			*/
/*  mode	: KB::ShowAs	: Design mode				*/
/*  (returns)	: void		:					*/

void	KBGrid::showAs
	(	KB::ShowAs	_showAs
	)
{
	/* Clear the default tab order list as the tab order may have	*/
	/* changed while in design mode.				*/
	KBObject::showAs  (_showAs) ;
	m_defTabOrder.clear () ;

	/* Find out how wide the rowmark width is, if there in one.	*/
	/* This will be the left-hand column and the width is needed to	*/
	/* offset the headers.						*/
	m_rmOffset = -1 ;
	LITER
	(	KBNode,
		getParent()->getChildren(),
		r,
		if (r->isRowMark() != 0)
		{
			m_rmOffset = r->isRowMark()->geometry().right() ;
			break	;
		}
	)

	if (m_rmOffset < 0) m_rmOffset = 0 ;
}

/*  KBGrid								*/
/*  orderByExpr	: Order columns by expression (or not ...)		*/
/*  byExp	: bool		: True to order by expression		*/
/*  (returns)	: void		:					*/

void	KBGrid::orderByExpr
	(	bool	byExpr
	)
{
	/* If there is only a single column then there is nothing to	*/
	/* reorder ....							*/
	if (m_visItems.count() <= 1)
		return	;

	/* If the default tab order list is empty, then this is the	*/
	/* first call (maybe since leaving design mode), in which case	*/
	/* record the tab order - an int dictionary is a convenient way	*/
	if (m_defTabOrder.count() == 0)
		LITER
		(	KBItem,
			m_visItems,
			i,
			m_defTabOrder.insert (i->getTabOrder(), i)
		)

	/* Now get a list of items orders by expression or by default	*/
	/* tab order. We use a sorter list in either case. For order	*/
	/* by expression we insert labelled with the expression; for	*/
	/* tab order label with the tab order (fixed so that lexical	*/
	/* ordering gives the right result.				*/
	KBSorter<KBItem> sorter ;

	if (byExpr)
	{
		LITER
		(	KBItem,
			m_visItems,
			item,
			sorter.append (item->getAttrVal("expr"), item)
		)
	}
	else
	{
		QIntDictIterator<KBItem> iter (m_defTabOrder) ;
		KBItem	*item ;

		while ((item = iter.current()) != 0)
		{
			sorter.append (QString().sprintf("%04ld", iter.currentKey()), item) ;
			iter	+= 1 ;
		}
	}

	sorter.sort () ;

	/* Clear the items list and reinsert the items in the new	*/
	/* order, and adjust the column widths to match. The item tab	*/
	/* order values are updated .....				*/
	clearItems () ;


	for (uint idx = 0 ; idx < sorter.count() ; idx += 1)
	{
		KBItem	*item	= sorter.at(idx) ;
		appendItem (item) ;

		if (item->getTabOrder () != 0)
			item->setTabOrder (idx + 1) ;
	}

	m_grid->adjustItems (0) ;

	/* ... and then we fix up the tab order so that we get		*/
	/* left-to-right tabbing.					*/
	KBNavigator *navi = getNavigator  () ;
	if (navi != 0) navi->setupTabOrder() ;
}

/*  KBGrid								*/
/*  setItemsVisible							*/
/*		: Show or hide columns					*/
/*  visible	: QValueList<bool> : Column visibility flags		*/
/*  pack	: bool		   : Pack down visible columns		*/
/*  (returns)	: void		   :					*/

void	KBGrid::setItemsVisible
	(	QValueList<bool> &visible,
		bool
	)
{
	clearItems () ;

	for (uint slot = 0 ; slot < m_allItems.count() ; slot += 1)
	{
		KBItem	*item	= m_allItems.at(slot) ;
		bool	vis	= visible[slot]	      ;

		item->setAllVisible (vis ) ;
		if (vis) appendItem (item) ;
	}

	
	m_grid->adjustItems (0) ;
}

/*  KBGrid								*/
/*  setItemsEnabled							*/
/*		: Enable or disable columns				*/
/*  enabled	: QValueList<bool> : Column enable flags		*/
/*  (returns)	: void		   :					*/

void	KBGrid::setItemsEnabled
	(	QValueList<bool> &enabled
	)
{
	for (uint slot = 0 ; slot < m_allItems.count() ; slot += 1)
		m_allItems.at(slot)->setAllEnabled (enabled[slot]) ;
}

uint	KBGrid::rowMarkOffset ()
{
	return	m_rmOffset	;
}

void	KBGrid::adjustItems ()
{
	if (m_grid != 0) m_grid->adjustItems (0) ;
}

bool	KBGrid::noSort ()
{
	return	m_nosort.getBoolValue() ;
}

#if	! __KB_RUNTIME

/*  KBGrid								*/
/*  propertyDlg	: Show property dialog					*/
/*  iniAttr	: cchar *	: Initial attribute			*/
/*  (returns)	: bool		: Success				*/

bool	KBGrid::propertyDlg
	(	cchar	*iniAttr
	)
{
	if (!KBNode::basePropertyDlg ("Grid Header", iniAttr)) return false ;

	m_grid->setCtrlGeometry
	(	QRect
		(	0,
			0,
			getBlock()->geometry().width(),
			geometry().height()
		),
		-1
	)	;

	setChanged () ;
	getLayout  ()->addSizer (getSizer(), false) ;

	return	true  ;
}
#endif

NEWNODE(Grid, (cchar *)0, 0, KNF_FORM|KNF_BLOCK|KNF_STATIC)
