/***************************************************************************
    file	         : kb_highlighter.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/

#include	<qfile.h>
#include	<qdom.h>
#include	<qdir.h>
#include	<qlabel.h>

#include	"kb_highlighter.h"
#include	"kb_textedit.h"
#include	"kb_locator.h"


static	QStringList	*hlNames ;



/*  KBHLSection	:							*/
/*  color	: Get colour for region					*/
/*  idx		: uint		: Region				*/
/*  (returns)	: QColor	: Colour				*/

QColor	KBHLSection::color
	(	uint		idx
	)
{
	if (idx >= m_highlight.count())
		return	QColor	() ;
	return	m_highlight.at(idx)->color () ;
}

/*  KBHLSection	:							*/
/*  font	: Get font for region					*/
/*  idx		: uint		: Region				*/
/*  kbsh	: KBSyntaxHighlighter *					*/
/*				: Syntax highlighter			*/
/*  (returns)	: QFont		: Font					*/

QFont	KBHLSection::font
	(	uint			idx,
		KBSyntaxHighlighter	*kbsh
	)
{
	if (idx < m_highlight.count())
		switch (m_highlight.at(idx)->style())
		{
			case KBHLHighlight::Bold	: return kbsh->bold  () ;
			case KBHLHighlight::Italic	: return kbsh->italic() ;
			default				: break	;
		}

	return	kbsh->normal() ;
}

/*  ------------------------------------------------------------------	*/

class	KBMarkRegion
{
public	:
	int		m_markAt	;
	int		m_markLength	;

	inline	KBMarkRegion ()
		:
		m_markAt	(0),
		m_markLength	(0)
	{
	}

	inline	KBMarkRegion
		(	QRegExp	&regexp,
			int	bit
		)
		:
		m_markAt	(regexp.pos(bit)),
		m_markLength	(regexp.cap(bit).length())
	{
//		fprintf
//		(	stderr,
//			"KBMarkRegion::KBMarkRegion: %d for %d\n",
//			m_markAt,
//			m_markLength
//		)	;
	}
}	;

/*  ------------------------------------------------------------------	*/

KBHLHighlight::KBHLHighlight
	(	const QDomElement	&elem
	)
{
	QString	style	= elem.attribute ("style") ;

	m_name	= elem.attribute ("name") ;
	m_color	= QColor (elem.attribute ("color")) ;

	if	(style == "b") m_style = Bold	;
	else if	(style == "i") m_style = Italic	;
	else		       m_style = Normal ;
}

/*  ------------------------------------------------------------------	*/

KBHLSection::KBHLSection
	(	const QDomElement		&elem,
		const QDict<KBHLHighlight>	&highlights
	)
{
	m_esolp	    = elem.attribute ("esolp").toInt() ;
	m_reFrom    = QRegExp (elem.attribute ("from" )) ;
	m_reUntil   = QRegExp (elem.attribute ("until")) ;

	QStringList  hl = QStringList::split(';', elem.attribute ("highlight")) ;
	for (uint idx = 0 ; idx < hl.count() ; idx += 1)
	{	KBHLHighlight *hp = highlights.find (hl[idx]) ;
		if (hp != 0) m_highlight.append(hp) ;
	}
}

/*  ------------------------------------------------------------------	*/

QDict<KBHLHighlighter>	*KBHLHighlighter::m_highlighters ;

KBHLHighlighter::KBHLHighlighter
	(	const QString	&specName
	)
{
	QFile	specFile (specName) ;

//	fprintf
//	(	stderr,
//		"KBHLHighlighter::KBHLHighlighter: [%s]\n",
//		(const char *)specName
//	)	;

	if (!specFile.open (IO_ReadOnly))
	{
		fprintf
		(	stderr,
			"KBHLHighlighter::KBHLHighlighter: cannot open [%s]\n",
			(const char *)specName
		)	;
		return	;
	}

	QDomDocument doc ;
	if (!doc.setContent (&specFile))
        {
		fprintf
		(	stderr,
			"KBHLHighlighter::KBHLHighlighter: cannot parse [%s]\n",
			(const char *)specName
		)	;
		return	;
	}

	QDomElement root = doc.documentElement() ;

	for
	(	QDomNode node	= root.firstChild () ;
		!node.isNull () ;
		node		= node.nextSibling()
	)
	{
		QDomElement elem = node.toElement () ;
		if (elem.isNull()) continue ;

		if (elem.tagName() == "highlight")
		{
			KBHLHighlight *hl = new KBHLHighlight (elem) ;
			m_highlights.insert (hl->name(), hl) ;
			continue ;
		}

		if (elem.tagName() == "section")
		{
			KBHLSection *s = new KBHLSection (elem, m_highlights) ;
			if (s->ok()) m_sections.append (s) ;
			continue ;
		}
	}

	if (m_highlighters == 0)
	{
		m_highlighters	= new QDict<KBHLHighlighter> ;
		hlNames		= new QStringList ;
	}

	m_highlighters->insert (root.attribute ("name"), this) ;
	hlNames->append (root.attribute ("name")) ;
}

/*  KBHLHighlighter							*/
/*  highlight	: Highlight paragraph					*/
/*  text	: const QString &	: Paragraph text		*/
/*  esolp	: int			: End state of last paragraph	*/
/*  kbsh	: KBSyntaxHighlighter *	: Syntax highlighter		*/
/*  (returns)	: int			: End state of this paragraph	*/

int	KBHLHighlighter::highlight
	(	const QString		&text,
		int			esolp,
		KBSyntaxHighlighter	*kbsh
	)
{
	int	markLast	= 0 ;
	int	offset		= 0 ;

	/* Main loop iterates over the text. The "markLast" variable	*/
	/* contains the last point up to which we formatted the text.	*/
	while (offset < (int)text.length())
	{
		KBHLSection			*bestSection	= 0 ;
		int				matchAt		= 0 ;
		int				matchEnd	= 0 ;
		QValueList<KBMarkRegion>	matches		;

//		fprintf
//		(	stderr,
//			"KBHLHighlighter::highlight: offset=%d esolp=%d\n",
//			offset,
//			esolp
//		)	;

		/* Normally, "esolp" is zero, but if we are being	*/
		/* called as the continuation of the previous paragraph	*/
		/* and this is the first time round the loop, "esolp"	*/
		/* is value for the section that is being continued.	*/
		if (esolp > 0)
		{
			for
			(	KBHLSection *s = m_sections.first () ;
				s != 0 ;
				s  = m_sections.next  ()
			)
			{
				if (s->esolp() == esolp)
				{
					bestSection = s ;
					break	;
				}
			}

			esolp	= 0 ;

			/* Found the continued section. All continued	*/
			/* sections should have an "until" pattern. Set	*/
			/* up as if we matched "from" at the very start	*/
			/* of the paragraph.				*/
			if (bestSection != 0)
				matches.append (KBMarkRegion()); 
		}

		/* This is the usual case. It can also occur if a	*/
		/* paragraph continuation did not find the section that	*/
		/* is being continued (which is really an error).	*/
		if (bestSection == 0)
		{
			int	bestNext	= 0x7fffff ;

			/* Loop over the sections, finding the one	*/
			/* which gives the earliest match. If two give	*/
			/* the same match then the first in the list	*/
			/* will be used.				*/
			for
			(	KBHLSection *s = m_sections.first () ;
				s != 0 ;
				s  = m_sections.next  ()
			)
			{
				int found = s->from().search(text, offset) ;
				if (found >= 0)
					if (found < bestNext)
					{	bestNext    = found ;
						bestSection = s	;
					}
			}

			/* If nothing matches then drop out of the main	*/
			/* scanning loop ....				*/
			if (bestSection == 0)
				break	;

			/* .... else note the matched range and the	*/
			/* section to be formatted.			*/
			matchAt		= bestNext ;
			matchEnd	= matchAt  + bestSection->from().cap(0).length() ;
	
			matches.clear () ;
			for (int idx = 1 ; idx <= bestSection->from().numCaptures() ; idx += 1)
				matches.append (KBMarkRegion (bestSection->from(), idx)) ;
		}

		uint		mcount	  = matches.count() ;
		KBMarkRegion	&lastMark = matches[mcount - 1] ;

//		fprintf
//		(	stderr,
//			"KBHLHighlighter::highlight: bs=%p nm=%d\n",
//			(void *)bestSection,
//			mcount
//		)	;

		if (mcount == 0)
			break	;

		/* If there is any space from the end of the last 	*/
		/* format then reset it to the default format.		*/
		if (matches[0].m_markAt > markLast)
		{
//			fprintf
//			(	stderr,
//				"KBHLHighlighter::highlight: default %d for %d\n",
//				markLast,
//				matches[0].m_markAt - markLast
//			)	;

			kbsh->setFormat
			(	markLast,
				matches[0].m_markAt - markLast,
				kbsh->normal(),
				Qt::black
			)	;
		}

		if (mcount > 1)
			for (uint idx = 0 ; idx < mcount - 1 ; idx += 1)
			{
				KBMarkRegion	&thisMark = matches[idx] ;

//				fprintf
//				(	stderr,
//					"KBHLHighlighter::highlight: %d: set %d for %d\n",
//					idx,
//					thisMark.m_markAt,
//					thisMark.m_markLength
//				)	;

				kbsh->setFormat
				(	thisMark.m_markAt,
					thisMark.m_markLength,
					bestSection->font (idx, kbsh),
					bestSection->color(idx)
				)	;

				if (thisMark.m_markAt + thisMark.m_markLength > markLast)
					markLast = thisMark.m_markAt + thisMark.m_markLength ;
			}


		/* See if the section has an "until" pattern. If so and	*/
		/* it is *not* matched then format to the end of the	*/
		/* paragraph and then return the continuation. If found	*/
		/* then extend the match and format to this position.	*/
		if (!bestSection->until().pattern().isEmpty())
		{
			int next = bestSection->until().search(text, matchEnd) ;
			if (next < 0)
			{
				kbsh->setFormat
				(	lastMark.m_markAt,
					text.length() - lastMark.m_markAt,
					bestSection->font (mcount - 1, kbsh),
					bestSection->color(mcount - 1)
				)	;
				return	bestSection->esolp() ;
			}

			matchEnd	      = next + bestSection->until().cap(0     ).length() ;
			lastMark.m_markLength = next + bestSection->until().cap(mcount).length() - lastMark.m_markAt ;

//			fprintf
//			(	stderr,
//				"KBHLHighlighter::highlight: until %d/%d\n",
//				matchEnd,
//				lastMark.m_markLength
//			)	;
		}

//		fprintf
//		(	stderr,
//			"KBHLHighlighter::highlight: %d: %d for %d\n",
//			mcount - 1,
//			lastMark.m_markAt,
//			lastMark.m_markLength
//		)	;

		kbsh->setFormat
		(	lastMark.m_markAt,
			lastMark.m_markLength,
			bestSection->font (mcount - 1, kbsh),
			bestSection->color(mcount - 1)
		)	;

		if (lastMark.m_markAt + lastMark.m_markLength > markLast)
			markLast = lastMark.m_markAt + lastMark.m_markLength ;
		offset	 = matchEnd	;
	}

	/* If there is any left-over text then reset it to the default	*/
	/* format.							*/
	if (markLast < (int)text.length())
		kbsh->setFormat
		(	markLast,
			text.length() - markLast,
			kbsh->normal(),
			Qt::black
		)	;

	return	0 ;
}


KBHLHighlighter
	*KBHLHighlighter::find
	(	const QString	&highlight
	)
{
	return	m_highlighters == 0 ? 0 : m_highlighters->find (highlight) ;
}


/*  ------------------------------------------------------------------  */

/*  KBSyntaxHighlighter							*/
/*  KBSyntaxHighlighter							*/
/*		: Constructor for syntax highlighter			*/
/*  editor	: KBTextEdit *	  : Editor to be highlighted		*/
/*  highlight	: const QString & : Highlight scheme			*/
/*  font	: const QFont &	  : Font to be used			*/
/*  (returns)	:KBSyntaxHighlighter					*/

KBSyntaxHighlighter::KBSyntaxHighlighter
	(	KBTextEdit	*editor,
		const QString	&highlight,
		const QFont	&font
	)
	:
	QSyntaxHighlighter	(editor->textEdit())
{
	static	bool	first	= true ;
	if (first)
	{	loadHighlights () ;
		first	= false	  ;
	}
	
	m_highlighter	= KBHLHighlighter::find (highlight) ;
	setFont	(font)	;
	if (m_highlighter == 0)
		fprintf
		(	stderr,
			"KBSyntaxHighlighter::KBSyntaxHighlighter: [%s] not found\n",
			(const char *)highlight
		)	;
}

/*  KBSyntaxHighlighter							*/
/*  setFont	: Set font						*/
/*  font	: const QFont &	: Font to be used			*/
/*  (returns)	: void		:					*/

void	KBSyntaxHighlighter::setFont
	(	const QFont	&font
	)
{
	m_normal	= font	;
	m_bold		= font	;
	m_italic	= font	;

	m_bold  .setWeight (QFont::Bold) ;
	m_italic.setItalic (true) ;
}

int	KBSyntaxHighlighter::highlightParagraph
	(	const QString	&text,
		int		esolp
	)
{
//	fprintf
//	(	stderr,
//		"KBSyntaxHighlighter::highlightParagraph: [%s] %d\n",
//		(const char *)text,
//		esolp
//	)	;

	if (m_highlighter != 0)
		return	m_highlighter->highlight (text, esolp, this) ;

	return	0 ;
}

void	KBSyntaxHighlighter::loadHighlights ()
{
	QString		hlPath	= locateDir ("appdata", "highlights/python.hilight") ;
	QDir		hlDir	= QDir	    (hlPath + "/highlights") ;
	QStringList	hlFiles = hlDir.entryList ("*.hilight") ;

	for (uint idx = 0 ; idx < hlFiles.count() ; idx += 1)
		new KBHLHighlighter (hlPath + "/highlights/" + hlFiles[idx]) ;
}

QStringList
	KBSyntaxHighlighter::highlightNames ()
{
	return	hlNames == 0 ? QStringList() : *hlNames ;
}
