/***************************************************************************
    file	         : kb_keymapper.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/

#include	<stdio.h>
#include	<stdlib.h>

#include	<qfile.h>
#include	<qintdict.h>
#include	<qtimer.h>
#include	<qregexp.h>


#include	"kb_classes.h"
#include	"kb_dom.h"
#include	"kb_textedit.h"
#include	"kb_locator.h"
#include	"kb_intelli.h"
#include	"kb_methdict.h"
#include	"kb_noobjdlg.h"
#include	"kb_options.h"
#include	"kb_object.h"

#include	"kb_keymapper.moc"


#define		__DEBUG_KEYMAP	(0)
#if		__DEBUG_KEYMAP
#include	"kb_dprintf.h"
#endif


/*  KBKeyMapperMap							*/
/*  --------------							*/
/*  This class is used to represent a specific binding in a key map.	*/
/*  There will be one such element for each unique key prefix, which	*/
/*  contains information on what to do with that key prefix.		*/

class	LIBKBASE_API KBKeyMapperMap
{
	QIntDict<KBKeyMapperMap>
				m_map		;

	enum	Binding
	{	BindNone,
		BindFunction,
		BindHelper
	}	;

	Binding			m_binding	;
	int			m_bindCode	;
	QString			m_bindText	;

public	:

	KBKeyMapperMap	()	;

	void		bindKeyMapper	(int, KBKeyMapperMap *) ;
	void		bindFunction	(int, const QString & = QString::null) ;
	void		bindHelper	(int, const QString & = QString::null) ;

	bool		activate	(KBKeyMapper *) ;
	KBKeyMapperMap	*applyKey	(int)	;
}	;

/*  ------------------------------------------------------------------	*/

/*  KBKeyMapperMap							*/
/*  KBKeyMapperMap							*/
/*		: Constructor for key mapper binding object		*/
/*  (returns)	: KBKeyMapperMap :					*/

KBKeyMapperMap::KBKeyMapperMap ()
{
	m_map.setAutoDelete (true) ;
	m_binding = BindNone	   ;
}

/*  KBKeyMapperMap							*/
/*  setEventNode: Set node and language asssociated with event		*/
/*  node	: KBNode *	  : Event node				*/
/*  language	: const QString & : Scripting language			*/
/*  (returns)	 : void		    :					*/

void	KBTextEditMapper::setEventNode
	(	KBNode		*node,
		const QString	&language
	)
{
	m_eventNode	= node		;
	m_language	= language	;
}

/*  KBKeyMapperMap							*/
/*  bindKeyMapper: Bind submap to key					*/
/*  bindCode	 : int		    : Function code			*/
/*  mapper	 : KBKeyMapperMap * : Submap mapper			*/
/*  (returns)	 : void		    :					*/

void	KBKeyMapperMap::bindKeyMapper
	(	int		key,
		KBKeyMapperMap	*mapper
	)
{
	m_map.insert (key, mapper) ;
}

/*  KBKeyMapperMap							*/
/*  bindFunction: Bind a function to the binding object			*/
/*  bindCode	: int		  : Function code			*/
/*  bindText	: const QString & : Function text			*/
/*  (returns)	: void		  :					*/

void	KBKeyMapperMap::bindFunction
	(	int		bindCode,
		const QString	&bindText
	)
{
	m_binding  = BindFunction ;
	m_bindCode = bindCode	  ;
	m_bindText = bindText	  ;
}

/*  KBKeyMapperMap							*/
/*  bindHelper	: Bind a helper to the binding object			*/
/*  bindCode	: int		  : Helper code				*/
/*  bindText	: const QString & : Helper text				*/
/*  (returns)	: void		  :					*/

void	KBKeyMapperMap::bindHelper
	(	int		bindCode,
		const QString	&bindText
	)
{
	m_binding  = BindHelper	;
	m_bindCode = bindCode	;
	m_bindText = bindText	;
}

/*  KBKeyMapperMap							*/
/*  activate	: Activate binding object				*/
/*  mapper	: KBKeyMapper * : Mapper to invoke			*/
/*  (returns)	: bool		 : True if function activated		*/

bool	KBKeyMapperMap::activate
	(	KBKeyMapper	*mapper
	)
{
	DPRINTF
	((	"KBKeyMapperMap::activate: mapper=%p m_binding=%d\n",
		mapper,
		m_binding
	))	;

	switch (m_binding)
	{
		case BindHelper   :
			/* If the binding is to a helper then invoke	*/
			/* the helper. The return code indicates if	*/
			/* the caller should clear the current mapping.	*/
			return	mapper->helper   (m_bindCode, m_bindText) ;

		case BindFunction :
			/* Invoke the function. The return code		*/
			/* indicates if the caller should clear the	*/
			/* current mapping.				*/
			return	mapper->function (m_bindCode, m_bindText) ;

		default	:
			/* Otherwise, we are presumably part-way along	*/
			/* a key binding.				*/
			break	;
	}

	return	false	;
}


/*  KBKeyMapperMap							*/
/*  applyKey	: Apply key to binding					*/
/*  key		: int		    : Key from QKeyEvent		*/
/*  (returns)	: KBKeyMapperMap * : Next binding or null of none	*/

KBKeyMapperMap
	*KBKeyMapperMap::applyKey
	(	int		key
	)
{
	/* This is a simple lookup. The caller uses this method to step	*/
	/* through key binding sequences.				*/
	return	m_map.find (key) ;
}



/*  ------------------------------------------------------------------  */

/*  ------------------------------------------------------------------  */

QIntDict<KBKeyMapperMap>	KBKeyMapper::m_map	;

/*  KBKeyMapper								*/
/*  KBKeyMapper	: Constructor for key mapper base class			*/
/*  parent	: QObject *	: Parent object				*/
/*  (returns)	: KBKeyMapper	:					*/

KBKeyMapper::KBKeyMapper
	(	QObject		*parent
	)
	:
	QObject	(parent)
{
	m_currentMap	= 0	;
	m_gobble	= false	;
}

/*  KBKeyMapper								*/
/*  ~KBKeyMapper: Destructor for key mapper base class			*/
/*  (returns)	 :		:					*/

KBKeyMapper::~KBKeyMapper ()
{
}


/*  KBKeyMapper								*/
/*  clearKeyMap	: Clear key maps					*/
/*  (returns)	: void		:					*/

void	KBKeyMapper::clearKeyMap ()
{
	/* FIX ME! We should make sure that the m_currentMap pointers	*/
	/* of _all) extant KBKeyMapper objects are reset!		*/
	m_map.clear  ()  ;
	m_currentMap = 0 ;
}

/*  KBKeyMapper								*/
/*  findMapperMap: Find mapper element for specified key string		*/
/*  keyString	 : QValueList<int> & : Key string			*/
/*  (returns)	 : KBKeyMapperMap * : Mapper element			*/

KBKeyMapperMap
	*KBKeyMapper::findMapperMap
	(	const QValueList<int>	&keyString
	)
{
	KBKeyMapperMap	*map	;

	DPRINTF
	((	"KBKeyMapper::findMapperMap: 0: [%08x]\n",
		keyString[0]
	))	;

	if ((map = m_map.find (keyString[0])) == 0)
	{
		map	= new KBKeyMapperMap () ;
		m_map.insert (keyString[0], map) ;
	}

	for (uint idx = 1 ; idx < keyString.count() ; idx += 1)
	{
		DPRINTF
		((	"KBKeyMapper::findMapperMap: %d: [%08x]\n",
			idx,
			keyString[idx]
		))	;
		KBKeyMapperMap *next = map->applyKey (keyString[idx]) ;

		if (next == 0)
		{
			next = new KBKeyMapperMap () ;
			map->bindKeyMapper (keyString[idx], next) ;
		}

		map	= next	;
	}

	return	map	;
}

void	KBKeyMapper::error ()
{
	fprintf
	(	stderr,
		"KBKeyMapper::error: Called\n"
	)	;
}

/*  KBKeyMapper								*/
/*  bindFunction: Bind key sequence to function				*/
/*  keyString	: QValueList<int> : Key sequence			*/
/*  bindCode	: int		  : Function code			*/
/*  bindText	: const QString & : Function text			*/
/*  (returns)	: void		  :					*/

void	KBKeyMapper::bindFunction
	(	const QValueList<int>	&keyString,
		int			bindCode,
		const QString		&bindText
	)
{
	findMapperMap(keyString)->bindFunction (bindCode, bindText) ;
}

/*  KBKeyMapper								*/
/*  bindHelper	: Bind key sequence to helper				*/
/*  keyString	: QValueList<int> : Key sequence			*/
/*  bindCode	: int		  : Helper code				*/
/*  bindText	: const QString & : Helper text				*/
/*  (returns)	: void		  :					*/

void	KBKeyMapper::bindHelper
	(	const QValueList<int>	&keyString,
		int			bindCode,
		const QString		&bindText
	)
{
	findMapperMap(keyString)->bindHelper   (bindCode, bindText) ;
}

/*  KBKeyMapper							*/
/*  mapKeysToKeys: Map character kde sequence to code key sequence	*/
/*  keys	 : const QString & : Character sequence			*/
/*  (returns)	 : QValueList<int> : Code sequence			*/

QValueList<int>
	KBKeyMapper::keysToKeys
	(	const QString	&keys
	)
{
	struct	KeysToCode
	{	char	key [8]	;
		int	code	;
	}	;

	static	KeysToCode	mapKeysToKeys[] =
	{
		{	"A",		Qt::Key_A		},
		{	"B",		Qt::Key_B		},
		{	"C",		Qt::Key_C		},
		{	"D",		Qt::Key_D		},
		{	"E",		Qt::Key_E		},
		{	"F",		Qt::Key_F		},
		{	"G",		Qt::Key_G		},
		{	"H",		Qt::Key_H		},
		{	"I",		Qt::Key_I		},
		{	"J",		Qt::Key_J		},
		{	"K",		Qt::Key_K		},
		{	"L",		Qt::Key_L		},
		{	"M",		Qt::Key_M		},
		{	"N",		Qt::Key_N		},
		{	"O",		Qt::Key_O		},
		{	"P",		Qt::Key_P		},
		{	"Q",		Qt::Key_Q		},
		{	"R",		Qt::Key_R		},
		{	"S",		Qt::Key_S		},
		{	"T",		Qt::Key_T		},
		{	"U",		Qt::Key_U		},
		{	"V",		Qt::Key_V		},
		{	"W",		Qt::Key_W		},
		{	"X",		Qt::Key_X		},
		{	"Y",		Qt::Key_Y		},
		{	"Z",		Qt::Key_Z		},

		{	"ESC",		Qt::Key_Escape		},

		{	"F1",		Qt::Key_F1		},
		{	"F2",		Qt::Key_F2		},
		{	"F3",		Qt::Key_F3		},
		{	"F4",		Qt::Key_F4		},
		{	"F5",		Qt::Key_F5		},
		{	"F6",		Qt::Key_F6		},
		{	"F7",		Qt::Key_F7		},
		{	"F8",		Qt::Key_F8		},
		{	"F9",		Qt::Key_F9		},
		{	"F10",		Qt::Key_F10		},
		{	"F11",		Qt::Key_F11		},
		{	"F12",		Qt::Key_F12		},

		{	"Return",	Qt::Key_Return		},
		{	"Enter",	Qt::Key_Enter		},

		/* Add more here as needed ...				*/
		{	"",		0			}
	}	;

	/* Scan the string key by key. converting to Qt key codes. In	*/
	/* the string, characters stand for themselves, except that	*/
	/* braces can be used for special cases, like "{F1}".		*/
	QValueList<int>	seq	;
	int		idx1	= 0;

	while (idx1 < (int)keys.length())
	{
		QString	key	;
		int	mod	= 0 ;
		int	idx2	;

		for (;;)
		{
			if (keys.at(idx1) == ' ')
			{	idx1	+= 1		;
				continue		;
			}
			if (keys.mid(idx1, 5).lower() == "ctrl-" )
			{	mod	|= Qt::CTRL	;
				idx1	+= 5		;
				continue		;
			}
			if (keys.mid(idx1, 6).lower() == "shift-")
			{	mod	|= Qt::SHIFT	;
				idx1	+= 6		;
				continue		;
			}
			if (keys.mid(idx1, 4).lower() == "alt-"  )
			{	mod	|= Qt::ALT	;
				idx1	+= 4		;
				continue		;
			}

//			Mac OS/X decodes the Apple key as ctrl- and the
//			Ctrl key as meta-, so to make life simple we
//			treat them as synonymous. See the code in
//			KBTextEditMapper::processEvent.
//
//			if (keys.mid(idx1, 5).lower() == "meta-" )
//			{	mod	|= Qt::META	;
//				idx1	+= 5		;
//				continue		;
//			}
			break	;
		}

		if (keys.at(idx1) == "{")
		{
			if ((idx2 = keys.find ('}', idx1 + 1)) < 0)
				break	;

			key	= keys.mid (idx1 + 1, idx2 - idx1 - 1) ;
			idx1	= idx2 + 1 ;
		}
		else
		{	key	= keys.mid (idx1, 1) ;
			idx1    += 1 ;
		}

		DPRINTF
		((	"KBKeyMapper::keysToKeys: mapping [%s][%08x]\n",
			key.latin1(),
			mod
		))	;
		for (KeysToCode *kp = &mapKeysToKeys[0] ; kp->code != 0 ; kp += 1)
			if (kp->key == key)
			{	seq.append (kp->code|mod) ;
				break	;
			}
	}

	return	seq	;
}

/*  KBKeyMapper								*/
/*  loadKeyMap	: Load a key map specification				*/
/*  kName	: const QString & : Specification file name		*/
/*  (returns)	: QString	  : Error or NULL if OK			*/

QString	KBKeyMapper::loadKeyMap
	(	const QString	&kName
	)
{
	QFile	kFile	(kName) ;
	if (!kFile.open (IO_ReadOnly))
		return	QString(TR("%1: unable to open")).arg(kName) ;

	QDomDocument	kDoc	;
	if (!kDoc.setContent (&kFile))
		return	QString(TR("%1: unable to parse")).arg(kName) ;

	DOMITER_BEGIN(kDoc.documentElement(),"function",kElem)
	{
		QString		kKeys = kElem.attribute ("keys") ;
		QString		kCode = kElem.attribute ("code") ;
		QString		kText = kElem.attribute ("text") ;

		DPRINTF
		((	"KBKeyMapper::loadKeyMap: Function[%s]->[%s][%s]\n",
			(const char *)kKeys,
			(const char *)kCode,
			(const char *)kText
		))	;
		bindFunction
		(	keysToKeys (kKeys),
			codeToCode (kCode),
			kText
		)	;
	}
	DOMITER_END

	DOMITER_BEGIN(kDoc.documentElement(),"helper",kElem)
	{
		QString		kKeys = kElem.attribute ("keys") ;
		QString		kCode = kElem.attribute ("code") ;
		QString		kText = kElem.attribute ("text") ;

		DPRINTF
		((	"KBKeyMapper::loadKeyMap: Helper  [%s]->[%s][%s]\n",
			(const char *)kKeys,
			(const char *)kCode,
			(const char *)kText
		))	;
		bindHelper
		(	keysToKeys (kKeys),
			codeToCode (kCode),
			kText
		)	;
	}
	DOMITER_END

	DOMITER_BEGIN(kDoc.documentElement(),"option",kElem)
	{
		QString	value	= kElem.attribute("value") ;

		if (kElem.attribute ("name") == "gobble")
		{	m_gobble = value.toInt() ;
			continue ;
		}
	}
	DOMITER_END

	return	QString::null	;
}

/*  KBKeyMapper								*/
/*  useAtOuterLevel							*/
/*		: See if key should be processed at outer level		*/
/*  key		: int		: Key code from QKeyEvent		*/
/*  mod		: int		: Modifiers				*/
/*  (returns)	: bool		: Key consumed by mapper		*/

bool	KBKeyMapper::useAtOuterLevel
	(	int		,
		int		
	)
{
	return	true	;
}

/*  KBKeyMapper								*/
/*  applyKey	: Apply key to the mapper				*/
/*  key		: int		: Key code from QKeyEvent		*/
/*  mod		: int		: Modifiers				*/
/*  (returns)	: bool		: Key consumed by mapper		*/

bool	KBKeyMapper::applyKey
	(	int		key,
		int		mod
	)
{
	DPRINTF
	((	"KBKeyMapper::applyKey: key=%d mod=%04x map=%p\n",
		key,
		mod,
		(void *)m_currentMap
	))	;

	if (m_currentMap == 0)
	{
		/* At the outermost level, check if we should use the	*/
		/* key ...						*/
		if (!useAtOuterLevel (key, mod))
			return	false	;

		/* OK, got a usable key. See if the key is in the map,	*/
		/* if not then results depend in whether it is a	*/
		/* modified character key, and whether we are gobbling	*/
		/* unmatched keys.					*/
		m_currentMap = m_map.find (key | mod) ;

		DPRINTF
		((	"KBKeyMapper::applyKey: [%08x] map=%p\n",
			key | mod,
			m_currentMap
		))	;

		if (m_currentMap == 0)
			return	(mod == Qt::CTRL ) &&
				(key >= Qt::Key_A) &&
				(key <= Qt::Key_Z) && m_gobble ;
	}
	else
	{
		/* At the inner levels, we handle all keys. If the key	*/
		/* leads nowhere then report an error.			*/
		m_currentMap = m_currentMap->applyKey (key | mod) ;

		if (m_currentMap == 0)
		{	error	()	;
			return	true	;
		}
	}

	/* We are inside the mappings. See if this mapping consumes the	*/
	/* key, in which case we reset the current mapping pointer.	*/
	if (m_currentMap->activate (this))
	{
		m_currentMap = 0 ;
		return	true	 ;
	}

	/* Current mapping did not consume the key, so we consume the	*/
	/* key but leave the mapper in place.				*/
	return	true	;
}

/*  KBKeyMapper								*/
/*  clearKey	: Clear any partial mapping				*/
/*  (returns)	: void		:					*/

void	KBKeyMapper::clearKey ()
{
	m_currentMap = 0 ;
}



/*  ------------------------------------------------------------------  */

/*  KBTextEditMapper							*/
/*  KBTextEditMapper							*/
/*		: Constructor for KBTextEdit-specific key mapper	*/
/*  editor	: KBTextEdit *	   : Editor				*/
/*  (returns)	: KBTextEditMapper :					*/

KBTextEditMapper::KBTextEditMapper
	(	KBTextEdit	*editor
	)
	:
	KBKeyMapper	(editor),
	m_editor 	(editor)
{
	m_helper  	= 0	;
	m_eventNode     = 0	;

	QString	keymap	= getenv ("REKALL_KEYMAP") ;

	if (keymap.isEmpty()) keymap = KBOptions::getKeymap() ;
	if (keymap.isEmpty()) keymap = locateFile ("appdata", "keymaps/keymap.null") ;

	DPRINTF
	((	"KBTextEditMapper::KBTextEditMapper: \"%s\"\n",
		keymap.latin1()
	))	;

	QString r = loadKeyMap (keymap) ;
	if (!r.isNull())
	{
		if (keymap.isEmpty())
			fprintf
			(	stderr,
				"KBTextEditMapper::KBTextEditMapper:: %s\n",
				(const char *)r
			)	;
		else	KBError::EError
			(	TR("Cannot load user keymap"),
				QString(TR("Path: %1\nError: %2")).arg(keymap).arg(r),
				__ERRLOCN
			)	;
	}
}

/*  KBTextEditMapper							*/
/*  ~KBTextEditMapper							*/
/*		: Destructor for tkEditor-specific key mapper		*/
/*  (returns)	 :		 :					*/

KBTextEditMapper::~KBTextEditMapper ()
{
	if (m_helper) delete m_helper  ;
}


/*  KBTextEditMapper							*/
/*  showHelper	: Show helper popup					*/
/*  scanner	: KBIntelliScan * : Scanner with scan results		*/
/*  (returns)	: void		  :					*/

void	KBTextEditMapper::showHelper
	(	KBIntelliScan	*scanner
	)
{
	KBNode	*node	= scanner->node  () ;
	QString	klass	= scanner->klass () ;
	QString	legend	;

	if (node != 0)
		legend	= QString(TR("%1: %2")).arg(klass).arg(node->getName()) ;
	else	legend	= QString(TR("Unknown %1 object")).arg(klass) ;

	m_helper = KBIntelli::create
			(	m_editor,
				legend,
				scanner->methods(),
				scanner->prefix ()
			)	;

	/* Note the current editor position, the matched text and the	*/
	/* helper type. These are needed if the user selects a helper	*/
	/* entry for substitution.					*/
	m_editor->getCursorPosition (&m_helperPara, &m_helperCol) ;

	connect
	(	m_helper,
		SIGNAL	(sigChosen (KBMethDictEntry *, bool)),
		SLOT	(slotChosen(KBMethDictEntry *, bool))
	)	;
}

/*  KBTextEditMapper							*/
/*  hideHelper	: Hide helper if present				*/
/*  (returns)	: void		:					*/

void	KBTextEditMapper::hideHelper ()
{
	DPRINTF
	((	"KBTextEditMapper::hideHelper: m_helper=[%p]\n",
		(void *)m_helper
	))	;

	/* If the helper is present delete it *later*, since we may	*/
	/* have been called via a helper signal.			*/
	if (m_helper)
	{	m_helper->deleteLater () ;
		m_helper = 0 ;
	}
}

#if	! __KB_RUNTIME

/*  KBTextEditMapper							*/
/*  scanForMethods : Scan for methods					*/
/*  (returns)	   : KBIntelliScan * : Scan results or null if none	*/

KBIntelliScan *KBTextEditMapper::scanForMethods ()
{
	int	col	= m_editor->currentColumn() ;
	QString	text	= m_editor->currentText  ().left(col) ;

	DPRINTF
	((	"KBTextEditMapper::scanForMethod: %d in [%s]\n",
		col,
		text.latin1()
	))	;

	extern	QString	fixLanguageName (const QString &language) ;
	KBIntelliScan *scanner = KBIntelliScan::getScanner (fixLanguageName (m_language)) ;
	QString	      selfName = scanner->getSelfName(m_editor->text()) ;

	if (!scanner->scanForMethods (selfName, m_eventNode, text))
		return	0 ;

	m_scanOffset = scanner->offset () ;
	m_scanPrefix = scanner->prefix () ;

	return	scanner ;
}

#endif

/*  KBTextEditMapper							*/
/*  slotScan	: Fire off the intelli-scanner				*/
/*  (returns)	: void		:					*/

void	KBTextEditMapper::slotScan ()
{
#if	! __KB_RUNTIME
	KBIntelliScan *scanner = scanForMethods () ;
	if (scanner != 0) showHelper (scanner) ;
#endif
}

/*  KBTextEditMapper							*/
/*  useAtOuterLevel							*/
/*		: See if key should be processed at outer level		*/
/*  key		: int		: Key code from QKeyEvent		*/
/*  mod		: int		: Modifiers				*/
/*  (returns)	: bool		: Key consumed by mapper		*/

bool	KBTextEditMapper::useAtOuterLevel
	(	int		key,
		int		mod
	)
{
	DPRINTF
	((	"KBTextEditMapper::useAtOuterLevel: %d %d [%08x]\n",
		key,
		Qt::Key_Period,
		key | mod
	))	;

#if	! __KB_RUNTIME
	/* If the intelli-helper is to be displayed automagically, we	*/
	/* use a zero-timeout timer so that the character is inserted	*/
	/* before the scan takes place.					*/
	if (KBOptions::getAutoIntelli ())
		if ((key == Qt::Key_Period) && (mod == 0))
		{
			hideHelper () ;
			QTimer::singleShot (0, this, SLOT(slotScan())) ;
			return	false	 ;
		}
#endif
	return	KBKeyMapper::useAtOuterLevel (key, mod) ;
}

/*  KBTextEditMapper							*/
/*  codeToCode	: Convert text representation code code to action code	*/
/*  name	: const QString & : Text representation			*/
/*  (returns)	: int		  : Code				*/

int	KBTextEditMapper::codeToCode
	(	const QString	&name
	)
{
	struct	CodeToCode
	{	const char	*name	;
		Action		code	;
	}	;

	static	CodeToCode	mapCodeToCode[] =
	{
		{	"StartOfDoc",	StartOfDoc		},
		{	"EndOfDoc",	EndOfDoc		},
		{	"StartOfLine",	StartOfLine		},
		{	"EndOfLine",	EndOfLine		},
		{	"NextWord",	NextWord		},
		{	"PreviousWord",	PreviousWord		},
		{	"DelWordLeft",	DelWordLeft		},
		{	"DelWordRight",	DelWordRight		},
		{	"DelLine",	DelLine			},
		{	"Cut",		Cut			},
		{	"Copy",		Copy			},
		{	"Paste",	Paste			},
		{	"ScrollDown",	ScrollDown		},
		{	"ScrollUp",	ScrollUp		},
		{	"SearchForward",SearchForward		},
		{	"SearchBack",	SearchBack		},
		{	"Undo",		Undo			},
		{	"Redo",		Redo			},

		/* Add more cases here to match the Enum ...		*/
		{	0,		NoAction		}
	}	;

	for (CodeToCode *cp = &mapCodeToCode[0] ; cp->name != 0 ; cp += 1)
		if (cp->name == name)
			return	cp->code ;

	return	0 ;
}

/*  KBTextEditMapper							*/
/*  helper	: Execute an editor helper				*/
/*  code	: int		  : Helper code				*/
/*  text	: const QString & : Helper text				*/
/*  (returns)	: bool		  : True to clear callers mapping	*/

bool	KBTextEditMapper::helper
	(	int		,
		const QString	&text
	)
{
#if	! __KB_RUNTIME

	DPRINTF
	((	"KBTextEditMapper::helper: %s\n",
		text.latin1()
	))	;

	/* ShowMethods							*/
	/* -----------							*/
	/* Show all possible matches for a method. We scan for an	*/
	/* immediately preceding method and then get all descriptions	*/
	/* for which the method text is a prefix.			*/
	if (text == "ShowMethods")
	{
#if	! __KB_RUNTIME
		hideHelper () ;
		slotScan   () ;
		return	true  ;
#endif
	}

	/* FindObject							*/
	/* ----------							*/
	/* Helper used to locate one object relative to another. The	*/
	/* staring point is "m_eventNode" if set.			*/ 
	if (text == "FindObject")
	{
		if (m_eventNode == 0) return true ;

		KBIntelli::destroy	() ;

		KBRouteToNodeDlg rdlg	(m_eventNode, QString::null, TR("Locate object ...")) ;
		QString		 route	;

		switch (rdlg.exec())
		{
			case 1 :
				route	= rdlg.routeToNode (false) ;
				break	;

			case 2 :
				route	= rdlg.routeToNode (true ) ;
				break	;

			default	:
				return	true	;
		}

		DPRINTF
		((	"KBTextEditMapper::helper: route=[%s]\n",
			route.latin1()
		))	;

		m_editor->insert (route) ;
		return	true	;
	}

	if (text == "FindTextShow")
	{
		m_editor->showFindText    (true ) ;
		return	true	;
	}
	if (text == "ReplaceTextShow")
	{
		m_editor->showReplaceText (true ) ;
		return	true	;
	}
	if (text == "FindReplaceTextHide")
	{
		m_editor->showFindText    (false) ;
		m_editor->showReplaceText (false) ;
		return	true	;
	}
#endif
	return	true	;
}

/*  KBTextEditMapper							*/
/*  function	: Execute an editor function				*/
/*  code	: int		  : Function code			*/
/*  text	: const QString & : Function text			*/
/*  (returns)	: bool		  : True to clear callers mapping	*/

bool	KBTextEditMapper::function
	(	int		code,
		const QString	&text
	)
{
	RKTextEdit::CursorAction  cursorAction	= (RKTextEdit::CursorAction)-1	;

	/* First check for the cases which map directly onto editor	*/
	/* actions ....							*/
	switch (code)
	{
		case StartOfDoc  : cursorAction = RKTextEdit::MoveHome		; break ;
		case EndOfDoc	 : cursorAction = RKTextEdit::MoveEnd		; break ;
		case StartOfLine : cursorAction = RKTextEdit::MoveLineStart	; break ;
		case EndOfLine	 : cursorAction = RKTextEdit::MoveLineEnd	; break	;
		case NextWord	 : cursorAction = RKTextEdit::MoveWordForward	; break	;
		case PreviousWord: cursorAction = RKTextEdit::MoveLineEnd	; break	;
		case ScrollUp	 : cursorAction = RKTextEdit::MovePgUp		; break	;
		case ScrollDown	 : cursorAction = RKTextEdit::MovePgDown	; break	;

		default		 : break  ;
	}

	if (cursorAction != (RKTextEdit::CursorAction)-1)
	{
		m_editor->moveCursor (cursorAction, false) ;
		return	true ;
	}

	if ((code == SearchForward) || (code == SearchBack))
	{
//**		getFindReplace()->exec(m_editor, false) ;
		return true ;
	}

	switch (code)
	{
		case DelLine	 :
			m_editor->delLine() ;
			break	 ;

		case Cut	 :
			m_editor->del   () ;
			break	 ;

		case Copy	 :
			m_editor->copy	() ;
			break	 ;

		case Paste	 :
			m_editor->paste	() ;
			break	 ;

		case Undo	 :
			m_editor->undo	() ;
			break	 ;

		case Redo	 :
			m_editor->redo	() ;
			break	 ;

		default	:
			DPRINTF
			((	"KBTextEditMapper::function: >function[%d][%s]\n",
				code,
				(cchar *)text
			))	;
			break	;
	}

	return	true	;
}

/*  KBTextEditMapper							*/
/*  checkChangeLine							*/
/*		: See if line has changed since helper was displayed	*/
/*  (returns)	: void		:					*/

void	KBTextEditMapper::checkChangeLine ()
{
	DPRINTF
	((	"KBTextEditMapper::checkChangeLine: %d <-> %d\n",
		m_helperPara,
		m_editor->currentPara()
	))	;

	if (!m_helper)
		return	;

	if (m_helperPara != m_editor->currentPara())
	{	hideHelper () ;
		return	;
	}

	int	ccol	= m_editor->currentColumn() ;
	QString	text	= m_editor->currentText  () ;

	if (ccol < m_scanOffset)
	{	hideHelper () ;
		return	;
	}

	m_helper->setPrefix (text.mid (m_scanOffset, ccol - m_scanOffset)) ;
}

/*  KBTextEditMapper							*/
/*  processEvent: Process events for the editor				*/
/*  e		: QEvent *	: The event				*/
/*  (returns)	: bool		: True if consumed			*/

bool	KBTextEditMapper::processEvent
	(	QEvent		*e
	)
{
	switch (e->type())
	{
		case QEvent::KeyPress		:
		case QEvent::AccelOverride	:
		{
			DPRINTF
			((	"KBTextEditMapper::processEvent: key/accel event\n"
			))	;

			/* Trap keystrokes and feed them through the	*/
			/* kep mapper. On return, start a timer if	*/
			/* there is a helper then start a timer. This	*/
			/* is needed so that we can remove it if the	*/
			/* user changes line.				*/
			QKeyEvent *k	= (QKeyEvent *)e ;

			if (m_helper) switch (k->key())
			{
				case Qt::Key_ParenRight  :
				case Qt::Key_Escape	 :
					delete	m_helper ;
					break	;

				default	:
					break	;
			}

			int	mod	= 0 ;
			int	state	= k->state()	;
			if ((state & Qt::ControlButton) != 0) mod |= Qt::CTRL	;
			if ((state & Qt::ShiftButton  ) != 0) mod |= Qt::SHIFT	;
			if ((state & Qt::AltButton    ) != 0) mod |= Qt::ALT	;

//			Decode Ctrl and Meta as the same, to make life easy
//			on Mac OS/X. See also the code in KBKeyMapper::keysToKeys
//
//			if ((state & Qt::MetaButton   ) != 0) mod |= Qt::META	;
			if ((state & Qt::MetaButton   ) != 0) mod |= Qt::CTRL	;

			bool	rc	= applyKey (k->key(), mod) ;
			DPRINTF
			((	"KBTextEditMapper::processEvent: key rc=%d\n",
				rc
			))	;
			if (rc) k->accept() ;

			/* If there is a helper, start a timer to check	*/
			/* if the key causes the line to change.	*/
			if (m_helper)
				QTimer::singleShot(50, this, SLOT(checkChangeLine ())) ;
			return	  rc	;
		}

		case QEvent::MouseButtonPress	:
		case QEvent::MouseButtonRelease	:
		case QEvent::MouseButtonDblClick:
			/* All mouse clicks delete the helper		*/
			DPRINTF
			((	"KBTextEditMapper::processEvent: MOUSE EVENT\n"
			))	;
			hideHelper ()	;
			return	false	;

//		case QEvent::FocusOut		:
//			if (m_helper) m_helper->setClosePending() ;
//			return	false	;

		case QEvent::FocusIn		:
			return	false	;

		default	:
			break	;
	}

	return	false	;
}

/*  KBTextEditMapper							*/
/*  slotChosen	: User has chosen a function from the helper		*/
/*  entry	: KBMethDictEntry * : Selected entry			*/
/*  ctrlDown	: bool		    : Control key down			*/
/*  (returns)	: void		    :					*/

void	KBTextEditMapper::slotChosen
	(	KBMethDictEntry		*entry,
		bool			ctrlDown
	)
{
	XPRINTF
	((	"KBTextEditMapper::slotChosen: ctrlDown=%d\n",
		ctrlDown
	))	;

	hideHelper ()	;

	/* Paranoia check. The helper should be dismissed if the user	*/
	/* changes line, but lets be certain.				*/
	if (m_helperPara != m_editor->currentPara())
	{
		DPRINTF
		((	"KBTextEditMapper::slotChosen: line changed!!\n"
		))	;
		return	;
	}

	/* Grab the current line and locate the matched text back from	*/
	/* the column at which the helper was invoked. Should always be	*/
	/* there since the helper is dismissed by keystrokes.		*/
	QString	line	= m_editor->currentText() ;

	DPRINTF
	((	"KBTextEditMapper::slotChosen: [%s][%s] [%d:%d][%d:%d]\n",
		entry->substitute().latin1(),
		entry->name      ().latin1(),
		m_helperPara,
		m_scanOffset,
		m_helperPara,
		m_helperCol
	))	;

	/* Replace the text from the matched text up to the helper	*/
	/* column with the substitution text. Maybe we should move the	*/
	/* cursor .....							*/
	m_editor->setSelection
	(	m_helperPara,
		m_scanOffset,
		m_helperPara,
		m_editor->currentColumn()
	)	;

	m_editor->insert
	(	ctrlDown ?
			entry->substitute() :
			entry->name      (),
		QTextEdit::RemoveSelected
	)	;
}
