/***************************************************************************
    file	         : kb_layout.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/

#include	<limits.h>

#include	<qevent.h>
#include	<qwidget.h>
#include	<qptrlist.h>

#include	"kb_classes.h"
#include	"kb_type.h"
#include	"kb_value.h"
#include	"kb_location.h"
#include	"kb_block.h"
#include	"kb_sizer.h"
#include	"kb_layout.h"
#include	"kb_options.h"
#include	"kb_gui.h"

#if		! __KB_RUNTIME

#include	"kb_formcopier.h"
#include	"kb_component.h"
#include	"kb_compaccessdlg.h"

/*  KBLayout								*/
/*  initSizer	: Initialise sizers on entry to design mode		*/
/*  (returns)	: void		:					*/

void	KBLayout::initSizer()
{
	while (m_sizerList.count() > 0)
	{
		m_sizerList.at(0)->getObject()->setMonitorSelect (false) ;
		m_sizerList.at(0)->setState(KBSizer::sbIdle);
		m_sizerList.remove((uint)0);
	}
}

/*  KBLayout								*/
/*  releaseSizer: Release sizer at end of update			*/
/*  (returns)	: void		:					*/

void	KBLayout::releaseSizer ()
{
	/* Tell all selected sizers to note the new size and position	*/
	/* settings, and return the leading sizer to the leading state	*/
	/* (as it was in the active state for the duration of the	*/
	/* update).							*/
	for (uint idx = 0 ; idx < m_sizerList.count() ; idx += 1)
		m_sizerList.at(idx)->accept (false) ;

	m_sizerList.at(0)->setState(KBSizer::sbLeader) ;
}

/*  KBLayout								*/
/*  dropSizer	: Drop sizer from list					*/
/*  sizer	: KBSizer *	: Sizer in question			*/
/*  (returns)	: void		:					*/

void	KBLayout::dropSizer
	(	KBSizer	*sizer
	)
{
	if (sizer != 0)
	{
		sizer->getObject()->setMonitorSelect (false) ;
		m_sizerList.remove (sizer) ;
		setGUIEnables	 ()	 ;
	}
}

/*  KBLayout								*/
/*  add		: User has selected a sizer tool			*/
/*  sizer	: KBSizer *	: Selected sizer			*/
/*  multi	: bool		: Selected for multiple update		*/
/*  (returns)	: KBLimit	: Movement limits			*/

KBLimit	KBLayout::addSizer
	(	KBSizer	*sizer,
		bool	multi
	)
{
	KBLimit	limit	(SHRT_MIN, SHRT_MAX,SHRT_MIN, SHRT_MAX) ;

	/* If this is _not_ a multiple update then remove all existing	*/
	/* sizers from the list, reseting each to the idle state.	*/
	if (!multi) initSizer() ;

	if (sizer == 0) return limit ;

	/* Make sure the specified sizer is not in the list (in case	*/
	/* of multiple selection), then put it at the front of the list	*/
	/* whence it becomes the leading sizer.				*/
	m_sizerList.remove (   sizer) ;
	m_sizerList.insert (0, sizer) ;

	/* Inform all listed sizers other then the new one that they	*/
	/* are now followers, and inform this one that it is active.	*/
	for (uint idx1 = 1 ; idx1 < m_sizerList.count() ; idx1 += 1)
		m_sizerList.at(idx1)->setState (KBSizer::sbFollower) ;

	sizer->getObject()->setMonitorSelect (true) ;
	sizer->setState (KBSizer::sbActive) ;
	setGUIEnables   () ;

	/* Scan though all sizers to determin the limits on any		*/
	/* movement with their parent.					*/
	for (uint idx2 = 0 ; idx2 < m_sizerList.count() ; idx2 += 1)
	{
		KBLimit here = m_sizerList.at(idx2)->getObject()->getMoveLimit() ;

		if (here.left  > limit.left ) limit.left   = here.left  ;
		if (here.right < limit.right) limit.right  = here.right ;
		if (here.up    > limit.up   ) limit.up     = here.up    ;
		if (here.down  < limit.down ) limit.down   = here.down  ;
	}

	return	limit	;
}

/*  KBLayout								*/
/*  trackMove	: Pass movement tracking to active sizers		*/
/*  dx		: int		: X-movement				*/
/*  dy		: int		: Y-movement				*/
/*  move	: bool		: Operation is movement (not resize)	*/
/*  (returns)	: void		:					*/

void	KBLayout::trackMove
	(	int	dx,
		int	dy,
		bool	move
	)
{
	for (uint idx = 0 ; idx < m_sizerList.count() ; idx += 1)
		m_sizerList.at(idx)->doResize (dx, dy, move) ;
}

/*  KBLayout								*/
/*  autoCtrlrect: Get new control area					*/
/*  object	: KBObject *	 : object in which control will appear	*/
/*  bState	: Qt::ButtonState: Button state when option selected	*/
/*  defrect	: QRect		 : Default area				*/
/*  (returns)	: QRect		 : New control area			*/

QRect	KBLayout::autoCtrlRect
	(	KBObject	*object,
		Qt::ButtonState	bState,
		QRect		defrect
	)
{
	/* If there are no sizers on the list then just return the	*/
	/* default rectangle; nothing else clever we can do.		*/
	if (m_sizerList.count() == 0)
		return	defrect	;

	/* If the new control is to go into a different object, then	*/
	/* just use the relative position of the existing leading	*/
	/* object.							*/
	if (m_sizerList.at(0)->getObject()->getParent() != object)
		return	m_sizerList.at(0)->getPosition () ;

	/* Case: more than one item on the sizer list. In this case	*/
	/* use the size of the most recent, and offset by the offset	*/
	/* between the most recent two.					*/
	if (m_sizerList.count() > 1)
	{
		QRect	rect0	= m_sizerList.at(0)->getPosition() ;
		QRect	rect1	= m_sizerList.at(1)->getPosition() ;
		QRect	rect	= rect0 ;

		rect.moveBy
		(	rect0.left() - rect1.left(),
			rect0.top () - rect1.top ()
		)	;
		return	rect	;
	}

	/* Case: one item on the sizer list. Use the same size and	*/
	/* offset right, unless the shift key was down in which case	*/
	/* offset down.							*/
	if (m_sizerList.count() == 1)
	{
		QRect	rect	= m_sizerList.at(0)->getPosition() ;
		int	space	= KBOptions::getCtrlSpace       () ;

		if ((bState & Qt::ShiftButton) == 0)
			rect.moveBy (rect.width() + space,  0) ;
		else	rect.moveBy (0, rect.height() + space) ;

		return	rect	;
	}

	/* Nothing on the sizer list so use the supplied default ...	*/
	return	defrect	;
}

/*  KBLayout								*/
/*  doCtrlAlign	: Handle control alignment				*/
/*  align	: KBCtrlAlign	: Requested alignment			*/
/*  (returns)	: void		:					*/

void	KBLayout::doCtrlAlign
	(	KB::CtrlAlign	align
	)
{
	if (m_sizerList.count() < 2) return ;

	QRect	pLeader	= m_sizerList.at(0)->getPosition() ;
	KBSizer	*sizer	= m_sizerList.first () ;

	while ((sizer = m_sizerList.next()) != 0)
	{
		QRect	pOther	= sizer->getPosition() ;

		switch (align)
		{
			case KB::AlignTop    :
				pOther.moveTopLeft
				(	QPoint
					(	pOther .left (),
						pLeader.top  ()
					)
				) ;
				break	;

			case KB::AlignLeft   :
				pOther.moveTopLeft
				(	QPoint
					(	pLeader.left (),
						pOther .top  ()
					)
				) ;
				break	;

			case KB::AlignBottom :
				pOther.moveBottomRight
				(	QPoint
					(	pOther .right (),
						pLeader.bottom()
					)
				) ;
				break	;

			case KB::AlignRight  :
				pOther.moveBottomRight
				(	QPoint
					(	pLeader.right (),
						pOther .bottom()
					)
				) ;
				break	;

			case KB::SameHeight  :
				pOther.setHeight (pLeader.height()) ;
				break	;

			case KB::SameWidth   :
				pOther.setWidth  (pLeader.width ()) ;
				break	;

			case KB::SameSize    :
				pOther.setHeight (pLeader.height()) ;
				pOther.setWidth  (pLeader.width ()) ;
				break	;

			default	:
				break	;
		}

		sizer->getObject()->setGeometry (pOther) ;
		sizer->accept (false)	;
	}
}

/*  KBLayout								*/
/*  setGUIEnables: Enable/disable GUI elements				*/
/*  (returns)	 : void		:					*/

void	KBLayout::setGUIEnables ()
{

	if (m_currGUI != 0)
	{
		m_currGUI->setEnabled (KBaseGUI::GRCopy,  m_sizerList.count() > 0) ;
		m_currGUI->setEnabled (KBaseGUI::GRAlign, m_sizerList.count() > 1) ;
		m_currGUI->setEnabled (KBaseGUI::GRDelta, m_design ? m_dDesign : m_dData) ;
	}
}

/*  KBLayout								*/
/*  doCopy	: Copy selected items					*/
/*  (returns)	: void		:					*/

void	KBLayout::doCopy ()
{
	if (m_sizerList.count() == 0)
		return ;

	KBFormCopier::self()->clearCopy () ;

	LITER
	(	KBSizer,
		m_sizerList,
		sizer,

		KBObject *o = sizer->getObject() ;
		KBFormCopier::self()->addToCopy (o->replicate (0), o->objType())
	)
}

/*  KBLayout								*/
/*  doCut	: Cut selected items					*/
/*  (returns)	: void		:					*/

void	KBLayout::doCut ()
{
	if (m_sizerList.count() > 0)
	{
		setChanged ()	;
		doCopy	   ()	;

		while (m_sizerList.count() > 0)
		{
			fprintf
			(	stderr,
				"KBLayout::doCut: destroy %p\n",
				(void *)m_sizerList.at(0)->getObject()
			)	;
			delete	m_sizerList.at(0)->getObject() ;
		}
	}
}

/*  KBLayout								*/
/*  doPaste	: Paste set of nodes					*/
/*  (returns)	: void			:				*/

void	KBLayout::doPaste ()
{
	if (m_sizerList.count() == 0)
	{	m_parent->isObject()->pasteObjects () ;
		return	;
	}

	if (m_sizerList.count() == 1)
	{	m_sizerList.at(0)->getObject()->pasteObjects() ;
		return	;
	}

	KBError::EWarning
	(	TR("Cannot paste when several objects are selected"),
		QString::null,
		__ERRLOCN
	) ;
}

/*  KBLayout								*/
/*  doSaveComponent							*/
/*		: Save selected objects as a component			*/
/*  location	: KBLocation &	: Owning documents location		*/
/*  (returns)	: void		:					*/

void	KBLayout::doSaveComponent
	(	KBLocation	&location
	)
{
	QString		 svName	  = location.server() ;
	QString		 objName  ;
	QString		 comment  ;

	/* Run the component save-as dialog, which will return the	*/
	/* desired object name, server, and optionally a comment.	*/
	bool		   toFile ;
	KBComponentSaveDlg csDlg
			   (	objName,
				svName,
				comment,
				location.dbInfo(),
				&toFile
			   )	;

	if (!csDlg.exec()) return ;


	QPtrList<KBObject> copies	;
	QRect		range	(0, 0, 0, 0) ;

	copies.setAutoDelete	(true) ;

	LITER
	(	KBSizer,
		m_sizerList,
		sizer,

		KBObject *copy	= (KBObject *)sizer->getObject()->replicate(0) ;
		QRect	 r	= copy->geometry() ;
		range = range.unite  (r) ;

		copies.append (copy) ;
	)

	int	dx	= 20 - range.x() ;
	int	dy	= 20 - range.y() ;

	range.moveBy (dx, dy) ;

	LITER
	(	KBObject,
		copies,
		copy,

		QRect  r = copy->geometry(QPoint (dx, dy)) ;
		copy->setGeometry (r) ;
	)

	extern	QString	kbXMLEncoding  () ;

	QString	 text	= QString
			  (	"<?xml version=\"1.0\" encoding=\"%1\"?>\n"
			  	"<KBComponent w=\"%2\" h=\"%3\" type=\"%4\" notes=\"%5\">\n" 
			  )
			  .arg	(kbXMLEncoding())
			  .arg	(range.width () + 40)
			  .arg	(range.height() + 40)
			  .arg	((uint)m_parent->objType())
			  .arg	(comment) ;


	LITER
	(	KBObject,
		copies,
		copy,
		copy->printNode(text, 2, false)
	)

	fprintf	  (stderr, "Component:[[[\n%s\n]]]\n", (cchar *)text) ;

	/*  .... and save the component where requested. The special	*/
	/* case is if the used specified save-to-file.			*/
	if (toFile)
	{
		saveComponentToFile (objName, text) ;
		return	;
	}

	KBLocation	newlocn
			(	location.dbInfo(),
				"component",
				svName,
				objName
			)	;
	KBError		error	;

	if (!newlocn.save (QString::null, QString::null, text, error))
		error.DISPLAY() ;

}

/*  KBLayout								*/
/*  snapToGrid	: Snap selected controls to the grid			*/
/*  (returns)	: void		:					*/

void	KBLayout::snapToGrid ()
{
	if (m_parent->showingDesign ())
	{
		LITER
		(	KBSizer,
			m_sizerList,
			sizer,
			sizer->snapToGrid ()
		)
		setChanged () ;
	}
}

/*  KBLayout	:							*/
/*  doMultiProp	: Invoke multiple-object properties dialog		*/
/*  (returns)	: void		:					*/

void	KBLayout::doMultiProp ()
{
	if (m_sizerList.count() > 1)
	{
		QPtrList<KBNode> nodeList ;

		for (uint idx = 0 ; idx < m_sizerList.count() ; idx += 1)
			nodeList.append (m_sizerList.at(idx)->getObject()) ;

		if (m_sizerList.at(0)->getObject()->doMultiProp (nodeList))
			setChanged (true) ;
	}
}

/*  KBLayout	:							*/
/*  doSingleProp: Invoke object properties dialog			*/
/*  (returns)	: void		:					*/

void	KBLayout::doSingleProp ()
{
	/* HACK ALERT:							*/
	/* Changed to invoke the design popup menu. This routine should	*/
	/* be renamed.							*/
	if (m_sizerList.count() > 0)
	{
		QMouseEvent me
			    (	QEvent::MouseButtonPress,
				QPoint(),
				QCursor::pos(),
				QMouseEvent::RightButton,
				QMouseEvent::RightButton
			    )	;

		m_sizerList.at(0)->doDesignPopup (&me) ;
	}
}

/*  KBLayout								*/
/*  clear	: Clear layout						*/
/*  design	: bool		: Design mode				*/
/*  (returns)	: void		:					*/

void	KBLayout::clear
	(	bool	design
	)
{
	m_sizerList  .clear ()	;
	if (!m_design) setChanged (false) ;
	m_design = design	;
}

#endif

/*  KBLayout								*/
/*  setChanged	: Set document changed status				*/
/*  changed	: bool		  : True if changed			*/
/*  thing	: const QString & : Name of changed thing if any	*/
/*  (returns)	: void		  :					*/

void	KBLayout::setChanged
	(	bool		changed,
		const QString	&thing
	)
{
	bool	 &rChanged = m_design ? m_dDesign : m_dData ;

//	fprintf
//	(	stderr,
//		"KBLayout::setChanged: c=%d rc=%d [%s]\n",
//		changed,
//		rChanged,
//		(cchar *)thing
//	)	;

	if (!changed)
		m_changeList.clear  () ;
	if (!thing.isEmpty())
		if (m_changeList.find (thing) == m_changeList.end())
			m_changeList.append (thing) ;

	if (changed == rChanged)
		return ;

	rChanged = changed ;
	if (m_currGUI != 0)
		m_currGUI->setEnabled (KBaseGUI::GRDelta, changed) ;
}

/*  KBLayout								*/
/*  getChanged	: Get object changed status				*/
/*  both	: bool		: Check both data and design changes	*/
/*  list	: QStringList & : Changed list				*/
/*  (returns)	: cchar *	: Change message or null if none	*/

cchar	*KBLayout::getChanged
	(	bool		both,
		QStringList	&list
	)
{
//	fprintf
//	(	stderr,
//		"KBLayout::getChanged: [%s]\n",
//		(cchar *)m_changeList.join("][")
//	)	;

	list	= m_changeList	;

	if (both)
		if (m_dDesign && m_dData)
			return	"data and design" ;

	if (( m_design || both) && m_dDesign) return "design"  ;
	if ((!m_design || both) && m_dData  ) return "data"	   ;

	return	0 ;
}



/*  KBLayout								*/
/*  setUnMorphedItem							*/
/*		: Note which control is not morphed			*/
/*  item	: KBItem *	: Item or null				*/
/*  drow	: uint		: Display row number			*/
/*  (returns)	: void		:					*/

//void	KBLayout::setUnMorphedItem
//	(	KBItem		*item,
//		uint		drow
//	)
//{
//	fprintf
//	(	stderr,
//		"KBLayout::setUnMorphedItem: [%s][%d]\n",
//		item == 0 ? "NONE" : (cchar *)item->getName(),
//		drow
//	)	;
//
//	/* NOTE: This routine keeps track of the unmorphed control when	*/
//	/* there are one or more morphing control in a form.		*/
//	if (m_unMorphedItem != 0)
//		if ((m_unMorphedItem != item) || (m_unMorphedDRow != drow))
//			m_unMorphedItem->reMorphItem (m_unMorphedDRow) ;
//
//	m_unMorphedItem = item ;
//	m_unMorphedDRow = drow ;
//}
