/***************************************************************************
    file	         : kb_layoutitem.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/

#include	<qpainter.h>
#include	<qpixmap.h>
#include	<qapplication.h>


#include	"kb_layoutitem.h"
#include	"kb_attrgeom.h"
#include	"kb_block.h"
#include	"kb_display.h"


#define		__DEBUG_LAYOUTITEM	(0)
#if		__DEBUG_LAYOUTITEM
#include	"kb_dprintf.h"
#endif


/*  KBCtrlValidator							*/
/*  ---------------							*/
/*  Class used to handle validator blobs.				*/

class	KBCtrlValidator : public QWidget
{
	const KBItem::ValFlag	&m_mode		;
	QPixmap			m_graphicGood	;
	QPixmap			m_graphicBad	;
	const bool		&m_isValid	;

public	:

	KBCtrlValidator
	(	QWidget			*,
		const KBItem::ValFlag	&,
		const bool		&
	)	;

	virtual	void	paintEvent	(QPaintEvent *)	;

	inline	void	setGraphicGood
		(	const QPixmap	&pm
		)
	{
		m_graphicGood	= pm	;
	}
	inline	void	setGraphicBad
		(	const QPixmap	&pm
		)
	{
		m_graphicBad	= pm	;
	}
			
}	;


/*  KBCtrlValidator							*/
/*  KBCtrlValidator							*/
/*		: Constructor for validator flag			*/
/*  parent	: QWidget *	  : Parent widget			*/
/*  mode	: KBItem::ValFlag : Validator mode			*/
/*  isValid	: const bool &	  : Validity flag			*/
/*  (returns)	: KBCtrlValidator :					*/

KBCtrlValidator::KBCtrlValidator
	(	QWidget			*parent,
		const KBItem::ValFlag	&mode,
		const bool		&isValid
	)
	:
	QWidget		(parent ),
	m_mode		(mode   ),
	m_isValid	(isValid)
{
}

/*  KBCtrlValidator							*/
/*  paintEvent	: Paint validator					*/
/*  p		: QPaintEvent *	: The event				*/
/*  (returns)	: void		:					*/

void	KBCtrlValidator::paintEvent
	(	QPaintEvent	*
	)
{
	QPainter painter   (this) ;

	int	 w	   = width () ;
	int	 h	   = height() ;

	/* Clear the validator area and return if no validation. Then	*/
	/* either draw a red/greed blob or use the specified pixmap.	*/
	painter.setBackgroundColor (parentWidget()->paletteBackgroundColor()) ;
	painter.eraseRect  	   (0, 0, w, h) ;

	if (m_mode != KBItem::VFShow)
		return	;

	const QPixmap &pm = m_isValid ? m_graphicGood : m_graphicBad ;
	if (pm.isNull())
		painter.fillRect   (0, 3, 3, h - 6, QBrush(m_isValid ? Qt::green : Qt::red)) ;
	else	painter.drawPixmap (0, 0, pm) ;
}



/*  ------------------------------------------------------------------	*/

/*  KBLayoutItem							*/
/*  KBLayoutItem: Constructor for layout item				*/
/*  widget	: QWidget *	     : Widget to be managed		*/
/*  display	: KBDisplay *	     : Owning display			*/
/*  geom	: const KBAttrGeom & : Geometry information		*/
/*  (returns)	: KBLayoutItem	     :					*/

KBLayoutItem::KBLayoutItem
	(	QWidget			*widget,
		KBDisplay		*display,
		const KBAttrGeom	&geom,
		KBAttrGeom::MgmtMode	mgmtMode
	)
	:
	QWidgetItem	(widget),
	m_display	(display),
	m_geom		(geom),
	m_mgmtMode	(mgmtMode)
{
	DPRINTF
	((	"KBLayoutItem::KBLayoutItem: %08x/%08x\n",
		m_geom.alignment() & Qt::AlignHorizontal_Mask,
		m_geom.alignment() & Qt::AlignVertical_Mask
	))	;

	setAlignment	(m_geom.alignment()) ;

	m_rtti	= RTTIWidget	;
}

KBLayoutItem::~KBLayoutItem ()
{
	DPRINTF
	((	"KBLayoutItem::~KBLayoutItem: called: layout=%p\n",
		(void *)layout()
	))	;
}

/*  KBLayoutItem							*/
/*  rtti	: Get RTTI code						*/
/*  (returns)	: int		: Code					*/

int	KBLayoutItem::rtti ()
{
	return	m_rtti	;
}

/*  KBLayoutItem							*/
/*  setRtti	: Set RTTI code						*/
/*  rtti	: int		: Code					*/
/*  (returns)	: void		:					*/

void	KBLayoutItem::setRtti
	(	int		rtti
	)
{
	m_rtti	= rtti	;
}

QRect	KBLayoutItem::adjustGeometry
	(	const QRect	&rect
	)
{
	/* For controls embedded in a dynamic layout we need to handle	*/
	/* alignment and stuff. This code is taken from the Qt source	*/
	extern	QSize	qSmartMaxSize (const QWidgetItem *i, int align = 0) ;

	int	x	= rect.x     () ;
	int	y	= rect.y     () ;
	int	w	= rect.width () ;
	int	h	= rect.height() ;
	QSize	s	= rect.size  ().boundedTo(qSmartMaxSize(this)) ;

	int	al	= alignment() ;
	int	ah	= QApplication::horizontalAlignment (al) ;

	if ((al & (Qt::AlignHorizontal_Mask|Qt::AlignVertical_Mask)) != 0)
	{
        	QSize pref = widget()->sizeHint().expandedTo(widget()->minimumSize()) ;

		if ((al & Qt::AlignHorizontal_Mask) != 0)
			s.setWidth(QMIN(s.width(), pref.width())) ;

		if ((al & Qt::AlignVertical_Mask  ) != 0)
			if (hasHeightForWidth())
				s.setHeight(QMIN(s.height(), heightForWidth(s.width()))) ;
			else	s.setHeight(QMIN(s.height(), pref.height   ())) ;
        }

	if	((ah & Qt::AlignRight ) != 0) x += (w - s.width ())	;
	else if ((ah & Qt::AlignLeft  ) == 0) x += (w - s.width ()) / 2	;

	if	((al & Qt::AlignBottom) != 0) y += (h - s.height())	;
	else if ((al & Qt::AlignTop   ) == 0) y += (h - s.height()) / 2	;

	return	QRect(x,y,w,h) ;
}

/*  KBLayoutItem							*/
/*  designGeometry							*/
/*		: Get raw design geometry				*/
/*  (returns)	: QRect		: Geometry				*/

QRect	KBLayoutItem::designGeometry ()
{
	/* Note that this is only ever used in dynamic layouts, so we	*/
	/* don't have to concern outselves with stretch and float.	*/
	return	QRect (m_geom.x(), m_geom.y(), m_geom.w(), m_geom.h()) ;
}


/*  KBLayoutItem							*/
/*  mapGeometry	: Map design geometry into parent geometry		*/
/*  rect	: const QRect & : Parent geometry			*/
/*  (returns)	: QRect		: Mapped geometry			*/

QRect	KBLayoutItem::mapGeometry
	(	const QRect	&rect
	)
{
	/* The geometry for the widget is mapped against the parent	*/
	/* geometry, to process for stretching and floating.		*/
	return	m_geom.mapGeometry (rect) ;
}

/*  KBCtrlLayoutItem							*/
/*  setGeometry	: Set geometry						*/
/*  rect	: const QRect &	: Required geometry			*/
/*  (returns)	: void		:					*/

void	KBLayoutItem::setGeometry
	(	const QRect	&rect
	)
{
	m_display->setWidgetGeometry (widget(), rect) ;
}


/*  ------------------------------------------------------------------	*/

KBCtrlLayoutItem::KBCtrlLayoutItem
	(	QWidget			*widget,
		KBDisplay		*display,
		const KBAttrGeom	&geom,
		KBAttrGeom::MgmtMode	mgmtMode,
		uint			drow,
		KBBlock			*block
	)
	:
	KBLayoutItem	(widget, display, geom, mgmtMode),
	m_drow		(drow),
	m_block		(block)
{
	m_rtti		= RTTICtrl ;

	m_validator	= 0	;
	m_helper	= 0	;
	m_isValid	= true	;
	m_showHelper	= false	;
	m_mode		= KBItem::VFNone ;
}

KBCtrlLayoutItem::~KBCtrlLayoutItem ()
{
	DPRINTF
	((	"KBCtrlLayoutItem::~KBCtrlLayoutItem: m_validator=%p\n",
		(void *)m_validator
	))	;
}

void	KBCtrlLayoutItem::setCtrlGeometry
	(	const QRect	&rect
	)
{
	m_ctrlRect = rect	;
}

/*  KBCtrlLayoutItem							*/
/*  mapGeometry	: Map design geometry into parent geometry		*/
/*  rect	: const QRect & : Parent geometry			*/
/*  (returns)	: QRect		: Mapped geometry			*/

QRect	KBCtrlLayoutItem::mapGeometry
	(	const QRect	&rect
	)
{
	/* Map the geometry for the widget as in the base class, then	*/
	/* shift for multiple row offseting.				*/
	QRect	mapped	= m_geom.mapGeometry (rect, m_ctrlRect) ;

//	mapped.moveBy
//		(	m_block->getDisplayDX() * m_drow,
//			m_block->getDisplayDY() * m_drow
//		)	;

	return	mapped	;
}


/*  KBCtrlLayoutItem							*/
/*  setHelper	: Set (or clear) helper widget				*/
/*  helper	: QWidget *	: Helper, or null to clear		*/
/*  (returns)	: void		:					*/

void	KBCtrlLayoutItem::setHelper
	(	QWidget		*helper
	)
{
	m_helper     = helper	;
	m_showHelper = false	;
	if (m_helper != 0) m_helper->hide ()  ;
}

/*  KBCtrlLayoutItem							*/
/*  showHelper	: Show or hide the helper, if any			*/
/*  show	: bool		: True to show				*/
/*  (returns)	: void		:					*/

void	KBCtrlLayoutItem::showHelper
	(	bool		show
	)
{
	DPRINTF
	((	"KBCtrlLayoutItem::showHelper: m_helper=%p show=%d\n",
		(void *)m_helper,
		show
	))	;

	if (m_helper != 0)
	{
		m_showHelper = show ;

		if (show)
			m_helper->show () ;
		else	m_helper->hide () ;

		setGeometry (m_rect) ;
	}
}

void	KBCtrlLayoutItem::showValidator
	(	bool		show
	)
{
	if (m_validator != 0)
		if (show)
			m_validator->show() ;
		else	m_validator->hide() ;
}


/*  KBCtrlLayoutItem							*/
/*  setValidatorMode							*/
/*		: Set mode for validator flag				*/
/*  item	: KBItem *	: Item for which setting		*/
/*  (returns)	: void		:					*/

void	KBCtrlLayoutItem::setValidatorMode
	(	KBItem		*item
	)
{
	QPixmap		graphicGood	;
	QPixmap		graphicBad	;
	KBItem::ValFlag	mode		= item->validatorMode (graphicGood, graphicBad) ;

	DPRINTF
	((	"KBCtrlLayoutItem::setValidatorMode: mode=%d (v=%p)\n",
		mode,
		(void *)m_validator
	))	;

	bool	delta	= false	;

	switch (m_mode = mode)
	{
		case KBItem::VFShow  :
		case KBItem::VFBlank :
			/* For show or blank (which is an empty shown)	*/
			/* validator, create one if needed.		*/
			if (m_validator == 0)
			{
				m_validator = new KBCtrlValidator
						  (	m_display->getDisplayWidget(),
							m_mode,
							m_isValid
						  )	;
				m_validator->resize (6, 20) ;
				delta	= true	;
			}
			m_validator->setGraphicGood (graphicGood) ;
			m_validator->setGraphicBad  (graphicBad ) ;
			break	;

		default	:
			/* Otherwise, destroy one if present.		*/
			if (m_validator != 0)
			{	delete	m_validator ;
				m_validator = 0	    ;
				delta	= true	    ;
			}
			break	;
	}

	/* Delta is set if anything changes, in which case redo the	*/
	/* geometry to get things shown correctly ...			*/
	if (delta)
		setGeometry (m_rect) ;

	/* ... and show the validator if there is one.			*/
	if (m_validator != 0)
		m_validator->show () ;
}

void	KBCtrlLayoutItem::setValid
	(	bool		valid
	)
{
	if (m_validator != 0)
	{
		m_isValid = valid	;
		m_validator->update ()	;
	}
}

bool	KBCtrlLayoutItem::isValidating ()
{
	return	m_validator != 0	;
}

/*  KBCtrlLayoutItem							*/
/*  setGeometry	: Set geometry						*/
/*  rect	: const QRect &	: Required geometry			*/
/*  (returns)	: void		:					*/

void	KBCtrlLayoutItem::setGeometry
	(	const QRect	&rect
	)
{
	QRect	r	= m_mgmtMode == KBAttrGeom::MgmtDynamic ?
					adjustGeometry (rect) :
					rect ;

	int	x	= r.x     ()	;
	int	y	= r.y     ()	;
	int	w	= r.width ()	;
	int	h	= r.height()	;

	DPRINTF
	((	"KBCtrlLayoutItem::setGeometry: (%d,%d,%d,%d) v=%p h=%p sh=%d\n",
		x, y,
		w, h,
		(void *)m_validator,
		(void *)m_helper,
		m_showHelper
	))	;

	/* The geometry is saved; it is used if a helper button is	*/
	/* subsequently shown or hidden.				*/
	m_rect	= rect	;

	/* If there is a validator then it is moved to the top-left	*/
	/* position of the specified geometry, and the geometry x-pos	*/
	/* and width adjusted for the size of the validator		*/
	if (m_validator != 0)
	{
		DPRINTF
		((	"KBCtrlLayoutItem::setGeometry: vw=%d\n",
			m_validator->width()
		))	;

		m_display->setWidgetGeometry
		(	m_validator,
			QRect
			(	x,
				y,
				m_validator->width (),
				h > 20 ? 20 : h
			)
		)	;

		x += m_validator->width() ;
		w -= m_validator->width() ;
	}

	/* If there is a helper and the helper is currently shows, move	*/
	/* the helper appropriately, and reduce the geometry width for	*/
	/* the helper.							*/
	if ((m_helper != 0) && m_showHelper)
	{
		DPRINTF
		((	"KBCtrlLayoutItem::setGeometry: hw=%d\n",
			m_helper->width ()
		))	;

		m_display->setWidgetGeometry
		(	m_helper,
			QRect
			(	x + w - m_helper->width(),
				y,
				m_helper->width (),
				m_helper->height()
			)
		)	;

		w -= m_helper->width () ;
	}

	/* The x, y, ..., settings now contain the geometry for the	*/
	/* control after allowing for a validator and/or helper.	*/
	DPRINTF
	((	"KBCtrlLayoutItem::setGeometry: finally (%d,%d,%d,%d)\n",
		x, y,
		w, h
	))	;

	m_display->setWidgetGeometry (widget(), QRect (x, y, w, h)) ;
}

void	KBCtrlLayoutItem::dropValidator ()
{
	DELOBJ	(m_validator) ;
}
