/***************************************************************************
    file	         : kb_parse.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/

#include	<stdio.h>
#include	<stdlib.h>
#include	<stdarg.h>


#include	<qpopupmenu.h>
#include	<qfile.h>
#include	<qtextstream.h>
#include	<qregexp.h>

#include	"kb_classes.h"
#include	"kb_type.h"
#include	"kb_value.h"
#include	"kb_object.h"
#include	"kb_slot.h"
#include	"kb_test.h"
#include	"kb_location.h"
#include	"kb_macro.h"
#include	"kb_nodereg.h"
#include	"kb_popupmenu.h"

#include	"kb_parse.h"

#include	"kb_form.h"
#include	"kb_report.h"

#if		! __KB_RUNTIME
#include	"kb_toolbox.h"
#endif


#define		__DEBUG_PARSE	(0)
#if		__DEBUG_PARSE
#include	"kb_dprintf.h"
#endif


class	KBAttrSortList	: public QPtrList<KBAttr>
{

protected	:

	virtual	int	compareItems (QCollection::Item, QCollection::Item) ;
}	;

int	KBAttrSortList::compareItems
	(	QCollection::Item	item1,
		QCollection::Item	item2
	)
{
	return	(int)((KBAttr *)item1)->getOrder() -
		(int)((KBAttr *)item2)->getOrder() ;
}

/*  ------------------------------------------------------------------  */

static	QPtrList<NodeSpec>	*nodeSpecs	;
static	QDict<uint>		nodeFlags	;


/*  registerNode: Register a node type					*/
/*  nodeName	: cchar *	: XML node name				*/
/*  msgText	: cchar *	: Text for popup menu			*/
/*  popupFunc	: MKPOPUP	: Popup menu creation function		*/
/*  nodeFunc	: MKNODE	: Node creation function		*/
/*  flags	: uint		: Applicability flags			*/
/*  (returns)	:		:					*/

LIBKBASE_API
	void	registerNode
	(	cchar		*nodeName,
		cchar		*msgText,
		MKPOPUP		popupFunc,
		MKNODE		nodeFunc,
		uint		flags
	)
{
	DPRINTF
	((	"registerNode: [%12s] [%s]\n",
		nodeName,
		(cchar *)msgText
	))	;

	NodeSpec *ns = new NodeSpec 	;

	ns->m_nodeName  = nodeName	;
	ns->m_msgText   = msgText 	;
	ns->m_popupFunc = popupFunc	;
	ns->m_nodeFunc  = nodeFunc	;
	ns->m_wizardFunc= 0		;
	ns->m_flags     = flags		;
	ns->m_id	= -1		;

	if (nodeSpecs == 0) nodeSpecs = new QPtrList<NodeSpec> ;
	nodeSpecs->append (ns)		;
}

/*  GetNodeFlags: Get flags for named node type				*/
/*  name	: const QString & : Node name				*/
/*  (returns)	: uint		  : Flags, zero if node not found	*/

LIBKBASE_API
	uint	GetNodeFlags
	(	const QString	&name
	)
{
	uint	*flags = nodeFlags.find (name)	;

	if (flags == 0)
	{
		if (name == "KBForm"  ) return KNF_FORM   ;
		if (name == "KBReport") return KNF_REPORT ;
		if (name == "KBQuery" ) return KNF_QUERY  ;
		return	0 ;
	}

	return	*flags	;
}

/*  LoadNodeFunc: Load node definition functions			*/
/*  pNodeFns	: uint		   : Number checked previously		*/
/*  mask	: int		   : Node type mask			*/
/*  nodeDict	: QDict<NodeSpec>& : Node definition dictionary		*/
/*  (returns)	: uint		   : Number now checked			*/

LIBKBASE_API
	uint LoadNodeFuncs
	(	uint		pNodeFns,
		int		mask,
		QDict<NodeSpec>	&nodeDict
	)
{
#if	0
	static	bool	first	= true	;

	if (first)
	{
		registerKBaseNodes ()	;
		first	= false		;
	}
#endif

	while (pNodeFns < nodeSpecs->count())
	{
		const NodeSpec *spec = nodeSpecs->at(pNodeFns) ;

		if ((spec->m_flags & mask ) != 0)
		{
			nodeDict .insert (spec->m_nodeName, spec) ;
			nodeFlags.insert (spec->m_nodeName, &spec->m_flags) ;

#if	! __KB_RUNTIME
			if (spec->m_msgText != 0)
				KBToolBox::self()->appendNode (mask, spec) ;
#endif
		}

		pNodeFns += 1 ;
	}

	return	pNodeFns ;
}

/*  ------------------------------------------------------------------  */

/*  KBSAXHandler							*/
/*  KBSAXHandler: Constructor for base parser class			*/
/*  what	: cchar *	 : What is being parsed			*/
/*  location	: KBLocation &	 : Database location			*/
/*  parent	: KBNode *	 : Initial parent node			*/
/*  (returns)	: KBSAXHandler	 :					*/

KBSAXHandler::KBSAXHandler
	(	cchar			*what,
		const KBLocation	&location,
		KBNode			*parent
	)
	:
	m_what		(what),
	m_location	(location)
{
	m_kbTOS		= parent	;
	m_kbTop		= 0		;
	m_kbSlot	= 0		;
	m_kbTest	= 0		;
	m_kbEvent	= 0		;
	m_kbMacro	= 0		;
	m_kbAttr	= 0		;
	m_state		= Normal	;
	m_bGotErr	= false		;
}

/*  KBSAXHandler							*/
/*  parseFile	: Parse document					*/
/*  xmlSource	: QXmlInputSource & : XML input source			*/
/*  (returns)	: bool		    : Success				*/

bool	KBSAXHandler::parse
	(	QXmlInputSource	&xmlSource
	)
{
	QXmlSimpleReader xmlReader		;

	xmlReader.setContentHandler(this     )	;
	xmlReader.parse		   (xmlSource)	;

	/* See if an error was logged. If so then delete any parse tree	*/
	/* that was forthcoming, then return failure.			*/
	if (m_bGotErr)
	{
		if (m_kbTop != 0) delete m_kbTop ;
		return	false	;
	}

	/* See if any parse tree at all resulted. It will be missing if	*/
	/* the document is empty.					*/
	if (m_kbTop == 0)
	{
		m_lError  = KBError
			  (	KBError::Error,
			  	QString(TR("%1 is empty")).arg(m_what),
			  	QString::null,
			  	__ERRLOCN
			  )	;
		return	false	;
	}

	return	true	;
}

/*  KBSAXHandler							*/
/*  parseFile	: Parse document					*/
/*  document	: const QString &: Document path			*/
/*  (returns)	: bool		 : Success				*/

bool	KBSAXHandler::parseFile
	(	const QString	&document
	)
{
	/* Check that the document actually exists, if not then we can	*/
	/* harldy parse it!						*/
	if (!QFile::exists (document))
	{
		m_lError  = KBError
			  (	KBError::Error,
			  	QString(TR("%1 \"%2\" does not exist"))
			  		.arg(m_what)
			  		.arg(document),
			  	QString::null,
			  	__ERRLOCN
			  )	;
		return	false	;
	}


	/* Set up the Qt SAX parser and parse the document ...		*/
	QFile		 xmlFile   (document )	;
	QXmlInputSource	 xmlSource (xmlFile  )	;

	return	parse	(xmlSource) ;
}

/*  KBSAXHandler							*/
/*  parseText	: Parse document					*/
/*  document	: QByteArray &	: Document text				*/
/*  (returns)	: bool		: Success				*/

bool	KBSAXHandler::parseText
	(	QByteArray	&document
	)
{
	/* Set up the Qt SAX parser and parse the document ...		*/
	QTextStream	 xmlText   (document, IO_ReadOnly) ;
	QXmlInputSource	 xmlSource (xmlText ) ;

	return	parse	(xmlSource) ;
}

/*  KBSAXHandler							*/
/*  setErrMessage: Set an error message					*/
/*  e		 : const QXmlParseException &				*/ 
/*				: SAX parser exception information	*/
/*  (returns)	 : void		:					*/

void	KBSAXHandler::setErrMessage
	(	const QXmlParseException &e
	)
{
	m_lError = KBError
		   (	KBError::Error,
			QString  (TR("%1 parsing error at line %2, column %3"))
			   		.arg(m_what)
					.arg(e.lineNumber  ())
					.arg(e.columnNumber()),
			e.message(),
			__ERRLOCN
		   )	  ;
	m_bGotErr  = true ;
}

/*  KBSAXHandler							*/
/*  setErrMessage: Set an error message					*/
/*  msg		 : const QString&: Error message			*/
/*  arg		 : const QString&: Error message argument		*/
/*  (returns)	 : void		 :					*/

void	KBSAXHandler::setErrMessage
	(	const QString	&msg,
		const QString	&arg
	)
{
	m_lError  = KBError
		  (	KBError::Error,
			QString(TR("Error parsing %1"))	.arg(m_what),
			QString(TR(msg))		.arg(arg),
			__ERRLOCN
		  )	;
	m_bGotErr = true;
}

/*  KBSAXHandler							*/
/*  setErrMessage: Set an error message					*/
/*  msg		 : const QString&: Error message			*/
/*  state	 : ParseState	 : Error state				*/
/*  (returns)	 : void		 :					*/

void	KBSAXHandler::setErrMessage
	(	const QString	&msg,
		ParseState	state
	)
{
	QString	arg	;

	switch (state)
	{
		case Normal	 : arg = "Normal"	; break ;
		case InEvent	 : arg = "InEvent"	; break ;
		case InEvent2	 : arg = "InEvent2"	; break ;
		case InTest	 : arg = "InTest"	; break ;
		case InTest2	 : arg = "InTest2"	; break ;
		case InSlot	 : arg = "InSlot"	; break ;
		case InSlotLink	 : arg = "InSlotLink"	; break ;
		case InSlotCode	 : arg = "InSlotCode"	; break ;
		case InMacro	 : arg = "InMacro"	; break ;
		case InMacroInstr: arg = "InMacroInstr"	; break ;
		case InMacroArg	 : arg = "InMacroArg"	; break ;
		case InAttribute : arg = "InAttribute"	; break ;
		default		 :
			arg = QString("Unknown (%1)").arg(state) ;
			break	 ;
	}

	setErrMessage (msg, QString(": state %1").arg(arg)) ;
	m_bGotErr = true;
}

/*  KBSAXHandler							*/
/*  endElement	: Handle end of element					*/
/*  URI		: const QString &	: Namespace URI			*/
/*  localName	: const QString &	:				*/
/*  qName	: const QString &	: Element name			*/
/*  (returns)	: bool			: Success			*/

bool	KBSAXHandler::endElement
	(	const	QString		&,
		const	QString		&,
		const	QString		&qName
	)
{
	DPRINTF
	((	"KBSAXHandler::endElement: [%s] [%p][%p][%p][%p]\n",
		(cchar *)qName,
		(void  *)m_kbTOS,
		(void  *)m_kbSlot,
		(void  *)m_kbEvent,
		(void  *)m_kbAttr
	))	;

	if (m_kbTOS == 0)
	{	setErrMessage (TR("Internal error: "), TR("Empty parse stack")) ;
		return	false	 ;
	}

	/* Check for special states related to processing slots, macros	*/
	/* and events.							*/
	switch (m_state)
	{
		case InSlot	 :
			/* End of a slot. Tidy the code and return to	*/
			/* the normal state.				*/
			m_kbSlot->tidy ()	;
			m_kbSlot  = 0		;
			m_state	  = Normal	;
			return	  true 		;

		case InSlotLink	  :
			/* End of a slot link. Just pop back to the	*/
			/* main slot state.				*/
			m_state	  = InSlot	;
			return	  true		;

		case InSlotCode   :
			/* End of slot code. Just pop back to the main	*/
			/* slot state.					*/
			m_state	  = InSlot	;
			return	  true		;

		case InMacro	  :
			/* End of a macro. Store the instructions in	*/
			/* assocated event and return to the normal	*/
			/* state.					*/
			if (m_kbTest  != 0) m_kbTest ->setMacro (m_kbMacro) ;
			if (m_kbEvent != 0) m_kbEvent->setMacro (m_kbMacro) ;
			m_kbMacro = 0		;
			m_kbEvent = 0		;
			m_kbTest  = 0		;
			m_state	  = Normal	;
			return	  true		;
		
		case InMacroInstr :
			/* End of a macro instruction. Append it to the	*/
			/* macro (checking for errors), then clean up	*/
			/* and return to the main macro state.		*/
			if (!m_kbMacro->append
				(	m_macroAction,
					m_macroArgs,
					m_macroComment,
					m_lError
				))
			{
				setErrMessage
				(	QString(TR("Invalid macro '%1' ignored")).arg(m_macroAction),
					""
				)	;
				return	false	;
			}

			m_macroAction  = QString::null ;
			m_macroComment = QString::null ;
			m_macroArgs.clear() 	;
			m_state	  = InMacro	;
			return	  true		;

		case InMacroArg	   :
			/* End of a macro argument. Add the accumulated	*/
			/* characters to the macro arguments and return	*/
			/* to the macro instruction state.		*/
			m_macroArgs.append (m_chars)	;
			m_chars   = QString::null	;
			m_state	  = InMacroInstr;
			return	  true 		;

		case InEvent	   :
		case InEvent2	   :
			/* End of an event. Tidy up the stored text and	*/
			/* return to the normal state.			*/
			m_kbEvent->tidy ()	;
			m_kbEvent = 0		;
			m_state	  = Normal	;
			return	  true		;

		case InTest	   :
		case InTest2	   :
			/* End of a test. Tidy up the stored text and	*/
			/* return to the normal state.			*/
			m_kbTest ->tidy ()	;
			m_kbTest  = 0		;
			m_state	  = Normal	;
			return	  true		;

		case InAttribute	   :
			/* End of an attribute. Tidy up the stored	*/
			/* text and return to the normal state.		*/
			m_kbAttr->setValue (m_kbAttr->getValue().stripWhiteSpace()) ;
			m_kbAttr  = 0		;
			m_state	  = Normal	;
			return	  true		;

		default	:
			break	;
	}

	if (qName == "extra")
		return	true	;

	m_kbTOS = m_kbTOS->endParse () ;
	return	true ;
}

/*  KBSAXHandler							*/
/*  processNode	: Process an XML node					*/
/*  qName	: const QString	&	 : Node name			*/
/*  aList	: const QDict<QString>	 : Attributes			*/
/*  nDict	: const QDict<NodeSpec> &: Node specifications		*/
/*  (returns)	: bool			 : Success			*/

bool	KBSAXHandler::processNode
	(	const QString		&qName,
		const QDict<QString>	&aList,
		const QDict<NodeSpec>	&nDict
	)
{
	DPRINTF
	((	"KBSAXHandler::processNode: [%s]\n",
		(cchar *)qName
	))	;

	m_chars	     = QString::null ;


	/* SLOTS							*/
	/* -----							*/

	/* First there is some special processing to handle slots, 	*/
	/* which appear as child nodes of objects in the XML. The first	*/
	/* and main case is the "slot" element itself. This cannot be	*/
	/* a top-level element, cannot be nested, and must have a	*/
	/* "name" attribute.						*/
	if (qName == "slot")
	{
		if ((m_kbTOS == 0) || (m_kbTOS->isObject() == 0))
		{	setErrMessage (TR("Slot outside object"), "") ;
			return	false ;
		}

		const QString	*name	= aList.find("name") ;
		const QString	*l2	= aList.find("l2"  ) ;

		if (name == 0)
		{	setErrMessage (TR("Missing slot name"), "") ;
			return	false ;
		}

		if (m_state != Normal)
		{	setErrMessage (TR("Nested slot elements"), m_state) ;
			return	false ;
		}

		m_kbSlot = new KBSlot
			   (	m_kbTOS->isObject(),
				*name,
				l2 == 0 ? false : l2->toUInt()
			   )	;
		m_state	 = InSlot ;
		return	 true	  ;
	}

	/* Next is a "slotlink" which specifies a linkage from the slot	*/
	/* to another target object. This must be inside a slot, and	*/
	/* must have all four attributes (link name, target name,	*/
	/* target enabled and enabled) specified.			*/
	if (qName == "slotlink")
	{
		if (m_state != InSlot)
		{	setErrMessage (TR("Slot link outside slot"), m_state) ;
			return	false ;
		}

		const QString	*name	= aList.find("name"   ) ;
		const QString	*target	= aList.find("target" ) ;
		const QString	*event	= aList.find("event"  ) ;
		const QString	*enabled= aList.find("enabled") ;

		if ((name == 0) || (target == 0) || (event == 0) || (enabled == 0))
		{	setErrMessage (TR("Malformed slot linkage"), "") ;
			return	false ;
		}

		m_kbSlot->addLinkage
		(	*name,
			*target,
			*event,
			enabled->toInt()
		)	;

		m_state = InSlotLink ;
		return	true	     ;
	}

	/* Lastly there is the code of the slot. The code appears as	*/
	/* embedded text, so provided that we are inside a slot, set	*/
	/* the in-slot-code flag.					*/
	if (qName == "slotcode")
	{
		if (m_state != InSlot)
		{	setErrMessage (TR("Slot code outside slot"), m_state) ;
			return	false ;
		}

		m_state	= InSlotCode  ;
		return	true	      ;
	}


	/* EVENTS							*/
	/* ------							*/
	/* Now similarly for nested event code. This cannot be a	*/
	/* top-level element, cannot be nested, and must have a "name"	*/
	/* attribute.							*/
	if ((qName == "event") || (qName == "event2"))
	{
		if ((m_kbTOS == 0) || (m_kbTOS->isObject() == 0))
		{	setErrMessage (TR("Event outside object"), "") ;
			return	false ;
		}

		const QString	*name	= aList.find("name") ;
		if (name == 0)
		{	setErrMessage (TR("Missing event name"), "") ;
			return	false ;
		}

		if (m_kbEvent != 0)
		{	setErrMessage (TR("Nested event elements"), "") ;
			return	false ;
		}

		KBAttr	*attr = m_kbTOS->getAttr(*name) ;
		if ((attr == 0) || (attr->isEvent() == 0))
		{	setErrMessage (TR("No such event attribute"), "") ;
			return	false ;
		}

		m_kbEvent = attr->isEvent() ;
		m_state	  = qName == "event" ? InEvent : InEvent2 ;
		return	true ;
	}

	/* TESTS							*/
	/* ------							*/
	/* Now similarly for nested test code. This cannot be a		*/
	/* top-level element, cannot be nested, and must have a "name"	*/
	/* attribute.							*/
	if ((qName == "test") || (qName == "test2"))
	{
		if ((m_kbTOS == 0) || (m_kbTOS->isObject() == 0))
		{	setErrMessage (TR("Test outside object"), "") ;
			return	false ;
		}

		const QString	*name	= aList.find("name"   ) ;
		const QString	*comment= aList.find("comment") ;
		if (name == 0)
		{	setErrMessage (TR("Missing test name"), "") ;
			return	false ;
		}

		if (m_kbTest != 0)
		{	setErrMessage (TR("Nested test elements"), "") ;
			return	false ;
		}

		KBAttr	*attr = m_kbTOS->getAttr(*name) ;
		if ((attr != 0) && (attr->isTest() == 0))
		{	setErrMessage (TR("Test exists as attribute"), "") ;
			return	false ;
		}

		if (attr == 0)
			m_kbTest = new KBTest  (m_kbTOS, *name) ;
		else	m_kbTest = attr->isTest() ;

		if (comment != 0)
			m_kbTest->setComment (*comment) ;

		m_state	  = qName == "test" ? InTest : InTest2 ;
		return	true ;
	}

	/* MACROS							*/
	/* ------							*/
	/* Code for handling macros. A macro is defined by a "macro"	*/
	/* node, which references an event, and should contain a number	*/
	/* of "instruction" nodes that specify the instructions. These	*/
	/* in turn contain "arg" nodes for each macro argument.		*/
	if (qName == "macro")
	{
		if ((m_kbTOS == 0) || (m_kbTOS->isObject() == 0))
		{	setErrMessage (TR("Macro outside object"), "") ;
			return	false ;
		}

		const QString	*name	= aList.find("name"   ) ;
		const QString	*test	= aList.find("test"   ) ;
		const QString	*comment= aList.find("comment") ;
		bool		isTest	= (test != 0) && (test->toInt() != 0) ;

		if (name == 0)
		{	setErrMessage (TR("Missing macro name"), "") ;
			return	false ;
		}

		if (m_kbMacro != 0)
		{	setErrMessage (TR("Nested macro elements"), "") ;
			return	false ;
		}

		if (isTest)
		{
			KBAttr	*attr = m_kbTOS->getAttr(*name) ;
			if ((attr != 0) && (attr->isTest() == 0))
			{	setErrMessage (TR("Macro test exists as attribute"), "") ;
				return	false ;
			}

			if (attr == 0)
				m_kbTest = new KBTest  (m_kbTOS, *name) ;
			else	m_kbTest = attr->isTest() ;

			if (comment != 0)
				m_kbTest->setComment (*comment) ;

			m_kbMacro = new KBMacroExec
					(	m_location.dbInfo(),
						m_location.server(),
						"tests"
					)	;

			m_kbMacro->setName (*name) ;
		}
		else
		{
			KBAttr	*attr = m_kbTOS->getAttr(*name) ;
			if ((attr == 0) || (attr->isEvent() == 0))
			{	setErrMessage (TR("No such macro attribute"), "") ;
				return	false ;
			}

			m_kbEvent = attr->isEvent() ;

			m_kbMacro = new KBMacroExec
					(	m_location.dbInfo(),
						m_location.server(),
						"standard"
					)	;

			m_kbMacro->setName (*name) ;
		}

		m_state	  = InMacro 	    ;
		return	true ;
	}

	if (qName == "instruction")
	{
		if (m_state != InMacro)
		{	setErrMessage (TR("Macro code outside macro"), m_state) ;
			return	false ;
		}

		const QString	*action	= aList.find("action" ) ;
		const QString	*comment= aList.find("comment") ;

		if (action == 0)
		{	setErrMessage (TR("Unnamed macro code"), "") ;
			return	false ;
		}

		m_macroAction	= *action ;
		m_macroComment	= comment == 0 ? QString::null : *comment ;
		m_macroArgs.clear ()	  ;
		m_state	= InMacroInstr	  ;
		return	true ;
	}

	if (qName == "arg")
	{
		if (m_state != InMacroInstr)
		{	setErrMessage (TR("Macro argument outside macro"), m_state) ;
			return	false ;
		}

		m_state	= InMacroArg  ;
		return	true	      ;
	}


	/* ATTRIBUTES							*/
	/* ----------							*/

	/* This element allows setting attributes (properties) from	*/
	/* XML text elements rather than XML element attributes. It is	*/
	/* mainly intended for setting things like the "notes" in	*/
	/* component definition. Beware that the value is *not* present	*/
	/* when the node is created, since at that point the SAX parser	*/
	/* will not have reached the attribute node.			*/
	if (qName == "attribute")
	{
		if ((m_kbTOS == 0) || (m_kbTOS->isObject() == 0))
		{	setErrMessage (TR("Attribute outside object"), "") ;
			return	false ;
		}

		const QString	*name	= aList.find("name") ;
		if (name == 0)
		{	setErrMessage (TR("Missing attribute name"), "") ;
			return	false ;
		}

		if (m_kbAttr != 0)
		{	setErrMessage (TR("Nested attribute elements"), "") ;
			return	false ;
		}

		KBAttr	*attr = m_kbTOS->getAttr(*name) ;
		if (attr == 0)
		{	setErrMessage (TR("No such attribute"), "") ;
			return	false ;
		}

		m_kbAttr  = attr	;
		m_state	  = InAttribute	;
		return	true ;
	}

	/* EXTRA							*/
	/* -----							*/

	/* Sometimes we need to be able to pass extended values to	*/
	/* properties (eg., geometry cell setup). This is handled by	*/
	/* and "extra" node which should contain an attribute which	*/
	/* names the property, plus a set of additional attributes that	*/
	/* give the information.					*/

	if (qName == "extra")
	{
		if ((m_kbTOS == 0) || (m_kbTOS->isObject() == 0))
		{	setErrMessage (TR("Extra attribute outside object"), "") ;
			return	false ;
		}

		if (!m_kbTOS->setExtraAttribs (aList))
		{	setErrMessage (TR("Error setting extra attribute" ), "") ;
			return	false ;
		}

		return	true	;
	}

	/* Done the special stuff, now do the normal processing of	*/
	/* KBNode-derived objects.					*/
	NodeSpec *spec = nDict.find (qName) ;
	if (spec == 0)
	{
		setErrMessage (TR("Unknown XML element \"%1\""), qName) ;
		return	false	 ;
	}


	m_kbTOS = (*spec->m_nodeFunc) (m_kbTOS, aList, 0) ;
	m_kbTOS->startParse () ;

	if (m_kbTop == 0) m_kbTop = m_kbTOS ;
	return	true ;

}

/*  KBSAXHandler							*/
/*  characters	: Handle a block of characters				*/
/*  code	: const QString & : Characters				*/
/*  (returns)	: bool		  : Success				*/

bool	KBSAXHandler::characters
	(	const QString		&code
	)
{
	if (m_state == InSlotCode)
	{
		m_kbSlot ->setCode (code, true) ;
		return	true ;
	}

	if (m_state == InEvent )
	{
		m_kbEvent->setCode (code, true) ;
		return	true ;
	}

	if (m_state == InEvent2)
	{
		m_kbEvent->setCode2 (code, true) ;
		return	true ;
	}

	if (m_state == InTest  )
	{
		m_kbTest ->setCode (code, true) ;
		return	true ;
	}

	if (m_state == InTest2 )
	{
		m_kbTest ->setCode2 (code, true) ;
		return	true ;
	}

	if (m_state == InAttribute)
	{
		m_kbAttr->setValue (m_kbAttr->getValue() + code) ;
		return	true ;
	}

	m_chars	+= code	;
	return	true	;
}

bool	KBSAXHandler::error
	(	const QXmlParseException &e
	)
{
	setErrMessage (e) ;
	return	true	;
}

bool	KBSAXHandler::warning
	(	const QXmlParseException &e
	)
{
	setErrMessage (e) ;
	return	true	;
}

bool	KBSAXHandler::fatalError
	(	const QXmlParseException &e
	)
{
	setErrMessage (e) ;
	return	false	 ;
}


/*  ------------------------------------------------------------------  */

/*  makeDesignMenu							*/
/*		: Insert node entries into popup design menu		*/
/*  nodeSpec	: QDict<NodeSpec>& : Set of possible node specificators	*/
/*  popup	: QPopupMenu *	   : Popup menu being constructed	*/
/*  receiver	: QObject *	   : Received for menu signals		*/
/*  flags	: uint		   : Inclusion flags			*/
/*  (returns)	: void		   :					*/

LIBKBASE_API
	void	makeDesignMenu
	(	QDict<NodeSpec>	&nodeSpec,
		QPopupMenu	*popup,
		QObject		*receiver,
		uint		flags,
		Qt::ButtonState	*bState
	)
{
	QDictIterator<NodeSpec> nsIter	(nodeSpec) ;
	NodeSpec		*nsItem	;

	QPtrList<NodeSpec>	extra	;
	bool			sep	= false	;

	/* Non-data entries, for instance for buttons and labels, are	*/
	/* inserted first. We also skip any that do not have the	*/
	/* appropriate (typically KF_FORM or KF_REPORT) flag set.	*/
	for (nsIter.toFirst () ; (nsItem = nsIter.current()) != 0 ; nsIter += 1)
	{
		if ((nsItem->m_flags & KNF_DATA ) != 0)
		{
			continue ;
		}

		/* Items flagged as extra are noted for inclusion later	*/
		/* in their own sub-menu.				*/
		if ((nsItem->m_flags & KNF_EXTRA) != 0)
		{
			extra.append (nsItem) ;
			continue ;
		}

		DPRINTF
		((	"makeDesignMenu: [%12s] [%p][%p] [%08x:%08x]\n",
			(cchar *)nsItem->m_msgText,
			(void  *)nsItem->m_popupFunc,
			(void  *)nsItem->m_wizardFunc,
			nsItem->m_flags,
			flags
		))	;

		if ((nsItem->m_flags &   flags) == 0)
			continue ;

		/* If there is no text then this node is parsed but	*/
		/* does not have a popup menu item.			*/
		if (nsItem->m_msgText == 0)
			continue	;

		/* Node spec defines its own callback function to build	*/
		/* the popup menu, so use that.				*/
		if (nsItem->m_popupFunc != 0)
		{
			NodeSpec    *d	;
			KBPopupMenu *s	= (*nsItem->m_popupFunc) (popup, receiver, bState, &d) ;
			popup->insertItem (TR(nsItem->m_msgText), s) ;
			continue	;
		}


		/* Node spec defines the menu text so insert straight	*/
		/* into the popup menu. Set the separator flag so that	*/
		/* a separator will appear between non-data and data	*/
		/* entries.						*/
		popup->insertItem
		(	TR(nsItem->m_msgText),
			receiver,
			SLOT (newNode(int)),
			0,
			nodeSpecToId(nsItem)
		)	;

		sep	= true	;
	}

	/* Now scan for data items. This is essentially the same as the	*/
	/* loop above ...						*/
	for (nsIter.toFirst () ; (nsItem = nsIter.current()) != 0 ; nsIter += 1)
	{
		if ((nsItem->m_flags & KNF_DATA ) == 0)
		{
			continue ;
		}

		if ((nsItem->m_flags & KNF_EXTRA) != 0)
		{
			extra.append (nsItem) ;
			continue ;
		}

		DPRINTF
		((	"makeDesignMenu: [%12s] [%p][%p] [%08x:%08x]\n",
			(cchar *)nsItem->m_msgText,
			(void  *)nsItem->m_popupFunc,
			(void  *)nsItem->m_wizardFunc,
			nsItem->m_flags,
			flags
		))	;

		if ((nsItem->m_flags &   flags) == 0)
			continue ;

		if (nsItem->m_popupFunc != 0)
		{
			NodeSpec    *d	;
			KBPopupMenu *s	= (*nsItem->m_popupFunc) (popup, receiver, bState, &d) ;
			popup->insertItem (TR(nsItem->m_msgText), s) ;
			continue ;
		}

		if (nsItem->m_msgText == 0)
			continue	;

		/* Add a separator into the menu before the first data	*/
		/* item added.						*/
		if (sep)
		{	popup->insertSeparator () ;
			sep	= false	;
		}

		popup->insertItem
		(	nsItem->m_msgText,
			receiver,
			SLOT (newNode(int)),
			0,
			nodeSpecToId(nsItem)
		)	;
	}

	if (extra.count() > 0)
	{
		QPopupMenu *em = new QPopupMenu (popup) ;

		LITER
		(	NodeSpec,
			extra,
			exItem,

			em->insertItem
			(	exItem->m_msgText,
				receiver,
				SLOT (newNode(int)),
				0,
				nodeSpecToId(exItem)
			)	;
		)

		popup->insertItem (TR("Extra"), em) ;
	}
}

