/***************************************************************************
    file	         : kb_urlrequest.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/


#include	"kb_classes.h"
#include	"kb_deleter.h"
#include	"kb_object.h"
#include	"kb_slot.h"
#include	"kb_script.h"

#if		__KB_AUTOMOC
#include	"kb_urlrequest.moc"
#else		
#include	"kb_urlrequest.h"
#endif


#define		__DEBUG_URLREQUEST	(0)
#if		__DEBUG_URLREQUEST
#include	"kb_dprintf.h"
#endif


/*  KBURLRequest							*/
/*  KBURLRequest: Constructor for HTTP requestor			*/
/*  (returns)	: KBURLRequest :					*/

KBURLRequest::KBURLRequest
	(	KBObject	*notify,
		const QString	&slotName
	)
	:
	QHttp		(notify),
	m_notify	(notify),
	m_slotName	(slotName)
{
	DPRINTF
	((	"KBURLRequest::KBURLRequest: slot=[%s]\n",
		slotName.latin1()
	))	;

	m_URLSetHostID	= -1	;
	m_URLGetID	= -1	;
	m_active	= false	;
	m_slot		= 0	;
	m_inSlot	= false	;
	m_immolate	= false	;

	connect
	(	this,
		SIGNAL	(requestStarted	 (int)),
		SLOT	(slotURLStarted  (int))
	)	;
	connect
	(	this,
		SIGNAL	(requestFinished (int, bool)),
		SLOT	(slotURLFinished (int, bool))
	)	;
	connect
	(	this,
		SIGNAL	(dataReadProgress(int, int)),
		SLOT	(slotURLProgress (int, int))
	)	;
	connect
	(	this,
		SIGNAL	(readyRead	 (const QHttpResponseHeader &)),
		SLOT	(slotURLReady	 (const QHttpResponseHeader &))
	)	;
	connect
	(	notify->getRoot(),
		SIGNAL	(sigShowingAs	 (KB::ShowAs)),
		SLOT	(slotShowingAs	 (KB::ShowAs))
	)	;
}

/*  KBURLRequest							*/
/*  ~KBURLRequest: Destructor for URL downloader			*/
/*  (returns)	 :		:					*/

KBURLRequest::~KBURLRequest ()
{
	DPRINTF
	((	"KBURLRequest::~KBURLRequest: called\n"
	))	;

	disconnect(this) ;
}

/*  KBURLRequest							*/
/*  slotShowingAs: Document display mode changes			*/
/*  showAs	 : KB::ShowAs	: New mode				*/
/*  (returns)	 : void		:					*/

void	KBURLRequest::slotShowingAs
	(	KB::ShowAs	showAs
	)
{
	DPRINTF
	((	"KBURLRequest::slotShowingAs: called: showAs=%d m_inSlot=%d\n",
		showAs,
		m_inSlot
	))	;

	/* Normally we just halt when the document from data view, but	*/
	/* if we are in a slot then pend it until we exit.		*/
	if (showAs != KB::ShowAsData)
		if (m_inSlot)
			m_immolate = true ;
		else	halt () ;
}

/*  KBURLRequest							*/
/*  notifySlot	: Set notification back through slot			*/
/*  state	: KBURLRequest::State : Current state			*/
/*  message	: const QString &     : Notification message		*/
/*  (returns)	: void		      :					*/

void	KBURLRequest::notifySlot
	(	KBURLRequest::State	state,
		const QString		&message
	)
{
	if (m_slot != 0)
	{
		KBScriptError	*rc	= 0 ;
		KBValue		resval	;
		KBValue		args[2]	;

		m_inSlot = true		;
		args[0]	 = (int)state	;
		args[1]	 = message	;

		m_slot->eventSignal (m_notify, "httpreq", 2, args, resval, rc, 0) ;

		if (rc != 0) 
			KBScriptError::processError (rc, KBScriptError::Normal) ;

		m_inSlot = false	;

		if (m_immolate) halt ()	;
	}
}

void	KBURLRequest::setProgress
	(	const QString	&progress
	)
{
	notifySlot (ProgressMessage, progress) ;
}

void	KBURLRequest::setProgress
	(	int		done,
		int		total
	)
{
	notifySlot (Progress, QString("%1:%2").arg(done).arg(total)) ;
}

/*  KBURLRequest							*/
/*  setURLError: Display HTTP error					*/
/*  (returns)	: void		:					*/

void	KBURLRequest::setURLError ()
{
	QString	msg	= TR("Unrecognised error occurred") ;

	switch (error())
	{
		case QHttp::NoError :
			msg	= TR("No error occured") ;
			break	;

		case QHttp::HostNotFound :
			msg	= TR("Host not found") ;
			break	;

		case QHttp::ConnectionRefused :
			msg	= TR("Connection to host refused") ;
			break	;

		case QHttp::UnexpectedClose :
			msg	= TR("Host unexpectedly closed connection") ;
			break	;

		case QHttp::InvalidResponseHeader :
			msg	= TR("Invalid response from host") ;
			break	;

		case QHttp::WrongContentLength :
			msg	= TR("Host sent wrong content length") ;
			break	;

		case QHttp::Aborted :
			msg	= TR("Transfer aborted") ;
			break	;

		case QHttp::UnknownError :
			msg	= TR("Unknown error occurred") ;
			break	;

		default	:
			break	;
	}

	notifySlot (Error, msg) ;
	halt	   ()		;
}

/*  KBURLRequest							*/
/*  slotURLStarted							*/
/*		: QHttp object signals start of an operation		*/
/*  id		: int		: Operation identifier			*/
/*  (returns)	: void		:					*/

void	KBURLRequest::slotURLStarted
	(	int		id
	)
{
	if (id == m_URLSetHostID)
	{
		DPRINTF
		((	"KBURLRequest::slotURLStarted: Connecting to host\n"
		))	;
		setProgress (TR("Connecting to remote host")) ;
		return	;
	}
	if (id == m_URLGetID)
	{
		DPRINTF
		((	"KBURLRequest::slotURLStarted: retrieveing data\n"
		))	;
		setProgress (QString(TR("Retrieving %1")).arg(m_what)) ;
		return	;
	}
}

/*  KBURLRequest							*/
/*  slotURLFinished							*/
/*		: QHttp object signals completion of an operation	*/
/*  id		: int		: Operation identifier			*/
/*  hadError	: bool		: Error occurred			*/
/*  (returns)	: void		:					*/

void	KBURLRequest::slotURLFinished
	(	int		id,
		bool		hadError
	)
{
	/* First case: ::setHost() completes. This is actually the	*/
	/* connection to the server, so we display that (or an error).	*/
	/* If the connection was OK then the QHttp object will move on	*/
	/* automatically to download the database specification.	*/
	if (id == m_URLSetHostID)
	{
		DPRINTF
		((	"KBURLRequest::slotURLFinished: Connected to host: err=%d\n",
			hadError
		))	;
		if (hadError)
		{	setURLError () ;
			return	;
		}
		m_URLSetHostID	= -1	;
		setProgress (TR("Connected to remote host")) ;
		return	;
	}

	/* Second case: ::get() completes. This is the end of the data	*/
	/* download, possibly with an error.				*/
	if (id == m_URLGetID)
	{
		DPRINTF
		((	"KBURLRequest::slotURLFinished: retrieved data: err=%d\n",
			hadError
		))	;
		if (hadError)
		{	setURLError () ;
			return	;
		}
		m_URLGetID	= -1	;
		m_active	= false	;
		setProgress (QString(TR("Retrieved %1")).arg(m_what)) ; ;

		DPRINTF
		((	"KBURLRequest::slotURLFinished: [%s]\n",
			m_data.data()
		))	;

		notifySlot (Finished, m_data.array()) ;
		return	;
	}
}

/*  KBURLRequest								*/
/*  slotURLProgress							*/
/*		: QHttp object reports progress				*/
/*  done	: int		: Download completed so far		*/
/*  total	: int		: Total download			*/
/*  (returns)	: void		:					*/

void	KBURLRequest::slotURLProgress
	(	int		done,
		int		total
	)
{
	DPRINTF
	((	"KBURLRequest::slotURLProgress: got %d of %d\n",
		done,
		total
	))	;
	setProgress (done, total) ;
}

/*  KBURLRequest								*/
/*  slotURLReady							*/
/*		: Data available from HTTP object			*/
/*  (returns)	: void		:					*/

void	KBURLRequest::slotURLReady
	(	const QHttpResponseHeader	&resp
	)
{
	DPRINTF
	((	"KBURLRequest::slotURLReady: resp.rc=%d\n",
		resp.statusCode()
	))	;

	if (resp.statusCode() != 200)
	{
		notifySlot
		(	Finished, 
			QString(TR("%1 (code %2)"))
				.arg(resp.reasonPhrase())
				.arg(resp.statusCode  ())
		)	;

		halt	() ;
		return	;
	}

	m_data.append (readAll()) ;
}

/*  KBURLRequest								*/
/*  exec	: Execute download					*/
/*  url		: const QString & : Specification URL			*/
/*  what	: const QString & : What is being downloaded		*/
/*  (returns)	: void		  :					*/

void	KBURLRequest::exec
	(	const QString	&url,
		const QString	&what
	)
{
	m_url	= QUrl(url)	;
	m_what	= what		;
	m_slot	= 0		;

	const QPtrList<KBSlot> &slist = m_notify->getSlots() ;
	LITER
	(	KBSlot,
		slist,
		slot,
		if (slot->name() == m_slotName)
		{	m_slot	 = slot ;
			break	 ;
		}
	)

	if (m_slot == 0)
	{
		KBError::EError
		(	QString(TR("Slot %1 not found")).arg(m_slotName),
			QString::null,
			__ERRLOCN
		)	;
		return	;
	}

	/* See if the specification originates on an HTTP server, in	*/
	/* which case set up to download the specification.		*/
	if (m_url.protocol().lower() == "http")
	{
		QString	text	= m_url.path() ;
		if (!m_url.query().isEmpty())
			text   += "?" + m_url.query() ;

		m_active	= true	;

		DPRINTF
		((	"KBURLRequest::exec: host=[%s] port=[%d] get=[%s]\n",
			m_url.host  ().latin1(),
			m_url.port  (),
			text .latin1()
		))	;

		int	port	= m_url.port()	;
		if (port < 0) port = 80		;
		m_URLSetHostID	= setHost (m_url.host(), port) ;
		m_URLGetID	= get     (text) ;

		return	;
	}

	KBError::EError
	(	TR("Unknown download protocol"),
		m_url.protocol(),
		__ERRLOCN
	)	;

	halt ()	;
}

void	KBURLRequest::halt ()
{
	stop ()	 ;

	m_active = false ;
	KBDeleter::addObject (this) ;
}
