/***************************************************************************
    file	         : kb_writer.h
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/

#ifndef	__KB_WRITER_H
#define	__KB_WRITER_H


#include	<qpixmap.h>
#include	<qprinter.h>
#include	<qpainter.h>
#include	<qptrlist.h>
#include	<qwidget.h>

#include	"libkbase_exports.h"

#include	"kb_location.h"
#include	"kb_dom.h"
#include	"tk_print.h"


/*  The classes in this file implement output writing, either to an on-	*/
/*  screen widget or a printer.						*/
/*									*/
/*  The basic mechanism is to an instances of classes derived from	*/
/*  KBWriterItem to a single instance of KBWriter, which can then be	*/
/*  used to render the output. This file, and kb_writer.cpp define and	*/
/*  implement a basic set of KBWriterItem classes; others can be	*/
/*  defined and implemented elsewhere as neccessary as database		*/
/*  controls are defined.						*/


/*  KBWriterItem							*/
/*  ------------							*/
/*  This is a base class for data which is to be written onto a widget	*/
/*  and which must therefore be retained for repaints.			*/

class	LIBKBASE_API	KBWriterItem
{
protected :

	KBWriter 	*m_writer	;
	QRect		m_rect		;

	KBNode		*m_node		;
	uint		m_row		;

	bool		m_frameSet	;
	int		m_frameStyle	;
	int		m_frameWidth	;

	void		drawFrame	(QPainter &, QRect, int, int, int, const QColorGroup &) ;

public	:

	enum	RTTI
	{
		WriterText	= 1,
		WriterBox	= 2,
		WriterPixmap	= 3,
		WriterBG	= 4
	}	;

	virtual	void	paintEvent	(QPaintEvent *, QPainter &) = 0 ;

	virtual	RTTI	rtti		() = 0	;
	virtual	QString	describe	(bool = true)    ;

	void		setParent	(KBNode *, uint) ;
	virtual void	setFrame	(int, int)	 ;

	KBNode		*node		()	;
	uint		row		()	;
	QRect		rect		()	;
	int		frameStyle	()	;
	int		frameWidth	()	;

	KBWriterItem 			(KBWriter *, QRect) ;
	virtual	~KBWriterItem 		() = 0	;
}	;


/*  KBWriterText							*/
/*  ------------							*/
/*  Class used to hold text which is to be written onto a widget. Holds	*/
/*  position and style as well as the text.				*/

class	LIBKBASE_API	KBWriterText : public KBWriterItem
{
	const QPalette	*m_pal		;
	const QFont	*m_font		;
	QString		m_text		;
	int		m_align		;
	bool		m_fSubs		;
	int		m_extra		;
	bool		m_useExtra	;

	void		calculateExtra	()	;

public	:

	virtual	RTTI	rtti		()	;

	virtual	void	paintEvent	(QPaintEvent *, QPainter &) ;
	virtual	QString	describe	(bool = true) ;
	virtual void	setFrame	(int, int)    ;

	int		extra		() ;
	const QString	&text		() ;

	KBWriterText
	(	KBWriter	*writer,
		QRect		rect,
		const QPalette	*pal,
		const QFont	*font,
		const QString 	&text,
		int		align,
		bool		fSubs
	)	;
	virtual	~KBWriterText	() ;

}	;


/*  KBWriterBox								*/
/*  -----------								*/
/*  Class used to describe a simple box.				*/

class	LIBKBASE_API	KBWriterBox : public KBWriterItem
{
public	:

	virtual	RTTI	rtti		()	;

	virtual	void	paintEvent	(QPaintEvent *, QPainter &) ;
	virtual	QString	describe	(bool = true) ;

	KBWriterBox			(KBWriter *, QRect) ;
	virtual	~KBWriterBox		() ;
}	;


/*  KBWriterPixmap							*/
/*  --------------							*/
/*  Class used to hold an image						*/

class	LIBKBASE_API	KBWriterPixmap : public KBWriterItem
{
	QPixmap		m_pixmap 	;
	int		m_scaling	;

public	:

	virtual	RTTI	rtti		()	;

	virtual	void	paintEvent 	(QPaintEvent *, QPainter &) ;
	virtual	QString	describe	(bool = true) ;

	KBWriterPixmap			(KBWriter *, QRect, const QPixmap &, int) ;
	virtual	~KBWriterPixmap		() ;
}	;


/*  KBWriteBG								*/
/*  ---------								*/
/*  Class used to hold background color information			*/

class	LIBKBASE_API	KBWriterBG : public KBWriterItem
{
	QString		m_color	;

public	:

	virtual	RTTI	rtti		()	;

	virtual	void	paintEvent 	(QPaintEvent *, QPainter &) ;
	virtual	QString	describe	(bool = true) ;

	KBWriterBG			(KBWriter *, QRect, const QString &) ;
	virtual	~KBWriterBG		() ;
}	;



/*  KBWriter								*/
/*  --------								*/
/*  General writer class. The main function of this class is to		*/
/*  accumulate output elements; these will then be written onto, for	*/
/*  instance, a QWidget or a Q/KPrinter.				*/

typedef	QPtrList<KBWriterItem>	KBWPage	;

class	LIBKBASE_API	KBWriter : public QWidget
{
	KBLocation		m_docLocation	;
	bool			m_report	;

	uint			m_lMargin	;
	uint			m_rMargin	;
	uint			m_tMargin	;
	uint			m_bMargin	;

	uint			m_vRows		;
	uint			m_vCols		;
	uint			m_vRowGap	;
	uint			m_vColGap	;
	int			m_vRowHeight	;
	int			m_vColWidth	;
	uint			m_vAtRow	;
	uint			m_vAtCol	;
	bool			m_vBorders	;

	int			m_pWidth	;
	int			m_pHeight	;

	QPoint			m_offset	;
	int			m_foot		;
	QValueList<int>		m_pageList	;

	QPtrList<KBWPage>	m_pages		;
	KBWPage			*m_page		;
	uint			m_cPNum		;

	TKPrinterSettings	m_settings	;

	TKPrinter		*m_printer	;
	QPainter		*m_painter	;

	QDomElement	getPrinterSpec	(const QString &) ;

	void		startPage   	()	;

protected :

	virtual	void	paintEvent (QPaintEvent *) ;

public	:

	KBWriter	(QWidget *, KBLocation  &) ;
       ~KBWriter	()			;

	virtual	void	clear		()	;
	virtual	void	newPage		()	;
	virtual	bool	spaceAvailable	(int )	;
	virtual	void	showPage	(uint)	;
	virtual	bool	pageIsEmpty	()	;
	virtual	uint	numPages	()	;

	bool		setup		(const QString &, bool, uint, uint, uint, uint, bool) ;
	bool		setup		(bool)	;
	bool		setupVirtual	(uint, uint, uint, uint, bool, bool) ;

	void		add		(KBWriterItem  *, QRect &) ;
	void		adjust		(QRect &) ;

	QPoint		setOffset 	(bool, QPoint   ) ;
	QSize		getSize		()	;
	int		reserve		(int)	;
	QString		textSub		(const QString &) ;
	void		printDoc	(const QString &, int) ;
	QString		describe 	()	;

	inline	void	setReport()
	{
		m_report = true		;
	}

	inline	bool	asReport ()
	{
		return	m_report	;
	}

	KBWriterItem	*findItem	(uint, KBNode *, uint) ;
}	;

extern	LIBKBASE_API	double	pixelsPerMM	() ;
extern	LIBKBASE_API	void	getPixelPageSize(cchar *, int &, int &) ;

#endif	// __KB_WRITER_H
