/*  -*- C++ -*-
    This file is part of the KDE libraries
    Copyright (C) 1997 Tim D. Gilman (tdgilman@best.org)
              (C) 1998-2001 Mirko Boehm (mirko@kde.org)
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/
/* 
	RKDatePick is a 100% pure Qt port of KDatePick, which 
	is part of the KDELibs. Class, function and variable names have been changed 
	to make them consistent with those found in Rekall
*/


#include <qlayout.h>
#include <qframe.h>
#include <qpainter.h>
#include <qdialog.h>
#include <qstyle.h>
#include <qtoolbutton.h>
#include <qcombobox.h>
#include <qtooltip.h>
#include <qfont.h>
#include <qvalidator.h>
#include <qpopupmenu.h>
#include <qapplication.h>
#include <qtoolbar.h>
#include <qlineedit.h>

#if	__KB_AUTOMOC
#include "rk_datepicker.moc"
#else
#include "rk_datepicker.h"
#endif

#include "rk_dategridview.h"
#include "rk_datevalidator.h"
#include "rk_datepickerimages.h"
#include "rk_yearselector.h"
#include "rk_popupframe.h"

void RKDatePicker::fillWeeksCombo( const QDate& date )
{
	m_pSelectWeekComboBox->clear();

 	QDate day( date.year(), 1, 1 );
	int lastMonth = 12;
	QDate lastDay( date.year(), lastMonth, QDate( date.year(), lastMonth, 1 ).daysInMonth() );

	for ( ; day <= lastDay; day = day.addDays( 7 ) )
	{
		int year = 0;
		QString week = QString( tr( "Week %1" ).arg( day.weekNumber( &year) ) );
    
		if ( year != date.year() ) 
			week += "*";  
		
		m_pSelectWeekComboBox->insertItem(week);
	}
}

RKDatePicker::RKDatePicker( QWidget* parent, QDate date, const char* name )
  : QFrame( parent, name )
{
	init( date );
}

RKDatePicker::RKDatePicker( QWidget* parent, QDate date, const char* name, WFlags f )
  : QFrame( parent, name, f )
{
	init( date );
}

RKDatePicker::RKDatePicker( QWidget* parent, const char* name )
  : QFrame( parent, name )
{
	init( QDate::currentDate() );
}

void RKDatePicker::init( const QDate& date )
{
	QBoxLayout* pTopLayout = new QVBoxLayout(this);

	m_pNavigationLayout = new QHBoxLayout( pTopLayout);
	m_pNavigationLayout->addStretch();
	
	m_pYearBackwardToolButton = new QToolButton( this );
	m_pYearBackwardToolButton->setAutoRaise( true );
	m_pNavigationLayout->addWidget( m_pYearBackwardToolButton );
  
	m_pMonthBackwardToolButton = new QToolButton( this );
	m_pMonthBackwardToolButton ->setAutoRaise( true );
	m_pNavigationLayout->addWidget( m_pMonthBackwardToolButton );
	m_pNavigationLayout->addSpacing( 6 );

	m_pSelectMonthToolButton = new QToolButton( this );
	m_pSelectMonthToolButton ->setAutoRaise( true );
	m_pNavigationLayout->addWidget( m_pSelectMonthToolButton );
  
	m_pSelectYearToolButton = new QToolButton( this );
	m_pSelectYearToolButton->setToggleButton( true );
	m_pSelectYearToolButton->setAutoRaise( true );
	m_pNavigationLayout->addWidget( m_pSelectYearToolButton );
	m_pNavigationLayout->addSpacing( 6 );

	m_pMonthForwardToolButton = new QToolButton( this );
	m_pMonthForwardToolButton ->setAutoRaise( true );
	m_pNavigationLayout->addWidget( m_pMonthForwardToolButton );
  
	m_pYearForwardToolButton = new QToolButton( this );
	m_pYearForwardToolButton ->setAutoRaise( true );
	m_pNavigationLayout->addWidget( m_pYearForwardToolButton );
	m_pNavigationLayout->addStretch();

	m_pDateLineEdit = new QLineEdit( this );
	m_pDateValidator = new RKDateValidator( this );
	m_pDateGridView = new RKDateGridView(this);
	m_FontSize = QApplication::font().pointSize();
  
	if ( m_FontSize == -1 )
		m_FontSize = QFontInfo( QApplication::font() ).pointSize();

	m_FontSize++; 

	m_pSelectWeekComboBox = new QComboBox( false, this );  
	m_pTodayToolButton = new QToolButton( this );
	m_pTodayToolButton->setIconSet( QPixmap::fromMimeSource( "today.png" ) );

	QToolTip::add( m_pYearForwardToolButton, tr( "Next year" ) );
	QToolTip::add( m_pYearBackwardToolButton, tr( "Previous year" ) );
	QToolTip::add( m_pMonthForwardToolButton, tr( "Next month" ) );
	QToolTip::add( m_pMonthBackwardToolButton, tr( "Previous month" ) );
	QToolTip::add( m_pSelectWeekComboBox, tr( "Select a week" ) );
	QToolTip::add( m_pSelectMonthToolButton, tr( "Select a month" ) );
	QToolTip::add( m_pSelectYearToolButton, tr( "Select a year" ) );
	QToolTip::add( m_pTodayToolButton, tr( "Select the current day" ) );

  	setFontSize( m_FontSize );
	m_pDateLineEdit->setValidator( m_pDateValidator );
	m_pDateLineEdit->installEventFilter( this );

	m_pYearForwardToolButton->setIconSet( QPixmap::fromMimeSource( "2rightarrow.png" ) );
	m_pYearBackwardToolButton->setIconSet( QPixmap::fromMimeSource( "2leftarrow.png" ) );
	m_pMonthForwardToolButton->setIconSet( QPixmap::fromMimeSource( "1rightarrow.png" ) );
	m_pMonthBackwardToolButton->setIconSet( QPixmap::fromMimeSource( "1leftarrow.png" ) );

	connect( m_pDateGridView, SIGNAL( dateChanged( QDate ) ), SLOT( slotDateChanged( QDate ) ) );
	connect( m_pDateGridView, SIGNAL( gridViewClicked() ), SLOT( slotGridViewClicked() ) );
	connect( m_pMonthForwardToolButton, SIGNAL( clicked() ), SLOT( slotMonthForwardClicked() ) );
	connect( m_pMonthBackwardToolButton, SIGNAL( clicked() ), SLOT( slotMonthBackwardClicked() ) );
	connect( m_pYearForwardToolButton, SIGNAL( clicked() ), SLOT( slotYearForwardClicked() ) );
	connect( m_pYearBackwardToolButton, SIGNAL( clicked() ), SLOT( slotYearBackwardClicked() ) );
	connect( m_pSelectWeekComboBox, SIGNAL( activated( int ) ), SLOT( slotSelectWeekClicked( int ) ) );
	connect( m_pTodayToolButton, SIGNAL( clicked() ), SLOT( slotTodayButtonClicked() ) );
	connect( m_pSelectMonthToolButton, SIGNAL( clicked() ), SLOT( slotSelectMonthClicked() ) );
	connect( m_pSelectYearToolButton, SIGNAL( toggled( bool ) ), SLOT( slotSelectYearClicked() ) );
	connect( m_pDateLineEdit, SIGNAL( returnPressed() ), SLOT( slotLineEditEnterPressed()));
  
	m_pDateGridView->setFocus();

	pTopLayout->addWidget( m_pDateGridView );

	QBoxLayout* pBottomLayout = new QHBoxLayout( pTopLayout );
	pBottomLayout->addWidget( m_pTodayToolButton );
	pBottomLayout->addWidget( m_pDateLineEdit );
	pBottomLayout->addWidget( m_pSelectWeekComboBox );

	m_pDateGridView->setDate( date );
	slotDateChanged( date );  
}

RKDatePicker::~RKDatePicker()
{
}

bool RKDatePicker::eventFilter(QObject* pObject, QEvent* pEvent )
{
   if ( pEvent->type() == QEvent::KeyPress ) 
   {
      QKeyEvent* pKeyEvent = ( QKeyEvent* )pEvent;

      if ( ( pKeyEvent->key() == Qt::Key_Prior ) ||
           ( pKeyEvent->key() == Qt::Key_Next )  ||
           ( pKeyEvent->key() == Qt::Key_Up )    ||
           ( pKeyEvent->key() == Qt::Key_Down ) )
       {
          QApplication::sendEvent( m_pDateGridView, pEvent );
          m_pDateGridView->setFocus();
          return true; 
       }
   }

   return QFrame::eventFilter( pObject, pEvent );
}

void RKDatePicker::resizeEvent( QResizeEvent* pEvent )
{
  QWidget::resizeEvent( pEvent );
}

void RKDatePicker::slotDateChanged( QDate date )
{
	m_pDateLineEdit->setText( date.toString( Qt::ISODate ) );
    m_pSelectMonthToolButton->setText( date.shortMonthName( date.month() ) );
    fillWeeksCombo(date);

    QDate firstDay(date.year(), 1, 1);
    m_pSelectWeekComboBox->setCurrentItem( ( date.dayOfYear() + firstDay.dayOfWeek() - 2 ) / 7 );

    m_pSelectYearToolButton->setText( yearString(date, false ) );

    emit( dateChanged( date ) );
}

void RKDatePicker::slotGridViewClicked()
{
	emit( dateSelected( m_pDateGridView->date() ) );
	emit( gridViewClicked() );
}

const QDate& RKDatePicker::date() const
{
	return m_pDateGridView->date();
}

bool RKDatePicker::setDate(const QDate& date)
{
    if( date.isValid() )
    {
        m_pDateGridView->setDate( date );  
        return true;
    }
    else
    {
        return false;
    }
}

void RKDatePicker::slotMonthForwardClicked()
{
    QDate temp;
    temp = m_pDateGridView->date().addMonths( 1 );

    setDate( temp );
}

void RKDatePicker::slotMonthBackwardClicked()
{
    QDate temp;
    temp = m_pDateGridView->date().addMonths( -1 );

    setDate( temp );
}

void RKDatePicker::slotYearForwardClicked()
{
    QDate temp;
    temp = m_pDateGridView->date().addYears( 1 );

    setDate( temp );
}

void RKDatePicker::slotYearBackwardClicked()
{
    QDate temp;
    temp = m_pDateGridView->date().addYears( -1 );

    setDate( temp );
}

void RKDatePicker::slotSelectedWeekClicked( int week )
{
	QDate date = m_pDateGridView->date();
	int year = date.year();

	date.setYMD( year, 1, 1 );
	date = date.addDays( week * 7 - date.dayOfWeek() + 1 );

	setDate( date );
}

void RKDatePicker::slotSelectMonthClicked()
{
	QDate date = m_pDateGridView->date();
	int i;
	int month = 12;

	QPopupMenu popup( m_pSelectMonthToolButton );

	for ( i = 1; i <= month; i++ )
		popup.insertItem( QDate::shortMonthName( i ) );

	popup.setActiveItem( date.month() - 1 );

	if ( ( month = popup.exec( m_pSelectMonthToolButton->mapToGlobal( QPoint( 0, 0 ) ), 
		date.month() - 1 ) ) == -1 ) 
		return;  

	int day = date.day();

	date.setYMD( date.year(), month, QMIN( day, date.daysInMonth() ) );
  
	setDate( date );
}

void RKDatePicker::slotSelectYearClicked()
{
	if (m_pSelectYearToolButton->state() == QButton::Off)
	{
		return;
	}

	int year;
	RKPopupFrame* pPopup = new RKPopupFrame( this );
	RKYearSelector* pYearSelector = new RKYearSelector( pPopup );
  
	pYearSelector->resize( pYearSelector->sizeHint() );
	pYearSelector->setYear( m_pDateGridView->date().year() );
	pYearSelector->selectAll();
	pPopup->setMainWidget( pYearSelector );
	connect( pYearSelector, SIGNAL( shutDown( int )), pPopup, SLOT( close( int ) ) );
	pYearSelector->setFocus();

	if( pPopup->exec( m_pSelectYearToolButton->mapToGlobal( 
		QPoint( 0, m_pSelectMonthToolButton->height() ) ) ) )
    {
		QDate date;
		int day;
      
		year = pYearSelector->year();
		date = m_pDateGridView->date();
		day = date.day();
      
		date.setYMD(year, date.month(), QMIN( day, date.daysInMonth() ) );
      	setDate( date );
    } 
	else 
	{
		QApplication::beep();
    }
  
	m_pSelectYearToolButton->setOn( false );
	delete pPopup;
}

void RKDatePicker::setEnabled( bool enable )
{
	QWidget* pWidgets[]= 
	{
		m_pYearForwardToolButton, 
		m_pYearBackwardToolButton, 
		m_pMonthForwardToolButton, 
		m_pMonthBackwardToolButton,
		m_pSelectMonthToolButton, 
		m_pSelectYearToolButton,
		m_pDateLineEdit, 
		m_pDateGridView, 
		m_pSelectWeekComboBox, 
		m_pTodayToolButton 
	};
  
	const int widgetCount = sizeof( pWidgets ) / sizeof( pWidgets[ 0 ] );
	int count;
  
	for( count = 0; count < widgetCount; ++count )
    {
		pWidgets[ count ]->setEnabled( enable );
    }
}

void RKDatePicker::slotLineEditEnterPressed()
{
	QDate temp;
  
	if( m_pDateValidator->date( m_pDateLineEdit->text(), temp ) == QValidator::Acceptable )
    {
        emit( dateEntered( temp ) );
        setDate( temp );
    } 
	else 
	{
		QApplication::beep();
    }
}

void RKDatePicker::slotTodayButtonClicked()
{
	setDate( QDate::currentDate() );
}

QSize RKDatePicker::sizeHint() const
{
	return QWidget::sizeHint();
}

void RKDatePicker::setFontSize( int size )
{
	QWidget* pButtons[]= 
	{
		// m_pYearBackwardToolButton,
		// m_pMonthBackwardToolButton,
		m_pSelectMonthToolButton,
		m_pSelectYearToolButton,
		// m_pMonthForwardToolButton,
		// m_pYearForwardToolButton
	};
  
	const int buttonCount = sizeof( pButtons ) / sizeof( pButtons[ 0 ] );
	int count;
	QFont font;
	QRect rect;
  
	m_FontSize = size;
  
	for( count = 0; count < buttonCount; ++count )
    {
		font = pButtons[ count ]->font();
		font.setPointSize( size );
		pButtons[ count ]->setFont( font );
    }

	QFontMetrics metrics( m_pSelectMonthToolButton->fontMetrics() );

	for ( int i = 1; i <= 12; ++i )
    {
		QString str = QDate::shortMonthName( i );

		rect = metrics.boundingRect( str );
		m_MaxMonthRect.setWidth( QMAX( rect.width(), m_MaxMonthRect.width() ) );
		m_MaxMonthRect.setHeight( QMAX( rect.height(),  m_MaxMonthRect.height() ) );
    }

	QSize metricBound = style().sizeFromContents( 
		QStyle::CT_ToolButton, m_pSelectMonthToolButton, m_MaxMonthRect );
	m_pSelectMonthToolButton->setMinimumSize( metricBound );

	m_pDateGridView->setFontSize( size );
}

void RKDatePicker::setCloseButton( bool enable )
{
    if ( enable == ( m_pCloseToolButton != 0L ) )
        return;

    if ( enable ) 
	{
        m_pCloseToolButton = new QToolButton( this );
        m_pCloseToolButton->setAutoRaise(true);
        m_pNavigationLayout->addSpacing( 6 );
        m_pNavigationLayout->addWidget( m_pCloseToolButton );
        QToolTip::add( m_pCloseToolButton, tr( "Close" ) );
        m_pCloseToolButton->setPixmap( QPixmap::fromMimeSource( "remove.png" ) );
        
		connect( m_pCloseToolButton, SIGNAL( clicked() ),
			topLevelWidget(), SLOT( close() ) );
    }
    else 
	{
        delete m_pCloseToolButton;
        m_pCloseToolButton = 0L;
    }

    updateGeometry();
}

bool RKDatePicker::hasCloseButton() const
{
    return ( m_pCloseToolButton );
}

QString RKDatePicker::yearString( const QDate& date, bool bShort ) const
 {
	QString retVal;
  
	retVal.setNum( date.year() );
 
	if ( bShort && retVal.length() == 4 )
		retVal = retVal.right( 2 );
  
	return retVal;
 }

 QString RKDatePicker::monthString( const QDate& date, bool bShort ) const
 {
	QString retVal;
  
	retVal.setNum( date.month() );
 
	if ( ! bShort && retVal.length() == 1 )
		retVal.prepend( ' ' );
  
	return retVal;
 }

 QString RKDatePicker::dayString( const QDate& date, bool bShort ) const
 {
	QString retVal;
  
	retVal.setNum( date.day() );
 
	if ( ! bShort && retVal.length() == 1 )
		retVal.prepend( ' ' );
  
	return retVal;
 }



