/***************************************************************************
    file	         : kb_kjsregister.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/

#include	<stdio.h>

#include	<kjs/global.h>
#include	<kjs/interpreter.h>
#include	<kjs/lookup.h>
#include	<kjs/object.h>
#include	<kjs/operations.h>
#include	<kjs/types.h>
#include	<kjs/ustring.h>
#include	<kjs/value.h>

#include	"kb_classes.h"
#include	"kb_item.h"

#include	"kb_idbase.h"
#include	"kb_objectproxy.h"
#include	"kb_itemproxy.h"
#include	"kb_kjsregister.h"
#include	"kb_kjsscript.h"


#define		__DEBUG_KJSREGISTER	(0)
#if		__DEBUG_KJSREGISTER
#include	"kb_dprintf.h"
#endif


static	QDict<MKFUNC*>	mkfuncMap	;


/*  registerClass: Register a rekall class				*/
/*  kbname	 : cchar *	 : Rekall name				*/
/*  mkFunc	 : MKFUNC *	 : Pointer to factory function		*/
/*  aliases	 : cchar **	 : Possible aliases list		*/
/*  (returns)	 : void		:					*/

void	registerClass
	(	cchar		*kbname,
		MKFUNC		*mkFunc,
		cchar		**aliases
	)
{
	DPRINTF
	((	"KJS: registerClass: [%s]->[%p]\n",
		(cchar *)kbname,
		(void  *)mkFunc
	))	;

	/* Use a pointer-to-pointer for the factory function, since	*/
	/* QDict cannot be used directory (it tries to create a		*/
	/* function destructor:).					*/
	MKFUNC	**mkFuncPtr = new MKFUNC * ;

	*mkFuncPtr	= mkFunc ;
	mkfuncMap.insert (kbname, mkFuncPtr) ;


	/* Add any aliases if present ...				*/
	if (aliases != 0)
		for (const char **ap = &aliases[0] ; *ap != 0 ; ap += 1)
			mkfuncMap.insert (*ap, mkFuncPtr) ;
}


/*  registerClasses: Register all classes				*/
/*  (reutrns)	   : void	:					*/

void	registerClasses ()
{
#define	RCLASS(Class)	\
	extern	void	register##Class	()	;	\
	register##Class	()			;

	RCLASS	(KBObject	)
	RCLASS	(KBItem		)
	RCLASS	(KBBlock	)
	RCLASS	(KBFormBlock	)
	RCLASS	(KBForm		)
	RCLASS	(KBButton	)
	RCLASS	(KBLabel	)
	RCLASS	(KBLabel	)
	RCLASS	(KBField	)
	RCLASS	(KBChoice	)
	RCLASS	(KBCheck	)
	RCLASS	(KBListBox	)
	RCLASS	(KBRichText	)
	RCLASS	(KBFramer	)
	RCLASS	(KBLinkTree	)
	RCLASS	(KBLink		)
	RCLASS	(KBStack	)
	RCLASS	(KBStackPage	)
	RCLASS	(KBTabber	)
	RCLASS	(KBTabberPage	)
}


/*  ------------------------------------------------------------------  */

/*  makeProxy	: Make proxy object					*/
/*  interp	: KBKJSInterpreter *	: Interpreter			*/
/*  node	: KBNode *		: Node to be proxied		*/
/*  (returns)	: KBObjectProxy *	: Proxy object			*/

KBObjectProxy
	*makeProxy
	(	KBKJSInterpreter	*interp,
		KBNode			*node
	)
{
	uint	lc = interp->languageCode() ;

	/* If the node is actually an object then see if it already	*/
	/* contains a corresponding script object, in which case we	*/
	/* have a corresponding persistent python object.		*/
	if (node->isObject())
	{
		KBKJSScriptObject *so = (KBKJSScriptObject *)node->isObject()->scriptObject(lc) ;
		if (so != 0) return so->object() ;
	}


	/* First see if the rekall node is present in the factory map,	*/
	/* in which case construct a proxy using said factory.		*/
	MKFUNC		**mkFunc ;
	KBObjectProxy	*proxy	 = 0 ;

	if ((mkFunc = mkfuncMap.find (node->getElement())) != 0)
	{
		DPRINTF
		((	"makeProxy: [%s][%s]->[%p]\n",
			(cchar *)node->isObject()->getName(),
			(cchar *)node->getElement(),
			(void  *)(*mkFunc)
		))	;

		proxy	= (*mkFunc) (interp, node) ;
	}

	/* If not in the factory map then see if it is an item or an	*/
	/* object, in which case create the appropriate proxy. This	*/
	/* gives a catchall for rekall objects which have no special	*/
	/* scripting interface (other than that which they inherit).	*/
	if (proxy == 0)
		if (node->isItem  () != 0)
		{
			DPRINTF
			((	"makeProxy: [%s][%s]->Item\n",
				(cchar *)node->isObject()->getName(),
				(cchar *)node->getElement()
			))	;

			proxy	= new KBItemProxy (interp, node->isItem  ()) ;
		}

	if (proxy == 0)
		if (node->isObject() != 0)
		{
			DPRINTF
			((	"makeProxy: [%s][%s]->Object\n",
				(cchar *)node->isObject()->getName(),
				(cchar *)node->getElement()
			))	;

			proxy	= new KBObjectProxy (interp, node->isObject()) ;
		}

	if (proxy != 0)
	{
		/* If the node is an object then set this script object	*/
		/* so that we have persistence of objects. This will	*/
		/* increment the reference count.			*/
		if (node->isObject() != 0)
			node->isObject()->setScriptObject (lc, new KBKJSScriptObject (proxy)) ;

		/* We also increment the reference coutn before return	*/
		/* so that the caller can use proxy->deref() and decide	*/
		/* whether to delete or not.				*/
		proxy->ref()	;
		return	proxy	;
	}

	/* Hmmm, this really should not happen ....			*/
	fprintf
	(	stderr,
		"makeProxy: [%s][%s]: *** NO PROXY ***\n",
		(cchar *)node->getElement(),
		node->isObject() == 0 ?
			"" :
			(cchar *)node->isObject()->getName()
	)	;
	return	0 ;
}

