/***************************************************************************
    file	         : kb_objectproxy.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/


#include	<qvariant.h>

#include	<kjs/global.h>
#include	<kjs/interpreter.h>
#include	<kjs/lookup.h>
#include	<kjs/object.h>
#include	<kjs/operations.h>
#include	<kjs/types.h>
#include	<kjs/ustring.h>
#include	<kjs/value.h>
#include	<kjs/array_instance.h>


#include	"kb_classes.h"
#include	"kb_form.h"
#include	"kb_config.h"

#include	"kb_idbase.h"
#include	"kb_objectproxy.h"
#include	"kb_eventsproxy.h"
#include	"kb_slotsproxy.h"
#include	"kb_kjsregister.h"

#include	"kb_kjsscript.h"
#include	"kb_slotnotifier.h"


#define		__DEBUG_OBJECTPROXY	(0)
#if		__DEBUG_OBJECTPROXY
#include	"kb_dprintf.h"
#endif



cchar	*decodeError
	(	const KBError		&error
	)
{
	static  QString	aQString ;
	const	QString	&details = error.getDetails() ;

	aQString = error.getMessage() ;

	if (!details.isEmpty())
	{	aQString += ": "    ;
		aQString += details ;
	}

	return	aQString ;
}

/*  ------------------------------------------------------------------	*/

/*  KBObjectProxy							*/
/*  KBObjectProxy: Constructor for rekall object proxy			*/
/*  interp	 : KBKJSInterpreter *	: Intepreter object		*/
/*  object	 : KBObject *		: Rekall object			*/
/*  (returns)	 : KBObjectProxy	:				*/

KBObjectProxy::KBObjectProxy
	(	KBKJSInterpreter	*interp,
		KBObject		*object
	)
	:
	KJS::ObjectImp	(),
	m_interp	(interp),
	m_object	(object)
{
	m_slots	 = 0 ;
	m_events = 0 ;
}

/*  KBObjectProxy							*/
/* ~KBObjectProxy: Destructor for rekall object proxy			*/
/*  (returns)	 :		:					*/

KBObjectProxy::~KBObjectProxy ()
{
	if (m_slots  != 0)
	{	m_slots ->deref() ;
		delete	m_slots	  ;
	}
	if (m_events != 0)
	{	m_events->deref() ;
		delete	m_events  ;
	}
}

/*  KBObjectProxy							*/
/*  addBindings	: Add proxy bindings to specified object		*/
/*  exec	: KJS::ExecState *	: Execution state		*/
/*  object	: KJS::Object &		: Object to which to add	*/
/*  (returns)	: void			:				*/

void	KBObjectProxy::addBindings
	(	KJS::ExecState	*exec,
		KJS::Object	&object
	)
{
	static	KBMethodSpec methods	[] =
	{
	{	"getAttr",	MethodImp::id_getAttr,		0	},
	{	"getConfig",	MethodImp::id_getConfig,	0	},
	{	"setAttr",	MethodImp::id_setAttr,		0	},
	{	"resize",	MethodImp::id_resize,		0	},
	{	"getNamedCtrl",	MethodImp::id_getNamedCtrl,	0	},
	{	"setEnabled",	MethodImp::id_setEnabled,	0	},
	{	"setVisible",	MethodImp::id_setVisible,	0	},
	{	"isEnabled",	MethodImp::id_isEnabled,	0	},
	{	"isVisible",	MethodImp::id_isVisible,	0	},
	{	"getName",	MethodImp::id_getName,		0	},
	{	"getElement",	MethodImp::id_getElement,	0	},
	{	"getType",	MethodImp::id_getType,		0	},
	{	"width",	MethodImp::id_width,		0	},
	{	"height",	MethodImp::id_height,		0	},
	{	"getQTProperty",MethodImp::id_getQTProperty,	0	},
	{	"setFont",	MethodImp::id_setFont,		0	},
	{	"getParent",	MethodImp::id_getParent,	0	},
	{	"getBlock",	MethodImp::id_getBlock,		0	},
	{	"getForm",	MethodImp::id_getForm,		0	},
	{	"lastError",	MethodImp::id_lastError,	0	},
	{	"getControls",	MethodImp::id_getControls,	0	},
	{	"getRoot",	MethodImp::id_getRoot,		0	},
	{	"setFGColor",	MethodImp::id_setFGColor,	0	},
	{	"setBGColor",	MethodImp::id_setBGColor,	0	},
	{	"notify",	MethodImp::id_notify,		0	},
	{	"toString",	MethodImp::id_toString,		0	},
	{	0,		0,				0	}
	}	;

	for (KBMethodSpec *ms = &methods[0] ; ms->m_method != 0 ; ms += 1)
		object.put
		(	exec,
			KJS::Identifier(ms->m_method),
			KJS::Object    (new MethodImp(ms, this))
		)	;
}

KJS::Value
	KBObjectProxy::fromKBValue
	(	KJS::ExecState		*exec,
		const KBValue		&value
	)
{
	if (value.isNull())
		return	KJS::Null() ;

	switch (value.getType()->getIType())
	{
		case KB::ITFixed	:
			return	KJS::Number (value.getRawText().toInt   ()) ;

		case KB::ITFloat	:
			return	KJS::Number (value.getRawText().toDouble()) ; 

		case KB::ITBinary	:
		case KB::ITString	:
			return	KJS::String (value.getRawText()) ;

		case KB::ITBool		:
			return	KJS::Boolean(value.isTrue() ? 1 : 0) ;

		case KB::ITDate	    :
		case KB::ITTime	    :
		case KB::ITDateTime :
			return	KJS::String (value.getDateTime()->defFormat (value.getType()->getIType())) ;

		case KB::ITUnknown  :
		case KB::ITRaw	    :
		case KB::ITDriver   :
			return	KJS::String (value.getRawText()) ;

		case KB::ITNode		:
		{
			KBObjectProxy *proxy = makeProxy
					       (	dynamic_cast<KBKJSInterpreter *>(exec->interpreter()),
							value.getNode()
					       )	;
			KJS::Object   object (proxy) ;

			proxy->addBindings   (exec, object) ;
			proxy->deref	     ()	;

			return	object ;
		}

		default	:
			break	;
	}

	return	KJS::String(value.getRawText()) ;
}

KBValue	KBObjectProxy::fromKJSValue
	(	KJS::ExecState		*exec,
		const KJS::Value        &value
	)
{
	switch (value.type())
	{
		case KJS::NullType	:
			return	KBValue () ;

		case KJS::BooleanType	:
			return	KBValue (value.toBoolean(exec)) ;

		case KJS::StringType	:
			return	KBValue (value.toString (exec).qstring()) ;

		case KJS::NumberType :
		{
			double	d = value.toNumber(exec) ;
			if (int(d) != d) return d ;
			return	(int) d ;
		}

		default	:
			DPRINTF
			((	"KBObjectProxy::fromKJSValue: t=%d\n",
				value.type()
			))	;

			break	;
	}

	return	KBValue	() ;
}

KBObject
	*KBObjectProxy::toKBObject
	(	KJS::ExecState		*exec,
		const KJS::Value	&value
	)
{
	KJS::ObjectImp	*kjsobj	= value.toObject(exec).imp() ;
	if (kjsobj == 0) return 0 ;

	KBObjectProxy	*rklobj	= dynamic_cast<KBObjectProxy *>(kjsobj) ;
	if (rklobj == 0) return 0 ;

	return	rklobj->m_object  ;
}

/*  KBObjectProxy							*/
/*  get		: Get property value					*/
/*  exec	: KJS::ExecState *	: Execution state		*/
/*  property	: KJS::Identifier &	: Property name			*/
/*  (returns)	: KJS::Value		: Result			*/

KJS::Value
	KBObjectProxy::get
	(	KJS::ExecState		*exec,
		const KJS::Identifier	&property
	)
	const
{
	QString	p	= property.qstring() ;

	DPRINTF
	((	"KBObjectProxy::get: %s.%s: %s\n",
		m_object->getElement().latin1(),
		m_object->getName   ().latin1(),
		p.latin1()
	))	;


	if (p == "__events__")
	{
		if (m_events == 0)
		{	((KBObjectProxy *)this)->m_events = new KBEventsProxy (m_interp, m_object) ;
			m_events->ref() ;
		}

		KJS::Object    object (m_events) ;
		m_events->addBindings (exec, object) ;
		return	object ;
	}

	if (p == "__slots__" )
	{
		if (m_slots  == 0)
		{	((KBObjectProxy *)this)->m_slots  = new KBSlotsProxy  (m_interp, m_object) ;
			m_slots ->ref() ;
		}

		KJS::Object    object (m_slots ) ;
		m_slots ->addBindings (exec, object) ;
		return	object ;
	}

	if (!m_interp->clientSide())
	{
		KBValue	value	;
		bool	exists	= m_object->getKBProperty (p, value) ;

		if (exists) return fromKBValue (exec, value) ;
	}

	return	KJS::ObjectImp::get (exec, property) ;
}

/*  KBObjectProxy							*/
/*  put		: Set property value					*/
/*  exec	: KJS::ExecState *	: Execution state		*/
/*  property	: KJS::Identifier &	: Property name			*/
/*  value	: const KJS::Value &	: Value				*/
/*  attr	: int			:				*/
/*  (returns)	: void			:				*/

void	KBObjectProxy::put
	(	KJS::ExecState		*exec,
		const KJS::Identifier	&property,
		const KJS::Value	&value,
		int			attr
	)
{
	QString	p = property.qstring() ;

	DPRINTF
	((	"KBObjectProxy::put: [%s]=[%s] cs=%d has=%d\n",
		p.latin1(),
		fromKJSValue(exec, value).getRawText().latin1(),
		m_interp->clientSide(),
		m_object->hasKBProperty(p)
	))	;


	/* If we are not running client-side emulation and the object	*/
	/* has the specified property then set it. If not, then fall	*/
	/* through to set an attribute.					*/
	if (!m_interp->clientSide())
		if (m_object->hasKBProperty(p))
		{
			KBValue	v	= fromKJSValue (exec, value)	 ;
			bool	ok	= m_object->setKBProperty (p, v) ;

			if (ok) return	;

			// TODO: Raise exception here
			//
			return	;
		}

	KJS::ObjectImp::put (exec, property, value, attr) ;
}


/*  ------------------------------------------------------------------  */

/*  KBObjectProxy							*/
/*  MethodImp	: 							*/
/*  MethodImp	: Object implementation					*/
/*  mthodSpec	: KBMethodSpec *  : Method specification		*/
/*  object	: KBObjectProxy * : Associated proxy			*/
/*  (return)	: MethodImp	  :					*/

KBObjectProxy::MethodImp::MethodImp
	(	KBMethodSpec		*methodSpec,
		const KBObjectProxy	*object
	)
	:
	KJS::ObjectImp	(),
	m_methodSpec	(methodSpec),
	m_object	(object)
{
}

/*  KBObjectProxy							*/
/*  checkArgs	: Check arguments					*/
/*  exec	: KJS::ExecState *	: Execution state		*/
/*  args	: const KJS::List &	: Argument list			*/
/*  argSpec	: cchar *		: Argument specification	*/
/*  (returns)	: bool			: Arguments OK			*/

bool	KBObjectProxy::checkArgs
	(	KJS::ExecState		*exec,
		const KJS::List		&args,
		cchar			*argSpec
	)
{
	int	argIdx	= 0	;
	bool	opt	= false	;

	if (argSpec == 0)
		return	true	;

	while (*argSpec != 0)
	{
		/* Vertical bar indicates that remaining arguments	*/
		/* are optional.					*/
		if (*argSpec == '|')
		{
			opt	 = true	;
			argSpec += 1	;
			continue	;
		}

		/* If we run beyond the arguments then we are OK or	*/
		/* are in error depending whether we are into any	*/
		/* optional arguments.					*/
		if (argIdx >= args.size())
		{
			if (opt) return true ;;
			return false	;
		}

		KJS::Type type = args[argIdx].type() ;

		/* Actual type matching is a bit vague since JS does a	*/
		/* lot of implicit type conversions, so we ought to	*/
		/* allow some flexibility.				*/
		switch (*argSpec)
		{
			case 's' :
				/* Any simple type can be considered to	*/
				/* be a string.				*/
				if
				(	(type != KJS::NullType	 ) &&
					(type != KJS::BooleanType) &&
					(type != KJS::StringType ) &&
					(type != KJS::NumberType )
				)
					return	false ;
				break	;

			case 'n' :
				/* Allow just numbers. Maybe we should	*/
				/* allow Boolean as 0/1 but maybe not!	*/
				if
				(	(type != KJS::NumberType )
				)
					return	false ;
				break	;

			case 'b' :
				/* For boolean allow JS numbers and 	*/
				/* boolean values.			*/
				if
				(	(type != KJS::NumberType ) &&
					(type != KJS::BooleanType)
				)
					return	false ;
				break	;

			case 'O'	:
				/* Object required. This one is the	*/
				/* easy case.				*/
				if
				(	(type != KJS::ObjectType )
				)
					return	false ;
				break	;

			case 'x' :
				/* ANything goes ...			*/
				break	;

			default	 :
				/* Treat anything else as an error.	*/
				/* This should not occur in debugged	*/
				/* code.				*/
				fprintf
				(	stderr,
					"KBObjectProxy::MethodImp::checkArgs: unknown: [[[[%c]]]\n",
					*argSpec
				)	;
				return	false ;
		}

		argSpec	+= 1 ;
		argIdx	+= 1 ;
	}

	/* Reached the end of the specification without error, so the	*/
	/* arguments are OK provided that there are none left.		*/
	if (argIdx >= args.size())
		return	true ;

	return	false	;
}

/*  KBObjectProxy							*/
/*  MethodImp	: 							*/
/*  checkArgs	: Check arguments					*/
/*  exec	: KJS::ExecState *	: Execution state		*/
/*  args	: const KJS::List &	: Argument list			*/
/*  argSpec	: cchar *		: Argument specification	*/
/*  (returns)	: KJS::Value		: Possible error		*/

KJS::Value
	KBObjectProxy::MethodImp::checkArgs
	(	KJS::ExecState		*exec,
		const KJS::List		&args,
		cchar			*argSpec
	)
{
	DPRINTF
	((	"KBObjectProxy::MethodImp::checkArgs: %s: %s/%s\n",
		m_methodSpec->m_method,
		argSpec,
		m_methodSpec->m_argSpec
	))	;

	/* If no specific string is passed then use the instance	*/
	/* default. If this is not set either then there is no checking	*/
	/* so return success whatever.					*/
	if (argSpec == 0)
		argSpec	= m_methodSpec->m_argSpec ;

	if ((argSpec != 0) && (argSpec[0] == '!'))
	{
		if (m_object->m_interp->clientSide())
		{
			KJS::Value err	= KJS::Error::create
					  (	exec,
						KJS::GeneralError,
						QString	("Method %1.%2 on %3 is not implemented client-side")
							.arg	(m_object    ->m_object->getElement())
							.arg	(m_methodSpec->m_method)
							.arg	(m_object    ->m_object->getName   ())
							.latin1	()
					  )	;
			exec->setException(err)	;
			return	err ;
		}

		argSpec	+= 1	;
	}

	if (!KBObjectProxy::checkArgs (exec, args, argSpec))
	{
		KJS::Value err	= KJS::Error::create
				  (	exec,
					KJS::GeneralError,
					QString	("Bad arguments: %1:%2.%3")
						.arg	(m_object->m_object->getElement())
						.arg	(m_object->m_object->getName   ())
						.arg	(m_methodSpec->m_method)
					.latin1	()
			  	  )	;
		exec->setException(err)	;
		return	err ;
	}

	return	KJS::Value() ;
}

/*  KBObjectProxy							*/
/*  MethodImp	: 							*/
/*  fromKBNode	: Build value corresponding to a node			*/
/*  exec	: KJS::ExecState *	: Execution state		*/
/*  node	: KBNode *		: Node in question		*/
/*  (returns)	: KJS::Value		: Encapsulated node		*/

KJS::Value
	KBObjectProxy::MethodImp::fromKBNode
	(	KJS::ExecState		*exec,
		KBNode			*node
	)
{
	if (node != 0)
	{
		KBObjectProxy *proxy = makeProxy (m_object->interpreter(), node) ;
		KJS::Object   object (proxy) ;

		proxy->deref	     ()	     ;
		proxy->addBindings   (exec, object) ;

		return	object	;
	}

	return	KJS::Null()	;
}

/*  MethodImp								*/
/*  callBase	: Invoke a method					*/
/*  exec	: KJS::ExecState *	: Execution state		*/
/*  obj		: KJS::Object &		: Apparently "self"		*/
/*  args	: const KJS::List	: Argument list			*/
/*  (returns)	: KJS::Value		: Result			*/

KJS::Value
	KBObjectProxy::MethodImp::callBase
	(	KJS::ExecState		*exec,
		KJS::Object		&,
		const KJS::List		&args
	)
{
	KBObject *object = m_object->m_object ;

	DPRINTF
	((	"KBObjectProxy::MethodImp::call: id=%d name=%s\n",
		m_methodSpec->m_id,
		m_methodSpec->m_method
	))	;

	switch (m_methodSpec->m_id)
	{
		case id_getAttr		:
			return	KJS::String(object->getAttrVal(kjsStringArg (exec, args, 0))) ;

		case id_getConfig	:
			{
			QString	name	= kjsStringArg (exec, args, 0) ;
			TITER
			(	Config,
				object->getChildren(),
				c,
				if (c->attrib() == name)
					return	KJS::String(c->value()) ;
			)
			}
			return	KJS::String(QString::null) ;

		case id_getName		:
			return	KJS::String(object->getName()) ;

		case id_getElement	:
		case id_getType		:
			return	KJS::String(object->getElement()) ;

		case id_setAttr		:
			{
			QString	name	= kjsStringArg (exec, args, 0) ;
			QString	value	= kjsStringArg (exec, args, 1) ;
			object->setAttrVal (name, value) ;
			}
			return	KJS::Null() ;

		case id_resize		:
			{
			QRect g = object->geometry() ;
			g.setWidth  (args.at(0).toInteger(exec)) ;
			g.setHeight (args.at(1).toInteger(exec)) ;
			object->setGeometry  (g)	 ;
			return	KJS::Null() ;
			}

		case id_getNamedCtrl	:
			{
			return	fromKBNode
				(	exec,
					object->getNamedNode
					(	kjsStringArg (exec, args, 0),
						args.at(1).toBoolean (exec)
					)
				)	;
			}

		case id_setEnabled	:
			object->setEnabled(args.at(0).toBoolean(exec)) ;
			return	KJS::Null() ;

		case id_setVisible	:
			object->setVisible(args.at(0).toBoolean(exec)) ;
			return	KJS::Null() ;

		case id_isEnabled	:
			return	KJS::Boolean(object->isEnabled()) ;

		case id_isVisible	:
			return	KJS::Boolean(object->isVisible()) ;

		case id_width	:
			return	KJS::Number (object->geometry().width ()) ;

		case id_height	:
			return	KJS::Number (object->geometry().height()) ;

		case id_getQTProperty :
		{
			QVariant v = object->property (kjsStringArg (exec, args, 0)) ;
			QString  t = v.toString() ;
			if (t.isNull()) return KJS::Null() ;
			return	KJS::String(t) ;
		}

		case id_setFont :
			object->setCtrlFont (kjsStringArg (exec, args, 0)) ;
			return	KJS::Null() ;

		case id_getParent :
			return	fromKBNode  (exec, object->getParent()) ;

		case id_getBlock :
			return	fromKBNode  (exec, object->getBlock ()) ;

		case id_getForm :
			return	fromKBNode  (exec, object->getForm  ()) ;

		case id_getRoot :
			return	fromKBNode  (exec, object->getRoot  ()) ;

		case id_setFGColor:
			object->setCtrlFGColor  (kjsStringArg (exec, args, 0)) ;
			return	KJS::Null() ;

		case id_setBGColor:
			object->setCtrlBGColor  (kjsStringArg (exec, args, 0)) ;
			return	KJS::Null() ;

		case id_lastError :
		{
			cchar	*error	= decodeError (object->lastError()) ;
			if (error == 0) return KJS::Null () ;
			return	KJS::String(QString(error)) ;
		}

		case id_getControls :
		{
			KJS::List children ;

			TITER
			(	Item,
				object->getChildren(),
				i,
				if (i->isBlock() == 0)
					children.append (fromKBNode (exec, i)) ;
			)

			return	KJS::Object::dynamicCast
				(	exec->interpreter()->builtinArray().construct
					(	exec,
						children
				)	)	;
		}

		case id_notify	 :
		{
			if (args.size() < 1)
			{
				KBError::EError
				(	TR("Notifier invoked with insufficient arguments"),
					QString(TR("Got %1, expected at least 1")).arg(args.size()),
					__ERRLOCN
				)	;
			
				return	KJS::Null() ;
			}

			/* First argument should be the caller,		*/
			/* followed by the notification event. Other	*/
			/* arguments are gathered up into a separate	*/
			/* vector.					*/
			QString	  event	  = kjsStringArg (exec, args, 0) ;
			uint	  __nArgs = args.size() - 1 ;
			KBValue	  *__args = new KBValue[__nArgs] ;

			for (uint idx = 0 ; idx < __nArgs ; idx += 1)
				__args[idx] = KBObjectProxy::fromKJSValue (exec, args[idx + 1]) ;

			KBScriptError *rc = 0 ;
			KBSlotNotifier::self()->fire (object, event, __nArgs, __args, rc) ;

			if (rc != 0)
				KBScriptError::processError (rc, KBScriptError::Normal) ;

			return	KJS::Number (0) ;
		}

		case id_toString :
			return	KJS::String
				(	QString("[%1.%2]")
						.arg(object->getElement())
						.arg(object->getName   ())
				)	;

		default	:
			break	;
	}

	return	KJS::Number(-1)	;
}

/*  MethodImp								*/
/*  call	: Invoke a method					*/
/*  exec	: KJS::ExecState *	: Execution state		*/
/*  obj		: KJS::Object &		: Apparently "self"		*/
/*  args	: const KJS::List	: Argument list			*/
/*  (returns)	: KJS::Value		: Result			*/

KJS::Value
	KBObjectProxy::MethodImp::call
	(	KJS::ExecState		*exec,
		KJS::Object		&object,
		const KJS::List		&args
	)
{
	KJS::Value	error	= checkArgs (exec, args) ;
	if (error.isValid()) return error ;

	return	callBase (exec, object, args) ;
}

KBObjectProxy
	*makeObjectProxy
	(	KBKJSInterpreter	*interp,
		KBNode			*node
	)
{
	return	new KBObjectProxy (interp, (KBObject *)node) ;
}

void	registerKBObject	()
{
	registerClass	("KBObject", makeObjectProxy) ;
}

