/***************************************************************************
    file	         : kb_pyvalue.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
			   (C)	2005
			   	Abel Deuring
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/


#include "kb_pyvalue.h"
#include "kb_pyscript.h"
#include "kb_pybase.h"

#include "structmember.h"

#include "kb_type.h"
#include "kb_value.h"
#include "kb_options.h"


#include <qregexp.h>

#define		__DEBUG_PYVALUE		(0)
#if		__DEBUG_PYVALUE
#include	"kb_dprintf.h"
#endif


static	bool	   __pyCodecLoaded	= false	;
static	QTextCodec *__pyCodec		= 0	;


/*  makeCodec	: Return coded for current python encoding		*/
/*  (returns)	: QTextCodec *	: Codec					*/

static	QTextCodec *makeCodec ()
{
	/* Codec is kept over calls, only make it when there is not one	*/
	/* already.							*/
	if (__pyCodecLoaded)
	{
		QString	pyenc	= KBOptions::getPythonEncoding() ;
		if (!pyenc.isEmpty()) __pyCodec = QTextCodec::codecForName(pyenc) ;
		__pyCodecLoaded = true ;
	}

	return	__pyCodec ;
}

/*  resetCodec	: Reset (ie., clear) codec				*/
/*  (returns)	: void		:					*/

void	resetCodec ()
{
	/* Just clear the codec pointer. A new one will be set up next	*/
	/* time "makeCodec" is called.					*/
	__pyCodec	= 0	;
	__pyCodecLoaded	= false	;
}


/*  ------------------------------------------------------------------	*/


/* kb_qStringToPyString							*/
/*		: convert a QString into a Python string		*/
/* qs		: QString to be converted				*/
/* (returns)	: Python string object 					*/
/* xxx allow explicit conversion into a Python Unicode object??		*/

PyObject
	LIBPYSCRIPT_API	*kb_qStringToPyString
	(	const QString	&qs
	)
{
	if (qs.isNull())
		return	PY_NONE	;

	QTextCodec *codec = makeCodec() ;
	if (codec == 0)
		return	PyString_FromString(__C(qs.ascii())) ;
	
	return	PyString_FromString(codec->fromUnicode(qs)) ;
}


/* kb_pyStringToQString							*/
/*		: convert a Python String into a QString		*/
/* po		: Python object to be converted				*/
/* error	: on return true, if a Python exception occurred	*/
/*		: the caller must clear or process the exception	*/
/* (returns)	: Python string object 					*/

QString
	LIBPYSCRIPT_API kb_pyStringToQString
	(	PyObject	*po,
		bool		&error
	)
{
	error = false ;

	if (po == 0)
		return	QString::null	;

	/* First case, see if the python object is already UniCode, in	*/
	/* which case force it into UTF8, which can then be converted	*/
	/* into a QString.						*/
	if (PyUnicode_Check (po))
	{
		PyObject *pUtf8 = PyUnicode_EncodeUTF8
				  (	PyUnicode_AS_UNICODE(po),
					PyUnicode_GET_SIZE  (po),
					"strict"
				  )	;
		if (pUtf8 == 0)
		{
			error	= true	;
			return ""	;
		}
		
		QString res = QString::fromUtf8 (PyString_AsString(pUtf8)) ;
		Py_DECREF (pUtf8) ;
		return	  res	  ;
	}

	/* Second case, if the python object is a string. If we have a	*/
	/* QT codec for the selected python encoding, the use that to	*/
	/* covert to a UniCode QString; if not then fall back on a	*/
	/* simple conversion and hope that the characters are OK.	*/
	if (PyString_Check(po))
	{
		QTextCodec *codec = makeCodec() ;
		if (codec == 0)
			return	PyString_AsString(po)	;
		return codec->toUnicode (PyString_AsString(po)) ;
	}
	
	/* Not UniCode, nor a string. See if the object can be		*/
	/* converted to a Python string; if not then fail, else convert	*/
	/* the string. The fallback logic is the same as above.		*/
	PyObject *asStr = PyObject_Str (po) ;
	if (asStr == 0)
	{
		error	= true	;
		return	""	;
	}
	
	QString	   res    ;
	QTextCodec *codec = makeCodec() ;
	if (codec == 0)
		res	= PyString_AsString(asStr)	;
	else	res	= codec->toUnicode (PyString_AsString(asStr)) ;

	Py_DECREF(asStr) ;
	return	res ;
}

/*  pyEncodeSourcecode							*/
/*		: Encode source code prior to compiling			*/
/*  s		: QString	: Source code				*/
/*  (returns)	: QCString	: Encoded source			*/
QCString
	LIBPYSCRIPT_API pyEncodeSourcecode
	(	QString		s
	)
{
	/* If we have a QT codec then use this, otherwise fall back on	*/
	/* conversion to Latin-1.					*/
	QTextCodec *codec = makeCodec() ;
	return	codec == 0 ? (QCString)(s.latin1()) : codec->fromUnicode(s) ;
}

/*  ------------------------------------------------------------------	*/

//  Two combined problems:
//
//  (1) The QT codecs can produce encoded strings which the corresponding
//  Python codecs can't decode. Example: A Unicode string containing
//  Umlauts, cyrillic caracters and Chinese symbols converted by a QT codec
//  into BIG5 will raise an error when fed to the Python codec.
//
//  (2) If Py_CompileString detects an encoding error,it simply raises
//  the meaningless exception "MemoryError". See bugs 979739 and 1115379
//  in Python's sourceforge bug tracker.
//
//  http://sourceforge.net/tracker/index.php?func=detail&aid=979739&group_id=5470&atid=105470
//
//  Most importantly, this Memory error does not provide a line number
//  for the error. So let's check for possible encoding errors before
//  passing the source code to the interpreter: check for every source
//  code line, if it will be accepted by the Python codec.


/*  checkScriptEncoding							*/
/*		: Check script encoding as aboce			*/
/*  location	: KBLocation &		: Source location for errors	*/
/*  source	: const QString &	: Source code			*/
/*  encname	: const char *		: Python encoder name		*/
/*  pError	: KBError &		: Return error			*/
/*  (returns)	: bool			: Success			*/

bool	LIBPYSCRIPT_API
	checkScriptEncoding
	(	KBLocation		&location,
		const QString		&source,
		const char		*encname,
		KBError			&pError
	)
{
	extern	QString	saveCompileError (KBLocation &, cchar *) ;

	DPRINTF
	((	"checkScriptEncoding: %s\n",
		encname
	))	;

	/* If we do not have a coded then always succeed (because the	*/
	/* "pyEncodeSourcecode" routine requires a codec to do the	*/
	/* encoding). This really should be checked at the caller end	*/
	/* but this keeps the "makeCodec()" function static in this	*/
	/* file.							*/
	if (makeCodec() == 0)
		return	true	;

	QRegExp lbreak	= QRegExp("(\\r\\n|\\r|\\n)") ;
	int	start	= 0	;
	int	stop	= 0	;
	int	lineno	= 0	;

	while ((stop = lbreak.search (source, stop)) >= 0)
	{
		QString	 code  = source.mid
				 (	start, 
					stop - start + lbreak.matchedLength()
				 )	;
		QCString line  = pyEncodeSourcecode (code) ;
		PyObject *test = PyUnicode_Decode
			         (	__C(line), 
					line.length(),
					encname,
					"strict"
			         )	;

		if (test == 0)
		{
			DPRINTF
			((	"checkScriptEncoding: test=0: [%s]\n",
				code.latin1()
			))	;

			
			QString	eText	= saveCompileError
					  (	location,
						"Unknown python compilation error occurred"
					  )	;

			pError	= KBError
				  (	KBError::Error,
					TR("Error compiling python script"),
					QString("line %1: %2")
						.arg(lineno)
						.arg(eText),
					__ERRLOCN
				  )	;

			return	false	;
		}

		Py_DECREF (test) ;
		stop   += lbreak.matchedLength() ;
		start   = stop	 ;
		lineno += 1	 ;
	}

	return	true	;
}

/*  ------------------------------------------------------------------	*/
	
/* PyKBValue: Python class that represents/contains a KBValue object 	*/

#define FIXED		0
#define FLOAT		1
#define	DATE		2
#define	TIME		3
#define	DATETIME	4
#define STRING		5
#define	BINARY		6
#define	BOOL		7


#define TYPE_ENTRY(x) {x, #x}
static struct
{
	int		num;
	const char*	name;
} kbtypes[] =
{
	TYPE_ENTRY(FIXED),
	TYPE_ENTRY(FLOAT),
	TYPE_ENTRY(DATE),
	TYPE_ENTRY(TIME),
	TYPE_ENTRY(DATETIME),
	TYPE_ENTRY(STRING),
	TYPE_ENTRY(BINARY),
	TYPE_ENTRY(BOOL),
	{-1, 	0}
};

typedef struct
{
	PyObject_HEAD
	KBValue	*value;
} pyKBValue;


static PyObject *PyKBValue_new(PyTypeObject *type, PyObject *args, PyObject *kw)
{
	pyKBValue *self;
	self = (pyKBValue*)type->tp_alloc(type, 0);
	if (self == 0)
		return 0;
	
	self->value = new KBValue();
	
	return (PyObject*) self;
}

static int PyKBValue_init(pyKBValue *self, PyObject *args, PyObject *kw)
{
	PyObject 	*pyVal, *asStr;
	int 		type = STRING;
	long		ival;
	double		dval;
	QString		sval;
	QByteArray	bval;
	bool		error;
	
	if (!PyArg_ParseTuple(args, "O|i", &pyVal, &type))
		return -1;
	
	switch (type)
	{
		case	FIXED:
			ival = PyInt_AsLong(pyVal);
			if (PyErr_Occurred())
				return -1;
			delete self->value;
			self->value = new KBValue(ival, &_kbFixed);
			break;

		case	FLOAT:
			dval = PyFloat_AsDouble(pyVal);
			if (PyErr_Occurred())
				return -1;
			delete self->value;
			self->value = new KBValue(dval, &_kbFloat);
			break;

		case	DATE:
			// xxx remember: Python 2.4 has a special DateTime objects
			PyErr_SetString(PyKBRekallError, 
				"conversion to KB::ITDate type not yet implemented");
			return -1;
		
		case	TIME:
			// xxx remember: Python 2.4 has a special DateTime objects
			PyErr_SetString(PyKBRekallError, 
				"conversion to KB::ITTime type not yet implemented");
			return -1;
		
		case	DATETIME:
			// xxx remember: Python 2.4 has a special DateTime objects
			PyErr_SetString(PyKBRekallError, 
				"conversion to KB::ITDateTime type not yet implemented");
			return -1;
		
		case	STRING:
			sval = kb_pyStringToQString(pyVal, error);
			if (error)
				return -1;
			delete self->value;
			self->value = new KBValue(sval, &_kbString);
			break;
		
		case	BINARY:
			asStr = PyObject_Str(pyVal);
			if (asStr == 0)
				return -1;
			bval.duplicate(PyString_AS_STRING(asStr), 
					PyString_Size(asStr));
			delete self->value;
			self->value = new KBValue(bval, &_kbBinary);
			break;
			
		case	BOOL:
			delete self->value;
			self->value = new KBValue(PyObject_IsTrue(pyVal),
						  &_kbBool);
			break;
			
		default:
			PyErr_SetString(PyExc_TypeError, "invalid KB type");
			return -1;
	}
	if (self->value == 0)
	{
		PyErr_NoMemory();
		return -1;
	}
	return 0;
}

static void PyKBValue_dealloc(pyKBValue *self)
{
	delete self->value;
	self->ob_type->tp_free((PyObject*) self);
}

static PyMethodDef PyKBValue_methods[] =
{
	{0}
};

static PyMemberDef PyKBValue_members[] = 
{
	{0}
};

PyTypeObject pyKBValueType =
{
	PyObject_HEAD_INIT(NULL)
	0,				/* ob_size 		*/
	"RekallMain.KBValue",		/* tp_name 		*/
	sizeof(pyKBValue),		/* tp_basicsize		*/
	0,				/* tp_itemsize		*/
	(destructor) PyKBValue_dealloc,	/* tp_dealloc		*/
	0,				/* tp_print		*/
	0,				/* tp_getattr		*/
	0,				/* tp_setattr		*/
	0,				/*tp_compare		*/
	0,				/*tp_repr		*/
	0,				/*tp_as_number		*/
	0,				/*tp_as_sequence	*/
	0,				/*tp_as_mapping		*/
	0,				/*tp_hash 		*/
	0,				/*tp_call		*/
	0,				/*tp_str		*/
	0,				/*tp_getattro		*/
	0,				/*tp_setattro		*/
	0,				/*tp_as_buffer		*/
	Py_TPFLAGS_DEFAULT,		/*tp_flags		*/
	"xxx no docs yet",		/* tp_doc		*/
	0,				/* tp_traverse		*/
	0,				/* tp_clear 		*/
	0,				/* tp_richcompare 	*/
	0,				/* tp_weaklistoffset 	*/
	0,				/* tp_iter 		*/
	0,				/* tp_iternext 		*/
	PyKBValue_methods,		/* tp_methods 		*/
	PyKBValue_members,		/* tp_members 		*/
	0,				/* tp_getset 		*/
	0,				/* tp_base 		*/
	0,				/* tp_dict 		*/
	0,				/* tp_descr_get 	*/
	0,				/* tp_descr_set 	*/
	0,				/* tp_dictoffset 	*/
	(initproc)PyKBValue_init,	/* tp_init 		*/
	0,				/* tp_alloc 		*/
	PyKBValue_new,			/* tp_new 		*/

};

KBValue 
	LIBPYSCRIPT_API *kb_pyKBValueCheck(PyObject *po)
{
	if (po->ob_type == &pyKBValueType)
	{
		return ((pyKBValue*)po)->value;
	}
	else	return 0;
}


void	initPyValue(PyObject *module)
{
	if (PyType_Ready(&pyKBValueType) < 0)
		return;
	
	int i = 0;
	while (kbtypes[i].num >= 0)
	{
		PyObject	*val = PyInt_FromLong(kbtypes[i].num);
		if (-1 == PyDict_SetItemString 
				(
					pyKBValueType.tp_dict,
					kbtypes[i].name,
					val
				))
			return	; // need to signal an error!!!
		Py_DECREF(val)	;
		i += 1 ;
	}
	
	Py_INCREF(&pyKBValueType);
	PyModule_AddObject(module, "KBValue", (PyObject*) &pyKBValueType);
	
}
